/* JavaGame is a game engine used to teach computer programming.
 * There is a corresponding JavaScript version to use in a web page.
 *
 * JavaGame copyright 2020 Itty Bitty Computers and released at this time
 * to the public as open source. There are no warranties of any kind.
 */                                               // 2020 July 25

import java.io.File;               // these all used by the mechanism..
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowListener;
import java.awt.event.MouseListener;
import java.awt.event.KeyListener;
import java.awt.event.WindowEvent;
import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;
import java.awt.MouseInfo;
import java.awt.Graphics;
import java.awt.Insets; // not available at win creation
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;
import javax.swing.JOptionPane;
import javax.swing.Timer;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.net.HttpURLConnection;
import java.net.URL;

public class JavaGame extends JFrame implements MouseListener, KeyListener,
    WindowListener {
  private static final long serialVersionUID = 1L; // unneed but Java insists

 /**
  * All GameEngine (non-widget) control is done in this class.
  */

  // First some public data...

// [Shared] GameEngine constants..
  final static String GameTitle = "GameEngine",        RevDate = "20/7/25";

  final static boolean LockStart = false,     Do_JS = false,
      Logging = false, GrafLog = false, Dev_Log = false,
      StopSprites = true, ScrnSee = true;

  final static int Qlog = 0, y2020 = -304, FrameTime = 100,
      MaxWgts = 1024, Max_Log = 333, Fust_Log = 222, Win_Hi = 640, Win_Wi = 800,
      myPL = 1,  WgtNameHi = 16, Clik2time = 999;

// [Predefined] GameEngine Macros & Data..              -- 2020 July 6
  public final int[] Wgt_Macros = {253,4, 0,3,1,12,12,0xFFFFFF,0, // checkbox..
        2,3,1,12,12,0x0,-7, 2,1,2,16,12,0xFF0000,-6, 4,4,16,16,24,0x0,-66,
      254,3, 4,2,0,12,16,0x0,-66,                    // labelled text entry..
        0,0,8,20,32,0xFFFFFF,0x42616B46,  // = 'BakF'
        1540,3,12,12,20,0x0,-77,
      252,9, 4,5,8,16,24,0x0,-66,    // pushbutton (tall=22,color=0xCCCCCC)..
        0,1,1,2,30,0xFFFFFF,0, 0,1,1,18,2,0xFFFFFF,0,
        0,17,2,2,30,0x999999,0, 0,2,28,17,2,0x999999,0,
        0,0,0,1,32,0x0,0, 0,0,30,20,1,0x0,0, 0,0,0,20,1,0x0,0,
        0,19,0,1,32,0x0,0,
      251,5, 2,0,0,16,16,0x0,-5, 0,0,0,1,16,0x0,0, // color swatch (t/w=16)..
        0,0,15,16,1,0x0,0, 0,0,0,16,1,0x0,0, 0,15,0,1,16,0x0,0,
      250,28, 0,0,0,1,1024,0x0,0,  0,0,0,256,1,0x0,0,  // marching ants border..
        0,31,0,3,1024,0x0,0,  0,0,31,256,3,0x0,0, // good for 256(tall) x 1K(wi)
        2,15,0,3,256,0xFFFFFF,-2, 2,0,15,32,3,0xFFFFFF,-2,
        2,0,0,1,256,0xFFFFFF,-2,  2,0,0,32,1,0xFFFFFF,-2,
        2,31,256,3,256,0xFFFFFF,-2, 2,31,512,3,256,0xFFFFFF,-2,
        2,31,768,3,256,0xFFFFFF,-2, 2,32,31,32,3,0xFFFFFF,-2,
        2,64,31,32,3,0xFFFFFF,-2, 2,96,31,32,3,0xFFFFFF,-2,
        2,0,256,1,256,0xFFFFFF,-2, 2,0,512,1,256,0xFFFFFF,-2,
        2,0,768,1,256,0xFFFFFF,-2, 2,32,0,32,1,0xFFFFFF,-2,
        2,64,0,32,1,0xFFFFFF,-2, 2,96,0,32,1,0xFFFFFF,-2,
        2,128,31,32,3,0xFFFFFF,-2, 2,160,31,32,3,0xFFFFFF,-2,
        2,192,31,32,3,0xFFFFFF,-2, 2,224,31,32,3,0xFFFFFF,-2,
        2,128,0,32,1,0xFFFFFF,-2, 2,160,0,32,1,0xFFFFFF,-2,
        2,192,0,32,1,0xFFFFFF,-2, 2,224,0,32,1,0xFFFFFF,-2,
      248,1, 5,0,0,32,32,-1,0,                 // tool wgt template (stub) **
      247,3, 5,0,0,64,64,-1,0,                 // V+H-scro'd panel (s=64/64)
        0x28F4,0,52,64,12,0x999999,0, 0x28F6,52,0,12,64,0x999999,0,
      246,1, 0,0,0,12,64,0xFFFFFF,0,     // H-scrollbar thumb (in 12/64 box)
      244,1, 0,0,0,64,12,0xFFFFFF,0,     // V-scrollbar thumb (in 64/12 box)
      245,2, 5,0,0,64,64,-1,0,                 // V-scrolled panel (s=64/78)
        0x28F4,0,64,64,12,0x999999,0,
      241,7, 4,10,8,16,252,0x0,-66,               // alert box (t=64,w=240)..
        1276,36,192,20,40,0xCCCCCC,1, 1276,36,408,20,60,0xCCCCCC,2,
        0,0,0,2,240,0x0,0, 0,0,238,64,2,0x0,0,
        0,0,0,64,2,0x0,0, 0,62,0,2,240,0x0,0,
      240,8, 254,8,8,40,208,-1,0,  2,9,408,16,16,0x0,-4, // InpDlg (t=64,c=wh)..
        1276,8,192,20,40,0xCCCCCC,1, 1276,36,172,20,60,0xCCCCCC,2, // (w=240)
        0,0,0,2,240,0x0,0, 0,0,238,64,2,0x0,0,
        0,0,0,64,2,0x0,0, 0,62,0,2,240,0x0,0,
      249,3, 2,3,1,12,12,0xFFFFFF,-8,  // radio button (tall=16 color=-1)..
        2,3,1,12,12,0x0,-8,  4,4,16,16,44,0x0,-66,
      255,0,  0,0};
  public final int[] DragArrows = {256,896,1984,256,256, // enb by +2048
      0x2108,0x610C,0xFFFE,0x610C,0x2108,256,256,1984,896,256,0};
  public final int[] NoCoGr = {0x55555555,0xAAAAAAAA, 0x55555555,0xAAAAAAAA,
        0x55555555,0xAAAAAAAA, 0x55555555,0xAAAAAAAA, 0x55555555,0xAAAAAAAA,
        0x55555555,0xAAAAAAAA, 0x55555555,0xAAAAAAAA, 0x55555555,0xAAAAAAAA,
        0,0};
  public final static int[] ChkBoxz = {
      4095,4095,3075,3075,3075,3075,3075,3075,3075,3075,4095,4095};
  public final static int[] ChkBitz = {
      0x380,0x700,0x1E00,0x3C00,0x1E00,0x780,0x1E0,0x70,0x30,24,4,0, 0,0,0,0};

  public final static int[] Circ4f = {0,12,30,30,12, 0};
  public final static int[] Circ6h = {12,18,33,33,18,12, 0};
  public final int[] Circ6a6 = {12,2,0,0,16,12, 12,16,0,32,2,12, 8,16,32,31,2,4,
      0,0,33,31,2,0, 0,2,33,1,16,0, 4,2,1,0,16,8, 0};
  public final static int[] Circ5f = {0,28,62,62,62,28, 0};
  public final static int[] Circ7h = {28,34,65,65,65,34,28, 0};
  public final int[] Circ7a8 = {12,2,1,0,64,32,24, 12,2,1,0,64,32,24,
        16,32,64,65,1,2,4, 16,32,64,65,1,2,4,   12,2,1,0,64,32,24,
      12,2,1,0,64,32,24,  16,32,64,65,1,2,4, 16,32,64,65,1,2,4, 0};
  public final static int[] Cir12f = {0x0F0,0x3FC,0x7FE,0x7FE,0xFFF,0xFFF,
      0xFFF,0xFFF,0x7FE,0x7FE,0x3FC,0x0F0, 0};
  public final static int[] Cir12h = {0x0F0,0x30C,0x402,0x402,0x801,0x801,
      0x801,0x801,0x402,0x402,0x30C,0x0F0,  0x0F0,0x30C,0x402,0x4F2,0x9D9,0x9E9,
      0x9F9,0x9F9,0x4F2,0x402,0x30C,0x0F0, 0};
  public final int[] Cir12a2 = {0x0F0,0x30C,0x402,0x402,0x801,0x801,
      0x801,0x801,0x402,0x402,0x30C,0x0F0, 0, 0};
  public final static int[] Cir13f = {0x1F0,0x7FC,0xFFE,0xFFE,0x1FFF,0x1FFF,
      0x1FFF,0x1FFF,0x1FFF,0xFFE,0xFFE,0x7FC,0x1F0, 0};
  public final static int[] Cir13h = {0x1F0,0x60C,0x802,0x802,0x1001,0x1001,
      0x1001,0x1001,0x1001,0x802,0x802,0x60C,0x1F0, 0};
  public final int[] Cir13a2 = {0x1F0,0x60C,0x802,0x802,0x1001,0x1001,
      0x1001,0x1001,0x1001,0x802,0x802,0x60C,0x1F0, 0, 0};
  public final static int[] Cir16f = {0x7E0,0x1FF8,0x3FFC,0x7FFE,0x7FFE,0xFFFF,
      0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0x7FFE,0x7FFE,0x3FFC,0x1FF8,0x7E0, 0};
  public final static int[] Cir16h = {0x7E0,0x1818,0x2004,0x4002,0x4002,0x8001,
      0x8001,0x8001,0x8001,0x8001,0x8001,0x4002,0x4002,0x2004,0x1818,0x7E0, 0};
  public final static int[] Cir16a2 = {0x7E0,0x1818,0x2004,0x4002,0x4002,0x8001,
      0x8001,0x8001,0x8001,0x8001,0x8001,0x4002,0x4002,0x2004,0x1818,0x7E0,
      0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0};

  final int[] AimIndex = {4,5,1,13,12,15,3,7}; // 1=+h,3=-h,4=+v,12=-v
  final int[] SmallAnts = {0xE3806, 0x387007, // off-bits:15, +/-n:4, patln:4
      0xF0F008, 0x3C1E009, 0xF83E00A, 0x3F7E00B, 0xE3806, 0xE3816, 0x387007,
      0xF0F0F8, 0xF0F008,  0xF0F018,  0x3C1E009, 0xF83E0FA, 0xF83E00A, 0};

  public final int[] Antz = { // max tall = 32, max wide = 256..
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E,
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E,
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E,
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E,
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E,
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E,
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E,
      0x0F0F0F0F,0x87878787, 0xC3C3C3C3,0xE1E1E1E1,
      0xF0F0F0F0,0x78787878, 0x3C3C3C3C,0x1E1E1E1E, // [63]
      1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0,
      1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0,
      1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0,
      1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0,
      1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0,
      1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0,
      1,1,1,1,0,0,0,0, 1,1,1,1,0,0,0,0}; // (+8 needed for animation)

  public static boolean UserMode = false; // (Java only..)

  // Default Monaco-like font +0
  private final int[] MonoFnt = { 0x4D6F6E6F, 0xC0209, 1, 6, 0, 6, // 'Mono'
                               0, 0, 190, 0, 0,              // !
     0, 14, 0, 14, 0,          40, 124, 40, 124, 40,
     76, 146, 511, 146, 100,   66, 37, 146, 328, 132,
     96, 150, 137, 118, 208,   0, 0, 14, 0, 0,
     0, 124, 130, 257, 0,      0, 257, 130, 124, 0,
     8, 42, 28, 42, 8,         16, 16, 124, 16, 16,
     0, 512, 448, 192, 0,      16, 16, 16, 16, 16,
     0, 192, 192, 0, 0,        0, 384, 96, 24, 6,
     56, 68, 130, 68, 56,      0, 4, 254, 0, 0,              // 0
     132, 194, 162, 146, 140,  68, 130, 146, 146, 108,
     48, 40, 36, 254, 32,      78, 138, 138, 138, 114,
     120, 148, 146, 146, 96,   2, 2, 226, 26, 6,
     108, 146, 146, 146, 108,  12, 146, 146, 82, 60,
     0, 204, 204, 0, 0,        0, 716, 460, 0, 0,
     0, 16, 40, 68, 130,       40, 40, 40, 40, 40,
     0, 130, 68, 40, 16,       4, 2, 162, 18, 12,
     120, 132, 306, 330, 60,   224, 56, 38, 56, 224,         // @A
     130, 254, 146, 146, 108,  124, 130, 130, 130, 68,
     130, 254, 130, 68, 56,    254, 146, 146, 130, 130,
     254, 18, 18, 18, 2,       124, 130, 130, 146, 244,
     254, 16, 16, 16, 254,     0, 130, 254, 130, 0,
     96, 128, 128, 128, 126,   254, 16, 40, 68, 130,
     254, 128, 128, 128, 128,  254, 4, 24, 4, 254,
     254, 4, 24, 32, 254,      124, 130, 130, 130, 124,
     254, 18, 18, 18, 12,      124, 130, 130, 386, 636,      // P
     254, 18, 50, 82, 140,     76, 146, 146, 146, 100,
     2, 2, 254, 2, 2,          126, 128, 128, 128, 126,
     14, 48, 192, 48, 14,      30, 224, 24, 224, 30,
     198, 40, 16, 40, 198,     6, 8, 240, 8, 6,
     194, 162, 146, 138, 134,  0, 0, 511, 257, 0,
     0, 6, 24, 96, 384,        0, 257, 511, 0, 0,
     4, 2, 1, 2, 4,            128, 128, 128, 128, 128,
     0, 2, 4, 0, 0,            64, 168, 168, 168, 240,       // `a
     254, 136, 136, 136, 112,  112, 136, 136, 136, 136,
     112, 136, 136, 136, 254,  112, 168, 168, 168, 48,
     16, 252, 18, 2, 4,        112, 648, 648, 648, 504,
     254, 8, 8, 8, 240,        0, 136, 250, 128, 0,
     512, 520, 506, 0, 0,      254, 16, 48, 72, 128,
     0, 2, 254, 0, 0,          248, 8, 240, 8, 240,
     248, 16, 8, 8, 240,       112, 136, 136, 136, 112,
     1016, 136, 136, 136, 112, 112, 136, 136, 136, 1016,     // p
     248, 16, 8, 8, 16,        144, 168, 168, 168, 72,
     0, 8, 126, 136, 128,      120, 128, 128, 128, 248,
     8, 48, 192, 48, 8,        56, 192, 48, 192, 56,
     136, 80, 32, 80, 136,     536, 608, 384, 96, 24,
     136, 200, 168, 152, 136,  0, 16, 238, 257, 0,
     0, 0, 511, 0, 0,          0, 257, 238, 16, 0,
     4, 2, 2, 4, 2 }; //~MonoFnt

  private final int[] TextFnt = { 0x54657874, 0xC0309, 1, 3, // 'Text'
      100,100,101,104,109,114,124,132, 134,137,140,145,150,152,155,157,
      159,164,169,174,179,184,189,194, 199,204,209,210,212,218,224,230,
      234,244,253,259,266,273,279,285, 293,301,304,309,317,323,332,340,
      347,353,360,367,372,379,387,394, 405,413,420,426,428,430,432,437,
      443,445,450,455,459,464,468,472, 477,483,486,488,494,497,506,512,
      517,522,527,532,536,539,545,552, 561,566,572,576,579,580,583,588,
      0x017E,0x000E,0x0000,0x000E,0x0048,0x01FE,0x0048,0x01FE, // 100..
      0x0048,0x008C,0x0112,0x03FF,0x0112,0x00E4,0x0018,0x0124,
      0x00A2,0x0052,0x002C,0x00D2,0x012A,0x0116,0x0092,0x0060,
      0x00C0,0x012C,0x0112,0x00AA,0x0044,0x00B0,0x0110,0x0100,
      0x0008,0x0006,0x00F0,0x030C,0x0402,0x0402,0x030C,0x00F0,
      0x0014,0x0008,0x003E,0x0008,0x0014,0x0020,0x0020,0x00F8,
      0x0020,0x0020,0x0400,0x0300,0x0020,0x0020,0x0020,0x0100,
      0x0000,0x01E0,0x001E,0x00FC,0x0102,0x0102,0x0102,0x00FC,
      0x0000,0x0104,0x01FE,0x0100,0x0000,0x0184,0x0142,0x0122,
      0x0112,0x008C,0x0084,0x0102,0x0112,0x0112,0x00EC,0x0060,
      0x0050,0x0048,0x0044,0x01FE,0x009E,0x0112,0x0112,0x0112, // 180..
      0x00E2,0x00F8,0x0114,0x0112,0x0112,0x00E0,0x0006,0x0002,
      0x01C2,0x0032,0x000E,0x00EC,0x0112,0x0112,0x0112,0x00EC,
      0x001C,0x0122,0x0122,0x00A2,0x007C,0x0108,0x0400,0x0308,
      0x0020,0x0020,0x0050,0x0050,0x0088,0x0088,0x0050,0x0050,
      0x0050,0x0050,0x0050,0x0050,0x0088,0x0088,0x0050,0x0050,
      0x0020,0x0020,0x0004,0x0162,0x0012,0x000C,0x01F0,0x020C,
      0x04E4,0x0512,0x050A,0x048A,0x01CA,0x0132,0x0084,0x0078,
      0x0100,0x0180,0x0160,0x0058,0x0046,0x0058,0x0160,0x0180,
      0x0100,0x0102,0x01FE,0x0112,0x0112,0x0112,0x00EC,0x0078,
      0x0084,0x0102,0x0102,0x0102,0x0084,0x004E,0x0102,0x01FE, // 260..
      0x0102,0x0102,0x0102,0x0084,0x0078,0x0102,0x01FE,0x0112,
      0x0112,0x013A,0x0186,0x0102,0x01FE,0x0112,0x0012,0x003A,
      0x0006,0x0078,0x0084,0x0102,0x0102,0x0102,0x0124,0x00EE,
      0x0020,0x0102,0x01FE,0x0112,0x0010,0x0010,0x0112,0x01FE,
      0x0102,0x0102,0x01FE,0x0102,0x0080,0x0100,0x0102,0x00FE,
      0x0002,0x0102,0x01FE,0x0112,0x0028,0x0046,0x0182,0x0102,
      0x0100,0x0102,0x01FE,0x0102,0x0100,0x0100,0x0180,0x0102,
      0x01FE,0x0106,0x0018,0x0060,0x0018,0x0106,0x01FE,0x0102,
      0x0102,0x01FE,0x0108,0x0010,0x0020,0x0042,0x01FE,0x0002,
      0x0078,0x0084,0x0102,0x0102,0x0102,0x0084,0x0078,0x0102, // 340..
      0x01FE,0x0122,0x0022,0x0022,0x001C,0x0078,0x0084,0x0102,
      0x0302,0x0502,0x0484,0x0078,0x0102,0x01FE,0x0122,0x0022,
      0x0062,0x01A2,0x011C,0x01CC,0x0092,0x0112,0x0124,0x00CE,
      0x0006,0x0002,0x0102,0x01FE,0x0102,0x0002,0x0006,0x0002,
      0x00FE,0x0102,0x0100,0x0100,0x0102,0x00FE,0x0002,0x0002,
      0x000E,0x0032,0x01C0,0x0032,0x000E,0x0002,0x0002,0x000E,
      0x0032,0x01C0,0x0032,0x000E,0x0032,0x01C0,0x0032,0x000E,
      0x0002,0x0102,0x0186,0x014A,0x0030,0x0030,0x014A,0x0186,
      0x0102,0x0002,0x0006,0x011A,0x01E0,0x011A,0x0006,0x0002,
      0x0186,0x0142,0x0122,0x0112,0x010A,0x0186,0x07FE,0x0402, // 420..
      0x001E,0x01E0,0x0402,0x07FE,0x0008,0x0004,0x0002,0x0004,
      0x0008,0x0400,0x0400,0x0400,0x0400,0x0400,0x0400,0x000C,
      0x0002,0x00C0,0x0128,0x0128,0x00F0,0x0100,0x0002,0x00FE,
      0x0108,0x0108,0x00F0,0x00F0,0x0108,0x0108,0x0098,0x00F0,
      0x0108,0x010A,0x01FE,0x0100,0x00F0,0x0128,0x0128,0x00B0,
      0x0108,0x01FC,0x010A,0x0002,0x02B0,0x0548,0x0548,0x0530,
      0x0208,0x0102,0x01FE,0x0008,0x0108,0x01F0,0x0100,0x0108,
      0x01FA,0x0100,0x0408,0x03FA,0x0102,0x01FE,0x0020,0x0158,
      0x0188,0x0100,0x0102,0x01FE,0x0100,0x0108,0x01F0,0x0008,
      0x0108,0x01F0,0x0008,0x0108,0x01F0,0x0100,0x0108,0x01F0, // 500..
      0x0108,0x0008,0x01F0,0x0100,0x00F0,0x0108,0x0108,0x0108,
      0x00F0,0x0408,0x07F8,0x0508,0x0108,0x00F0,0x00F0,0x0108,
      0x0508,0x07F8,0x0400,0x0108,0x01F0,0x0108,0x0008,0x0010,
      0x0190,0x0128,0x0128,0x00C8,0x0008,0x00FC,0x0108,0x0008,
      0x00F8,0x0100,0x0108,0x00F8,0x0100,0x0008,0x0018,0x0068,
      0x0180,0x0068,0x0018,0x0008,0x0008,0x0018,0x0068,0x0198,
      0x0068,0x0180,0x0068,0x0018,0x0008,0x0108,0x0198,0x0060,
      0x0198,0x0108,0x0408,0x0418,0x0268,0x0180,0x0068,0x0018,
      0x0198,0x0148,0x0128,0x0198,0x0020,0x03DC,0x0402,0x07FE, // 572..
      0x0402,0x03DC,0x0020,0x0010,0x0008,0x0018,0x0010,0x0008 }; //~TextFnt

  private final int[] ItalFnt = { 0x4974616C, 0xC0309, 0, 4, // 'Ital'
      100,100,105,109,116,123,135,143,145,151,157,163,168,170,173,174,
      180,187,193,201,208,216,224,231,238,245,252,256,261,268,275,282,
      288,301,309,318,328,338,348,358,368,380,387,395,406,413,426,438,
      447,456,465,475,484,493,503,511,523,535,544,554,561,565,572,577,
      583,585,592,598,603,610,615,623,630,637,640,646,652,655,663,669,
      675,682,689,695,699,703,709,715,724,730,736,741,747,748,753,758,
      0x0100,0x0040,0x0030,0x000C,0x0002,0x000C,0x0002,0x000C, // 100..
      0x0002,0x0140,0x00C0,0x0178,0x00CC,0x007A,0x000C,0x000A,
      0x0180,0x0300,0x01DC,0x0132,0x00DE,0x0023,0x0006,0x0100,
      0x0110,0x00A8,0x0064,0x0054,0x00EA,0x0152,0x012A,0x00AA,
      0x0046,0x0002,0x0002,0x01C0,0x0120,0x0120,0x009C,0x0072,
      0x01CA,0x0136,0x0010,0x0008,0x0006,0x07C0,0x0030,0x0008,
      0x0004,0x0002,0x0002,0x0400,0x0400,0x0200,0x0100,0x00C0,
      0x003E,0x0010,0x0014,0x0038,0x000C,0x001A,0x0004,0x0020,
      0x00E0,0x0030,0x0028,0x0020,0x0400,0x0300,0x0020,0x0020,
      0x0020,0x0100,0x0100,0x00C0,0x0020,0x0010,0x000C,0x0002,
      0x01C0,0x0130,0x010C,0x0102,0x00C2,0x0032,0x000E,0x0100, // 180..
      0x0100,0x01C0,0x0034,0x000C,0x0002,0x0100,0x0180,0x0140,
      0x0124,0x0122,0x0092,0x0012,0x000E,0x0180,0x0100,0x0104,
      0x0112,0x00F2,0x0012,0x000E,0x0040,0x0060,0x0050,0x0148,
      0x00C8,0x0034,0x000C,0x0002,0x0180,0x0110,0x011C,0x0112,
      0x00D2,0x0022,0x0002,0x0002,0x01C0,0x0130,0x0118,0x0114,
      0x00D2,0x0022,0x0002,0x0100,0x0084,0x0042,0x0022,0x0012,
      0x000A,0x0006,0x01C0,0x0120,0x012C,0x0112,0x00F2,0x0012,
      0x000E,0x0100,0x0110,0x012C,0x00A2,0x0062,0x0032,0x000E,
      0x0100,0x0000,0x0000,0x0008,0x0400,0x0300,0x0000,0x0000,
      0x0008,0x0020,0x0020,0x0050,0x0090,0x0088,0x0008,0x0008, // 260..
      0x0040,0x0050,0x0050,0x0050,0x0050,0x0050,0x0010,0x0080,
      0x0080,0x0048,0x0048,0x0050,0x0020,0x0020,0x0100,0x0044,
      0x0022,0x0022,0x0012,0x000E,0x0100,0x06C0,0x0430,0x05C8,
      0x0524,0x0014,0x0192,0x014A,0x012A,0x009A,0x0042,0x0036,
      0x0008,0x0100,0x0180,0x0040,0x0060,0x0050,0x0148,0x0144,
      0x01FE,0x0100,0x0100,0x01C0,0x0130,0x011A,0x0116,0x00D2,
      0x0032,0x000E,0x0040,0x01B0,0x0108,0x0104,0x0102,0x0082,
      0x0042,0x0006,0x000C,0x0002,0x0100,0x0100,0x01C0,0x0130,
      0x010A,0x0106,0x0082,0x0042,0x0036,0x0008,0x0100,0x0100,
      0x01C0,0x0130,0x011A,0x0116,0x00B2,0x000A,0x0006,0x0002, // 340..
      0x0100,0x0100,0x01C0,0x0030,0x001A,0x0016,0x0032,0x0002,
      0x0006,0x0002,0x0040,0x01B0,0x0108,0x0104,0x0102,0x0102,
      0x00E2,0x0026,0x002C,0x0002,0x0100,0x0100,0x01C0,0x0030,
      0x001E,0x0112,0x0112,0x01D0,0x0030,0x000E,0x0002,0x0002,
      0x0100,0x0100,0x01C0,0x0030,0x000E,0x0002,0x0002,0x0180,
      0x0100,0x0100,0x00C0,0x0030,0x000E,0x0002,0x0002,0x0100,
      0x0100,0x01C0,0x0030,0x001E,0x0162,0x0192,0x0108,0x0006,
      0x0002,0x0002,0x0100,0x0100,0x01C0,0x0130,0x010E,0x0102,
      0x0082,0x0100,0x0100,0x01C0,0x0030,0x000A,0x007E,0x0122,
      0x0110,0x01C8,0x0034,0x000E,0x0002,0x0002,0x0100,0x0100, // 420..
      0x01C0,0x0030,0x000E,0x001A,0x0160,0x00C0,0x0030,0x000E,
      0x0002,0x0002,0x0040,0x01B0,0x0108,0x0104,0x0102,0x0082,
      0x0042,0x0036,0x0008,0x0100,0x0100,0x01C0,0x0030,0x002A,
      0x0026,0x0022,0x0012,0x000E,0x0040,0x01B0,0x0708,0x0504,
      0x0102,0x0082,0x0042,0x0036,0x0008,0x0100,0x0100,0x01C0,
      0x0030,0x002A,0x0166,0x01A2,0x0022,0x0012,0x000E,0x0100,
      0x00C0,0x0100,0x011C,0x0112,0x00E2,0x0006,0x000C,0x0002,
      0x0100,0x0104,0x01C2,0x0032,0x000E,0x0002,0x0002,0x0006,
      0x0002,0x01C0,0x0130,0x010A,0x0106,0x0102,0x00C0,0x0030,
      0x000E,0x0002,0x0002,0x0100,0x00FA,0x0046,0x0022,0x0010, // 500..
      0x000A,0x0006,0x0002,0x0100,0x00FA,0x0006,0x001A,0x0120,
      0x00F2,0x000E,0x0032,0x0010,0x000E,0x0002,0x0002,0x0100,
      0x0100,0x0180,0x0040,0x0026,0x013A,0x0162,0x0190,0x0008,
      0x0006,0x0002,0x0002,0x0100,0x0100,0x01DA,0x0026,0x0012,
      0x0008,0x0006,0x0002,0x0002,0x0100,0x0180,0x0140,0x0144,
      0x0122,0x0112,0x008A,0x000A,0x0006,0x0002,0x0600,0x0500,
      0x00C0,0x0030,0x000C,0x0002,0x0002,0x0180,0x0060,0x0018,
      0x0006,0x0400,0x0400,0x0300,0x00C0,0x0030,0x000A,0x0006,
      0x0008,0x0004,0x0002,0x0006,0x0008,0x0400,0x0400,0x0400,
      0x0400,0x0400,0x0400,0x000C,0x0002,0x01C0,0x0130,0x0110, // 580..
      0x0088,0x01E8,0x0118,0x0080,0x0080,0x0170,0x011E,0x010A,
      0x00C8,0x0038,0x00E0,0x0110,0x0108,0x0088,0x0018,0x01E0,
      0x0110,0x0108,0x0088,0x0170,0x011A,0x0006,0x01E0,0x0130,
      0x0128,0x00A8,0x0018,0x0400,0x0400,0x0388,0x0078,0x000C,
      0x000A,0x0002,0x0002,0x0300,0x04B0,0x04C8,0x0548,0x0348,
      0x0030,0x0010,0x0180,0x0070,0x002E,0x0012,0x01C8,0x0138,
      0x0080,0x0188,0x0178,0x008A,0x0600,0x0400,0x0388,0x0078,
      0x0008,0x0002,0x0180,0x0070,0x002E,0x00E2,0x0118,0x0088,
      0x01E0,0x011A,0x0086,0x0100,0x00E8,0x0018,0x0190,0x0078,
      0x0010,0x01D0,0x0138,0x0188,0x0078,0x0028,0x0010,0x01E8, // 660..
      0x0118,0x00E0,0x0110,0x0108,0x0108,0x0088,0x0070,0x0400,
      0x0780,0x0568,0x0118,0x0108,0x0088,0x0078,0x01C0,0x0130,
      0x0510,0x0688,0x05C8,0x0070,0x0008,0x0180,0x0068,0x0018,
      0x0008,0x0008,0x0010,0x0180,0x0138,0x01C8,0x0018,0x0188,
      0x0178,0x008E,0x0008,0x0188,0x0178,0x0088,0x0040,0x01E0,
      0x0098,0x0008,0x0038,0x01C0,0x0080,0x0060,0x0018,0x0008,
      0x01F8,0x0080,0x0060,0x0010,0x01F8,0x0080,0x0060,0x0018,
      0x0100,0x0088,0x0088,0x00F0,0x0110,0x0088,0x0408,0x0418,
      0x02F0,0x0100,0x00C0,0x0038,0x0100,0x0190,0x0168,0x0218,
      0x0308,0x0020,0x03E0,0x0418,0x0404,0x0002,0x0002,0x01FE, // 740..
      0x0400,0x0600,0x0382,0x005E,0x0024,0x0060,0x0020,0x0020,
      0x0020,0x0030 }; //~ItalFnt

  private final int[] TinyFnt = { 0x54696E79, 0x90207, 0, 2, // 'Tiny'
      100,100,102,105,110,114,120,126,128,131,134,137,142,144,147,148,
      150,154,157,161,165,169,173,177,181,185,189,190,192,195,199,202,
      205,212,219,224,228,233,238,243,248,254,257,261,267,272,279,285,
      290,295,300,305,309,314,320,325,332,337,342,346,349,351,354,357,
      362,364,367,371,374,377,380,383,386,390,392,394,398,400,406,410,
      413,417,420,424,427,430,433,436,441,444,447,450,453,455,458,462,
      0x0000,0x005E,0x0006,0x0000,0x0006,0x0024,0x007E,0x0024, // 100..
      0x007E,0x0024,0x0024,0x004A,0x00FF,0x0032,0x000C,0x0052,
      0x002C,0x0038,0x004C,0x0032,0x0030,0x004C,0x005A,0x0036,
      0x0020,0x0050,0x0008,0x0006,0x0000,0x00FC,0x0102,0x0102,
      0x00FC,0x0000,0x000C,0x001E,0x000C,0x0010,0x0010,0x007C,
      0x0010,0x0010,0x0100,0x00C0,0x0010,0x0010,0x0010,0x0040,
      0x0070,0x000E,0x003C,0x0042,0x0042,0x003C,0x0044,0x007E,
      0x0040,0x0044,0x0062,0x0052,0x004C,0x0024,0x0042,0x004A,
      0x0034,0x0018,0x0014,0x007E,0x0010,0x002E,0x004A,0x004A,
      0x0032,0x003C,0x004A,0x004A,0x0030,0x0006,0x0062,0x001A,
      0x0006,0x0034,0x004A,0x004A,0x0034,0x000C,0x0052,0x0052, // 180..
      0x003C,0x0048,0x0100,0x00C8,0x0010,0x0028,0x0044,0x0028,
      0x0028,0x0028,0x0028,0x0044,0x0028,0x0010,0x0004,0x0052,
      0x000C,0x0038,0x0044,0x00B2,0x00AA,0x009A,0x0024,0x0018,
      0x0040,0x0060,0x0058,0x0016,0x0058,0x0060,0x0040,0x0042,
      0x007E,0x004A,0x004A,0x0034,0x003C,0x0042,0x0042,0x0026,
      0x0042,0x007E,0x0042,0x0042,0x003C,0x0042,0x007E,0x004A,
      0x0042,0x0066,0x0042,0x007E,0x004A,0x0002,0x0006,0x0018,
      0x0024,0x0042,0x0052,0x0072,0x0042,0x007E,0x0008,0x0008,
      0x007E,0x0042,0x0042,0x007E,0x0042,0x0020,0x0042,0x003E,
      0x0002,0x0042,0x007E,0x004A,0x0014,0x0062,0x0042,0x0042, // 260..
      0x007E,0x0042,0x0040,0x0060,0x0042,0x007E,0x0044,0x0008,
      0x0044,0x007E,0x0042,0x0042,0x007E,0x0048,0x0012,0x007E,
      0x0002,0x003C,0x0042,0x0042,0x0042,0x003C,0x0042,0x007E,
      0x0052,0x0012,0x000C,0x003C,0x0042,0x00C2,0x0142,0x003C,
      0x0042,0x007E,0x0052,0x0032,0x004C,0x0064,0x004A,0x0052,
      0x0026,0x0006,0x0042,0x007E,0x0042,0x0006,0x0002,0x003E,
      0x0040,0x0040,0x003E,0x0002,0x0006,0x001A,0x0060,0x001A,
      0x0006,0x0002,0x001E,0x0062,0x001C,0x0062,0x001E,0x0002,
      0x0042,0x0066,0x0018,0x0066,0x0042,0x0006,0x0048,0x0070,
      0x0048,0x0006,0x0062,0x0052,0x004A,0x0046,0x0000,0x01FE, // 340..
      0x0102,0x000E,0x0070,0x0102,0x01FE,0x0000,0x0004,0x0002,
      0x0004,0x0100,0x0100,0x0100,0x0100,0x0100,0x000C,0x0002,
      0x0068,0x0058,0x0070,0x0002,0x007E,0x0048,0x0030,0x0030,
      0x0048,0x0048,0x0030,0x004A,0x007E,0x0030,0x0068,0x0058,
      0x0008,0x007C,0x000A,0x0030,0x0148,0x00F8,0x0002,0x007E,
      0x0008,0x0070,0x0008,0x007A,0x0108,0x00FA,0x0002,0x007E,
      0x0010,0x0068,0x0002,0x007E,0x0008,0x0078,0x0008,0x0078,
      0x0008,0x0070,0x0008,0x0078,0x0008,0x0070,0x0030,0x0048,
      0x0030,0x0108,0x01F8,0x0048,0x0030,0x0030,0x0048,0x01F8,
      0x0000,0x0078,0x0010,0x0008,0x0050,0x0058,0x0028,0x0008, // 420..
      0x003C,0x0048,0x0038,0x0040,0x0078,0x0018,0x0060,0x0018,
      0x0018,0x0060,0x0018,0x0060,0x0018,0x0048,0x0030,0x0048,
      0x0138,0x00C0,0x0038,0x0068,0x0058,0x0048,0x0010,0x00EC,
      0x0102,0x0000,0x01FE,0x0102,0x00EC,0x0010,0x0010,0x0008,
      0x0010,0x0008 }; //~TinyFnt

  private final int[] NanoFnt = { 0x4E616E6F, 0x70105, 1, 2, // 'Nano'
      100,100,101,104,107,110,113,116,118,120,122,125,128,130,133,134,
      137,140,142,145,148,151,154,157,160,163,166,167,169,172,175,178,
      181,184,187,190,193,196,199,202,205,208,211,214,217,220,223,226,
      229,232,235,238,241,244,247,250,253,256,259,262,264,267,269,272,
      276,278,281,284,287,290,293,296,299,302,303,305,308,310,313,316,
      319,322,325,328,331,334,337,340,343,346,349,352,355,356,359,362,
      0x002E,0x0006,0x0000,0x0006,0x003E,0x0014,0x003E,0x002C, // 100..
      0x007E,0x0034,0x0012,0x0008,0x0024,0x0038,0x003E,0x0028,
      0x0004,0x0002,0x003E,0x0041,0x0041,0x003E,0x000A,0x0004,
      0x000A,0x0008,0x001C,0x0008,0x0040,0x0020,0x0004,0x0004,
      0x0004,0x0020,0x0060,0x0018,0x0006,0x001C,0x0022,0x001C,
      0x0004,0x003E,0x0032,0x002A,0x0024,0x0022,0x002A,0x0014,
      0x001C,0x0010,0x003E,0x002E,0x002A,0x0012,0x001C,0x002A,
      0x0012,0x0032,0x000A,0x0006,0x0014,0x002A,0x0014,0x0004,
      0x002A,0x001C,0x0028,0x0040,0x0028,0x0008,0x0014,0x0022,
      0x0014,0x0014,0x0014,0x0022,0x0014,0x0008,0x0002,0x002A,
      0x0004,0x003C,0x0042,0x005C,0x003C,0x0012,0x003C,0x003E, // 180..
      0x002A,0x0014,0x001C,0x0022,0x0014,0x003E,0x0022,0x001C,
      0x003E,0x002A,0x0022,0x003E,0x000A,0x0002,0x001C,0x0022,
      0x0032,0x003E,0x0008,0x003E,0x0022,0x003E,0x0022,0x0020,
      0x0020,0x001E,0x003E,0x0008,0x0036,0x003E,0x0020,0x0020,
      0x003E,0x0004,0x003E,0x003E,0x0002,0x003C,0x003E,0x0022,
      0x003E,0x003E,0x000A,0x000E,0x003E,0x0062,0x007E,0x003E,
      0x000A,0x0034,0x002C,0x002A,0x001A,0x0002,0x003E,0x0002,
      0x003E,0x0020,0x003E,0x001E,0x0020,0x001E,0x003E,0x0010,
      0x003E,0x0036,0x0008,0x0036,0x0006,0x0038,0x0006,0x0032,
      0x002A,0x0026,0x007F,0x0041,0x0006,0x0018,0x0060,0x0041, // 260..
      0x007F,0x0004,0x0002,0x0004,0x0040,0x0040,0x0040,0x0040,
      0x0002,0x0004,0x0010,0x0028,0x0038,0x003E,0x0028,0x0038,
      0x0010,0x0028,0x0028,0x0038,0x0028,0x003E,0x0038,0x0028,
      0x0018,0x0008,0x003C,0x000A,0x0050,0x0068,0x0038,0x003E,
      0x0008,0x0030,0x003A,0x0040,0x003A,0x003E,0x0010,0x0028,
      0x0002,0x003E,0x0038,0x0018,0x0038,0x0038,0x0008,0x0030,
      0x0010,0x0028,0x0010,0x0078,0x0028,0x0038,0x0038,0x0028,
      0x0078,0x0038,0x0008,0x0008,0x0020,0x0038,0x0008,0x0008,
      0x003E,0x0028,0x0038,0x0020,0x0038,0x0018,0x0020,0x0018,
      0x0038,0x0030,0x0038,0x0028,0x0010,0x0028,0x0058,0x0020, // 340..
      0x0018,0x0028,0x0038,0x0028,0x0008,0x0036,0x0041,0x007F,
      0x0041,0x0036,0x0008,0x0002,0x0002,0x0002 }; //~NanoFnt

  private final int[] BoldFnt = { 0x426F6C64, 0xC0309, 1, 4, // 'Bold'
      100,100,102,107,113,118,127,133,135,138,141,147,153,155,159,161,
      165,170,175,180,185,190,195,200,206,211,216,218,220,226,232,238,
      243,252,260,267,274,281,287,293,300,306,308,313,320,326,334,340,
      348,354,362,368,374,380,386,393,403,410,418,424,427,431,434,439,
      445,447,452,458,463,469,475,480,486,491,493,495,501,503,511,516,
      522,528,534,538,543,549,554,560,568,574,580,585,589,591,595,601,
      0x01BE,0x01BE,0x000E,0x000E,0x0000,0x000E,0x000E,0x0040, // 100..
      0x01E8,0x005E,0x01E8,0x005E,0x0008,0x009C,0x01B6,0x03FF,
      0x01B6,0x00E6,0x000C,0x001E,0x0312,0x00DE,0x00FC,0x01F8,
      0x0126,0x01E0,0x00C0,0x00EC,0x01FE,0x0132,0x01FE,0x00CC,
      0x0160,0x0016,0x000E,0x01F8,0x03FC,0x0402,0x0402,0x03FC,
      0x01F8,0x006C,0x0028,0x00FE,0x00FE,0x0028,0x006C,0x0060,
      0x0060,0x01F8,0x01F8,0x0060,0x0060,0x0580,0x0380,0x0060,
      0x0060,0x0060,0x0060,0x0180,0x0180,0x0180,0x00F0,0x003C,
      0x0006,0x00FC,0x01FE,0x0102,0x01FE,0x00FC,0x0000,0x0008,
      0x01FC,0x01FE,0x0000,0x018C,0x01C6,0x0162,0x013E,0x011C,
      0x008C,0x0186,0x0132,0x01FE,0x00EC,0x0060,0x0058,0x004C, // 180..
      0x01FE,0x01FE,0x009E,0x011E,0x0112,0x01F2,0x00E2,0x00FC,
      0x01FE,0x0112,0x01F2,0x00E4,0x0002,0x0182,0x01E2,0x007A,
      0x001E,0x0006,0x00EC,0x01FE,0x0112,0x01FE,0x00EC,0x009C,
      0x013E,0x0122,0x01FE,0x00FC,0x0198,0x0198,0x0598,0x0398,
      0x0060,0x0060,0x0090,0x0090,0x0198,0x0108,0x0090,0x0090,
      0x0090,0x0090,0x0090,0x0090,0x0108,0x0198,0x0090,0x0090,
      0x0060,0x0060,0x000C,0x0002,0x0162,0x013E,0x001C,0x0078,
      0x0084,0x0172,0x014A,0x014A,0x017A,0x0142,0x0044,0x0038,
      0x0180,0x01E0,0x0078,0x005E,0x004E,0x0078,0x01E0,0x0180,
      0x01FE,0x01FE,0x0112,0x0112,0x011E,0x01FC,0x00E0,0x0078, // 260..
      0x00FC,0x0186,0x0102,0x0102,0x01CE,0x00CC,0x01FE,0x01FE,
      0x0102,0x0102,0x0186,0x00FC,0x0078,0x01FE,0x01FE,0x0112,
      0x0112,0x0112,0x0102,0x01FE,0x01FE,0x0012,0x0012,0x0012,
      0x0002,0x0078,0x00FC,0x0186,0x0102,0x0122,0x01A6,0x01EC,
      0x01FE,0x01FE,0x0010,0x0010,0x01FE,0x01FE,0x01FE,0x01FE,
      0x00C0,0x01C0,0x0100,0x01FE,0x00FE,0x01FE,0x01FE,0x0070,
      0x007C,0x00EE,0x01C6,0x0102,0x01FE,0x01FE,0x0100,0x0100,
      0x0100,0x0100,0x01FE,0x01FE,0x0038,0x01E0,0x01E0,0x0038,
      0x01FE,0x01FE,0x01FE,0x01FE,0x0038,0x00E0,0x01FE,0x01FE,
      0x0078,0x00FC,0x0186,0x0102,0x0102,0x0186,0x00FC,0x0078, // 340..
      0x01FE,0x01FE,0x0022,0x0022,0x003E,0x001C,0x0078,0x00FC,
      0x0186,0x0102,0x0142,0x0186,0x03FC,0x0278,0x01FE,0x01FE,
      0x0022,0x0022,0x01FE,0x01DC,0x009C,0x019E,0x0132,0x0132,
      0x0136,0x00E4,0x0002,0x0002,0x01FE,0x01FE,0x0002,0x0002,
      0x00FE,0x01FE,0x0100,0x0100,0x01FE,0x00FE,0x0006,0x003E,
      0x00F0,0x0180,0x00F0,0x003E,0x0006,0x0006,0x003E,0x01F0,
      0x01F0,0x001E,0x001E,0x01F0,0x01F0,0x003E,0x0006,0x0102,
      0x0186,0x00FC,0x0030,0x00FC,0x0186,0x0102,0x0002,0x000E,
      0x001C,0x01F0,0x01F0,0x001C,0x000E,0x0002,0x0182,0x01C2,
      0x0172,0x013A,0x010E,0x0106,0x07FE,0x07FE,0x0402,0x0006, // 420..
      0x003C,0x00F0,0x0180,0x0402,0x07FE,0x07FE,0x0030,0x001C,
      0x0006,0x001C,0x0030,0x0400,0x0400,0x0400,0x0400,0x0400,
      0x0400,0x001C,0x001A,0x01D0,0x01E8,0x0128,0x01F8,0x01F0,
      0x01FE,0x01FE,0x0108,0x0108,0x01F8,0x00F0,0x00F0,0x01F8,
      0x0108,0x0108,0x0098,0x00F0,0x01F8,0x0108,0x0108,0x01FE,
      0x01FE,0x00F0,0x01F8,0x0128,0x0128,0x01B8,0x00B0,0x0010,
      0x01FC,0x01FE,0x0012,0x0002,0x00F0,0x05F8,0x0508,0x0508,
      0x07F8,0x03F8,0x01FE,0x01FE,0x0008,0x01F8,0x01F0,0x01FA,
      0x01FA,0x07FA,0x07FA,0x01FE,0x01FE,0x0060,0x01F8,0x0198,
      0x0100,0x01FE,0x01FE,0x01F8,0x01F8,0x0008,0x01F8,0x01F8, // 500..
      0x0008,0x01F8,0x01F0,0x01F8,0x01F8,0x0008,0x01F8,0x01F8,
      0x00F0,0x01F8,0x0108,0x0108,0x01F8,0x00F0,0x07F8,0x07F8,
      0x0108,0x0108,0x01F8,0x00F0,0x00F0,0x01F8,0x0108,0x0108,
      0x07F8,0x07F8,0x01F8,0x01F8,0x0008,0x0018,0x00B0,0x01B8,
      0x0168,0x01D8,0x00D0,0x0008,0x0008,0x00FE,0x01FE,0x0108,
      0x0008,0x00F8,0x01F8,0x0100,0x01F8,0x01F8,0x0018,0x0078,
      0x01C0,0x01C0,0x0078,0x0018,0x0018,0x00F8,0x01C0,0x0078,
      0x0078,0x01C0,0x00F8,0x0018,0x0108,0x0198,0x00F0,0x00F0,
      0x0198,0x0108,0x0018,0x0478,0x07C0,0x01C0,0x0078,0x0018,
      0x0188,0x01C8,0x0168,0x0138,0x0118,0x0060,0x03FC,0x079E, // 600..
      0x0402,0x01FE,0x01FE,0x0402,0x079E,0x03FC,0x0060,0x0040,
      0x0020,0x0020,0x0060,0x0040,0x0020 }; //~BoldFnt

  private final int[] HeadFnt = { 0x48656164, 0x12040E, 2, 5, // 'Head'
      100,100,103,109,119,128,143,154,157,161,166,172,182,185,190,193,
      200,209,215,224,233,242,251,260,269,278,287,290,293,303,313,323,
      332,346,358,369,380,391,401,410,422,432,435,443,454,463,476,487,
      499,509,521,532,543,553,563,574,590,601,612,622,626,633,637,645,
      655,658,667,676,685,694,703,709,718,727,730,733,742,745,759,768,
      778,787,796,802,811,816,825,834,848,857,866,874,880,883,889,900,
      0x030FE,0x033FE,0x030FE,0x0003E,0x0003E, // 100..
      0x00000,0x0003E,0x0003E,0x00006,0x00600,
      0x03E30,0x03F30,0x007F0,0x03E3E,0x03FB0,
      0x00FF0,0x0067E,0x00636,0x00030,0x00C78,
      0x03CFC,0x038FE,0x030C6,0x0FFFF,0x03186,
      0x03F9C,0x01F1C,0x00E10,0x00038,0x0007C,
      0x000C6,0x000C6,0x030C6,0x00C7C,0x00338,
      0x000C0,0x01E60,0x01E18,0x03306,0x03300,
      0x03300,0x01E00,0x01E00,0x00E00,0x01F80,
      0x03FBC,0x031FE,0x031E6,0x033E6,0x03F7E,
      0x01E3C,0x03F80,0x03780,0x02180,0x00026, // 150..
      0x0003E,0x0001E,0x01FC0,0x07FF0,0x1FFFC,
      0x3C01E,0x20002,0x3C01E,0x1FFFC,0x07FF0,
      0x01FC0,0x0000C,0x0006C,0x0003E,0x0003E,
      0x0007C,0x0002C,0x00300,0x00300,0x00300,
      0x00300,0x03FF0,0x03FF0,0x03FF0,0x00300,
      0x00300,0x00300,0x13800,0x0F800,0x07800,
      0x00300,0x00300,0x00300,0x00300,0x00300,
      0x03800,0x03800,0x03800,0x02000,0x03C00,
      0x01F00,0x003E0,0x000FC,0x0001E,0x00006,
      0x007F0,0x01FFC,0x03FFE,0x0380E,0x03006, // 200..
      0x0380E,0x03FFE,0x01FFC,0x007F0,0x00018,
      0x00018,0x00018,0x03FFC,0x03FFE,0x03FFE,
      0x03030,0x03C3C,0x03E3E,0x03F06,0x03706,
      0x033C6,0x031FE,0x030FC,0x03078,0x00C10,
      0x01C1C,0x0380E,0x030C6,0x030C6,0x039FE,
      0x03FFC,0x01FB8,0x00F00,0x00700,0x007C0,
      0x007E0,0x00678,0x0061E,0x03FFE,0x03FFE,
      0x03FFE,0x00600,0x00800,0x019FE,0x039FE,
      0x030C6,0x030C6,0x039C6,0x03FC6,0x01F86,
      0x00F00,0x007F0,0x01FFC,0x03FBC,0x038C6,
      0x030C6,0x030C6,0x03FCE,0x01F8C,0x00F00,
      0x00006,0x00006,0x03C06,0x03F06,0x01FC6,
      0x001F6,0x0007E,0x0001E,0x0000E,0x00F38,
      0x01FFC,0x03FFE,0x030C6,0x030C6,0x030FE,
      0x03FFC,0x01FB8,0x00F00,0x008F8,0x019FC,
      0x03BFE,0x0338E,0x03306,0x03B8E,0x01FFE,
      0x01FFC,0x007F0,0x03060,0x03060,0x03060,
      0x13060,0x1F060,0x0F060,0x00700,0x00700,
      0x00780,0x00F80,0x00CC0,0x00CC0,0x018C0,
      0x01860,0x01860,0x03030,0x00CC0,0x00CC0, // 300..
      0x00CC0,0x00CC0,0x00CC0,0x00CC0,0x00CC0,
      0x00CC0,0x00CC0,0x00CC0,0x03030,0x01860,
      0x01860,0x018E0,0x00CC0,0x00CC0,0x00D80,
      0x00780,0x00780,0x00700,0x00018,0x0001C,
      0x0001E,0x03306,0x03386,0x033C6,0x000FE,
      0x0007C,0x00038,0x003E0,0x00E30,0x01818,
      0x0180C,0x033CC,0x037E6,0x03676,0x03636,
      0x03336,0x037F6,0x01634,0x01E0C,0x01318,
      0x001F0,0x02000,0x03C00,0x03F80,0x01FF0,
      0x007FE,0x0067E,0x0061E,0x006FE,0x007F8,
      0x03FC0,0x03E00,0x03000,0x03FFE,0x03FFE,
      0x03FFE,0x030C6,0x030C6,0x030C6,0x030C6,
      0x038FE,0x03FFC,0x01FB8,0x00F00,0x003E0,
      0x00FF8,0x01FFC,0x03C1E,0x0380E,0x03006,
      0x03006,0x03006,0x0180E,0x01C1C,0x00C18,
      0x03FFE,0x03FFE,0x03FFE,0x03006,0x03006,
      0x03006,0x03006,0x03C0E,0x01FFC,0x00FF8,
      0x003F0,0x03FFE,0x03FFE,0x03FFE,0x030C6,
      0x030C6,0x030C6,0x030C6,0x030C6,0x030C6,
      0x03006,0x03FFE,0x03FFE,0x03FFE,0x000C6, // 400..
      0x000C6,0x000C6,0x000C6,0x000C6,0x00006,
      0x003E0,0x00FF8,0x01FFC,0x03C1C,0x0380E,
      0x03006,0x03006,0x03186,0x0198E,0x01F9C,
      0x03F9C,0x03F90,0x03FFE,0x03FFE,0x03FFE,
      0x000C0,0x000C0,0x000C0,0x000C0,0x03FFE,
      0x03FFE,0x03FFE,0x03FFE,0x03FFE,0x03FFE,
      0x00E00,0x01E00,0x03E00,0x03000,0x03000,
      0x03FFE,0x01FFE,0x00FFE,0x03FFE,0x03FFE,
      0x03FFE,0x003C0,0x001E0,0x007F8,0x00FFC,
      0x01F3E,0x03C1E,0x03806,0x03002,0x03FFE,
      0x03FFE,0x03FFE,0x03000,0x03000,0x03000,
      0x03000,0x03000,0x03000,0x03FFE,0x03FFE,
      0x03FFE,0x000FE,0x00FF8,0x03F80,0x03C00,
      0x03FE0,0x003FE,0x0001E,0x03FFE,0x03FFE,
      0x03FFE,0x03FFE,0x03FFE,0x03FFE,0x0003C,
      0x000F0,0x003C0,0x00780,0x01E00,0x03FFE,
      0x03FFE,0x03FFE,0x003E0,0x00FF8,0x01FFC,
      0x03C1E,0x03006,0x03006,0x03006,0x03006,
      0x03C1E,0x01FFC,0x00FF8,0x003E0,0x03FFE,
      0x03FFE,0x03FFE,0x00186,0x00186,0x00186, // 500..
      0x00186,0x001FE,0x000FC,0x00078,0x003E0,
      0x00FF8,0x01FFC,0x01C1E,0x03806,0x03006,
      0x03406,0x03E06,0x03C1E,0x01FFC,0x03FF8,
      0x067E0,0x03FFE,0x03FFE,0x03FFE,0x00186,
      0x00186,0x00186,0x00186,0x03F86,0x03FFE,
      0x03E7C,0x02038,0x00C38,0x01CFC,0x01CFE,
      0x030C6,0x031C6,0x031C6,0x031C6,0x0338E,
      0x01F9C,0x01F98,0x00600,0x00006,0x00006,
      0x00006,0x00006,0x03FFE,0x03FFE,0x03FFE,
      0x00006,0x00006,0x00006,0x00FFE,0x01FFE,
      0x03FFE,0x03800,0x03000,0x03000,0x03800,
      0x03FFE,0x01FFE,0x00FFE,0x00002,0x0001E,
      0x000FE,0x007F8,0x03F80,0x03C00,0x03F00,
      0x00FE0,0x001FE,0x0003E,0x0000E,0x00006,
      0x0003E,0x003FE,0x01FF0,0x03F00,0x03F00,
      0x01FF8,0x000FE,0x0001E,0x003FE,0x03FE0,
      0x03C00,0x03FC0,0x00FFC,0x000FE,0x0000E,
      0x02000,0x03006,0x03C1E,0x01F3E,0x00FF8,
      0x003E0,0x007F0,0x01F7C,0x03E1E,0x0380E,
      0x03002,0x00002,0x0000E,0x0003E,0x0007C, // 600..
      0x03FF0,0x03FC0,0x03FE0,0x001F8,0x0007E,
      0x0001E,0x00006,0x03806,0x03C06,0x03E06,
      0x03786,0x033C6,0x031E6,0x030F6,0x0303E,
      0x0301E,0x0300E,0x1FFFE,0x1FFFE,0x1FFFE,
      0x18006,0x00002,0x0001E,0x000FC,0x003E0,
      0x01F80,0x03C00,0x03000,0x18006,0x1FFFE,
      0x1FFFE,0x1FFFE,0x00100,0x001E0,0x000F8,
      0x0003E,0x0001E,0x0007C,0x001F0,0x00180,
      0x08000,0x08000,0x08000,0x08000,0x08000,
      0x08000,0x08000,0x08000,0x08000,0x08000,
      0x0003C,0x00036,0x00032,0x01E00,0x03F60,
      0x03F70,0x03330,0x03330,0x01FF0,0x03FF0,
      0x03FE0,0x02000,0x03FFE,0x03FFE,0x03FFE,
      0x01860,0x03030,0x03870,0x03FF0,0x01FE0,
      0x00FC0,0x00FC0,0x01FE0,0x03FF0,0x03870,
      0x03030,0x03030,0x038F0,0x018E0,0x008C0,
      0x00FC0,0x01FE0,0x03FF0,0x03870,0x03030,
      0x01860,0x03FFE,0x03FFE,0x03FFE,0x00F80,
      0x01FE0,0x03FF0,0x03B30,0x03330,0x03330,
      0x03BF0,0x01BE0,0x00BC0,0x00030,0x03FFC, // 700..
      0x03FFE,0x03FFE,0x00036,0x00030,0x08FC0,
      0x19FE0,0x3BFF0,0x33870,0x33030,0x39860,
      0x3FFF0,0x1FFF0,0x0FFF0,0x03FFE,0x03FFE,
      0x03FFE,0x00060,0x00030,0x00030,0x03FF0,
      0x03FF0,0x03FE0,0x03FF6,0x03FF6,0x03FF6,
      0x3FFF6,0x3FFF6,0x1FFF6,0x03FFE,0x03FFE,
      0x03FFE,0x00380,0x00FC0,0x01FE0,0x03E70,
      0x03810,0x02000,0x03FFE,0x03FFE,0x03FFE,
      0x03FF0,0x03FF0,0x03FF0,0x00020,0x00030,
      0x00030,0x03FF0,0x03FE0,0x03FE0,0x00030,
      0x00030,0x03FF0,0x03FF0,0x03FE0,0x03FF0,
      0x03FF0,0x03FF0,0x00060,0x00030,0x00030,
      0x03FF0,0x03FF0,0x03FE0,0x00780,0x01FE0,
      0x01FE0,0x03870,0x03030,0x03030,0x03870,
      0x01FE0,0x01FE0,0x00780,0x3FFF0,0x3FFF0,
      0x3FFF0,0x01860,0x03030,0x03870,0x03FF0,
      0x01FE0,0x00FC0,0x00FC0,0x01FE0,0x03FF0,
      0x03870,0x03030,0x01860,0x3FFF0,0x3FFF0,
      0x3FFF0,0x03FF0,0x03FF0,0x03FF0,0x00060,
      0x00070,0x00070,0x018C0,0x019E0,0x03BF0, // 800..
      0x03330,0x03330,0x03730,0x03F70,0x01E60,
      0x00C00,0x00030,0x01FFC,0x03FFC,0x03FFC,
      0x03030,0x01FF0,0x03FF0,0x03FF0,0x03000,
      0x03000,0x01800,0x03FF0,0x03FF0,0x03FF0,
      0x00010,0x000F0,0x007F0,0x03F80,0x03C00,
      0x03E00,0x01FE0,0x003F0,0x00070,0x00030,
      0x001F0,0x01FF0,0x03F00,0x03C00,0x03FF0,
      0x003F0,0x007F0,0x03FC0,0x03800,0x03F80,
      0x007F0,0x000F0,0x00010,0x02010,0x03030,
      0x03CF0,0x01FE0,0x00780,0x01FC0,0x03EF0,
      0x03870,0x03010,0x00010,0x300F0,0x307F0,
      0x3BF80,0x1FC00,0x0FE00,0x01FE0,0x003F0,
      0x00070,0x03070,0x03870,0x03C70,0x03F70,
      0x033F0,0x031F0,0x030F0,0x03070,0x00200,
      0x1FFF8,0x1FDFC,0x3F8FE,0x30006,0x20002,
      0x03FFE,0x03FFE,0x03FFE,0x30006,0x3F8FE,
      0x1FDFE,0x1FFFC,0x00700,0x00200,0x00600,
      0x00700,0x00180,0x00180,0x00380,0x00300,
      0x00700,0x00600,0x00600,0x00380,0x00080 }; //~HeadFnt

  private final int[] ScorFnt = { 0x53636F72, 0x12040E, 2, 5, // 'Scor'
      100,100,100,100,100,100,100,100,100,100,100,100,100,100,100,100,
      100,109,115,124,133,142,151,160,169,178,187,187,187,187,187,187,
      187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,
      187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,
      187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,
      187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,187,
      0x007F0,0x01FFC,0x03FFE,0x0380E,0x03006, // 100..
      0x0380E,0x03FFE,0x01FFC,0x007F0,0x00018,
      0x00018,0x00018,0x03FFC,0x03FFE,0x03FFE,
      0x03030,0x03C3C,0x03E3E,0x03F06,0x03706,
      0x033C6,0x031FE,0x030FC,0x03078,0x00C10,
      0x01C1C,0x0380E,0x030C6,0x030C6,0x039FE,
      0x03FFC,0x01FB8,0x00F00,0x00700,0x007C0,
      0x007E0,0x00678,0x0061E,0x03FFE,0x03FFE,
      0x03FFE,0x00600,0x00800,0x019FE,0x039FE,
      0x030C6,0x030C6,0x039C6,0x03FC6,0x01F86,
      0x00F00,0x007F0,0x01FFC,0x03FBC,0x038C6,
      0x030C6,0x030C6,0x03FCE,0x01F8C,0x00F00,
      0x00006,0x00006,0x03C06,0x03F06,0x01FC6,
      0x001F6,0x0007E,0x0001E,0x0000E,0x00F38,
      0x01FFC,0x03FFE,0x030C6,0x030C6,0x030FE,
      0x03FFC,0x01FB8,0x00F00,0x008F8,0x019FC,
      0x03BFE,0x0338E,0x03306,0x03B8E,0x01FFE,
      0x01FFC,0x007F0,0 }; //~ScorFnt

// [Shared] GameEngine working data (public state)..
public int MaxLogWgts = Max_Log, MaxLogAnim = Fust_Log>>1,
    FustFocLog = Fust_Log, FustKeyLog = Fust_Log;
public int[] thePixels = null;

public static GameWgt[] AllWgts = null; // init'd in game const'r

// [Shared] GameEngine working data (static).. // (used by static method)
public static int NextRefNo = 0, CallersID = 0,
    MaxWinSize = 0, WinHi = Win_Hi, WinWi = Win_Wi;

public static Timer TickTock = null;
public static JavaGame theWindow = null;

  public final static char Arro_LEFT = (char)28, // (internal, Java has none)
    Arro_RIGHT = (char)29, Arro_UP = (char)30, Arro_DOWN = (char)31,
    Esc_Key = (char)27;

  public int EdgeOffs = 0, SeenLog = 0; // bits log startup milestones (Jv)

// These names are used in converting to JavaScript..
  private final static String JSwgtMeths = " GetRefNum GetAnimBits "
        + " GetInfo GetColor GetParent GetPosn GetSize GetSelim "
        + " SetInfo SetColor SetParent SetPosn SetSize SetSelim "
        + " GetIxData PutIxData GetTextLn PutTextLn AddPrntsPos "
        + " IsVisible GetType SetFlags HideSho toString ",
      OrideGameMeths = " GameList StartUp toString Collided GotEnterKey "
        + " GotEscKey GotTabKey GotArrowKey KeyFld FocusEv BlurEv DoOften "
        + " ClickEvt MouseDn MouseUp StartDrag Dragging CancelDrag "
        + "Dropped . \nfunction o_Collided(whom,hitt)  {return true;}\n"
        + "function o_GotEnterKey(whom)  {return GotEnterKey(whom);}\n"
        + "function o_GotEscKey(whom)  {return GotEscKey(whom);}\n"
        + "function o_GotTabKey(whom)  {}\nfunction o_Dropped(whom)  {}\n"
        + "function o_GotArrowKey(whom,info)  {GotArrowKey(whom,info);}\n"
        + "function o_KeyFld(whom,info)  {KeyFld(whom,info);}\n"
        + "function o_FocusEv(whom,vert,horz)  {FocusEv(whom,vert,horz);}\n"
        + "function o_BlurEv(whom)  {BlurEv(whom);}\n"
        + "function o_DoOften()  {}\nfunction o_ClickEvt(whom,vert,horz)"
        + " {return ClickEvt(whom,vert,horz);}\n"
        + "function o_MouseDn(whom,vert,horz)  {MouseDn(whom,vert,horz);}\n"
        + "function o_MouseUp(whom,vert,horz)  {MouseUp(whom,vert,horz);}\n"
        + "function o_Dragging(whom,vert,horz)  {Dragging(whom,vert,horz);}\n"
        + "function o_StartDrag(whom,vert,horz)  {return 0;}\n"
        + "function o_CancelDrag(GameWgt whom)  {CancelDrag(whom);}\n",
        // + "function o_GameEnding()  {}\n"
      CantDoMeths = " GameEnding GetTextWide SetBlinker DrawMe "
        + " IsValidIDtxt GetIDtxt SwapData FixHaff HexyMore "
        + " FmtInfo FmtFront LogMe SetFocus CheckIt RadioBtn "
        + " FindElemt DeleElemt XchElemt AddElemt toStrMash ",
      JavaTypes = " public private static void int String float double "
        + " char JavaGame GameEvent GameWgt GameWgt[] String[] int[] ";

  // Handy (public utility) methods..

/**
  * A safe (no exceptions) non-OOPS way to extract a character from a string.
  *
  * @param  here  The (0-base) character position
  * @param  aStr  The string from which to extract that character
  *
  * @return       The character extracted, or '\0' if out of bounds
  */
  public static char CharAt(int here, String aStr)  { // Java version..
    if (here<0) return '\0';
    if (aStr == null) return '\0';
    try {
      if (here >= aStr.length()) return '\0';
      return aStr.charAt(here);}
    catch (Exception ex)  {return '\0';}} //~CharAt

 /**
  * A safe (no exceptions) non-OOPS way to get the length of a string.
  *
  * @param  aStr  The string to get the length of
  *
  * @return       Its length, or 0 if null
  */
  public static int StrLength(String aStr)  { // Java version..
    if (aStr == null) return 0;
    return aStr.length();} //~StrLength

 /**
  * Strongly-typed (not overloaded) string selector.
  * This is part of a collection of tools to build debug print lines.
  *
  * @param  whom  The selector
  * @param  tru   Returned if whom=true
  * @param  fls   Returned if whom=false
  *
  * @return       The selected string, tru if whom=true, else fls
  */
  public static String IffyStr(boolean whom, String tru, String fls)  {
    if (whom) return tru;
    return fls;} //~IffyStr

 /**
  * A safe (no exceptions) way to get the rest of a string.
  * Text past the end is considered empty.
  *
  * @param  here  The (0-base) position to start
  * @param  aStr  The string from which to extract that substring
  *
  * @return       The rest of the string, after here
  */
  public static String RestOf(int here, String aStr)  {
    int lxx;
    try {
    if (here<0) return "";
    lxx = StrLength(aStr);
    if (lxx==0) return "";
    if (here >= lxx) return "";
    return aStr.substring(here);
    } catch (Exception ex)  {}
    return "";  } //~RestOf

  /**
  * A safe (no exceptions) way to extract a substring.
  * Text past the end is empty.
  *
  * @param  here  The (0-base) position to start
  * @param  lxx   The desired length (the result could be less)
  * @param  aStr  The string from which to extract that substring
  *
  * @return       The substring
  */
  public static String Substring(int here, int lxx, String aStr)  {
    int nx = StrLength(aStr);
    try {
    if (here<0) return "";
    if (lxx <= 0) return "";
    if (nx==0) return "";
    if (here >= nx) return "";
    lxx = here+lxx;
    if (lxx >= nx) return RestOf(here,aStr);
    return aStr.substring(here,lxx);
    } catch (Exception ex)  {}
    return "";  } //~Substring

  /**
  * Formats an integer as hexadecimal and returns it with a prefix and suffix.
  * This is useful when printing numbers that you want to see the bits of.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The combined string
  */
  public static String SeeHex(String before, int whom, String after)  {
    int shft = 28, lxx = StrLength(before); // special-case URL encoding..
    if (lxx==1) if ((whom&-256)==0) if (StrLength(after)==0)
        if ((int)(CharAt(0,before)) == 37)  { // '%'
      if (whom>15)  {
        before = before + SeeHex("",whom>>4,"");
        whom = whom&15;} //~if
      else before = before + "0";} //~if
    if (whom<0)  {
      before = SeeHex(before,(whom>>28)&15,"");
      if (whom+0x00100000 >= 0)  { // remove leading 'F's..
        shft = 16;
        while ((1<<shft)+whom >= 0)  {
          shft = shft-4;
          if (shft==0) break;} //~while
        while (shft>0)  { // skip over (internal) leading zeros..
          if (((15<<shft)&whom) !=0) break;
          shft = shft-4;} //~while
        whom = whom&((16<<shft)-1);
        before = before + "-F";} //~if
      else whom = whom&0x0FFFFFFF;} //~if // (whom<0)
    if (whom>15) before = SeeHex(before,whom>>4,"");
    return before + Substring(whom&15,1,"0123456789ABCDEF")
        + after;} //~SeeHex

  /**
  * Extracts and sign-extends the second of a pair of half-integers.
  *
  * @param  here  The pair
  *
  * @return       The low half, sign-extended to a full int
  */
  public static int SignExtend(int here)  { // low 16 bits signed
    return (here&0x7FFF)-(here&0x8000);} //~SignExtend

  /**
  * Formats a 32-bit integer as two 16-bit values and returns it with a prefix
  * and suffix. This is useful when printing numbers that pack two smaller
  * integers (like board coordinates) into a single number.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The combined string
  */
  public static String LogVH(String before, int whom, String after)  {
    return before + (whom>>16) + "/" + SignExtend(whom) + after;} //~LogVH

/**
  * A safe (non-OOPS, no exceptions) way to replace every instance of
  * a particular string within a larger string.
  * An empty string cannot be replaced.
  *
  * @param  nuly    The replacement string
  * @param  prio    The string to replace
  * @param  theText The string in which to make the substitutions
  *
  * @return         theText as changed
  */
  public static String ReplacAll(String nuly, String prio,
      String theText)  {
    int lxx = StrLength(theText), nby = StrLength(prio),
        more = StrLength(nuly), whar, thar = 0;
    if (nby>0) if (lxx >= nby) try {  while (true)  {
      whar = theText.indexOf(prio,thar); // JS=Java
      if (whar<0) break;
      theText = Substring(0,whar,theText) + nuly + RestOf(whar+nby,theText);
      thar = whar+more;} //~while
    } catch (Exception ex)  {}
    return theText;} //~ReplacAll

  /**
  * A safe (no exceptions) way to find the (0-based) offset of a
  * particular string within a larger string, while ignoring zero
  * or more initial (non-overlapping) occurrences. An empty string
  * cannot be found.
  *
  * @param  whom  The number of occurrences to skip over
  * @param  aWord The string to look for
  * @param  aStr  The string in which to find it
  *
  * @return       The (indexOf) offset, or -1 if not found
  */
  public static int NthOffset(int whom, String aWord, String aStr)  {
    // skip over whom instances, 0-based offset; rtn = -2: can't find ""
    int lxx = StrLength(aWord), thar = 0, whar = -2;
    try {  if (lxx>0) if (whom >= 0) while (true)  {
        whar = aStr.indexOf(aWord,thar);
        if (whar<0) break;
        whom--;
        if (whom<0) break;
        thar = whar+lxx;} //~while
    } catch (Exception ex)  {}
    return whar;} //~NthOffset // T2 in FakeJ

/**
  * Adds two pairs of half-integers in a single int.
  *
  * @param  here  The first pair
  * @param  thar  The second pair
  *
  * @return       The pair of sums
  */
  public static int AddPair(int here, int thar)  {
    return (here&-0x10000)+(thar&-0x10000)
        + ((here+thar)&0xFFFF);} //~AddPair

  /**
  * Returns in a single int the negatives of a pair of half-integers.
  *
  * @param  here  The pair
  *
  * @return       Its negative
  */
  public static int PairNeg(int here)  {
    return ((-here)&0xFFFF)-(here&-0x10000);} //~PairNeg

/**
  * A safe (no exceptions) way to get the size of a (possibly null) array.
  *
  * @param  data  The array
  *
  * @return       Its length, or =0 if null
  */
  public static int SafeArySz(int[] data)  {
    if (data == null) return 0;
    return data.length;} //~SafeArySz

 /**
  * A safe (no exceptions) way to get a (possibly null) array element.
  * Returns a specified default value if the array is null or the given
  * index is out of bounds.
  *
  * @param  ix      The index into the array
  * @param  data    The array
  * @param  defalt  A default value to return if can't access
  *
  * @return       Its length, or =0 if null
  */
  public static int SafeAryElt(int ix, int[] data, int defalt)  {
    while (true)  {
      if (data == null) break;
      if (ix<0) break;
      if (ix >= data.length) break;
      defalt = data[ix];
      break;} //~while
    return defalt;} //~SafeAryElt

  /**
  * Formats an integer as hexadecimal if larger than 32K and returns it
  * with a prefix and suffix, for easier reading.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The result string
  */
  public static String HexIfMore(String before, int whom, String after)  {
    if (SignExtend(whom)==whom) return before + whom + after;
    return SeeHex(before + "0x",whom,after); } //~HexIfMore

  /**
  * Formats a 16-bit integer as 8.8 fixed-point and returns it with a prefix.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  *
  * @return       The result string
  */
  public static String FormFixt(String before, int whom)  {
    int frac = whom&255, zx = frac*10+6;   // assumes 8-bit fraction
    String info = "";                     // ..and signed 8+bit integer part
    boolean shorten = ((whom>25599)||(StrLength(before)==1));
    if (false) if ((whom&-0x8000) !=0) if (whom+0x10000 != (whom&0xFFFF))
      whom = SignExtend(whom); // caller did
    if (whom<0) return FormFixt(before + "-",-whom);
    before = before + (whom>>8);
    if (frac>0)  {
      before = before + ".";
      if (frac==128) info = "5";
      else if (frac==64) info = "25";
      else if (frac==192) info = "75";
      else if (frac==1)  {
        if (shorten) info = "01"; else
          info = "004";} //~if
      else if ((zx&255)<13) info = info + (zx>>8); // rounds to 1 digit
      else if (shorten)  { // round to 2 fractional digits..
        frac = (frac*100+127)>>8;
        if (frac<10) info = "0";
        else if (frac>99) frac = 99;
        info = info + frac;} //~if
      else if ((frac&31) !=0)  { // round to 3 fractional digits..
        frac = (frac*1000+127)>>8;
        if (frac<10) info = "00";
        else if (frac<100) info = "0";
        else if (frac>999) frac = 999;
        info = info + frac;} //~if
      else info = info + ((frac>>5)*125);} //~if
    return before + info;} //~FormFixt

  // Now some private data...

private String FindLog = "", myTmpStr = "";

// [BlueJ only] GameWgt subclasses here..

/**
 * Single color icon.
 *
 * Ginfo is used for animation info, bit fields, high to low:
 *     (frame rate)*16:6, (+ frames)-1:6, (step size):6, (pixel offset):14
 *   The frame rate is fixed-point, 2i.4f with a range [0 .. 3.935] where
 *     0.0 never changes the image, and 2.0 skips every other image.
 *   The number of frames determines how many frames show before the image
 *     repeats (starts over).
 *   The step size is the number of ints to advance when changing the image.
 *   The pixel offset is an offset into the image array to start the image.
 *
 *   Each frame (drawing cycle = 100ms) the fractional frame rate is added
 *     to the current value in the upper half of Gtype, then the integer part
 *     of the sum is subtracted off and multiplied by the step size and the
 *     pixel offset is reduced by that many units. If it goes negative, it
 *     is restarted at an offset = ((+ frames)-1)*(step size).
 *   It is not necessary for the step size to equal the image size. For
 *     example, the marching ants icon image used for rectangles is up to
 *     256 pixels wide, but the step size =1, so that the image shifts over
 *     one pixel each frame; since the pattern is diagonal bars, but only
 *     a single pixel (wide or tall) the pattern appears to be moving down
 *     or right one pixel per frame.
 */
private class GameIcon extends GameWgt  { // Single color icon..

 /**
  * Return the class name and RefNum for this widget, plus its ID code if any.
  *   See toStrMash for formatting.
  *
  * @return     The reference String for this widget
  */
  public String toString()  {                                     // (GameIcon)
    return toStrMash(super.toString(),"(GameIcon #");} //~toString

 /**
  * This is an internal utility for formatting a widget's information value,
  * GameIcon overrides this to append the formatted animation parts if !=0.
  *
  * @param  before  a prefix to add to the front
  * @param  tye     the widget's type
  *
  * @return         the composite display value
  */
  public String FmtInfo(String before, int tye)  { //         // (GameIcon)
    before = super.FmtInfo(before,tye);
    if ((Ginfo&-0x4000) !=0) before = before + "+" + ((Ginfo>>14)&63)
        + "#" + ((Ginfo>>20)&63) + FixHaff("/",(Ginfo>>22)&0x3F0);
    return before;} //~FmtInfo

 /**
  * This draws the GameIcon pixels on the window after advancing the offset
  * according to the animation specification. The boundaries of a masking
  * rectangle limit the drawing to (usually) the containing widget's dimensions.
  *
  * Ginfo is used for animation info, bit fields, high to low:
  *     (frame rate)*16:6, (+ frames)-1:6, (step size):6, (pixel offset):14
  *   The frame rate is fixed-point, 2i.4f with a range [0 .. 3.935] where
  *     0.0 never changes the image, and 2.0 skips every other image.
  *   The number of frames determines how many frames show before the image
  *     repeats (starts over).
  *   The step size is the number of ints to advance when changing the image.
  *   The pixel offset is an offset into the image array to start the image.
  *
  *   Each frame (drawing cycle = 100ms) the fractional frame rate is added
  *     to the current value in the upper half of Gtype, then the integer part
  *     of the sum is subtracted off and multiplied by the step size and the
  *     pixel offset is reduced by that many units. If it goes negative, it
  *     is restarted at an offset = ((+ frames)-1)*(step size).
  *
  * @param  top    do not draw above this position
  * @param  left   do not draw to the left of this position
  * @param  tall   do not draw below this position +top
  * @param  wide   do not draw to the right of this position +left
  */
  public void DrawMe(int top, int left, int tall, int wide)  { // (GameIco)
    int RealPrent = WhoCalled(0), // used only to verify hierarchy
        nx = AddPrntsPos(GetPosn()), myTop = nx>>16, myLft = SignExtend(nx),
        rx = myTop+Gtall, cx = myLft+Gwide, far = SafeArySz(Gdata),
        lxx = Gtall-1, info = Ginfo>>14, here, thar, logy = 0, why = 0;
    String msg = "";
    // CallersID = 0;
    while (true)  {
      if (Logging)  { // tye=2: "wIcon"
        if (logy>0)  {}
        else if (MaxLogWgts>0)  {
          logy = MaxLogWgts;
          MaxLogWgts--;} //~if
        else if (info !=0)  {
          nx = ((info&0x3F000)<<4)+Gtype; // factional counter
          if (MaxLogAnim>8) logy = MaxLogAnim;
          else if ((nx&-0x100000) !=0) logy = MaxLogAnim;
          if (logy>0) MaxLogAnim--;}} //~if
      if ((SeenLog&4)==0) if ((Gtype&1)==0) logy = 1234; // tye=2
      if (logy>0) msg = FmtFront("  (Dr.",logy>999,IffyStr(logy>999," "
          + top + "/" + left + " +" + tall + "/" + wide,"") + " + ");
      why++; // why = 1
      if (Gtop>0x3FFF) break; // hidden
      why++; // why = 2
      if (Parent==0) break;
      if (RealPrent !=0) if (RealPrent != Parent) break;
      why++; // why = 3
      if ((Gtype&254) !=2) break; // we also draw font chars +3..
      why++; // why = 4
      why++; // why = 5
      if (Gcolor<0) break; // single-layer hit rgn for sprite (undrawn)
      why++; // why = 6
      if (Gdata == null) break;
      why++; // why = 7
      if (lxx<0) break;
      if (lxx>31) lxx = 31;
      if (logy==1234) SeenLog = SeenLog|4;
      if (rx <= top+tall) tall = rx-1; // now it's edge, not size..
        else tall = top+tall-1;
      if (cx <= left+wide) wide = cx-1; // cx is win-rel right edge
        else wide = left+wide-1;
      cx = Ginfo&0x3FFF;
      if (info !=0)  { // animated..
        nx = ((info&0x3F000)<<4)+Gtype; // add fractional counter
        rx = info&63;
        while ((nx&-0x100000) !=0)  { // got integer part, go next image..
          if (cx<rx) cx = ((info>>6)&63)*rx; // restart at right end
            else cx = cx-rx; // otherwise step down toward front
          nx = nx-0x100000;} //~while // shouldn't, but could skip an image
        Ginfo = (info<<14)+cx;
        Gtype = nx;} //~if
      if (logy>0) msg = msg + cx;
      why++; // why = 8
      if ((Gtype&255)==3)  { // somebody else set this up..
        if (Gwide <= 0) break;
        why++; // why = 9
        if (cx+Gwide>far) break;} //~if
      if (myTop>top) top = myTop; // myTop/myLft are Pixels(=win)-relative
      if (myLft>left) left = myLft; // ..for drawing this icon/char
      if (top<0) top = 0;
      if (left<0) left = 0;
      if (logy>999)
        msg = msg + " => " + top + "/" + left + " +" + tall + "/" + wide;
      for (nx=0; nx <= Gwide-1; nx++) {
        info = SafeAryElt(nx+cx,Gdata,0);
        if (info==0) continue;
        here = nx+myLft;
        if (here<left) continue;
        if (here>wide) continue;
        for (rx=0; rx <= lxx; rx++) {
          thar = rx+myTop;
          if ((info&1) !=0)  { if (false)
            if (Logging) if (logy>999) System.out.println("   +++" + rx + "/" + nx
                + IffyStr((thar >= top) && (thar <= tall)," +V"," -v")
                + IffyStr((here<left) || (here>wide)," -h"," +H"));
            if (thar >= top) if (thar <= tall)
              PokePixel(Gcolor,thar,here);} //~if
          info = info>>1;
          if (info==0) break;}} //~for
      break;} //~while
    if (false) if (logy <= 0) if (why==2) logy = 1001;
    if (Logging||(logy==1234)) if (logy>0)  {
      System.out.println(msg + " = " + why         // (Dr.wFont + / (Dr.wIcon +
        + IffyStr(why==2," !prnt=" + Parent + "=" + RealPrent + " (RP)",""));
      if (logy>999) LogMe("  (DrMe) ");}} //~DrawMe

 /**
  * GameIcon constructor.
  */
  public GameIcon()  {
    if (Logging) System.out.println("(new GmIcon #" + RefNum + ")");
    Gtype = 2;}} //~GameIcon

/**
 * Font specification, mostly using GameIcon's drawing code.
 * Each character in the font is set up by CharSetup to be drawn as if
 * it were an icon.
 */
private class GameFont extends GameIcon  { // Font islike icon..

 /**
  * Return the class name and RefNum for this widget.
  *
  * @return     The reference String for this widget
  */
  public String toString()  {return "(GameFont #" + RefNum + ")";} //~toString

 /**
  * CharSetup takes a position, the ASCII (numeric code of a) character,
  * a color and Parent RefNum to alter this widget to resemble an icon
  * positioned within the parent container, so when it is drawn, the
  * character appears there. This is repeated for each character in a
  * GameTxLn's line of text.
  *
  * @param  top    the vertical position for this character
  * @param  left   the horizontal position for this character
  * @param  xCh    the character to be drawn there
  * @param  color  its color, 0x00RRGGBB
  * @param  whom   the parent (group) widget to insert this icon under
  * @param  logy   true to spit out many lines of console text
  *
  * @return        the pixel position ended = left+Gwide, or else err<0
  */
  public int CharSetup(int top, int left, int xCh, int color, int whom,
      boolean logy)  { // (this is noisy, avoid logging if posible)
    int here, thar, nx, lxx = SafeArySz(Gdata), tall = 0, rez = 0,
        info = 0, why = 0;
    if (Logging) if (MaxLogWgts+77>Max_Log) logy = true;
    while (true)  { // top/left are relative to Parent=whom
      why--; // why = -1
      if ((Gtype&255) !=3) break;
      why--; // why = -2
      if (Gdata == null) break;
      why--; // why = -3
      if (lxx<8) break;
      thar = SafeAryElt(4,Gdata,-1); // =0 if mono, >0 is end of index
      why--; // why = -4
      if (thar<0) break; // not a valid font table
      here = SafeAryElt(1,Gdata,0); // lnHi/desc/asc here
      why--; // why = -5
      if (here <= 0) break; // not a valid font table
      why = -why; // why = +5
      rez = ((here>>8)+here)&31;
      if (xCh<0) break; // caller wants line height
      why++; // why = +6
      rez = left;
      if (xCh<32) break;
      why++; // why = +7
      if (xCh>126) break;
      why++; // why = +8
      if (xCh==32)  {
        Ginfo = xCh;
        Gcolor = -1; // so it won't try to draw something
        rez = SafeAryElt(3,Gdata,-1)+left;
        break;} //~if
      tall = (here>>16)&31;
      here = SafeAryElt(2,Gdata,-1); // letter spacing
      why = -why; // why = -8
      if (here<0) break; // not a valid font table
      why--; // why = -9
      if (here>32) break; // assume we can't do that
      if (thar>0)  { // proportional..
        nx = xCh-28;
        why--; // why = -10
        if (nx>thar) break; // past end of short font, assume invalid
        thar = SafeAryElt(nx+1,Gdata,0);
        nx = SafeAryElt(nx,Gdata,0); // = offset to this char
        why--; // why = -11
        if (nx<100) break; // not in pixel rgn of tbl, assume invalid
        thar = thar-nx;} //~if // now it's the active pixel width
      else { // monospaced..
        thar = SafeAryElt(3,Gdata,0)-SafeAryElt(2,Gdata,0);
        nx = (xCh-33)*thar+SafeAryElt(5,Gdata,0);} //~else
      why = -12;
      if (thar<0) break; // we don't do kerning, assume invalid
      why--; // why = -13
      if (nx+thar>lxx) break; // ran off end, ditto
      why = -why; // why = +13
      rez = left;
      if (thar==0) break;
      rez = left+thar+here;
      why = 0;
      if (color<0) break; // just measuring
      Parent = whom; // (temp'ly set inside the font wgt, so it draws)
      Gtop = top; // top/left are relative to Parent=whom
      Gleft = left;
      Gtall = tall;
      Gcolor = color;
      Gwide = thar;
      Ginfo = nx;
      break;} //~while
    if (why<0)  {
      logy = true;
      rez = why;} //~if
    if (Logging) if (logy)  { // if (MaxLogWgts+77>Max_Log)
      System.out.println("  -ChrSet- #" + RefNum
          + ") " + xCh
          + " (~" + MaxLogWgts + ") = " + rez + "/" + why);
      if (false) LogMe("  (ChSt) ");
      MaxLogWgts--;} //~if
    return rez;} //~CharSetup // rtns end or -err

 /**
  * GameFont constructor.
  */
  public GameFont()  {
    if (Logging) System.out.println("(new GmFont #" + RefNum + ")");
    Gtype = 3;}} //~GameFont

/**
 * One line of text.
 *
 * Ginfo is used for cursor blinking, its bit fields, high to low:
 *     (cursor H-posn):12, (cursor height):5, (ref+ of the font):14
 *   The horizontal position is a pixel offset from the left edge of the
 *     widget to the cursor bar; the width of the cursor is one pixel.
 *   The top of the cursor is the top of the text widget; adding its height
 *     gives the bottom of the cursor bar.
 *   Fonts are generally instantiated early in the game definition process,
 *     so the RefNum of the font widget to be used is usually a small number;
 *     however, it could be anywhere in the load, and if there are more than
 *     10,000 widgets, the performance of GameEngine will be intolerably slow.
 *     14 bits is therefore quite adequate.
 *
 *   The cursor blinks at a fixed rate, but the upper bits of Gtype are still
 *     used to count the fractional frame rate.
 *
 *   To display the text, GameTxLn's DrawMe steps through the text line, one
 *     character at a time, then calls on the selected font's CharSetup method
 *     to prepare that character for display, then calls its DrawMe (inherited
 *     from GameIcon) to draw it and return a horizontal pixel position to
 *     draw the next character.
 *
 * Gselim (GameTxLn/GameScro only) = current selection
 */
private class GameTxLn extends GameWgt  { // one line of text..
  int Gselim; // = current selection

 /**
  * Returns the selection of this GameTxLn. The selection will eventually
  * encompass a block of characters, but for now it is only an insertion
  * point, a single character offset.
  *
  * @return       The selection character offset = Gselim
  */
  public int GetSelim()  {return Gselim;} //~GetSelim

 /**
  * Set the selection of a GameTxLn. The selection will eventually
  * encompass a block of characters, but for now it is only an insertion
  * point, a single character offset.
  *
  * @param  whar   The new selection to set Gselim
  */
  public void SetSelim(int whar)  {Gselim = whar;} //~SetSelim

 /**
  * Return the class name and RefNum for this widget, plus its ID code if any.
  *   See toStrMash for formatting.
  *
  * @return     The reference String for this widget
  */
  public String toString()  {                                     // (GameTxLn)
    return toStrMash(super.toString(),"(GameTxLn #");} //~toString

 /**
  * Set the cursor in a GameTxLn's text, either as a pixel offset, or else
  * as calculated from a character offset. The pixel offset is used when
  * the user clicked in the text field; the character offset is used when
  * the text has been edited (characters added or deleted, or else an arrow
  * key is used to move the cursor left or right). If whar specifies a
  * valid cursor position, both Gselim and Ginfo are set.
  *
  * @param  pix    True if the cursor position is a pixel offset
  * @param  whar   The new cursor position as a pixel or character offset
  */
  public void SetBlinker(boolean pix, int whar)  { // sets blink
    // pix=T: whar is +pix to left (in wgt); pix=F: whar is +chars to left
    GameFont whom = null;
    int cx, kx, nx, info = Ginfo&0x3FFF, lxx = SafeArySz(Gdata)-1,
        posn = 0, tall = 0, here = 0, thar = 0, why = 0;
    boolean logy = false;
    whom = (GameFont)GetIxWgt(info);
    if (whom != null)
      tall = whom.CharSetup(0,0,-1,-1,0,logy)&31; // gets line height
    if (tall>0) for (nx=0; nx <= lxx; nx++) { // why = +20 for each completed int
      info = SafeAryElt(nx,Gdata,0);
      if (info==0) break;
      for (cx=0; cx <= 3; cx++) {
        why++; // why = +1
        if (!pix) if (posn >= whar) break; // normal exit at end of whar chars
        kx = info&255;
        why++; // why = +2
        if (kx==0) break; // normal exit for partial int (end of text)
        if (kx<32) continue;
        if (kx>126) continue;
        why++; // why = +3
        info = info>>8;
        thar = whom.CharSetup(0,here,kx,-1,0,logy); // adds char width to here
        if (thar<0) break; // (err) font table probably broken
        why++; // why = +4 for each completed char
        if (pix) if (thar >= whar) break; // normal exit at end of whar pixels
        here = thar;
        posn++;} //~for // why = +16 for each completed int
      why = why+4;
      if (thar<0) break;
      if (pix)  {if (thar >= whar) break;} //~if
      else if (posn >= whar) break;} //~for // (nx: each int)
    if (thar >= 0)  {
      why = -why;
      if (thar>here) if (here+thar<whar+whar)  { // past halfway, advance to next
        here = thar; // relative to wgt
        posn++;} //~if
      here = (((here<<5)+tall)<<14)+(Ginfo&0x3FFF);
      Gselim = posn;
      Ginfo = here;} //~if
    if (Logging) if (logy) System.out.println("  (SetBlnk/" + IffyStr(pix,"T","F")
        + "+" + whar + ") => " + posn + SeeHex("/x",here," ") + thar
        + " " + tall + " = " + why);} //~SetBlinker

 /**
  * Give this widget focus, or else take focus away.
  * A widget that is not typeable, clickable, or dragable cannot accept focus.
  *
  * When a GameTxLn widget accepts focus, it used the horiozntal click position
  * to determine where to set the insertion point (cursor), and then it starts
  * it blinking. When it gives up focus, it turns the animation off.
  *
  * @param  nuly   true to give focus, false to take it
  * @param  vert   the vertical position of a mouse-click giving focus
  * @param  horz   the horizontal position of the mouse-click
  *
  * @return        true if focus accepted, false if can't or losing it
  */
  public boolean SetFocus(boolean nuly, int vert, int horz)  { // (GameTxLn)
    String msg = "";   // (vert,horz are wgt-relative; rtn F if refused/off)
    int why = 0, tye = Gtype&255, thar = Ginfo&0x3FFF, blnk = Ginfo>>14;
    boolean logy = (MaxLogWgts>0);
    GameFont whom = (GameFont)GetIxWgt(thar);
    Gtype = Gtype&0xFFFF;
    if (!nuly) while (true)  { // F: turn off the focus/selection..
      why--; // why = -1
      if (tye !=4) break;
      Ginfo = thar;
      Gselim = -1;
      why--; // why = -2
      break;} //~while
    else while (true)  {
      if (FustFocLog>0) if (MaxLogWgts<FustFocLog)  {
        MaxLogWgts = FustFocLog;
        FustFocLog = 0;
        logy = true;} //~if
      why++; // why = 1
      if (tye !=4) break;
      Ginfo = thar;
      why++; // why = 2
      if (Gdata == null) break;
      why++; // why = 3
      if (whom==null) break; // unspecified/non-existing font
      why++; // why = 4
      if ((whom.GetType()&15) !=3) break; // gotta be a font
      why++; // why = 5
      if (vert<0) if (horz<0) horz = blnk>>5; // *** sb right end
      if (Logging) msg = "  (SeFocus #" + RefNum + ") t=" + (Gtype&255)
          + " " + vert + "/" + horz + " " + blnk;
      SetBlinker(true,horz);
      blnk = Ginfo>>14;
      why = 0;
      break;} //~while
    if (Logging) if (logy)  {
      if (StrLength(msg)==0) msg = "  (sFoc #" + RefNum + ")";
      System.out.println(msg + " => " + blnk + " = " + why);
      MaxLogWgts--;} //~if
    return why==0;} //~SetFocus

 /**
  * This is an internal utility for formatting a widget's information value,
  * GameTxLn overrides this to append the formatted cursor parts if !=0.
  *
  * @param  before  a prefix to add to the front
  * @param  tye     the widget's type
  *
  * @return         the composite display value
  */
  public String FmtInfo(String before, int tye)  { //         // (GameTxLn)
    before = super.FmtInfo(before,tye);
    tye = Ginfo>>14;
    if ((Ginfo&(-0x4000)) !=0)
      before = before + "+" + (tye&31) + "/" + ((tye>>5)&0xFFF);
    return before + " s=" + Gselim;} //~FmtInfo

 /**
  * This draws the GameTxLn pixels on the window as a sequence of character
  * icons. The boundaries of a masking rectangle (usually) limit the drawing
  * to the containing widget's dimensions.
  *
  * Ginfo is used for cursor blinking, its bit fields, high to low:
  *     (cursor H-posn):12, (cursor height):5, (ref+ of the font):14
  *   The horizontal position is a pixel offset from the left edge of the
  *     widget to the cursor bar; the width of the cursor is one pixel.
  *   The top of the cursor is the top of the text widget; adding its height
  *     gives the bottom of the cursor bar.
  *
  *   The cursor blinks at a fixed rate, but the upper bits of Gtype are still
  *     used to count the fractional frame rate.
  *
  *   To display the text, GameTxLn's DrawMe steps through the text line, one
  *     character at a time, then calls on the selected font's CharSetup method
  *     to prepare that character for display, then calls its DrawMe (inherited
  *     from GameIcon) to draw it and return a horizontal pixel position to
  *     draw the next character.
  *
  * @param  top    do not draw above this position
  * @param  left   do not draw to the left of this position
  * @param  tall   do not draw below this position +top
  * @param  wide   do not draw to the right of this position +left
  */
  public void DrawMe(int top, int left, int tall, int wide)  { // (GameTxLn)
    int RealPrent = WhoCalled(0), // used only to verify hierarchy
        nx = AddPrntsPos(GetPosn()), myTop = nx>>16, myLft = SignExtend(nx),
        here, thar = Ginfo&0x3FFF, rx = myTop+Gtall, cx = myLft+Gwide, zx,
        lxx = SafeArySz(Gdata), blnk = Ginfo>>14, logy = 0, info = 0,
        xCh = 0, saw = 0, why = 0;
    String msg = "";
    GameFont whom = (GameFont)GetIxWgt(thar);
    // CallersID = 0;
    while (true)  {
      if (Logging)  { // tye=4: "wTxLn"  (Dr.wTxLn +
        if (logy>0)  {}
        else if (MaxLogWgts>0)  {
          logy = MaxLogWgts;
          MaxLogWgts--;} //~if
        else if (blnk !=0)  {
          if (MaxLogAnim>8) logy = MaxLogAnim;
          else if (((Gtype+0x10000)&0x30000)==0) logy = MaxLogAnim;
          if (logy>0) MaxLogAnim--;}} //~if
      if ((SeenLog&16)==0) logy = 1234;
      if (logy>0) msg = FmtFront("  (Dr.",false," +" + lxx
          + " " + blnk + "/" + (Gtype>>16));
      why++; // why = 1
      if (Gtop>0x3FFF) break; // hidden
      why++; // why = 2
      if (Parent==0) break;
      if (RealPrent !=0) if (RealPrent != Parent) break;
      why++; // why = 3
      if ((Gtype&255) !=4) break;
      why++; // why = 4
      if (Gdata == null) break;
      why++; // why = 5
      if (whom==null) break;
      why++; // why = 6
      if ((whom.GetType()&15) !=3) break; // gotta be a font
      why++; // why = 7
      if (lxx <= 0) break;
      why++; // why = 8
      if (logy==1234) SeenLog = SeenLog|16;
      if (rx <= top+tall) tall = rx-1; // now it's edge, not size..
        else tall = top+tall-1;
      if (cx <= left+wide) wide = cx-1;
        else wide = left+wide-1;
      if (myTop>top) top = myTop;
      if (myLft>left) left = myLft;
      if (top<0) top = 0;
      if (left<0) left = 0;
      if (logy>999)
        msg = msg + " => " + top + "/" + left + " +" + tall + "/" + wide;
      if (blnk !=0) if (Gcolor >= 0) while (true)  {
        here = myTop-1;
        Gtype = Gtype+0x10000;
        if ((Gtype&0x40000) !=0) break; // off phase of blink
        rx = (blnk&31)+here; // = btm of blinker
        // if (rx <= top) break;
        // if (myTop>tall) break;
        cx = (((blnk>>5)+nx)&0xFFF)-1; // H-posn of blinker
        if (here<top) nx = top;
          else nx = here;
        if (rx>tall) rx = tall; // nx..rx is V-coord
        if (logy>0) msg = msg + " {" + nx + IffyStr(nx<rx,"_","^") + rx
            + "/" + cx + IffyStr(cx+1<left,"-}",IffyStr(cx+2>wide,"+}","}"));
        if (cx<0) break;
        if (cx+1<left) break; // clipped (off-screen)..
        if (cx+2>wide) break;
        if (nx<rx) RectFill(0,nx,cx,rx,cx);
        break;} //~while
      here = 0; // can't use Gleft/myLft, cuz Dr.wFon adds Parent(=RefNum) loc
      if (logy>0) msg = msg + " '";
      for (nx=0; nx <= lxx-1; nx++) {
        if (here+myLft >= wide) break;
        info = SafeAryElt(nx,Gdata,0);
        if (info==0) break;
        for (cx=0; cx <= 3; cx++) {
          if (here >= wide) break;
          xCh = info&255;
          info = info>>8;
          if (xCh==0) break;
          if (xCh<32) continue;
          if (xCh>126) continue;
          if (logy>0) msg = msg + (char)xCh;
          thar = whom.CharSetup(0,here,xCh,Gcolor,RefNum,logy>999);
          if (thar<0) break; // (err) font table probably broken  // ..not myTop
          if (thar+myLft>left)  {
            zx = WhoCalled(RefNum);
            whom.DrawMe(top,left,tall-top,wide-left);} //~if
          else saw = (thar<<16)+myLft;
          here = thar;} //~for
        if (xCh==0) break;} //~for
      if (logy>0) msg = msg + "'";
      why = 0;
      break;} //~while
    if (false) if (logy <= 0) if (why==2) logy = 1001;
    if (Logging||(logy==1234)) if (logy>0)  {                // (Dr.wTxLn +
      System.out.println(msg + " = " + why + IffyStr(saw==0,"",LogVH(" <",saw,">"))
          + IffyStr(why==2," !prnt=" + Parent + "=" + RealPrent + " (RP)",""));
      if (logy>999) LogMe("  (DrMe) ");} //~if
    zx = WhoCalled(0);} //~DrawMe

 /**
  * GetTextWide returns the pixel width of this widget's text line.
  *
  * @return         the pixel width of the text line
  */
  public int GetTextWide()  { // gets it from SetBlink // ** redo frm tbl
    int posn = Gselim, info = Ginfo, here, thar = info&0x3FFF;
    boolean logy = (MaxLogWgts>0);
    here = StrLength(GetTextLn());
    SetBlinker(false,here);
    here = Ginfo>>19;
    Ginfo = info;
    Gselim = posn;
    if (Logging) if (logy)
      System.out.println("  (GetTxtWi #" + RefNum + ") = " + here);
    return here;} //~GetTextWide // only works in GameTx

 /**
  * Returns this widget's text line.
  *
  * @return       The text line = Gdata as a String
  */
  public String GetTextLn()  { // only works in GameTx
    int lxx = SafeArySz(Gdata), info, cx, nx, kx = 0;
    String aLine = "";
    int[] data = Gdata;
    if (lxx>0) for (nx=0; nx <= lxx-1; nx++) {
      if (data == null) break;
      if (nx >= data.length) break;
      info = data[nx];
      for (cx=0; cx <= 3; cx++) {
        kx = info&255;
        if (kx>126) kx = 0;
        else if (kx<32) kx = 0;
        if (kx==0) break;
        aLine = aLine + (char)kx;
        info = info>>8;} //~for
      if (kx==0) break;} //~for
    return aLine;} //~GetTextLn

 /**
  * Set this widget's text line.
  *
  * @param  aLine    The text to set Gdata to
  *
  * @return         the pixel width of the text line
  */
  public int PutTextLn(String aLine)  { // only works in GameTx
    boolean logy = (MaxLogWgts>0);
    int lxx = StrLength(aLine), kx = (lxx+3)>>2, nx = SafeArySz(Gdata),
        info = 0, why = 0;
    int[] data;
    while (true)  {
      why++; // why = 1
      if ((Gtype&255) !=4) break;
      why++; // why = 2
      if (lxx==0) if (GetIxData(0,0)==0) break; // no text = no data, done
      if (nx==0) data = SwapData(new int[kx+1]); // no data, create it
      else if (kx>nx) data = SwapData(new int[kx+4]); // enlarge existing data
      why++; // why = 3
      if (lxx>SafeArySz(Gdata)*4) break; // failed to enlarge
      data = Gdata;
      why++; // why = 4
      for (nx=lxx|3; nx >= 0; nx += -1) {
        kx = 0;
        if (nx<lxx) kx = (int)(CharAt(nx,aLine))&127;
        info = (info<<8)+kx;
        if ((nx&3)>0) continue;
        kx = nx>>2;
        if (kx<0) break;
        if (data != null) if (kx < data.length)
          data[kx] = info;
        info = 0;} //~for
      why = 0;
      break;} //~while
    kx = GetTextWide();
    if (Logging) if (logy)  {
      if (data == null) nx = 0;
        else nx = SafeArySz(data);
      System.out.println("  (Put.TxLn #" + RefNum + ") *" + nx + "/" + lxx + " => '"
          + GetTextLn() + "' w=" + kx + "/" + Gwide + " = " + why);} //~if
    MaxLogWgts--;
    return kx;} //~PutTextLn

  public GameTxLn()  {
    Gselim = -1; // no selection
    if (Logging) System.out.println("(new GmTxLn #" + RefNum + ")");
    Gtype = 4;}} //~GameTxLn

/**
 * Animated group head for a complex sprite.
 *
 * Ginfo is the packed (signed) velocity vector,
 *      (vertical*256):16, (horizontal*256):16
 *   Each velocity component is a 16-bit fixed-point number, 8i.8f,
 *     with eight bits of signed integer part and 8 bits fraction part,
 *     for a value range [-128.0 .. +127.996], where +2.0 means the sprite
 *     is moving down or right at two pixels per frame, about 20 pix/sec
 *     given a nominal frame rate of 10fps, and -0.1 means the sprite is
 *     moving up or left one pixel every ten frames, or 1 pix/sec.
 *
 *   Otherwise this is like a group (type=5), except that the components
 *     are expected to be one or more layers of simple widgets, rectangle or
 *     icon or text, each with their own color, (and the icons possibly with
 *     their own animation), all contained within the dimenstions of the
 *     sprite.
 *
 *   Curves or other acceleration must be programmed by periodically changing
 *     the velocity vector. For example, gravity is programmed by adding a
 *     constant positive value +32 feet/s/s = +10m/s/s to the vertical
 *     component each frame. If your game models vertical distance at 20 pix/m,
 *     then you want to add 10m/s every second, which is 1 m/s = 20 pix/s each
 *     frame (= 0.1 secs, use DoOften) to the vertical velocity, which is
 *     (20<<8)<<16 = Ginfo+0x14000000, or more simply, if your widget reference
 *     is named 'heavy' in Java:
 *
 *         heavy.SetInfo(heavy.GetInfo()+((20<<8)<<16));
 *
 *     In practice you may need to tweak these numbers for it to look right.
 *     Note that the (20<<8) shifts the integer 20 over eight places (the size
 *     of the fractional part) and then you shift it another 16 places into the
 *     upper half of the information value (because the horizontal component is
 *     in the low 16 bits).
 *
 *     For (very light) asteroid gravity, say 0.1 pix/s, you need to calculate
 *     the value of 0.1*256 = 25.6 rounded to 26, which is the fractional part
 *     (do not shift <<8): ...SetInfo(heavy.GetInfo()+(26<<16));
 *
 *   Of course your sprite does not need to move at all, or only get a new
 *     position on explicit call to SetPosn, so in such a case just leave
 *     Ginfo = 0.
 *
 *   If you want your sprite to have various colors, you would layer several
 *     icons of different colors, but the icon editor to do that in GameEngine
 *     does not yet exist. Real Soon Now.
 */
private class GameSprit extends GameWgt  { // Moving sprite

 /**
  * Return the class name and RefNum for this widget, plus its ID code if any.
  *   See toStrMash for formatting.
  *
  * @return     The reference String for this widget
  */
  public String toString()  {                                    // (GameSprit)
    return toStrMash(super.toString(),"(GameSprit #");} //~toString

 /**
  * Returns the specified color for this widget, = 0x00RRGGBB or -1 if none.
  * Not the color of a sprite root, but reach in and get it from
  * whatever rect or icon or text is last (top) child.
  *
  * @return       The specified color, or -1 if none
  */
  public int GetColor()  { //                                // (GameSprit)
    int inx, info, tye, colo = Gcolor, logy = 0;
    GameWgt nuly = null;
    for (inx=0; inx <= 99; inx++) {
      info = GetIxData(inx,0);
      if (info <= 0) break; // normal exit
      nuly = GetIxWgt(info);
      if (nuly==null) break;
      tye = nuly.GetType()&255;
      info = nuly.GetColor();
      if (Logging)
        System.out.println("(Sp.GetColo #" + RefNum + ") " + inx
            + ": t=" + tye + SeeHex(" c=",info,""));
      if (tye>4) continue;
      if ((tye&1) !=0) continue;
      colo = info;} //~for // save it if it's 0/2/4
    return colo;} //~GetColor

 /**
  * Set the color for this widget, = 0x00RRGGBB or -1 if none.
  * Never change the color of a sprite root, but reach in and
  * do it to whatever rect or icon or text is last (top) child.
  *
  * @param  color    The color to set Gcolor to, <0 is no color
  */
  public void SetColor(int color)  { //                      // (GameSprit)
    int inx, info, tye, logy = 0;
    GameWgt whom = null, nuly = null;
    for (inx=0; inx <= 99; inx++) {
      info = GetIxData(inx,0);
      if (info <= 0) break; // normal exit
      nuly = GetIxWgt(info);
      if (nuly==null) break;
      tye = nuly.GetType()&255;
      if (Logging)
        System.out.println("(Sp.SetColo #" + RefNum + ") " + inx
            + ": t=" + tye + SeeHex(" c=",nuly.GetColor(),""));
      if (tye>4) continue;
      if ((tye&1) !=0) continue;
      whom = nuly;} //~for // save it if it's 0/2/4
    if (whom != null) whom.SetColor(color); // set its color
      else Gcolor = color;} //~SetColor

 /**
  * This is an internal utility for formatting a widget's information value,
  * GameSprit overrides this to format it as two fixed-point numbers.
  *
  * @param  before  a prefix to add to the front
  * @param  tye     the widget's type
  *
  * @return         the composite display value
  */
  public String FmtInfo(String before, int tye)  { //        // (GameSprit)
    return FixHaff(before,Ginfo>>16) + FixHaff("/",SignExtend(Ginfo));} //~FmtInfo

 /**
  * This draws the GameSprit subwidgets on the window as just like the group
  * (type=5), except that the velocity is applied to the group position before
  * calling super to draw it.
  *
  * Ginfo is a velocity vector, each component a 16-bit fixed-point number,
  *     8i.8f, with eight bits of signed integer part and 8 bits fraction part,
  *     for a value range [-128.0 .. +127.996], where +2.0 means the sprite
  *     is moving down or right at two pixels per frame, about 20 pix/sec
  *     given a nominal frame rate of 10fps, and -0.1 means the sprite is
  *     moving up or left one pixel every ten frames, or 1 pix/sec.
  *
  * @param  top    do not draw above this position
  * @param  left   do not draw to the left of this position
  * @param  tall   do not draw below this position +top
  * @param  wide   do not draw to the right of this position +left
  */
  public void DrawMe(int top, int left, int tall, int wide)  { // (GameSpri)
    int whar, RealPrent = WhoCalled(0), // used only to verify hierarchy
        frax = Gtype>>16, thar = 0, logy = 0;
    String msg = "";
    // CallersID = 0;          // this only works properly w/parent origin @ 0,0
    if (Gtop>0x3FFF) return; // hidden
    if (RealPrent !=0) if (Parent !=0) if (RealPrent != Parent) return;
    if ((Gtype&255) !=6) return;
    if (Gdata == null) if (Gcolor<0) return;
    if ( JavaGame. SpritePause())  {
      super.DrawMe(top,left,tall,wide);
      return;} //~if
    if ((SeenLog&64)==0) logy = 12345;
    else if (Logging) if (logy>0)
      if ((logy&0xFFFF) != RefNum) if ((logy>>16) != RefNum) logy = 0;
    if (logy>0) msg = FmtInfo(" ",6) + FixHaff(" f=",(frax>>8)&255)
        + FixHaff("/",frax&255) + LogVH(" (",MaxWinSize,
        ")");
    if (Ginfo !=0)  { // we are moving, do it (fix.pt, 8-frac)..
      // frax = Gtype>>16;
      whar = (Ginfo>>16)+(Gtop<<8)+((frax>>8)&255); // new signed top
      if (logy>0) msg = msg + FixHaff(" => ",whar);
      if (whar<0) thar++; // off top
      thar = thar<<1;
      Gtop = whar>>8;
      if (Gtop+Gtall > WinHi) thar++; // off botm
      thar = thar<<1;
      frax = frax&255|((whar&255)<<8);
      whar = SignExtend(Ginfo)+(Gleft<<8)+(frax&255); // new signed left
      if (logy>0) msg = msg + FixHaff("/",whar);
      if (whar<0) thar++; // off left edge
      thar = thar<<1;
      Gleft = whar>>8;
      if (Gleft+Gwide > WinWi) thar++; // off right
      frax = frax&0xFF00|whar&255;
      if ((Gtype&256) !=0) if (thar>0)
        Gtype = Gtype|0x8000; // flag that we ran off the edge
      if (logy>0) msg = msg + " (" + thar + FixHaff(") f=",(frax>>8)&255)
          + FixHaff("/",frax&255);
      Gtype = (frax<<16)|Gtype&0xFFFF;} //~if
    if (Logging||(logy==12345)) if (logy>0)
      System.out.println("   (Sp.Dr #" + RefNum + ")" + msg
          + " => " + Gtop + "/" + Gleft);
    whar = WhoCalled(Parent);
    super.DrawMe(top,left,tall,wide);
    if (logy==12345) SeenLog = SeenLog|64;} //~DrawMe

  public GameSprit()  {
    if (Logging) System.out.println("(new GmSprt #" + RefNum + ")");
    Gtype = 6;}} //~GameSprit

/**
 * Scrolled panel with one or two scrollbars. [Not yet fully implemented]
 *
 * The scrolled panel is typically a group, the first widget indexed in the
 * Gdata array. The second indexed widget is the vertical scrollbar (details
 * TBD), and the third is the horizontal scrollbar. If a scrollbar is omitted,
 * its position in the array is zero.
 *
 * Gselim (GameTxLn/GameScro only) = scroll limits
 */
private class GameScro extends GameWgt  { // scrolling field..
  int Gselim; // = current scroll limits

 /**
  * Returns the scroll limits of a GameScro.
  *
  * @return       The scroll limits = Gselim
  */
  public int GetSelim()  {return Gselim;} //~GetSelim

 /**
  * Set the scroll limits of a GameScro.
  *
  * @param  whar   The new scroll limits for Gselim
  */
  public void SetSelim(int whar)  {Gselim = whar;} //~SetSelim

 /**
  * Return the class name and RefNum for this widget, plus its ID code if any.
  *   See toStrMash for formatting.
  *
  * @return     The reference String for this widget
  */
  public String toString()  {                                     // (GameScro)
    return toStrMash(super.toString(),"(GameScro #");} //~toString

 /**
  * This draws the GameScro subwidgets on the window as just like the group
  * (type=5), except that the scrollbar values are applied to the panel group
  * position before calling super to draw everything.
  *
  * @param  top    do not draw above this position
  * @param  left   do not draw to the left of this position
  * @param  tall   do not draw below this position +top
  * @param  wide   do not draw to the right of this position +left
  */
  public void DrawMe(int top, int left, int tall, int wide)  { // (GameScr)
    int RealPrent = WhoCalled(0), // used only to verify hierarchy
        nx = AddPrntsPos(GetPosn()), myTop = nx>>16, myLft = SignExtend(nx),
        here, thar, rx = myTop+Gtall, cx = myLft+Gwide, lxx = SafeArySz(Gdata),
        logy = 0, info = 0, why = 0;
    String msg = "";
    // CallersID = 0;
    while (true)  {
      why++; // why = 1
      if (Gtop>0x3FFF) break; // hidden
      why++; // why = 2
      if (RealPrent !=0) if (Parent !=0) if (RealPrent != Parent) break;
      why++; // why = 3
      if ((Gtype&255) !=7) break;
      why++; // why = 4
      why++; // why = 5
      if (Gcolor<0) break; // single-layer hit rgn for sprite (undrawn)
      why++; // why = 6
      if (Gdata == null) break;
      why++; // why = 7
      if (lxx <= 0) break;
      if (Logging) if (MaxLogWgts>0)
        msg = FmtFront("  (Dr.",false," +" + lxx);
      if (rx <= top+tall) tall = rx-1; // now it's edge, not size..
        else tall = top+tall-1;
      if (cx <= left+wide) wide = cx-1;
        else wide = left+wide-1;
      if (myTop>top) top = myTop;
      if (myLft>left) left = myLft;
      if (top<0) top = 0;
      if (left<0) left = 0;
      why++; // why = 8
      for (nx=0; nx <= Gwide-1; nx++) {
        info = SafeAryElt(nx+Ginfo,Gdata,0);
        if (info==0) continue;
        here = nx+myLft;
        if (here<left) continue;
        if (here>wide) continue;
        for (rx=0; rx <= lxx; rx++) {
          thar = rx+myTop;
          if ((info&1) !=0) if (thar >= top) if (thar <= tall)
            PokePixel(Gcolor,thar,here);
          info = info>>1;
          if (info==0) break;}} //~for
      break;} //~while // tye: "wScro"
    if (Logging) if (MaxLogWgts>0) System.out.println(msg + " = " + why);
    MaxLogWgts--;} //~DrawMe

  public GameScro()  {
    Gselim = 0; // no limits
    if (Logging) System.out.println("(new GmScro #" + RefNum + ")");
    Gtype = 7;}} //~GameScro

 /**
  * UsrGameJv is the source for the subclass generated by the Build button
  */
  private final static String UsrGameJv = "\nimport javagame.GameWgt;\n"
  + "import javagame.GameEvent;\nimport javagame.JavaGame;\n\n/**\n"
  + " * Subclass to handle events when creating a new game. 2020 July 25\n"
  + " *\n * The name of this class is the same as the name of your Game. If\n"
  + " * you change the name in GameMaker, then the Build button will create\n"
  + " * a new (or modify an existing) Java file with the new name.\n *\n"
  + " * Three sections of this code (beginning and ending with \"$\" tags)\n"
  + " * are generated when you click the \"Build\" button. You should avoid\n"
  + " * modifying these lines because your modifications will be lost when\n"
  + " * you click Build again. If you modify or move the tag lines the Build\n"
  + " * process may be thwarted or unpredictable. Move or rename your Java\n"
  + " * file and it will create a new clean copy in its place.\n */\n"
  + "public class UserGame extends GameEvent {\n"
  + "  JavaGame myGame = null; // soon: reference to the GameEngine\n"
  + "  GameWgt myUserGame = null; // reference to the game board widget\n"
  + "  /// ($) Begin generated class variables (do not modify) ($..)\n"
  + "  /// (..$) End generated class variables ($) (do not modify this line)\n"
  + "\n  public String toString() {return \"(UserGame)\";} //~toString\n\n"
  + "  // These called regardless:\n\n /**\n"
  + "  * (Generated code) GameList returns a text descriptor of all your\n"
  + "  * specified widgets. This code is created when you click the \"Build\"\n"
  + "  * button in GameMaker. It is called once before StartUp, then builds\n"
  + "  * the widgets from your text descriptor(s). You must call the super\n"
  + "  * to preserve the reference to the host JavaGame.\n  *\n"
  + "  * @param  whom  A reference to the host JavaGame, kept for later use\n"
  + "  *\n  * @return       The generated text string\n  */\n"
  + "  public String GameList(JavaGame whom) {\n"
  + "    String info = super.GameList(whom), data\n"
  + "    /// <$> Begin generated widget list (do not modify) <$..>\n"
  + "    /// <..$> End generated widget list <$> (do not modify this line)\n"
  + "    myGame = whom; // save a local reference to the GameEngine\n"
  + "    return data;} //~GameList\n\n /**\n"
  + "  * (Partly generated code) After the initial widgets have been created,\n"
  + "  * StartUp is called to do any other initialization you may have.\n"
  + "  * If you gave Java-correct names to any widgets, those variables are\n"
  + "  * assigned values by the generated code.\n  */\n"
  + "  public void StartUp() {\n"
  + "    if (myGame != null) myUserGame = myGame.FindListWgt(\"{UserGame}\");\n"
  + "    /// [$] Begin generated StartUp code (do not modify) [$..]\n"
  + "    /// [..$] End generated StartUp code [$] (do not modify this line)\n"
  + "    } //~StartUp\n\n /**\n"
  + "  * DoOften is called once each frame to do whatever game processing\n"
  + "  * needs to be done often. You should get in and get out, so that the\n"
  + "  * animation flows smoothly. Omit this method if you do not need it.\n"
  + "  */\n  // public void DoOften() {} //~DoOften\n\n /**\n"
  + "  * GameEnding is called once when the game window is about to close,\n"
  + "  * so you can do whatever needs to be done at the end.\n"
  + "  * Omit this method if you do not need it.\n  */\n"
  + "  // public void GameEnding() {} //~GameEnding\n\n"
  + "  // This is enabled by Gtype&256 (Collidable):\n\n /**\n"
  + "  * Collided is called once for each Sprite widget that is flagged as\n"
  + "  * Collidable (\"Bump\") when it touches any other widget also flagged\n"
  + "  * Collidable.\n  *\n  * @param  whom  The Sprite widget\n"
  + "  * @param  hitt  The other widget in the collision\n  *\n"
  + "  * @return       False if this collision is not satisfactory, or if\n"
  + "  *               you want to consider other (simultaneous) collisions\n"
  + "  *               for whom; true to terminate processing this Sprite.\n"
  + "  *               Note that other Sprites may yet have collisons pending\n"
  + "  *               in this same frame, resulting in additional calls\n"
  + "  *               on Collided.\n  */\n"
  + "  // public boolean Collided(GameWgt whom, GameWgt hitt) {\n"
  + "  //   if (keepLooking) return false;\n"
  + "  //   return true;} //~Collided // return true to stop looking\n\n"
  + "  // These events are enabled by Gtype&512 (Focusable):\n\n /**\n"
  + "  * GotEnterKey is called when the user types the EnterKey,\n"
  + "  * if FocusWgt is not null, or else if DefaultFocus is not null,\n"
  + "  * and that widget accepts keystrokes. If your game can open\n"
  + "  * a standard alert or input dialog, you should call super so\n"
  + "  * it can get its OK clicked. Your subclass can return either\n"
  + "  * true or false, because the first caller doesn't care.\n  *\n"
  + "  * @param  whom  The widget that has focus\n  *\n"
  + "  * @return       true if taken (by alert or input dlog)\n  */\n"
  + "  // public boolean GotEnterKey(GameWgt whom) {\n"
  + "  //   if (super.GotEnterKey(whom)) HandleDlogOK();\n"
  + "  //     else DoWhatever();\n"
  + "  //   return false;} //~GotEnterKey\n\n /**\n"
  + "  * GotEscKey is called when the user types the EscapeKey,\n"
  + "  * if FocusWgt is not null, or else if DefaultFocus is not null,\n"
  + "  * and that widget accepts keystrokes. If your game can open\n"
  + "  * a standard alert or input dialog, you should call super so\n"
  + "  * it can get its Cancel clicked. Your subclass can return either\n"
  + "  * true or false, because the first caller doesn't care.\n  *\n"
  + "  * @param  whom  The widget that has focus\n  *\n"
  + "  * @return       true if taken (by alert or input dlog)\n  */\n"
  + "  // public boolean GotEscKey(GameWgt whom) { // o'ride to do more\n"
  + "  //   if (super.GotEscKey(whom)) return true;\n  //   DoWhatever();\n"
  + "  //   return false;} //~GotEscKey\n\n /**\n"
  + "  * GotTabKey is called when the user types the TabKey,\n"
  + "  * if FocusWgt is not null, or else if DefaultFocus is not null,\n"
  + "  * and that widget accepts keystrokes.\n  *\n"
  + "  * @param  whom  The widget that has focus\n  */\n"
  + "  // public void GotTabKey(GameWgt whom) {} //~GotTabKey\n\n /**\n"
  + "  * GotArrowKey is called for any widget that has focus when the user\n"
  + "  * types one of the arrow keys. If FocusWgt is null and DefaultFocus\n"
  + "  * is null, or if that widget does not accept keystrokes, then\n"
  + "  * GotArrowKey is not called. The base class uses GotArrowKey to move\n"
  + "  * a dragable widget one pixel at time, so you should call super if\n"
  + "  * this is not yours.\n  *\n"
  + "  * @param  whom  The widget that has focus\n"
  + "  * @param  info  The arrow key that was pressed, one of the constant\n"
  + "  *               values JavaGame.Arro_UP, JavaGame.Arro_DOWN,\n"
  + "  *               JavaGame.Arro_LEFT, or JavaGame.Arro_RIGHT.\n  */\n"
  + "  // public void GotArrowKey(GameWgt whom, char info) {\n"
  + "  //   if (didSomethingWithIt) return;       // but call super..\n"
  + "  //   super.GotArrowKey(whom,info);} //~GotArrowKey\n\n /**\n"
  + "  * KeyFld is called for any widget that has focus when the user\n"
  + "  * types any other key. If FocusWgt is null and DefaultFocus is null,\n"
  + "  * or if that widget does not accept keystrokes, then KeyFld is not called.\n"
  + "  * The base class uses KeyFld to insert keystrokes in an entry text field,\n"
  + "  * so you should call super if this is not yours.\n  *\n"
  + "  * @param  whom  The widget that has focus\n"
  + "  * @param  info  The key that was typed\n  */\n"
  + "  // public void KeyFld(GameWgt whom, char info) {\n"
  + "  //   if (didSomethingWithIt) return;       // but call super..\n"
  + "  //   super.KeyFld(whom,info);} //~KeyFld\n\n /**\n"
  + "  * FocusEv is called for any widget that is \"Focusable\" when the user\n"
  + "  * clicks on it. Your code may also call FocusEv to give focus to a\n"
  + "  * widget for any other reason. You must call super to do the default\n"
  + "  * focus processing unless you choose to cancel it. The base class\n"
  + "  * calls the whom.SetFocus method for this widget, and the widget also\n"
  + "  * has the opportunity to cancel the focus.\n  *\n"
  + "  * Only one widget at a time can have focus, so BlurEv is called for\n"
  + "  * any previous widget with focus before any other widget can get it.\n"
  + "  *\n  * @param  whom  The widget to be given focus\n"
  + "  * @param  vert  The vertical offset into this widget where\n"
  + "  *                 it was clicked\n"
  + "  * @param  horz  The horizontal offset into this widget\n  */\n"
  + "  // public void FocusEv(GameWgt whom, int vert, int horz) {\n"
  + "  //   if (didSomethingWithIt) return;\n"
  + "  //   super.FocusEv(whom,vert,horz);} //~FocusEv\n\n /**\n"
  + "  * BlurEv is called for a widget when it loses focus. Your code may\n"
  + "  * also call BlurEv to unconditionally terminate focus for any other\n"
  + "  * reason. You should call super to do the default defocussing\n"
  + "  * (including telling the widget that it is losing focus) unless you\n"
  + "  * want to cancel the blur. If you cancel a blur that is called before\n"
  + "  * another widget is about to gain focus, then that focus is also\n"
  + "  * cancelled.\n  *\n"
  + "  * @param  whom  The widget losing focus, or null if unknown by caller\n"
  + "  */\n  // public void BlurEv(GameWgt whom) {\n"
  + "  //   if (cancellingThisBlur) return;\n  //   super.BlurEv(whom);\n"
  + "  //   // (do additional processing)\n  //   } //~BlurEv\n\n"
  + "  // This event is enabled by Gtype&1024 (Clickable):\n\n /**\n"
  + "  * ClickEvt is called for a widget that is \"Clickable\" when the user\n"
  + "  * clicks on it (mouse down followed by mouse up on the same widget).\n"
  + "  * FocusEv has already been called for this widget. The base class uses\n"
  + "  * ClickEvt to detect and process double-click events, so you should call\n"
  + "  * super if you want that to happen and this is not yours.\n  *\n"
  + "  * You would normally override ClickEvt to process button clicks.\n  *\n"
  + "  * The vertical and horizontal offsets are relative to the top-left\n"
  + "  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get\n"
  + "  * the position relative to the window.\n  *\n"
  + "  * @param  whom  The widget that was clicked\n"
  + "  * @param  vert  The vertical offset into this widget where\n"
  + "  *                 it was clicked\n"
  + "  * @param  horz  The horizontal offset into this widget\n  *\n"
  + "  * @return       true if taken (by 2-click or alert or input dlog)\n"
  + "  */\n  // public boolean ClickEvt(GameWgt whom, int vert, int horz) {\n"
  + "  //   if (super.ClickEvt(whom,vert,horz)) return true; // already did\n"
  + "  //   // DoWhatever();\n"
  + "  //   return false;} //~ClickEvt // (or true: nobody else cares)\n\n"
  + "  // These events are enabled by Gtype&2048 (Dragable):\n\n"
  + "  // DragonDrop (drag and drop) consists of the sequence of pressing the\n"
  + "  //   mouse button over a widget, then while holding the mouse button\n"
  + "  //   down, moving the mouse to some other place on the screen, then\n"
  + "  //   releasing the mouse button at that location. The default action\n"
  + "  //   in the base class for dragable widgets is to drag (move) the\n"
  + "  //   widget that was clicked to the new location. After a widget that\n"
  + "  //   is both dragable and focusable has been dragged with the mouse,\n"
  + "  //   you can fine-tune its final position using the arrow keys. You\n"
  + "  //   can specify other behavior for these events by overriding these\n"
  + "  //   methods.\n\n /**\n"
  + "  * MouseDn is called for a widget that is \"Dragable\" when the user\n"
  + "  * presses the (right) mouse button down on some widget. If the button\n"
  + "  * is released without moving the mouse, then it is just a ClickEvt\n"
  + "  * and no drag is attempted.\n  *\n"
  + "  * The vertical and horizontal offsets are relative to the top-left\n"
  + "  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get\n"
  + "  * the position relative to the window.\n  *\n"
  + "  * @param  whom  The widget that was clicked\n"
  + "  * @param  vert  The vertical offset into this widget where it\n"
  + "  *                 was clicked\n"
  + "  * @param  horz  The horizontal offset into this widget\n  */\n"
  + "  // public void MouseDn(GameWgt whom, int vert, int horz) {\n"
  + "  //   if (didSomethingWithIt) return;\n"
  + "  //   super.MouseDn(whom,vert,horz);} //~MouseDn\n\n /**\n"
  + "  * After a MouseDn is called for a widget that is \"Dragable\" and the\n"
  + "  * user moves the mouse at least eight pixels with the button still\n"
  + "  * pressed, then it becomes a drag and StartDrag is called. If you\n"
  + "  * override this method, you can either cancel the drag by returning\n"
  + "  * null (without calling super), or else substitute another (possibly\n"
  + "  * new) widget for (in the same location as) the clicked widget when you\n"
  + "  * call super, and the replacement will be what is dragged. If you do\n"
  + "  * this, there is some additional code needed in your MouseUp override\n"
  + "  * so that the Dropped event works correctly.\n  *\n"
  + "  * The vertical and horizontal offsets are relative to the top-left\n"
  + "  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get\n"
  + "  * the position relative to the window.\n  *\n"
  + "  * @param  whom  The widget that was clicked\n"
  + "  * @param  vert  The vertical offset into this widget where it\n"
  + "  *                 was clicked\n"
  + "  * @param  horz  The horizontal offset into this widget\n  *\n"
  + "  * @return       The widget being dragged, or null if cancelled\n  */\n"
  + "  // public GameWgt StartDrag(GameWgt whom, int vert, int horz) {\n"
  + "  //   // subclass is allowed to substitute another wgt for whom,\n"
  + "  //   //   then call super\n"
  + "  //   if (dontWantToStartDrag) return null;\n"
  + "  //   return super.StartDrag(whom,vert,horz);} //~StartDrag\n\n /**\n"
  + "  * After a StartDrag has been called for a Dragable widget and the user\n"
  + "  * continues to move the mouse, Dragging is called for the same widget.\n"
  + "  * You can override this method to do something with the dragged widget's\n"
  + "  * new position. Dragging in the base class updates the dragged widget's\n"
  + "  * position, so you must call super before you try to use that position.\n"
  + "  *\n"
  + "  * The vertical and horizontal offsets are relative to the top-left corner\n"
  + "  * of this widget before its position has been adjusted by super.\n  *\n"
  + "  * To get the position relative to the window, you should calculate\n"
  + "  * whom.AddPrntsPos(whom.GetPosn()) before calling super.\n  *\n"
  + "  * @param  whom  The widget that is being dragged\n"
  + "  * @param  vert  The vertical offset into this widget before it moves\n"
  + "  * @param  horz  The horizontal offset into this widget\n  */\n"
  + "  // public void Dragging(GameWgt whom, int vert, int horz) {\n"
  + "  //   super.Dragging(whom,vert,horz);\n  //   // DoWhatever();\n"
  + "  //   } //~Dragging\n\n /**\n"
  + "  * After a StartDrag has been called for a Dragable widget and the user\n"
  + "  * releases the mouse button, Dropped is called for the same widget.\n"
  + "  * If you override this method and do something with the dropped widget\n"
  + "  * that could invalidate further drags on it using the arrow keys, you\n"
  + "  * should also call BlurEv for it, or else make sure it cannot take\n"
  + "  * focus. The base class does nothing with Dropped, so it is not\n"
  + "  * necessary to call super.\n  *\n"
  + "  * @param  whom  The widget that was dragged\n  */\n"
  + "  // public void Dropped(GameWgt whom) {} //~Dropped\n\n /**\n"
  + "  * MouseUp is called for a widget that is \"Dragable\" when the user\n"
  + "  * releases the mouse button over the same widget that was clicked or\n"
  + "  * dragged. If the mouse is moved away from the clicked widget (and\n"
  + "  * there is no drag in process) then MouseUp is not called (and there\n"
  + "  * is no consequent ClickEvt).\n  *\n"
  + "  * The vertical and horizontal offsets are relative to the top-left\n"
  + "  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get\n"
  + "  * the position relative to the window.\n  *\n"
  + "  * @param  whom  The widget that was clicked\n"
  + "  * @param  vert  The vertical offset into this widget where it\n"
  + "  *                 was clicked\n"
  + "  * @param  horz  The horizontal offset into this widget\n  */\n"
  + "  // public void MouseUp(GameWgt whom, int vert, int horz) {\n"
  + "  //   if (didSomethingWithIt) return;\n"
  + "  //   super.MouseUp(whom,vert,horz);} //~MouseUp\n\n /**\n"
  + "  * This is the main() you start your game in.\n"
  + "  * DO NOT MODIFY THIS CODE.\n  */\n"
  + "  public static void main() { // (in UserGame)\n"
  + "    JavaGame whom = new JavaGame();\n"
  + "    GameEvent evh = new UserGame();\n"
  + "    String title = \"UserGame\", data = \"\";\n"
  + "    int tall = 0, wide = 0, colo = 0x0;\n"
  + "    if (whom != null) if (evh != null) {\n"
  + "      data = evh.GameList(whom);\n"
  + "      tall = JavaGame.SafeParseInt(JavaGame.NthItemOf(' ',5,data));\n"
  + "      wide = JavaGame.SafeParseInt(JavaGame.NthItemOf(' ',6,data));\n"
  + "      colo = JavaGame.SafeParseInt(JavaGame.NthItemOf(' ',7,data));\n"
  + "      whom.StartGame(tall,wide,colo,title,evh);\n"
  + "      whom.LogAlWgts(\"[main] \");}\n    System.out.println("
  + "\"{main} \" + tall + \"x\" + wide + \" '\" + title + \"'\");}\n\n"
  + " /**\n  * Constructor // RevDate = \"20/7/25\"\n  */\n"
  + "  public UserGame() {\n"
  + "    System.out.println(\"(new UserGame)\");}} //~UserGame\n";

  static GameWgt UserGameBd = null; // (unused)

  private int[] HitWgts = null; // (unused yet)
  private int[] StatWgts = null; // (unused yet)

 /**
  * WgtMacros is the (internal) data used to build complex widget structures
  * from simple user specifications. Normally we use a predefined array, but
  * this can be overridden by a specification included in the user data.
  */
  private int[] WgtMacros = Wgt_Macros;

 /**
  * UserEvHandler is set when your main() program calls StartGame, and used
  * everywhere to call your overridden methods.
  */
  private GameEvent UserEvHandler = null; // call InstallMyEvts to set this

 /**
  * FindSubWgt is not particularly speedy, but sometimes different places
  * need the same subwidget, so a successful find is cached so it can be
  * returned immediately if asked for again.
  *
  * GameEngine allows for multiple lists of widgets under the general game
  * board; SpriteList is defined to be a the first of them, and for speed
  * only its contents are considered for collision testing.
  *
  * Only one widget or group can have focus at a time, and "marching ants"
  * around it give a visual indication of focus. In the case of (rectangular)
  * groups, mAntsHost remembers who is currently showing ants, and mAntsWgt
  * is a reference to the ants group that is moved around from host to host.
  * Sprites each have their own ants data, customized to the shape of the
  * respective sprite.
  *
  * When your game is specified to use alert and input dialogs, theAlertDlg
  * and theInputDlg are references to the structures that display these
  * dialogs in your game window; DlgInputData is a reference to to the
  * actual input GameTxLn widget that accepts the user typed-in text, for
  * easy access by the code that processes that input.
  *
  * GameWgtsList is the top-level group of groups of widgets. When not
  * otherwise active, the marching ants group mAntsWgt is stored there.
  * Its first subwidget should be the SpriteList.
  */
  private GameWgt CashFindWgt = null, CashFoundWgt = null,
      SpriteList = null, mAntsHost = null, mAntsWgt = null,
      theAlertDlg = null, theInputDlg = null, DlgInputData = null,
      GameWgtsList = null, GameRoot = null, DontBump = null;

   /**
  * Dragging and rollover processing take time even if not currently
  * active, so if your game does not need one or the other of these,
  * the extra processing can be eliminated. The booleans WantsRollover
  * and WantsDrag remember that.
  *
  * The boolean flags ValidMouse and MouseIsDown remember the current
  * state of the mouse from frame to frame.
  */
  private boolean WantsRollover = false, WantsDrag = false,
      ValidMouse = false, MouseIsDown = false, KillDrag = false,
      PendingDraw = false, unSavedWgts = false;

   /**
  * MouseVert and MouseHorz are window-relative coordinates of the mouse
  * as of the most recent MouseDispatch.
  *
  * MonoFontNo,TextFontNo,TinyFontNo,BoldFontNo,ItalFontNo,HeadFontNo,
  * and NanoFontNo are the widget reference numbers for the respective
  * fonts in the current game environment. You can get them in GameMaker
  * using:
  *
  *      ChoseFont(null,code)
  *
  * where code is the ASCII code for the first letter of the name (except
  * TinyFontNo is 'Y' = 89).
  */
  private static int MouseVert = 0, MouseHorz = 0, LocalVert = 0,
      MonoFontNo = 0, TextFontNo = 0, TinyFontNo = 0, BoldFontNo = 0,
      ItalFontNo = 0, HeadFontNo = 0, NanoFontNo = 0, ScorFontNo = 0,
      TimerTix = 0, TimeBase = 0, DlogIsOpen = 0,
      InputDlgNo = 0, CashFindWhy = 0, Floatem = 0, InputWhy = 0;

   /**
  * theImpoList is saved from the list of widgets returned by your GameList
  * method. It is used in FindListWgt to search for widgets by descriptor.
  *
  * JSinfo and RecentJava are used in trnslating your game file to JavaScript.
  */
  private String theImpoList = "", JSinfo = "", RecentJava = "";

  /**
  * SpritePause is called by sprite widgets to see if they should pause
  * their motion (like when an alert or input dialog is open).
  *
  * @return         true to pause
  */
  public static boolean SpritePause() {
    if (!StopSprites) return false;
    return (DlogIsOpen>0);} //~SpritePause

  /**
  * GetGameValue is an access method to get one of the private data items
  * from outside the JavaGame game instance.
  *
  * @param  whom    A selector, 0..10
  *
  * @return         the selected value
  */
  public static int GetGameValue(int whom) {
    switch (whom) {
    case 0:
      whom = MonoFontNo;
      break;
    case 1:
      whom = TextFontNo;
      break;
    case 2:
      whom = TinyFontNo;
      break;
    case 3:
      whom = BoldFontNo;
      break;
    case 4:
      whom = ItalFontNo;
      break;
    case 5:
      whom = HeadFontNo;
      break;
    case 6:
      whom = NanoFontNo;
      break;
    case 7:
      whom = LocalVert;
      break;
    case 8:
      whom = MouseVert;
      break;
    case 9:
      whom = MouseHorz;
      break;
    case 10:
      whom = DlogIsOpen;
      break;
    case 11:
      whom = InputDlgNo;
      break;} //~switch
    return whom;} //~GetGameValue

 /**
  * Returns a (language-neutral) new int array of the specified length.
  * (Because the syntax is different in Java vs JavaScript)
  * Java = JS syntax for learning the length: myarray.length
  *
  * @param lxx  The length desired
  *
  * @return     The array new int[lxx]
  */
  public static int[] NewArray(int lxx) {
    return new  int[lxx];} //~NewArray

  /**
  * Converts a mouse click position from window-relative coordinates to
  * be relative to the widget clicked in. The horizontal offset is returned
  * directly, the vertical offset is cached in LocalVert for later recovery.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset relative to the window
  * @param  horz  The horizontal offset relative to the window
  *
  * @return       The horizontal offset into this widget
  */
  public int LocalizePos(GameWgt whom, int vert, int horz) { // frm MouseDisp
    int zx, whar = (vert<<16)+(horz&0xFFFF);
    LocalVert = 0;
    if (whom==null) return 0;
    whar = AddPair(whar,PairNeg(whom.AddPrntsPos(whom.GetPosn())));
    LocalVert = whar>>16;
    return SignExtend(whar);} //~LocalizePos

 /**
  * A safe (no exceptions) way to get the size of the game window.
  *
  * @return       Its (packed) (height<<16)+width
  */
  public static int GameWinSize() { // rtns packed window size
    return  WinHi*0x10000
        +  WinWi;} //~GameWinSize

  /**
  * GetIxWgt converts a widget reference number into a reference to that widget.
  *
  * @param  ix    an index into the AllWgts array = RefNum
  *
  * @return       the widget with that RefNum
  */
  public static GameWgt GetIxWgt(int ix) {
    if (AllWgts != null)
    if (ix<MaxWgts) if (ix>0) return AllWgts[ix];
    return null;} //~GetIxWgt

/**
  * A language-neutral way to get a (32-bit) time in milliseconds.
  * The first time this is called the caller sets TimeBase (initially =0),
  * so that all subsequent calls are based off that start value,
  * despite being less precision than normal Java.
  *
  * @return       The number of milliseconds since game start
  */
  public static int GetMilliSecs()  {
    return ((int)System.currentTimeMillis())  -TimeBase;} //~GetMilliSecs

/**
  * A safe (no exceptions) way to extract an integer number from a string.
  * If the number begins with "0x" then it is assumed to be hexadecimal.
  * Leading white space is ignored, and the first character that is not
  * part of the number ends the scan.
  *
  * @param  aStr  The string from which to extract that character
  *
  * @return       The number parsed, or 0 if none
  */
  public static int SafeParseInt(String aStr)  {
    int nCh, here, lxx = StrLength(aStr), valu = 0, mode = 0, seen = 0;
    for (here=0; here <= lxx-1; here++) {
      nCh = (int)(CharAt(here,aStr));
      if (nCh <= 0) break;
      if (valu==0)  {
        if (nCh <= 32)  {
          if (seen>0) break;
          continue;} //~if
        if (nCh==45)  { // '-'
          if (seen>0) break;
          if (mode !=0) break;
          mode++;
          continue;} //~if
        if ((nCh|0x20)==120)  { // 'x'
          if (mode<0) break;
          if (seen !=1) break;
          mode = mode-4;
          continue;}} //~if
      if (nCh<48) break; // '0'
      if (mode<0)  {
        if (nCh>96) nCh = nCh&0x5F; // convert lower case to caps
        if (nCh>70) break; // 'F'
        if (nCh>57) if (nCh<65) break; // '9'..'A'
        valu = valu*16+(nCh&15);
        if (nCh>57) valu = valu+9;} //~if // '9'
      else if (nCh <= 57)  {
        valu = valu*10+(nCh&15);
        seen++;} //~if
      else break;} //~for
    if ((mode&3)>0) valu = -valu;
    return valu;} //~SafeParseInt

  /**
  * A safe (no exceptions) way to count the number of (non-overlapping)
  * occurrences of a particular string within a larger string.
  * An empty string cannot be counted.
  *
  * @param  aWord The string to look for
  * @param  aStr  The string in which to find and count it
  *
  * @return       The number of occurrences found, or 0 if none
  */
  public static int Countem(String aWord, String aStr)  {
    int lxx = StrLength(aWord), whar, thar = 0, nby = 0;
    if (lxx==0) return 0;
    try {  while (true)  {
        whar = aStr.indexOf(aWord,thar); // JS=Java
        if (whar<0) break;
        thar = whar+lxx;
        nby++;} //~while
    } catch (Exception ex)  {}
    return nby;} //~Countem

  /**
  * Replace a substring.
  *
  * @param  nuly     The replacement string
  * @param  ofx      The starting offset to put it at
  * @param  sz       The length of text to replace
  * @param  theText  The string in which to make the substitutions
  *
  * @return          theText as changed
  */
  public static String Replace(String nuly, int ofx, int sz, String theText)  {
    return Substring(0,ofx,theText) + nuly + RestOf(ofx+sz,theText);} //~Replace

  /**
  * A safe (no exceptions) way to extract an item (word/line) from a string.
  * When getting words, excess white space is ignored, so the returned
  * string is empty only when requesting a word beyond the last. Otherwise
  * consecutive delimiters result in empty items being returned, so that
  * you always get the nth item of a multi-item string. Items past the end
  * or before the front (0 or less) are empty.
  *
  * @param  delim  Use '\n' to get a whole line, ' ' to get a word
  * @param  whom   The (1-base) item number to get
  * @param  aStr   The string from which to extract that item
  *
  * @return       The extracted string
  */
  public static String NthItemOf(char delim, int whom, String aStr)  {
    char xCh;
    int nCh = (int)(delim), lxx = StrLength(aStr), here, aCh;
    boolean xact = true, wordy = false, seen = false;
    String aWord = "";
    if (nCh >= 10) if (nCh <= 32)  {
      if (nCh == 32) wordy = true;
      xact = false;} //~if
    whom--; // comes in 1-based, now 0-base
    try {  if (whom >= 0) for (here=0; here <= lxx-1; here++) {
      xCh = CharAt(here,aStr);
      aCh = (int)(xCh);
      if (aCh <= 0) break;
      if (xact)  {
        if (aCh == nCh)  {
          whom--;
          if (whom<0) break;
          continue;}} //~if
      else if (aCh <= 32)  {
        if (wordy)  { // one or more spaces..
          if (seen) whom--;
          seen = false;
          if (whom<0) break;
          continue;} //~if
        else if (aCh < 32)  { // any ctl counts as \n or \r
          whom--;
          if (whom<0) break;
          continue;}} //~if
      else seen = true;
      if (whom <= 0) aWord = aWord + xCh;} //~for
    } catch (Exception ex)  {}
    return aWord;} //~NthItemOf

 /**
  * Converts an integer to String (for JS, Java not always auto'ly coerces).
  *
  * @param  whom  The integer
  *
  * @return       The same number as String
  */
  public static String CvInt2Str(int whom)  { // 'cuz JS is untyped
    return "" + whom;} //~CvInt2Str

/**
  * A safe way to read a whole text file.
  * The line-ends in the input file are converted to '\n' regardless of
  * how they are on the file.
  *
  * @param  filename The name of the text file
  *
  * @return          The text as read
  */
  public static String ReadWholeTextFile(String filename)  {
    char aCh;
    int nx, lxx = 0, nuly = 0; // adapted from what I found on StackOverflow..
    String content = "";
    File myFile = new File(filename); // for ex foo.txt
    FileReader reader = null;
    try {
      reader = new FileReader(myFile);
      lxx = (int)myFile.length();
      if (lxx <= 0) return "";
      char[] chars = new char[lxx];
      reader.read(chars);
      for (nx=0; nx <= lxx-1; nx++) { // convert Windoze line ends to internal..
        aCh = chars[nx];
        if (aCh == '\r')  {
          if (nx+1<lxx) if (chars[nx+1] == '\n') continue;
          aCh = '\n';}
        chars[nuly] = aCh;
        nuly++;} //~for
      content = new String(chars);
      if (nuly<lxx) content = content.substring(0,nuly);
      reader.close();
      reader = null;} //~try
    catch (Exception ex)  {}
    finally {
      try {if (reader !=null) reader.close();}
      catch (Exception ex)  {}} //~finally
    return content;} //~ReadWholeTextFile

 /**
  * A safe way to write a whole text file.
  *
  * @param  filename The name of the text file
  * @param  content  The text to be written
  *
  */
  public static void WriteWholeTextFile(String filename, String data)  {
    int lino, nx, lxx = data.length(), sofar = 0;
    String aLine;
    FileWriter myFile = null;
    BufferedWriter whom = null;
    PrintWriter doit = null;
    try {
      // data = ReplacAll("\r\n","\n",data); // unneeded
      myFile = new FileWriter(filename);
      whom = new BufferedWriter(myFile);
      doit = new PrintWriter(whom);
      for (lino=1; lino <= 999999; lino++) { // not the most efficient way to do it, but..
        if (sofar >= lxx) break; // normal exit at end
        aLine = NthItemOf('\n',lino,data);
        nx = aLine.length();
        if (nx>0) doit.print(aLine);
        sofar = sofar+nx;
        if (sofar >= lxx) break; // no \n on final line? OK
        doit.println();
        sofar++;} //~for // (lino)
      doit.close();
      doit = null;} //~try
    catch (Exception ex)  {}
    finally {
      try {if (doit !=null) doit.close();} //~try
      catch (Exception ex)  {}}} //~finally //~WriteWholeTextFile

 /**
  * A safe (no exceptions) way to compare two strings.
  * Works like compareTo but also works properly with null strings.
  *
  * @param  aStr  The first string to compare
  * @param  xStr  The other string to compare it to
  *
  * @return       -1/0/+1 as aStr<xStr or = or >
  */
  public static int SafeCompare(String aStr, String xStr)  {
    int whom = 0, az = 0, xz = 0;
    try {
    if (xStr != "") xz++;  if (aStr != "")  az++;
    if (az==0)  {if (xz>0) whom--;}
    else if (xz==0) whom++;
      else whom = aStr.compareTo(xStr);
    } catch (Exception ex)  {whom = 0;}
    return whom;} //~SafeCompare

 /**
  * Find the most significant non-zero bit in a 32-bit integer.
  * The least significant bit is easily extracted by (whom&-whom) but there
  * is no similarly trivial way to get the other end.
  *
  * @param  whom  an integer
  *
  * @return       the most significant bit of that number, or =0 if none
  */
  public static int TopBit(int whom)  { // rtns most significant non-zero bit
    int zx;
    while (true)  { // exit when found it
      if (whom==0) break; // none
      if (whom<0) whom = 0x80000000; // high bit
      zx = whom&-whom; // = its low bit, done when =
      if (zx==whom) break; // got it
      whom = whom-zx;} //~while
    return whom;} //~TopBit

 /**
  * A simple way to get a date+time string.
  *
  * @param  noDate   True: omit date
  *
  * @return          The date+time string
  */
  public static String TimeStamp(boolean noDate)  {
    String why = IffyStr(noDate,"HH:mm:ss","yy MMM dd, HH:mm:ss");
    DateFormat sdf = new SimpleDateFormat(why);
    Date now = new Date();
    return sdf.format(now);} //~TimeStamp

// Game Engine Code . . . . . . . . . . . . . . . . . . . . . . . . . . . .

 /**
  * This utility turns 4-letter text into four letters packed into an int,
  * if you have a GameTxLn widget to borrow for the purpose.
  *
  * @param  whom    the widget
  * @param  aStr    the string
  *
  * @return         the packed integer
  */
  public int PuTxGetData(GameWgt whom, String aStr) { // only frm MakeJSwgt
    int info;
    if (whom==null) return 0;
    info = whom.PutTextLn(aStr);
    return whom.GetIxData(0,0);} //~PuTxGetData

   /**
  * Safely pick out a single pixel from an icon pixel array.
  *
  * @param  rx     the pixel row, 0 is top
  * @param  cx     the pixel column, 0 is left edge
  * @param  pix    the pixel array
  *
  * @return        true if the pixel is opaque
  */
  public static boolean PeekIconPix(int rx, int cx, int[] pix) {
    if (rx<0) return false;
    if (cx<0) return false;
    if (rx>31) return false;
    if (pix == null) return false;
    if (cx >= pix.length) return false;
    return ((pix[cx]>>rx)&1) !=0;} //~PeekIconPix

   /**
  * Safely insert a single pixel into an icon pixel array.
  *
  * @param  abit   true to set the pixel opaque, false transparent
  * @param  rx     the pixel row, 0 is top
  * @param  cx     the pixel column, 0 is left edge
  * @param  pix    the pixel array
  */
  public static void PokeIconPix(boolean abit, int rx, int cx, int[] pix) {
    int zx = 0, mask;
    if (rx<0) return;
    if (cx<0) return;
    if (rx>31) return;
    if (pix == null) return;
    if (cx >= pix.length) return;
    if (abit) zx--;
    mask = 1<<rx;
    pix[cx] = pix[cx]&~mask|zx&mask;} //~PokeIconPix

  /**
  * A safe (no exceptions) non-OOPS way to convert a text number into
  * a fixed-point value with eights bits to the right of the binary point.
  *
  * @param  whom  a string containing a number
  *
  * @return       the same number as a fixed-point integer, or =0 if fails
  */
  public static int ParsFixPt(String whom) {
    int rez = 0, lxx = 0, sgn = 0, whar; // rtns 8-bit each, int+fraction
    String msg = whom; // for log
    whom = NthItemOf(' ',1,whom);
    if (Logging) msg = "  (ParsFxPt) '" + msg
        + IffyStr(StrLength(whom)==StrLength(msg),"'","' => " + whom);
    whar = NthOffset(0,"-",whom);
    if (whar >= 0) {
      whom = RestOf(whar+1,whom);
      if (Logging) msg = msg + " -" + whar + "/" + whom;
      sgn--;} //~if
    whar = NthOffset(0,".",whom);
    if (whar >= 0) {
      if (whar>0) rez = SafeParseInt(Substring(0,whar,whom))<<8;
      whom = RestOf(whar+1,whom);
      lxx = StrLength(whom);
      if (Logging) msg = msg + " ." + whar + "/" + whom + " =" + rez + " +" + lxx;
      if (lxx>4) {
        whom = Substring(0,4,whom);
        lxx = 4;} //~if
      whar = SafeParseInt(whom)<<8;
      if (Logging) msg = msg + SeeHex(" x",whar," ");
      if (lxx==4) whar = (whar+5000)/10000;
      else if (lxx==3) whar = (whar+500)/1000;
      else if (lxx==2) whar = (whar+50)/100;
        else whar = (whar+5)/10;
      if (Logging) msg = msg + " => " + whar;
      rez = rez+whar;} //~if
    else rez = SafeParseInt(whom)<<8;
    if (rez>0x8000) rez = rez&0x7FFF;
    else if (sgn==0) if (rez==0x8000) rez--;
    if (Logging)
      System.out.println(msg + SeeHex(IffyStr(sgn<0," = -x"," = x"),rez," "));
    if (sgn<0) rez = -rez;
    return rez;} //~ParsFixPt

  /**
  * A safe (no exceptions) way get the reference number from a (possibly null)
  * widget. If it is null, SafeWgtNum returns zero.
  *
  * @param  whom  the widget, or null if none
  *
  * @return       that widget's RefNum, or =0 if none
  */
  public static int SafeWgtNum(GameWgt whom) { // returns RefNum or 0
    if (whom==null) return 0;
    return whom.GetRefNum();} //~SafeWgtNum

  /**
  * LogWgtTree prints out to the console log an (indented) tree representation
  * of a specified widget with all its sub-widgets. It calls itself recursively
  * with deeper indentations for the sub-widgets.
  *
  * @param  whom   the top-level widget to log
  * @param  deep   the initial indentation, usually 0 or 1
  *
  * @return        an error code > why, or else =0 if no error
  */
  public static void LogWgtTree(GameWgt whom, int deep) {
    int nx, tye;
    String why;
    GameWgt nuly;
    if (whom==null) return;
    tye = whom.GetType();
    why = Substring(0,deep+deep,". . . . . . . . . . . . . . . . . . . . ");
    whom.LogMe(why);
    deep++;
    if (deep<20) if (((tye-5)&255)<3) for (nx=0; nx <= 99; nx++) {
      nuly = GetIxWgt(whom.GetIxData(nx,0));
      if (nuly==null) break;
      LogWgtTree(nuly,deep);}} //~LogWgtTree

  /**
  * Create a new widget with specified properties, or copy an existing widget,
  * or apply the properties to an existing widget.
  *
  * If the specified type is negative and the given widget is not null, then
  * that widget is mined for the type and all the parameter values given as 0.
  * In particular, a reference to the same data array is used in the copy,
  * unless a non-null data is supplied.
  *
  * If the type (without flags) is greater than 7, then the only way to
  * create a widget of that type is to create it yourself, then (optionally)
  * call NewGameWgt to populate the base class instance variables in it.
  *
  * @param  tye    the type of the new widget, or -1 to clone whom
  * @param  top    the vertical component of the top-left corner
  * @param  left   the horizontal component of the top-left corner
  * @param  tall   the vertical height
  * @param  wide   the horizontal width
  * @param  prnt   the parent reference number, or 0 if none
  * @param  color  the widget color, or -1 if none
  * @param  info   the information value
  * @param  data   an array to use for its Gdata, or null if none
  * @param  whom   a widget to copy or update, or null if none
  *
  * @return        the created widget, or null if it failed
  */
  public GameWgt NewGameWgt(int tye, int top, int left, int tall, int wide,
      int prnt, int color, int info, int[] data, GameWgt whom) {
    GameWgt[] theList = AllWgts;       // ** need equivalent JS code **
    int kx = 0, zx, nx = tye&255, rnum = -1;
    if (tye<0) if (whom != null) { // clone an existing wgt..
      tye = whom.GetType();
      nx = tye&255;
      zx = whom.GetPosn();
      if (top==0) top = zx>>16;
      if (left==0) left = SignExtend(zx);
      zx = whom.GetSize();
      if (tall==0) tall = zx>>16;
      if (wide==0) wide = zx&0xFFFF;
      if (color==0) color = whom.GetColor();
      if (info==0) info = whom.GetInfo();
      if (prnt==0) prnt = whom.GetParent();
      if (data == null) {
        data = whom.SwapData(data);
        if (whom.SwapData(data) != null) zx++;} //~if
      if (nx<8) whom = null;} //~if
    if (whom==null) switch (nx) {
      case 0: // static rect
      case 1: // pix image
      case 5: // wgt list
        whom = new GameWgt();
        break;
      case 2: // 1-color icon
        whom = new GameIcon();
        break;
      case 3: // font
        whom = new GameFont();
        break;
      case 4: // text line
        whom = new GameTxLn();
        break;
      case 6: // sprite
        whom = new GameSprit();
        break;
      case 7: // scrolled (text)
        whom = new GameScro();
        break;} //~switch
    if (whom==null) return null;
    if (theList == null) return null;
    rnum = whom.GetRefNum();
    if (rnum<0) return null;
    if (rnum >= MaxWgts) return null;
    AllWgts[rnum] = whom;
    whom.SetWgType(tye); // only works first time
    whom.SetPosn(top,left);
    whom.SetSize(tall,wide);
    whom.SetInfo(info);
    whom.SetParent(prnt);
    whom.SetColor(color);
    kx = whom.GetType();
    if (kx==tye) kx = 0;
      else whom.SetFlags(tye);
    if (tye>0) data = whom.SwapData(data);
    if (Logging) System.out.println("..nuGmWgt #" + rnum + "=" + tye
          + IffyStr(kx>0,"/" + kx + " "," ") + top + "/" + left + " " + tall
          + "/" + wide + SeeHex(" c=",color,whom.FmtInfo(" i=",tye))
          + " p=" + prnt+ " " + (data != null));
    return whom;} //~NewGameWgt

  /**
  * Recursively copy a widget with all its sub-widgets. This recognizes tool
  * widgets as a special case, so that these differences happen:
  *   (a) the parent is changed to Sprite list,
  *   (b) the top coordinates are globalized,
  *   (c) an unparented text (tool name) is preserved, and
  *   (d) the ID info is changed to `**DW`
  *
  * @param  whom   a widget to copy
  * @param  prnt   a parent reference number, or 0 to copy it from whom
  *
  * @return        the copied widget group, or null if it failed
  */
  public GameWgt CloneWgt(GameWgt whom, int prnt) { // rec'ly copy, except tool
    // tool=T: (a) change parent to Sprite list, (b) globalize top coords,
    // .. (c) preserve unparented text (tool name), and (d) change ID to `xxDW`
    // keep=T: preserve Gdata unchanged, otherwise assume it's wgts & clone them
    int deep = (prnt&(-0x10000))+0x10000, prep = prnt&0xFFFF, prio = prep,
        here, thar, info, zx, nx = 0, kx = 0, yx = 0, top = 0, left = 0,
        tye = 0, seen = 0, lxx = 0, why = 0;
    boolean tool = false, keep = false;
    int[] data = null;
    GameWgt nuly = null, whar = SpriteList;
    prnt = prep;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      nx = whom.GetRefNum();
      info = whom.GetInfo();
      // kx = whom.GetParent();
      tye = whom.GetType();
      yx = tye&255;
      if (deep>0x20020000) deep = deep-0x20000000; // below 1st-level tool
      if (yx<4) keep = true;  // special-case text parent inside tool..
      else if (yx==4) if (deep>0x20000000) { // if (kx != prnt) // prnt always !=
        if (info==0) keep = true; // no font: keep same tool name, else copy it..
        else if (prnt>0) if (info==TextFontNo) if (tye>yx) {
          prio = 0; // (-> nGmWg) this text is list label, so keep parent
          why = why+8;
          prnt = 0;}} //~if
      if (yx==5) if ((info&0xFFFF) == 0x4454) tool = true; // 'DT' = Drag Tool
      lxx = whom.GetIxData(-1,-1); // gets Gdata size
      why++; // why = 2
      if (lxx>0) if (!keep) if ((yx&-4)==4) // clone data array..
        data = new int[lxx];
      why++; // why = 3 => 5
      if (tool) {
        why++; // why = 4 => 6
        // if (kin==0x43724454) tye = tye|256; // 'CrDT' sprite tool already bump
        top = whom.AddPrntsPos(whom.GetPosn());
        left = SignExtend(top);
        top = top>>16;
        if (whar != null)
          prio = whar.GetRefNum(); // chg tool parent 2B sprite list
        deep = deep|0x20000000;
        info = info|7;} //~if
      else whar = null;
      nuly = NewGameWgt(-1,top,left,0,0,prio,0,info,data,whom);
      why = why+2; // why = 5/6
      if (nuly==null) break;
      prio = nx;
      nx = nuly.GetRefNum();
      kx = nuly.GetType();
      if (kx==tye) kx = 0;
        else nuly.SetFlags(tye);
      if (yx==4) if (prnt==0) if (deep>0x20000000) { // give it same parent,
        prnt = nuly.GetParent();
        whar = GetIxWgt(prnt); // ..but only if whom is there..
        if (whar != null) if (whar.FindElemt(prio)<0) {
          why = why|16;
          whar = null;}} //~if
      if (whar != null) zx = whar.AddElemt(nx); // gotta dele if cancel'd // ***
      if (data != null) // now clone array contents..
        for (lxx=lxx-1; lxx >= 0; lxx += -1) {
          zx = whom.GetIxData(lxx,0);
          if (zx>0) if (yx>4) {
            whar = GetIxWgt(zx); // don't clone fontless text inside tool..
            if (whar != null) if (((whar.GetType()&255) !=4)
                || (whar.GetInfo() !=0) || (deep<0x20000000)) {
              whar = CloneWgt(whar,nx+deep);
              if (whar != null) zx = whar.GetRefNum();}} //~if
          if (lxx >= 0)
            if (lxx < data.length)
              data[lxx] = zx;} //~for // (lxx)
      why = -why;
      break;} //~while
    if (Logging || (why>0)) System.out.println("  (ClonWgt #" + nx + "/" + prio
        + SeeHex(") t=x",tye, IffyStr(kx>0,SeeHex("/",kx," p=")," p=")) + prnt
        + "/" + prep + " = " + why + "/" + (deep>>16));
    return nuly;} //~CloneWgt

  /**
  * (Temporary) There are five fixed circle sizes that GameEngine knows
  * how to do. Real Soon Now, we should have code to generate all sizes
  * between 4 (the minimum size that can be distinguished from a rectangle)
  * and 32 pixels in diameter (the maximum that Game Engine can do a single
  * icon for). Until then you get 4-,5-,12-,13-, and 16-pixel diameters.
  * This function picks the size closest to what you ask for.
  *
  * @param  whom  the requested circle diameter, in pixels
  *
  * @return       the implemented diameter closest to whom
  */
  public static int KnownCircleSz(int whom) {
    if (whom >= 15) whom = 16; // rtns nearest supported size
    else if (whom<4) whom = 4;
    else if (whom>12) whom = 13;
    else if (whom>7) whom = 12;
    else if (whom>5) whom = 5;
    return whom;} //~KnownCircleSz

  /**
  * (Temporary) Given a valid circle diameter, this returns an array
  * of pixels for a circle of that diameter.
  *
  * @param  whom  the requested circle diameter, in pixels
  * @param  holo  true if the circle should be outline only
  *
  * @return       the pixel array for this diameter
  */
  public static int[] KnownCircleData(int whom, boolean holo) {
    if (holo) {
      if (whom >= 15) return Cir16h;
      if (whom <= 4) return Circ6h;
      if (whom>12) return Cir13h;
      if (whom>7) return Cir12h;
      return Circ7h;} //~if
    if (whom >= 15) return Cir16f;
    if (whom <= 4) return Circ4f;
    if (whom>12) return Cir13f;
    if (whom>7) return Cir12f;
    return Circ5f;} //~KnownCircleData

  /**
  * Given a widget that is either a circle icon or a group over a circle
  * icon, and a preferred diameter, choose the best circle size and update
  * the widget (and its sub-widget, if so) to be that size, with a data array
  * to match the selected size.
  *
  * BestCircData is recursive and private, it is expected that clients
  * will call CircleSize and not BestCircData directly.
  *
  * @param  whom   the widget to update
  * @param  tall   the desired diameter
  * @param  ins    if >0, the size of a space to center the widget in
  *
  * @return        the chosen size
  */
  private int BestCircData(GameWgt whom, int tall, int ins) {
    boolean uppy = tall<0; // tall: desired size (<0 force), info is best fit
    int prio = tall, info = KnownCircleSz(Math.abs(tall)),  // only frm CircleSiz
        cx = 0, kx = 0, rx = 0, yx = 0, zx = 0, nx = 0, diam = 0, why = 0;
    GameWgt nuly;   // ins is square to center it in, =0 if no indent; rtn diam
    int[] data = null;
    while (true) {                                  // only colors != B&W
      why++; // why = 1
      if (whom==null) break;
      nx = whom.GetRefNum();
      zx = whom.GetInfo()&0x7F00FFFF;
      rx = zx&0x7E00FFFC;
      kx = whom.GetType()&255;
      why++; // why = 2
      if (kx==6) { // recurse sprite..
        nuly = GetIxWgt(whom.GetIxData(0,0));
        if (nuly==null) break;
        yx = nuly.GetType()&255;
        if (yx !=2) break;
        yx = BestCircData(nuly,tall,0);
        if (yx>0) whom.SetSize(yx,yx);
        why = -why; // why = -2
        break;} //~if
      why++; // why = 3
      if (kx==5) { // recurse group for ants sublist..
        for (kx=0; kx <= 99; kx++) {
          nuly = GetIxWgt(whom.GetIxData(kx,0));
          if (nuly==null) break;
          yx = nuly.GetType()&255;
          if (yx !=2) if (yx !=5) continue;
          yx = BestCircData(nuly,tall,0);
          if (yx>diam) diam = yx;} //~for
        if ((zx&0x7E00FFFF) == 0x42004454) { // only 'C*DT' or 'B*DT', not *DW'
          yx = 0; // ..the draggable group has (large) offset inside container
          why++; // why = 4
          if (ins>diam) yx = (ins-diam)>>1;
          whom.SetPosn(yx,yx);} //~if
        why = -why; // why = -3/-4
        break;} //~if
      why++; // why = 4     // ** gotta rewrite this for full range of radii..
      tall = info; // new nominal
      kx = tall-4;
      if (kx<0) kx = 0;
      else if (kx>9) kx = kx+7;
      // zx = zx&0x7F00FFFF; // (not many places to put this..)
      if (rx == 0x42004454) { // 'C*DT' or 'B*DT'
        rx = ((kx+48)<<16)+zx;
        whom.SetInfo(rx);} //~if
      zx = info;
      if (zx<7) zx = zx+2; // zx: the icon size (ants outside smaller circles)
        // else info--;
      diam = zx;
      if (!uppy) if ((whom.GetSize()>>16) != zx) uppy = true;
      if (uppy) {
        whom.SetSize(zx,zx);} //~if
      cx = whom.GetColor();
      info = (info>>2)&6|info&1; // 16->4, 13->3, 12->2, 5->1, 4->0
      if (cx==0) zx = info+5; // black -> hollow
      else if (cx==0xFFFFFF) zx = info+10; // white -> ants
        else zx = info;
      why++; // why = 4
      info = 0;
      switch (zx) {
      case 0: // size=4, solid..
        data = Circ4f;
        break;
      case 1: // size=5, solid..
        data = Circ5f;
        break;
      case 2: // size=12, solid..
        data = Cir12f;
        break;
      case 3: // size=13, solid..
        data = Cir13f;
        break;
      case 4: // size=16, solid..
        data = Cir16f;
        break;
      case 5: // size=4, outline..
        data = Circ6h;
        break;
      case 6: // size=5, outline..
        data = Circ7h;
        break;
      case 7: // size=12, outline..
        data = Cir12h;
        break;
      case 8: // size=13, outline..
        data = Cir13h;
        break;
      case 9: // size=16, outline..
        data = Cir16h;
        break;
      case 10: // size=4, ants..
        data = Circ6a6;
        info = ((6*64+6)<<14)+0x40000000;
        break;
      case 11: // size=5, ants..
        data = Circ7a8;
        info = ((8*64+7)<<14)+0x40000000;
        break;
      case 12: // size=12, ants..
        data = Cir12a2;
        info = ((2*64+12)<<14)+0x40000000;
        break;
      case 13: // size=13, ants..
        data = Cir13a2;
        info = ((2*64+13)<<14)+0x40000000;
        break;
      case 14: // size=16, ants..
        data = Cir16a2;
        info = ((2*64+16)<<14)+0x40000000;
        break;} //~switch
      whom.SetInfo(info);
      data = whom.SwapData(data);
      why = -why;
      break;} //~while
    if (Logging || (why>0)) System.out.println("  (BstCirDat #" + nx + ") " + zx
        + " " + tall + "/" + prio + "/" + kx + IffyStr(why>0," = ",
          SeeHex(" i=x",info," = ")) + diam + "/" + why);
    return diam;} //~BestCircData

  /**
  * Given a widget that is either a circle icon or a group over a circle
  * icon, and a preferred diameter, choose the best circle size and update
  * the widget (and its sub-widget, if so) to be that size, with a data array
  * to match the selected size.
  *
  * CircleSize is the public access to BestCircData, which does the work.
  *
  * @param  whom   the widget to update
  * @param  rx     the desired diameter
  */
  public void CircleSize(GameWgt whom, int rx) { // whom sb draggable tool/wgt
    int prio = rx, tye, info, zx, nx = 0, yx = 0, tall = 0, wide = 0, why = 0;
    boolean doit = false; // true: also set container size
    GameWgt nuly = whom;
    while (true) {                // calls BestCircDat to force new radius rx
      why++; // why = 1
      if (whom==null) break;
      why++; // why = 2
      nx = whom.GetRefNum();
      tye = whom.GetType()&255;
      info = whom.GetInfo();
      if ((tye|4)==6) { // can be either sprite or icon
        if (tye==6) doit = true;
        yx = 0;} //~if
      else if ((info&0xFFFC) == 0x4454) { // 'DT'/'DW' (in Maker)..
        yx = (1-(info&1))*32; // offset: =32 if tool, =0 in game board..
        if (yx==0) doit = true;} //~if // game board only, also resize container
      else break;
      why++; // why = 3
      if (((tye-5)&255)<3) // should be always, unless icon
        nuly = GetIxWgt(whom.GetIxData(0,0)); // nuly is now the icon
      why++; // why = 4
      if (nuly==null) break;
      if (nuly==whom) whom = GetIxWgt(nuly.GetParent());
      why++; // why = 5
      if ((nuly.GetType()&255) !=2) break;
      why++; // why = 6
      info = nuly.GetColor();
      if (info==0xFFFFFF) nuly.SetColor(info-1); // BCD needs this..
      else if (info==0) nuly.SetColor(1);
      if (rx>32) rx = 0;
      if (rx<4) if (whom != null) {
        info = whom.GetInfo();
        zx = info&0x7E00FFFC;
        if (tye==6) // (Ball) sprite in the game..
          rx = KnownCircleSz(nuly.GetSize()&63);
        else if (zx == 0x42004454) { // 'C*DT'/'C*DW' or 'B*DT'/'B*DW'
          why = why+3; // why = 9 => 10/11/12
          rx = ((info>>16)&127)-44;
          if (rx>10) rx = rx-7;
          if (rx>4) rx = KnownCircleSz(rx);} //~if
        else whom = null;} //~if // icon in game bd is considered parentless
      why++; // why = +7
      if (rx<4) {
        why = why+6; // why = 13 => 14/15
        tall = nuly.GetSize();
        wide = SignExtend(tall);
        tall = tall>>16;
        rx = KnownCircleSz(tall);
        zx = KnownCircleSz(wide);
        if (Math.abs(zx-wide)<Math.abs(rx-tall)) rx = zx;
        if (rx==tall) if (rx==wide) if (nuly.GetInfo()+3 !=0)
          if (nuly.GetIxData(0,-1) >= 0) break;} //~if // probly OK
      why++; // why = 8
      if (whom != null) {
        why++; // why = 9
        zx = BestCircData(whom,-rx,yx);
        if (false) if (doit) if (yx==0) if (rx>0) whom.SetSize(rx,rx);} //~if
      else zx = BestCircData(nuly,-rx,yx); // (BCD sets size)
      why = -why;
      break;} //~while
    if (Logging || (why>0)) System.out.println("  (CircSiz #" + nx + ") r="
        + rx + "/" + prio + " i=" + yx + " = " + why);} //~CircleSize

  /**
  * One of the features of the GameMaker tool palette is a color swatch
  * representing the designated color of whatever tool or user game widget
  * is currently selected, or else a checkerboard gray if no color is set.
  * SwatchGetColor returns the currently visible swatch color (from the
  * widget you specify) as a number, 0x00RRGGBB, or else -1 if none.
  *
  * @param  whom    the swatch color widget to get this color from
  *
  * @return         the current swatch color, 0x00RRGGBB or -1 if none
  */
  public int SwatchGetColor(GameWgt whom) {
    int rez = -1, why = 0;
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      rez = whom.GetColor();
      why++; // why = 2
      nuly = GetIxWgt(whom.GetIxData(0,0));
      if (nuly==null) break;
      why++; // why = 3
      if (nuly.GetInfo()>0) break;
      why = 0;
      rez = -1;
      break;} //~while
    if (Logging) System.out.println(SeeHex("(SwaGetCo c=",rez," = " + why));
    return rez;} //~SwatchGetColor

  /**
  * One of the features of the GameMaker tool palette is a color swatch
  * representing the designated color of whatever tool or user game widget
  * is currently selected, or else a checkerboard gray if no color is set.
  * SetSwatchColor sets the swatch color widget you specify to a numeric
  * color value, 0x00RRGGBB, or else -1 if none.
  *
  * @param  whom    the swatch color widget to set to this swatch color
  * @param  info    the color to set it to, 0x00RRGGBB or -1 if none
  */
  public void SetSwatchColor(GameWgt whom, int info) {
    GameWgt nuly = null;
    int why = 0;
    if (whom==null) why--;
    else {
      nuly = GetIxWgt(whom.GetIxData(0,0));
      if (info<0) {
        if (nuly != null) nuly.SetInfo(0); // why = 0
          else why++; // why = 1
        whom.SetColor(0xFFFFFF);} //~if
      else {
        if (info==0xFFFFFE) info++;
        else if (info==1) info = 0;
        why = 2;
        if (nuly != null) nuly.SetInfo(32);
          else why++; // why = 3
        whom.SetColor(info);}} //~else
    if (Logging)
      System.out.println(SeeHex("(SetSwaCo c=",info," = " + why));} //~SetSwatchColor

  /**
  * ChoseFont is a multi-purpose font selector.
  *
  * If you pass it null for a widget, it converts one form of font selection
  * -- such as the negative of an existing font RefNum, or the numeric (ASCII)
  * value of the 1-letter font name -- into the other. The letter names are:
  *
  *   Bold=66, Head=72, Italic=73, Monosp=77, Nano=78, Text=84, tinY=89
  *
  * If you give it a GameTxLn widget or a radio button group or the text
  * tool or a user text widget on the game board and a negative number, it
  * will return the numeric font name associated with that widget; if you
  * give it the 1-letter name of a font, it will set that radio button or
  * widget to that font.
  *
  * @param  whom    the widget clicked on
  * @param  name    either a font name (number) or (negative) RefNum
  *
  * @return         the chosen/converted font name/RefNum
  */
  public int ChoseFont(GameWgt whom, int name) { // whom drag'le tool/wgt
    int zx, nx = 0, info = 0, rez = 0, why = 0; // ..or else the radio group;
    // ..rtns current setting; whom=null converts name <=> number
    GameWgt nuly = null; // ..rx is (int)('M'/'T'/'B'/'H'/'I') sets, 0 asks
    String msg = "";
    while (true) {
      why++; // why = 1
      if (whom==null) { // convert name <=> number..
        if (name<0) { // from -font wgt + to 1-char name..
          if (name+MonoFontNo==0) rez = 77; // 'M'
          else if (name+TextFontNo==0) rez = 84; // 'T'
          else if (name+ItalFontNo==0) rez = 73; // 'I'
          else if (name+BoldFontNo==0) rez = 66; // 'B'
          else if (name+HeadFontNo==0) rez = 72; // 'H'
          else if (name+TinyFontNo==0) rez = 89; // 'Y'
          else if (name+NanoFontNo==0) rez = 78; // 'N'
          else if (name+ScorFontNo==0) rez = 83; // 'S'
            else break;} //~if // why = 1 (rez=0)
        else if (name<8) { // from MOS font + to 1-char name..
          if (name==0) rez = 77; // 'M'
          else if (name==1) rez = 84; // 'T'
          else if (name==2) rez = 73; // 'I'
          else if (name==4) rez = 66; // 'B'
          else if (name==5) rez = 72; // 'H'
          else if (name==3) rez = 89; // 'Y'
          else if (name==6) rez = 78; // 'N'
          else if (name==7) rez = 83; // 'S'
            else break;} //~if // why = 1
        else if (name<65) break; // why = 1
        else if (name>90) break; // convert 1-char name to wgt+..
        else if (name==66) rez = BoldFontNo;
        else if (name==72) rez = HeadFontNo;
        else if (name==77) rez = MonoFontNo;
        else if (name==84) rez = TextFontNo;
        else if (name==73) rez = ItalFontNo;
        else if (name==89) rez = TinyFontNo;
        else if (name==78) rez = NanoFontNo;
        else if (name==83) rez = ScorFontNo;
          else break; // why = 1 (rez=0)
        if (rez>0) why = -why; // why = -1
        break;} //~if
      why++; // why = 2
      nx = whom.GetRefNum();
      info = whom.GetInfo();
      if ((info&0x7FFFFF00) == 0x52644200) { // 'RdB_' = set/get radio btns
        why = -why; // why = -2
        rez = whom.RadioBtn(name);
        break;} //~if
      if ((info&0x7FFFFFFC) == 0x54784454) // 'TxDT'/'TxDW'
        nuly = GetIxWgt(whom.GetIxData(0,0)); // should be the actual text
      else if ((whom.GetType()&255)==4) nuly = whom;
      if (nuly==null) break; // why = 2
      why++; // why = 3
      if ((nuly.GetType()&255) !=4) break;
      why++; // why = 4
      if (name >= 0) { // set it, to what?
        name = name&255;
        info = 0;
        if (name<8) name = ChoseFont(null,name);
        if (name>64) info = ChoseFont(null,name);
        if (info==0) break; // why = 4
        if (name>64) rez = name;
        nuly.SetInfo(info);
        why = -why; // why = -4
        break;} //~if
      why++; // why = 5: otherwise read out the font wgt+, make 1-char name..
      rez--;
      info = nuly.GetInfo();
      if (info <= 0) break; // why = 5
      rez = ChoseFont(null,-info);
      if (name+2==0) // name=-2 rtns MOS font +..
        for (info=0; info <= 7; info++) if (info>6) {
          why++; // why = 6
          rez = -rez;
          break;} //~if
        else if ((int)(CharAt(info,"MTIYBHNM"))==rez) {
          rez = info;
          break;} //~if
      if (why<6) why = 0;
      break;} //~while
    if (Logging || (why>0)) System.out.println("  (ChosFnt #" + nx
        + SeeHex(") n=",name," = ")
        + HexIfMore("",rez,"/") + why );
    return rez;} //~ChoseFont

  /**
  * An internal way to adjust one dimension of a newly created widget from
  * its default (created) size to conform to a user-specified size of its
  * containing widget.
  *
  * For a nominal size of (frex) 64, if the macro specifies 62, then the
  * returned value is the actual container size reduced by -2 (the difference)
  * unless the macro default is not close to the nominal (like being at the
  * other end of the created widget).
  *
  * @param  why    a reference number, for diagnostic log
  * @param  thar   the default size in the macro
  * @param  norm   the nominal size, against which to measure the default
  * @param  act    the actual size supplied by the container
  *
  * @return        the calculated adjusted size = act-(norm-thar)
  */
  private int AdjuDim(int why, int thar, int norm, int act) { // frm BuildThisM
    int zx = norm>>2;
    zx = norm-(zx>>1)-zx;
    if (thar>zx) thar = act-(norm-thar); // if (thar<act)
    return thar;} //~AdjuDim

  /**
  * An internal way to construct a 32-bit integer representation of a 4-char
  * string ID code. The code is enclosed in reverse-single-quotes '`' in the
  * input line, and if not there or invalid, 0 is returned.
  *
  * @param  aLine  the string to find the code in
  *
  * @return        the calculated (integer) ID code
  */
  private int EncodeID(String aLine) { // frm ImportGamLis,BuildThisMac
    // if (NthOffset(0,"`",aLine)<0) return 0; // no ID
    String xID = NthItemOf('`',2,aLine);
    int mID = 0, kx = StrLength(xID), nx, zx;
    myTmpStr = "";
    if (kx>2) if (kx<5) for (nx=0; nx <= kx-1; nx++) {
      zx = (int)(CharAt(nx,xID))&127;
      if (zx>32) if (zx<127) {
        mID = (mID<<8)+zx;
        continue;} //~if
      mID = 0;
      break;} //~for // (nx)
    if (mID<0x410000) mID = 0;
    if (mID>0) myTmpStr = xID;
    return mID;} //~EncodeID

  /**
  * Widgets are specified by the user text, each as ten or more integer numbers
  * on a single line, where the second is a type number (plus flags). Type
  * numbers > 224 are deemed to be macros, where multiple widgets are created
  * under a single group. BuildThisMacro adds as sub-widgets to an existing
  * group, whatever the specified macro calls for.
  *
  * @param  whom   the parent group widget to build this macro under
  * @param  aLine  the text line containing parameters this macro can use
  * @param  why    an error code base (the highest used in the caller)
  *                  so that returned error codes (if any) will not collide
  *
  * @return        an error code > why, or else =0 if no error
  */
  private int BuildThisMacro(GameWgt whom, String aLine, int why) {
    char delim = '\'';                      // frm ImportGamLis, why = 14
    int ix, kx, nx, zx, tye, top, left, tall, wide, color, info,
        here, thar, mID = 0, whar = 0, prnt = 0, more = 0, ctx = 0,
        nmx = 0, dimz = 0, lxx = 0;
    boolean seen = false;
    String aWord, aStr = "", theMacs = "", xID = "", msg = "";
    GameWgt nuly;
    GameEvent evh = UserEvHandler;
    int[] wgtmx = WgtMacros; // uses WgtMacros, ChkBoxz,ChkBitz, Antz
    int[] wgts = null;
    int[] data = null;
    while (true) { // once thru..
      why = 15;                                       // 254 labelled text entry
      if (whom==null) break;                          // 253 checkbox
      why++; // why = 16                              // 252 pushbutton
      prnt = whom.GetRefNum();                        // 251 color swatch
      dimz = whom.GetSize();                          // 250 marching ants
      tye = SafeParseInt(NthItemOf(' ',2,aLine))&255; // 249 radio button group
      msg = msg + " M=" + tye;                        // 240 input/alert dlog
      if (tye<222) break;
      if (NthOffset(0,"`",aLine)>0) { // got a macro ID..
        mID = EncodeID(aLine);
        xID = myTmpStr;
        if (mID>0)
            if ((whom.GetType()&0x3F00) !=0) if (whom.GetInfo()==0) {
          if (Logging)
            System.out.println("  (BTM/mID+0) #" + prnt + SeeHex(" = x",mID,""));
          whom.SetInfo(mID);}} //~if // NOT: mID = 0; (might be inner wgts)
      // why++; // why = 17
      while (true) { // loop until found..
        why = 17;
        if (whar<0) break;
        if (whar > wgtmx.length-2)
          break;
        nx = wgtmx[whar];
        lxx = wgtmx[whar+1];
        msg = msg + " @" + whar + ": " + nx + "+" + lxx;
        whar = whar+2;
        if (nx != tye) {
          whar = lxx*7+whar;
          continue;} //~if // only cont, otherwise break
        if (lxx<0) if (evh != null) {
          lxx = evh.UserMacro(lxx,mID,whom);
          whar = (lxx&0xFF00)*8-0x01000000;
          if (lxx>0) lxx = lxx&255;} //~if
        if (lxx <= 0) break;
        why++; // why = 18
        wgts = new int[lxx];
        if (wgts == null) break;
        data = whom.SwapData(wgts);
        why++; // why = 19
        why = 20;
        // nx = 255-nx; // switch (nx*8+lxx) { case
        for (zx=0; zx <= lxx-1; zx++) {
          why++; // why = 21
          if (whar>0) {
            if (whar > wgtmx.length-8)
              break; // Wgt_Macros = {253,4, 0,3,1,12,12,0xFFFFFF,0, // chekbx..
                // 254,3, 4,3,0,12,16,0x0,8,            // labelled text entry..
                //   0,0,8,16,32,0xFFFFFF,0, 0x2604,3,12,12,20,0x0,5,
            tye = wgtmx[whar];
            top = wgtmx[whar+1];
            left = wgtmx[whar+2];
            tall = wgtmx[whar+3];
            wide = wgtmx[whar+4];
            color = wgtmx[whar+5];
            info = wgtmx[whar+6];} //~if
          else if (evh != null) {
            ix = whar&0xFFFFF;
            tye = evh.UserMacro(ix,mID,whom); // +0
            ix++;
            top = evh.UserMacro(ix,mID,whom); // +1
            ix++;
            left = evh.UserMacro(ix,mID,whom); // +2
            ix++;
            tall = evh.UserMacro(ix,mID,whom); // +3
            ix++;
            wide = evh.UserMacro(ix,mID,whom); // +4
            ix++;
            color = evh.UserMacro(ix,mID,whom); // +5
            ix++;
            info = evh.UserMacro(ix,mID,whom); // +6
            whar++;} //~if
          else break;
          whar = whar+7;
          seen = false;
          data = null;
          aWord = "";
          // aStr = "";
          here = tye&255;
          if (here==4) if (info+64<0) info = ChoseFont(null,-info);
          if (here>222) { // queue up this macro for later processing..
            thar = NthOffset(0,String.valueOf(delim),aLine);
            ix = info;
            kx = StrLength(xID);
            if (kx<3) aWord = "";
              else aWord = " `" + Substring(0,kx-2,xID);
            if (info==1) aWord = "'OK'" + IffyStr(kx<3,"",aWord + "OK`");
            else if (info==2)
              aWord = "'Cancel'" + IffyStr(kx<3,"",aWord + "Cn`");
            else if (thar>0) aWord = RestOf(thar,aLine);
              else aWord = " `" + xID + "`";
            if ((nx&254)==240) { // nx/dimz: doing in this (outer) container
              if (here==252) if (info>0) if (info<3) if (mID>0) {
                if (info==1) info = (mID&(-0x10000)) | 0x4F4B; // "OK"
                else if (info==2) info = (mID&(-0x10000)) | 0x436E;} //~if // "Cn"
              if (dimz != 0x4000F0) { // macro nominal is 64/240..
                here = dimz>>16; // adjust coords to match (outer) container..
                top = AdjuDim(0,top,64,here);
                tall = AdjuDim(1,tall,64,here);
                here = dimz&0xFFFF;
                left = AdjuDim(2,left,240,here);
                wide = AdjuDim(3,wide,240,here);}} //~if
            aWord = "" + NextRefNo + " " + tye + " " + top + " " + left
                + " " + tall + " " + wide + " " + color + " " + ix // =info
                + " " + prnt + " 0 " + aWord;
            theMacs = theMacs + aWord + "\n";
            if (Logging) System.out.println("  [BTMm] '" + aWord + "'");
            tye = (tye&-256)|5;
            aWord = "";
            nmx--;} //~if
          else if (info+8==0) { // -8: radio button..
            if (color==0) data = Cir12h; // black radio btn overlay
            else if (color==0xFFFFFF) data = Cir12f;} //~if // white bkgnd
          else if (info+7==0) data = ChkBoxz; // -7: black box for checkbox
          else if (info+6==0) { // -6: red checkmark
            data = ChkBitz;
            info = 12;} //~if // (unchecked)
          else if (info+5==0) // -5: checkerboard "no color" pattern
            data = NoCoGr;
          else if (info+4==0) data = DragArrows; // -4: drag-arrow icon
          else if (info+2==0) { // -2: marching ants pattern..
            data = Antz;  // info: 6b(fr.frate),6b(+frs),6b(stpsz),14b(img.offs)
            info = ((top+left)&7)+0x40804000;} //~if
          // else if (nx==250) {} // create marching ants (max 32x32)..
          else if ((nx|4)==253) { // checkbox/radio (tall=16, color=-1)..
            if (here==4) aWord = NthItemOf(delim,2,aLine);} //~if
          // nx/dimz: doing in this container (nx=wgtmx[whar-2])..
          else if (nx==254) { // labelled text entry (tall=16, color=-1)..
            if (dimz>0x200000) { // tall>31 stacks label over entry panel..
              if (zx>0) { // entry panel; label is on top..
                if (dimz>0x280000) top = top+24;
                  else top = top+20;} //~if
              wide = dimz&0xFFFF;
              left = 0;
              more = 0;} //~if
            if (here==0) if (mID == 0x4E616D65) info = mID+1; // 'Name'
            if (here==4) {
              if (zx>0) {
                if (dimz>0x200000) {
                  wide = wide-8;
                  left = 4;} //~if
                else if (more>0) left = more+4;
                aWord = NthItemOf(delim,4,aLine);} //~if
              else if (info<0) { // skip the label..
                why--; // why = 20
                more = -1;
                continue;} //~if
              else aWord = NthItemOf(delim,2,aLine);} //~if
            else if (more<0) {
              wide = dimz&0xFFFF;
              left = 0;} //~if
            else if (more>0) left = more;} //~if
          else if (nx==252) { // pushbutton (tall=22, color=0xCCCCCC)..
            if (here==4) { // here = tye&255; // (4: this is button label)
              ix = info&0xFFFF;
              if (info==1) aWord = "OK";
              else if (info==2) aWord = "Cancel";
              else if (ix == 0x4F4B) aWord = "OK";
              else if (ix == 0x436E) aWord = "Cancel";
                else aWord = NthItemOf(delim,2,aLine);} //~if
            else if (more>0) { // assumes nominal 32 wide in template..
              left = AdjuDim(4,left,31,more);
              wide = AdjuDim(5,wide,31,more);} //~if
            seen = true;} //~if
          else if ((nx&254)==240) { // input/alert dlog (tall=64, color=wh)..
            if (tye==4) aWord = NthItemOf(delim,2,aLine);    // (nom.wide=240)
            if (dimz != 0x4000F0) { // macro nominal is 64/240..
              here = dimz>>16; // match posns & sizes to container dims..
              top = AdjuDim(6,top,64,here);
              tall = AdjuDim(7,tall,64,here);
              here = dimz&0xFFFF;
              left = AdjuDim(8,left,240,here);
              wide = AdjuDim(9,wide,240,here);}} //~if
          if (info<0) info = 0;
          kx = StrLength(aWord);
          if (mID>0) if ((tye&0x3F00) !=0) if ((info&-4)==0) {
            ix = info;
            if (info==0) {
              if ((mID&255)==0) mID = mID+46; // oops, partly used..
              if ((mID&0xFF00)==0) mID = mID+0x2E00;
              info = mID;
              mID = 0;} //~if
            else if (kx>0) {
              if (info==1) {
                info = (mID&(-0x10000)) | 0x4F4B; // "OK"
                mID = mID&-256;} //~if
              else if (info==2) {
                info = (mID&(-0x10000)) | 0x436E; // "Cn"
                mID = mID&~0xFF00;} //~if
              if ((mID&0xFFFF)==0) mID = 0;} //~if
            if (Logging) System.out.println("  (BTM/mID+" + ix + ") #" + NextRefNo
                + SeeHex(" = x",info,SeeHex("/",mID,"")) + " " + kx + " " + nx);
            seen = true;} //~if
          aStr = " t=" + tye + " i=" + info + " '" + aWord + "'"
              + IffyStr(nx==251,"",IffyStr(nx==253,""," lf=" + left
              + IffyStr(nx==252," wi=" + wide,IffyStr((nx&-2)==240," tp="
                + top + " ta=" + tall + " wi=" + wide,""))));
          if (Logging) {
            if (seen) System.out.println("  ++ " + aStr);
              else msg = msg + aStr;} //~if
          nuly = NewGameWgt(tye,top,left,tall,wide,prnt,color,info,data,null);
          why++; // why = 22
          if (nuly==null) break;
          why++; // why = 23
          ix = nuly.GetRefNum();
          if (ctx<0) break;
          if (ctx < wgts.length)
            wgts[ctx] = ix;
          ctx++;
          why++; // why = 24
          if (whar<0) if (evh != null)  // +7..
            ix = evh.UserMacro((whar-1)&0xFFFFF,mID,nuly); // ignore result
          if (data == null) if (kx>0) {
            // here = nuly.GetSize(); // = tall,wide
            nuly.SetSize(0,kx*10); // temp'ly, cuz GetTextWi looks
            thar = nuly.PutTextLn(aWord); // thar = nuly.GetTextWide();
            nuly.SetSize(tall,wide);
            // more = nuly.GetPosn()&0xFFFF; // =left
            if (thar>0) {
              if (nx>252) {
                thar = thar+4;
                if ((tye&0x600) !=0) if ((mID&0x3FFF) != 0x658) // `xxFx`
                  if (thar <= wide) thar = 0;
                if (thar>0) {
                  nuly.SetSize(0,thar);
                  more = left+thar;
                  msg = msg + " ^" + more;
                  if (whom != null) whom.SetSize(0,more);}} //~if
              else if (nx==252) if (whom != null) {
                kx = dimz&0xFFFF;
                ix = thar+6; // = a reasonable text width
                more = ix+8; // = the calc'd button width
                if (more<kx) { // adjust start of text to center it..
                  here = kx-more; // this many pixels to divide
                  more = more+here; // more is now actual button width
                  here = (here>>1)+left;
                  nuly.SetPosn(top,here);} //~if
                else whom.SetSize(0,more);
                if (ix>wide) nuly.SetSize(0,ix); }}} //~if
          why = 20;} //~for // (zx: each gen'd wgt)
        if (why>20) break;
        why--; // why = 19
        if (whom==null) break;
        if (mID>0) if (whom.GetInfo()==0) {
          if (Logging)
            System.out.println("  (BTM/mID+9) #" + prnt + SeeHex(" = x",mID,""));
          whom.SetInfo(mID);} //~if
        if (Logging) whom.LogMe(".. ");
        if (wgts != null) zx = wgts.length;
        for (zx=zx-1; zx >= 0; zx += -1)
          if (zx < wgts.length) {
            nuly = GetIxWgt(wgts[zx]);
            if (Logging) if (nuly != null) nuly.LogMe("  ..");} //~if
        why = 0;                                // ..(zx: each listed wgt)
        break;} //~while // (loop until found)
      break;} //~while // (once thru)
    if (why==0) why = nmx;
    if (why<0) {
      myTmpStr = theMacs;
      theMacs = " ==> ..\n" + theMacs;} //~if
    if (Logging) System.out.println("[BTMx #" + prnt + "]" + msg
        + " = " + why + theMacs);                             // why =
    return why;} //~BuildThisMacro

  /**
  * A search tool to find uniquely identified widgets built from the list
  * returned by the client GameList, so to assign references to them in
  * variables usable in the game.
  *
  * @param  msg   the unique string to search for
  *
  * @return       the first widget specified in theImpoList with that string
  */
  public GameWgt FindListWgt(String msg) { // in theImpoList
    int lino, info = 0;
    String aLine = "", data = theImpoList;
    GameWgt whom = null;
    for (lino=1; lino <= 9999; lino++) {
      info = 0;
      aLine = NthItemOf('\n',lino,data);
      if (StrLength(aLine)==0) break; // didn't find it
      if (NthOffset(0,msg,aLine)<0) continue;
      info = SafeParseInt(aLine); // = lino, except for additions
      whom = GetIxWgt(info);
      if (whom==null) break;
      break;} //~for
    if (Logging)
      System.out.println(IffyStr(whom==null,"(NOT) found #","(!) Found #")
          + info + " '" + msg + "' @ " + lino + ": '" + aLine + "'");
    return whom;} //~FindListWgt

  /**
  * A search tool to find uniquely identified widgets based on their Ginfo
  * values or a combination of type and info. The whom parameter is the
  * widget to start this search, and is returned if it matches the criteria.
  * Otherwise FindSubWgt searches its sub-widgets (if any) recursively.
  *
  * If the search number is larger than 16 bits, and it is equal to Ginfo
  * of the widget, then that widget is returned. Otherwise the upper half
  * of the search number is compared to the low 14 bits of Ginfo, and the
  * lower half of the search number is compared to the type+plus flags for
  * each respective widget. The usual search values are either a 4-char ID
  * value, or else a GameTxLn with specified flags and a particular font +,
  * so to pick out the parameters of a draggable widget or tool.
  *
  * Null is returned if the specified widget is not found.
  *
  * The code is not very robust, so sometimes it fails when the sought
  * widget is actually there. It probably needs to be rewritten, but
  * it seems to be good enough for now.
  *
  * @param  whom  the unique string to search for
  * @param  data  the unique value to search for
  * @param  deep  the current search depth, and if greater than 999,
  *                 a unique logging identifier
  *
  * @return       the widget found that matches data
  */
  public GameWgt FindSubWgt(GameWgt whom, int data, int deep) {
    // [this is too slow for searches through deeply nested wide trees]
    int fust = deep, nx, kx = 0, zx = 0, info = 0, tye = 0,  // rec'ly search
        logy = 0, seen = 0, why = 0;  // data = type+flgs+(info<<16) to find
    GameWgt nuly = null;
    while (true) { // once through, but recursive..
      why++; // why = 1
      if (whom==null) break; // only if caller did it
      info = whom.GetInfo();
      kx = whom.GetRefNum();
      why++; // why = 2
      if (deep>999) if ((deep&7)==0) fust = deep%1000; // only if (Logging)
      if (fust==0) { // if (CashFoundWgt != null)
        if (CashFindWgt==whom) if (CashFindWhy==data) {
          nuly = CashFoundWgt;
          fust--;} //~if
        if (deep>999) if (fust==0) FindLog = "";} //~if
      if (deep>999) {
        tye = whom.GetType();
        myTmpStr = "#" + kx + " ";
        if (NthOffset(0,myTmpStr,FindLog)<0) FindLog = FindLog
            + IffyStr(fust>0,"","(" + deep + ") ") + myTmpStr;
        else if (fust <= 0) FindLog = FindLog + "(" + deep + ") ";
        myTmpStr = "";
        if (fust<0) FindLog = FindLog + "$=" + SafeWgtNum(nuly) + "/"
            + SafeWgtNum(CashFindWgt) + SeeHex(" = x",CashFindWhy," ");} //~if
      if (fust<0) break; // same as prior
      why++; // why = 3
      nuly = whom;
      if (info>0x10000) if (info==data) break; // full-word success
      if (tye==0) tye = whom.GetType();
      info = (info&0x3FFF)<<16;
      why++; // why = 4
      if (tye+info == data) break; // split-word success
      why++; // why = 5
      nuly = null;
      if (((tye-5)&255)>2) break; // failed and no (more) recursion
      why++; // why = 6
      for (nx=0; nx <= 999; nx++) {
        zx = whom.GetIxData(nx,0);
        if (deep>999) {
          kx = zx;
          myTmpStr = "#" + zx + " ";
          if (kx>0) if (NthOffset(0,myTmpStr,FindLog)>0) {
            myTmpStr = myTmpStr + "= ";
            kx = -kx;} //~if
          myTmpStr = "+" + nx + " " + myTmpStr;
          FindLog = FindLog + myTmpStr;
          myTmpStr = "";
          if (kx<0) continue;} //~if
        else if (zx<0) continue;
        if (zx==0) {
          if (seen>0) break; // why = 6 (two consecutive zeroes)
          seen++;
          continue;} //~if
        seen = 0;
        kx = 0;
        nuly = GetIxWgt(zx);
        if (nuly==null) continue;
        nuly = FindSubWgt(nuly,data,deep+1);
        if (nuly==null) {
          if (deep>999) FindLog = FindLog + "=0 ";
          continue;} //~if
        kx = nuly.GetRefNum();
        why++; // why = 7
        break;} //~for // found it
      if (deep>999) {
        if (fust==0) FindLog = FindLog + "! ";
          else FindLog = FindLog + "(+) ";} //~if
      break;} //~while
    if (fust==0) if (why>2) if (whom != null) {
      zx = whom.GetRefNum();
      if (zx>0) {
        if (deep>999) FindLog = FindLog + " $C=" + zx
            + SeeHex("/x",data,"=#") + SafeWgtNum(nuly) + " ";
        CashFindWgt = whom;
        CashFindWhy = data;
        CashFoundWgt = nuly;}} //~if
    if (deep>999) {
      if (nuly==null) info = data;
      System.out.println(IffyStr(nuly==null,"(un) f","(*) F")  // ") Found #"
          + "ound #" + kx + SeeHex(" x",info," ") + deep + " = " + why);
      if (StrLength(FindLog)>0) {
        nx = 0;
        if (fust <= 0) nx--;
        else if (nuly==null) nx++;
        else if (NthOffset(0,"$",FindLog)>0) nx++;
        if (nx !=0) System.out.println("  .. " + FindLog);
        if (nx<0) FindLog = "";}} //~if
    return nuly;} //~FindSubWgt

  /**
  * Widgets specified by macro do not always get their dimensions adjusted
  * properly by AdjuDim. This is one of several ad-hoc fixups.
  *
  * The caller gives a negative number to adjust the child widget width
  * down from its parent, then this is added to the group widget size to
  * give a smaller (positive) number to set the size of the child widget.
  *
  * @param  whom   the parent group widget to look under
  * @param  plus   how to adjust it
  */
  private void Widen1stWgt(GameWgt whom, int plus) {
    int here, info, wide = plus, thar = 0, more = 0,
        tye = 0, nx = 0, rx = 0, cx = 0, zx = 0, why = 0;
    String msg = "";
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      rx = whom.GetRefNum();
      tye = whom.GetType()&255;
      why++; // why = 2
      if (plus>0) break;
      why++; // why = 3
      wide = SignExtend(whom.GetSize()+plus);
      if (wide<8) break;
      why++; // why = 3
      if (tye !=5) break;
      why++; // why = 4
      nuly = GetIxWgt(whom.GetIxData(0,0));
      if (nuly==null) break;
      why++; // why = 5
      nx = nuly.GetRefNum();
      more = nuly.GetSize();
      zx = nuly.GetType()&255;
      wide = wide-SignExtend(nuly.GetPosn());
      if (((more&0xFFFF)+16<wide)) nuly.SetSize(more>>16,wide);
      why++; // why = 6
      if (zx !=5) break;
      why++; // why = 7
      for (here=1; here <= 8; here++) {
        whom = GetIxWgt(nuly.GetIxData(here,0));
        if (whom==null) break;
        if ((whom.GetType()&0x1C00)==0) continue;
        cx = whom.GetRefNum();
        more = whom.GetSize();
        thar = wide-SignExtend(whom.GetPosn());
        whom.SetSize(more>>16,thar);
        why = 0;
        break;} //~for // otherwise why = 7
      break;} //~while
    if (Logging) System.out.println("(Widn1stWgt #" + rx + " t=" + tye
        + " / #" + nx + " t=" + zx + IffyStr(plus<0,") ",") +") + plus
        + " => " + wide + " => " + thar + LogVH(" / ",more," #")
        + cx + " = " + why);} //~Widen1stWgt

  /**
  * There are two ways to display marching ants around the edge of a widget.
  * AntsIcon is used for sprites with specially created animated icons (see
  * MkSpriteAnts) that trace around the odd shape of the sprite.
  *
  * @param  sho   true to display the ants, false to hide them
  * @param  whom  the group or sprite widget to turn the ants on or off for
  *
  * @return       true if successful
  */
  public boolean AntsIcon(boolean sho, GameWgt whom) { // = Antsify for icons
    int why = 0;          // much simpler, cuz it's all here & right size..
    String msg = "";
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      msg = "w=" + whom.toString();
      why++; // why = 2
      if (((whom.GetType()-5)&255)>2) break;
      why++; // why = 3
      nuly = FindSubWgt(whom,0x416E7449,0); // 'AntI'
      if (nuly==null) break;
      why++; // why = 4
      if (nuly != whom) {
        msg = msg + " -> " + nuly.toString();
        whom = nuly;} //~if
      if (whom==null) break;
      why++; // why = 5
      whom.HideSho(sho);
      if (sho) mAntsHost = whom;
      else if (whom==mAntsHost) {
        mAntsHost = null;
        msg = msg + " -mHo";} //~if
      why = 0;
      break;} //~while
    if (Logging)
      System.out.println("  (AntzIco/" + sho + ") " + msg + " = " + why);
    return why==0;} //~AntsIcon // T: did it

  /**
  * There are two ways to display marching ants around the edge of a widget.
  * Antsify is used for group rectangles which can contain the predefined
  * mAntsWgt group as a sub-widget. Each icon in the group is good for one
  * vertical or horizontal line of ants, up to 32 pixels tall or up to 256
  * pixels wide, and they are mostly pre-positioned for all edges except the
  * bottom and right. Antsify adjusts the positions of those two edges and
  * hides (or shows) the rest as needed for the rectangle size.
  *
  * If the group contains a group with ID 'AntI' it is assumed that AntsIcon
  * can do it.
  *
  * @param  sho   true to display the ants, false to hide them
  * @param  whom  the group widget to turn the ants on or off for
  */
  public void Antsify(boolean sho, GameWgt whom) { // 256x1K max for now
    int fx, hz, nx, px, rx = 0, zx = 0, sz = 0, why = 0;
    String msg = "";
    GameWgt nuly = null;
    if (whom != null) while (true) {
      rx = whom.GetRefNum();
      nuly = FindSubWgt(whom,0x416E7449,0); // 'AntI'
      if (nuly==null) break;
      if (nuly.GetParent() != rx) break; // oops, reached too deep
      sho = AntsIcon(sho,nuly);
      return;} //~while
    if (mAntsWgt != null) zx = mAntsWgt.GetRefNum();
    if (!sho) while (true) { // terminating ants..
      why--; // why = -1
      if (mAntsHost != null) {
        if (zx>0) {
          nx = mAntsHost.FindElemt(zx);
          if (nx >= 0) mAntsHost.DeleElemt(nx);
          else if (AntsIcon(false,mAntsHost)) why = -32;
            else why = -31;} //~if
        if (why+1<0) break;
        msg = "oH=" + mAntsHost.toString();
        mAntsHost = null;
        why--;} //~if // why = -2
      if (mAntsWgt != null) {
        if (GameWgtsList != null) px = GameWgtsList.AddElemt(zx);
        mAntsWgt.SetParent(0);
        mAntsWgt.HideSho(false);
        why = why-2; // why = -3/-4
        break;} //~if
      if (whom==null) break;
      if (whom==mAntsWgt) break;
      why = why-4; // why = -5..-8
      msg = msg + " w=" + whom.toString();
      if (mAntsWgt==null) { // only at startup..
        why = why-4; // why = -9..-12
        nuly = FindSubWgt(whom,0x416E747A,0); // 'Antz'
        if (Logging) if (nuly==null)
          nuly = FindSubWgt(whom,0x416E747A,1000);
        if (nuly==null) break;
        msg = msg + " a=" + nuly.toString();
        why = why-4; // why = -13..-16
        mAntsWgt = nuly;
        if (nuly != whom) mAntsHost = whom;} //~if // unlikely
      if (mAntsWgt != null) Antsify(false,mAntsWgt);
        else why = 4-why; // why = +9..20
      break;} //~while // why = -16..+20 // (terminating ants)
    else while (true) {
      why++; // why = 1
      if (whom==null) break;
      if (whom==GameWgtsList) break; // (+1, don't)
      if (whom==theAlertDlg) break;
      if (whom==theInputDlg) break;
      msg = msg + "w=" + whom.toString();
      why++; // why = 2
      if (mAntsHost != null) {
        msg = msg + " oH=" + mAntsHost.toString();
        if (mAntsHost==whom) break;
        Antsify(false,mAntsHost);} //~if
      why++; // why = 3
      if (zx==0) break;
      if (mAntsWgt==null) break;
      msg = msg + " a=" + mAntsWgt.toString();
      why++; // why = 4
      if (((whom.GetType()-5)&255)>2) break; // not a valid group type
      why++; // why = 5
      if (GameWgtsList==null) break;
      nx = GameWgtsList.FindElemt(zx);
      if (nx >= 0) GameWgtsList.DeleElemt(nx); // "(DelElt @"
      px = whom.AddElemt(zx);
      sz = whom.GetSize();
      mAntsHost = whom;
      mAntsWgt.SetParent(whom.GetRefNum());
      mAntsWgt.HideSho(true);
      mAntsWgt.SetSize(sz,0); // now adjust edges to match client..
      for (nx=0; nx <= 99; nx++) {
        zx = mAntsWgt.GetIxData(nx,0);
        if (zx==0) break;
        nuly = GetIxWgt(zx);
        if (nuly==null) break;
        fx = 0;
        px = nuly.GetPosn();
        hz = nuly.GetSize(); // (soon) =0 if this is top or left edge,
        if (Logging) if (MaxLogWgts>0)
          myTmpStr = LogVH(IffyStr(px<0x30000000,": ",": ^"),px&0x0FFFFFFF,
              LogVH(" + ",hz,""));
        if (px !=0) { // top/left (corner) edges never change..
          hz = (hz&2)-(hz&0x20000); // <0 if botm edge, >0 if right edge
          fx = 8;
          if (hz>0) px = (px&(-0x10000))|(sz-1)&0xFFFF;
          else if (hz<0) px = px&0xFFFF|(sz-0x10000)&(-0x10000);
            else fx = 0;
          if (px>0x30000000) {
            px = px-0x40000000;
            if (((sz-px)&0x80008000) !=0) px = px|0x40000000;
              else fx++;} //~if
          else if (((sz-px)&0x80008000) !=0) {
            px = px|0x40000000;
            fx++;}} //~if
        if (Logging) if (MaxLogWgts>0) {
          System.out.println("  (Antz) " + nx + myTmpStr + IffyStr(fx==0,"",
              IffyStr(px<0x40000000," => "," ^") + ((px>>16)&1023)
              + "/" + (px&0xFFFF)));
          myTmpStr = "";} //~if
        if (fx>0) nuly.SetPosn(px>>16,px&0xFFFF);} //~for
      why = 0;
      break;} //~while
    if (Logging) System.out.println("  (Antzify/" + sho + ") "
        + msg + " = " + why);} //~Antsify

  /**
  * Trace an icon outline to turn off the transparent segments of an ant
  * trail for MkSpriteAnts.
  *
  * @param  info  what we know about this particular length
  * @param  lxx   the trail length
  * @param  base  the row and column to start on
  * @param  wide  the nominal icon width
  * @param  whar  offset into pix for column=0
  * @param  pix   the pixel array for this icon
  */
  // private void DoAntTrail(int lxx, int base, int wide, int[] pix) { // ****

  /**
  * There are two ways to display marching ants around the edge of a widget.
  * AntsIcon is used for sprites with specially created animated icons that
  * trace around the odd shape of the sprite. MkSpriteAnts creates those
  * icons.
  *
  * The first step is to work through all the visible sub-widgets and
  * accumulate their opaque pixels into an array of pixels.
  *
  * Then the resulting blob of opaque pixels is examined from all sides to
  * find the edges, in a second icon-ready array. This will be the icon
  * data for the white outline [and for now, also the black outline that
  * blinks on and off in lieu of ants].
  *
  * Then [TBD RSN] the outline is traced and the number of pixels counted,
  * so to divide them into an integral number of 6- to 10-pixel white/black
  * ants (nominally four on then four off).
  *
  * The outline is then traced again (nominally eight times) to turn off
  * the white segments (the black is overlaid on the solid white outline,
  * so the transparent pixels show white), but advanced one more step each
  * iteration, so when the sequence is animated, the ants appear to march
  * around the composite sprite body. [TBD RSN]
  *
  * @param  whom  the sprite widget to create the ants for
  * @param  doit  true to display the created ants, false to hide them
  *
  * @return       true if successful
  */
  public boolean MkSpriteAnts(GameWgt whom, boolean doit) {
    int bitz = 0, prio, info, posn, vert, horz, yx, ix, kx, nx, rx = 0, zx = 0,
        lxx = 0, tall = 0, wide = 0, size = 0, more = 0, tye = 0, why = 0;
    String msg = "", txt = "";
    int[] data = null;
    int[] pix = null;
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      rx = whom.GetRefNum();
      tye = whom.GetType();
      why++; // why = 2
      if (((tye-5)&255)>1) break; // not valid sprite or group (tool) type
      why++; // why = 3
      info = whom.GetInfo()&0x7FFFFFFC;
      if (info == 0x53704454) { // 'SpDT'/DW'
        nuly = GetIxWgt(whom.GetIxData(0,0));
        if (Logging) System.out.println("  (MkSpAnt #" + rx + SeeHex(") x",tye," ")
            + doit + " => #" + SafeWgtNum(nuly));
        if (nuly==null) break;
        doit = MkSpriteAnts(nuly,doit);
        // if (doit) doit = AntsIcon(true,nuly); // already did it, if can
        return doit;} //~if
      why++; // why = 4
      lxx = whom.GetIxData(-1,-1); // = + icons inside this sprite
      size = whom.GetSize();
      tall = size>>16;
      wide = size&0xFFFF;
      if (((tall-1)&0xFFFF)>31) break; // unreasonably tall
      if (((wide-1)&0xFFFF)>62) break; // unreasonable width
      if (tall > WinHi) break;
      if (wide > WinWi) break;
      why++; // why = 5        // ..can't be bigger than window
      if (wide<2) wide++;
      data = new int[wide]; // create a blob to accumulate the opaque pixels
      if (data == null) break;
      why++; // why = 6
      for (zx=wide-1; zx >= 0; zx += -1)
        if (zx >= 0) if (zx < data.length)
          data[zx] = 0;     // not use RectFil..
      for (nx=0; nx <= lxx-1; nx++) { // 1st: accumulate the opaque pixels..................
        nuly = null;
        info = whom.GetIxData(nx,0);
        if (info <= 0) break;
        nuly = GetIxWgt(info);
        if (nuly==null) break;
        info = nuly.GetInfo();
        more = nuly.GetType();
        posn = nuly.GetPosn();
        size = nuly.GetSize();
        vert = posn>>16;
        horz = size&0xFFFF;
        zx = size>>16;
        if (Logging) {
          if (StrLength(msg)==0) msg = " [";
          msg = msg + " #" + info + LogVH("(",size,SeeHex(")t=",more,""));} //~if
        more = more&255;
        if (more==5) { // gotta test this before looping back on coords..
          if (Logging) msg = msg + SeeHex("i=x",info," w=") + why;
          if (info==0x416E7449) break;} //~if // 'AntI'
        if (zx <= 0) continue;
        if (horz==0) continue;
        if (vert>31) continue;
        if (vert+32 <= 0) continue;
        posn = SignExtend(posn);
        if (posn >= wide) continue;
        if (posn+horz <= 0) continue;
        if (more==0) { // rect..
          if (vert<0) {
            zx = zx+vert; // diminish v-size by pixels off the top
            if (zx <= 0) continue;
            vert = 0;} //~if
          bitz = -1;
          if (zx<32) if (vert+zx<32) bitz = bitz+(1<<zx); // trim v-size at btm
          if (vert>0) bitz = bitz<<vert; // shift pix down by v-posn
          zx = 0;
          if (posn>0) zx = posn;
          for (zx=zx; zx <= horz+posn; zx++) {
            if (zx<0) continue; // (can't)
            if (zx >= data.length) break;
            data[zx] = data[zx]|bitz;} //~for
          } //~if
        else if (more==2) { // ** (untested)
          zx = 0;
          if (posn<0) {
            zx = -posn;
            posn = 0;} //~if
          for (zx=zx; zx <= horz-1; zx++) {
            bitz = nuly.GetIxData(zx,0);
            if (vert<0) bitz = bitz>>(-vert);
              else bitz = bitz<<vert;
            ix = zx+posn;
            if (ix<0) continue; // (probly can't)
            if (ix >= data.length) break;
            data[ix] = data[ix]|bitz;} //~for
          } //~if
        else if (more==4) { // we don't do text yet // ***
          } //~if
        else nuly = null;
        if (nuly==null) break;} //~for
      why++; // why = 7
      if (nuly==null) break; // invalid data, otherwise it's the 'AntI' group
      why++; // why = 8
      whom = GetIxWgt(nuly.GetIxData(0,0)); // this will now be the outline
      if (whom==null) break; // ants wgts must already be there
      why++; // why = 9
      pix = whom.SwapData(null);
      lxx = SafeArySz(pix);
      if (lxx<wide) {
        lxx = wide;
        if (lxx<4) lxx = 8;
        pix = new int[lxx];} //~if
      tall = -1; // soon: actual width
      if (pix != null) { // 2nd: find the outline...............................
        for (nx=lxx-1; nx >= 0; nx += -1)
          if (nx >= 0) pix[nx] = 0;
        bitz = -1;
        prio = 0;      // first do right-to-left..
        for (nx=wide-1; nx >= 0; nx += -1) {
          if (nx<0) break;
          if (nx >= data.length) continue;
          zx = data[nx];
          if (zx==0) if (prio==0) continue; // no pix [yet]
          if (zx !=0) if (tall<0) tall = nx+2;
          ix = zx&-zx;
          kx = TopBit(zx);
          if (nx+3>wide) { // special-case singletons at right
            yx = ~(zx|bitz); // (now =0, but !=0 in next=last frame)
            if (yx !=0) {
              if ((yx<0)||(yx>kx)) {
                zx = yx|zx; // pretend singletons are 2-pix wide..
                kx = TopBit(yx);} //~if
              info = yx&-yx;
              if (info>0) if (info<ix) {
                zx = yx|zx;
                ix = info;}}} //~if
          info = 0;
          if (nx>0) info = data[nx-1]; // look-ahead
          if (prio !=0) if (zx !=0) { // if no overlap, kill prior..
            yx = prio&-prio;
            if (kx>0) if (kx<0x40000000) if (yx>kx+kx) prio = 0;
            if (prio>0) if (prio<0x40000000) if (ix>prio+prio) prio = 0;
            if (prio !=0) if (info !=0) if (ix==kx)
                if ( ((((info<<1)|(info>>1)|info)&kx) !=0) // ink on both sides,
                    == ((((prio-yx-yx)|prio)&kx) !=0)) { // ..or neither..
              if (ix<0) ix = 0x40000000; // stretch single pixel up or down..
              else if (kx<0x10000) kx = kx<<1;
                else ix = ix>>1;} //~if   // kitty-corner counts as single..
            if (prio>0) if (prio<ix) if (((prio+prio)&ix) !=0) ix = ix>>1;
            if (kx>0) if (yx>kx) kx = kx<<1;} //~if
          prio = ix|kx;
          info = bitz&zx|prio; // leave left edge open, close it next pass..
          bitz = (bitz&~info)|(ix-1)|prio|-kx;
          if (nx >= pix.length) break;
          pix[nx] = info;} //~for
        if (tall>0) if (tall<wide) wide = tall;
        bitz = -1;
        prio = 0;      // now left-to-right to fill in left edge(s)..
        for (nx=0; nx <= wide-1; nx++) {
          if (nx<0) break;
          if (nx >= data.length) break;
          zx = data[nx];
          if (zx==0) continue; // no pix [yet] // if (prio==0)
          if (nx >= pix.length) break;
          info = pix[nx];
          yx = 0;
          if (nx+1<wide) // widen remaining singletons..
            yx = ~(zx|bitz|prio); // (now =0, but !=0 in prev frame only)
          yx = zx&bitz|yx;
          if (yx !=0) {
            info = info|yx;
            pix[nx] = info;} //~if
          ix = info&-info;
          kx = TopBit(info);
          prio = ~bitz; // = pix in previous frame (soon: 2 frames back)
          yx = ix|kx;
          bitz = (bitz&~info)|(ix-1)|yx|-kx;} //~for
        why++; // why = 10
        prio = -1; // find a beginning pixel: index*64+bit+ ..
        bitz = 0; // total + pixels in the border
        for (nx=0; nx <= wide-1; nx++) {
          if (nx<0) break;
          if (nx >= pix.length) break;
          zx = pix[nx];
          if (zx==0) {
            if (prio<0) continue; // no pix [yet]
            break;} //~if // next zero is end of the pattern?
          yx = nx<<6;
          if (prio<0) while (true) {
            if ((zx&1) !=0) {
              prio = yx;
              break;} //~if
            yx++;
            zx = zx>>1;
            if (zx<0) zx = zx&0x7FFFFFFF;} //~while
          while (zx !=0) {
            if (zx<0) bitz++;
            zx = zx<<1;}} //~for
        size = 0;
        data = whom.SwapData(pix);
        whom.SetInfo(0); // no animation here
        whom = null;
        whom = GetIxWgt(nuly.GetIxData(1,0)); // this will now be the ants
        why++; // why = 11
        if (whom==null) break; // ants wgts must already be there
        why++; // why = 12
        data = whom.SwapData(null);
        size = SafeArySz(data);
        zx = 0;
        if (bitz>20) zx = 8;
        else if (bitz>5) zx = SmallAnts[(bitz-6)&15]&15;
        if (zx==0) zx = wide+2;
          else zx = zx*wide+2;
        if (size<zx) { // *** sb wide*8, one each, 4 pix on then 4 off
          size = zx;   // but we only blink on/off for now
          if (size<4) size = 8;
          data = new int[size];} //~if
        if (data != null) for (nx=0; nx <= wide+1; nx++) {
          if (nx<0) break;
          info = 0;
          if (nx<wide) if (nx < pix.length)
            info = pix[nx];
          if (nx < data.length)
            data[nx] = info;} //~for
        // if (bitz>5) DoAntTrail(bitz,prio,wide+1,data);
        why++; // why = 13
        pix = null;
        whom.SetInfo((wide+0x10040)<<14); // 2 steps, 1.0 frame rate
        data = whom.SwapData(data);} //~if
      why = -why;
      break;} //~while
    if (Logging) if (StrLength(msg)>0) msg = msg + " ]";
    System.out.println("  (MkSprAnts #" + rx + SeeHex(") x",tye," +") + lxx
        + IffyStr((why>0)&&(why<9),"","/" + size) + " w=" + wide
        + msg + " n=" + bitz + " = " + why);
    if (why>0) return false;
    doit = AntsIcon(doit,nuly);
    return doit;} //~MkSpriteAnts

  /**
  * Floats Alert & Input dialogs to top of sprite list.
  */
  public void FloatDlogs() {
    int inx, whom, why = 0;
    if (SpriteList==null) return;
    whom = SafeWgtNum(theInputDlg);
    if (whom>0) while (true) {
      inx = SpriteList.FindElemt(whom);
      why = inx<<16;
      if (inx<0) break;
      SpriteList.DeleElemt(inx);
      inx = SpriteList.AddElemt(whom);
      break;} //~while
    else why = -0x60000;
    whom = SafeWgtNum(theAlertDlg);
    if (whom>0) while (true) {
      inx = SpriteList.FindElemt(whom);
      why = why+inx+100;
      if (inx<0) break;
      SpriteList.DeleElemt(inx);
      inx = SpriteList.AddElemt(whom);
      why = why+inx+100;
      break;} //~while
    else why = why+7;
    Floatem = 0;
    if (Logging) System.out.println(LogVH("(FlotDlgs) ",why,""));} //~FloatDlogs

  /**
  * SetInputText is a utility to set up an input macro panel with its
  * prompt and predefined input text. Use GetInputTxt to capture the
  * input text after the user has clicked "OK" or otherwise accepted it.
  *
  * @param  whom    the group widget over a pair of GameTxLn widgets
  * @param  prompt  the prompt text, or "" to leave it unchanged
  * @param  data    the predefined input text
  * @param  doit    true to set the predefined input
  *
  * @return         true if successful
  */
  public boolean SetInputText(GameWgt whom, String prompt,
      String data, boolean doit) { // doit=T: set data value
    int lxx = StrLength(data), zx = StrLength(prompt), nx = 0, why = 0;
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      nx = whom.GetRefNum();
      why++; // why = 2
      if (MonoFontNo <= 0) break; // needed font not created..
      if (BoldFontNo <= 0) break;
      why++; // why = 3
      if (zx>0) {
        nuly = FindSubWgt(whom,(BoldFontNo<<16)+4,0);
        if (nuly==null) break; // can't find prompt
        zx = nuly.PutTextLn(prompt);
        if ((nuly.GetSize()&0xFFFF)<zx) nuly.SetSize(0,zx);} //~if
      else if (lxx>0) if (Logging) prompt = data; // (for log)
      if ((lxx>0)||doit) {
        why++; // why = 4
        nuly = FindSubWgt(whom,(MonoFontNo<<16)+0x0604,0);
        if (nuly==null) break; // can't find data field
        zx = nuly.PutTextLn(data);
        if ((nuly.GetSize()&0xFFFF)<zx) nuly.SetSize(0,zx);
        if (doit) DlgInputData = nuly;} //~if
      why = 0;
      break;} //~while
    if (Logging || (why>0)) {
      if (why==4) prompt = data;
      System.out.println("  (SetInText #" + nx + IffyStr(why==0," OK",
          " failed = " + why) + ") '" + prompt + "'");} //~if
    return why==0;} //~SetInputText

  /**
  * GetInputTxt is a utility to capture the input text after the user has
  * clicked "OK" or partial input any time before then. Use SetInputText
  * to set up the input panel with its prompt and predefined input text.
  *
  * @param  whom    the group widget over at least a GameTxLn input widget
  *
  * @return         the text from that widget
  */
  public String GetInputTxt(GameWgt whom) {
    int nx = 0, why = 0;
    String msg = "";
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      nx = whom.GetRefNum();
      why++; // why = 2
      if (MonoFontNo <= 0) break; // needed font not created..
      why++; // why = 3
      nuly = FindSubWgt(whom,(MonoFontNo<<16)+0x0604,0);
      if (nuly==null) break; // can't find data field
      msg = nuly.GetTextLn();
      why = 0;
      break;} //~while
    if (Logging || (why>0))
      System.out.println("  (GetInText #" + nx + IffyStr(why==0,"",
          " failed = " + why) + ") = '" + msg + "'");
    return msg;} //~GetInputTxt

  /**
  * Calls GetInputTxt to capture the input text, then converts it to a number.
  *
  * @param  whom    the group widget over at least a GameTxLn input widget
  *
  * @return         the numeric (integer) value from that widget's text
  */
  public int ParseInFld(GameWgt whom) {
    return SafeParseInt(GetInputTxt(whom));} //~ParseInFld

  /**
  * Open the Input Dialog (macro) widget with a specified prompt and
  * predefined input text.
  *
  * @param  prompt  the prompt text
  * @param  data    the predefined input text
  *
  * @return         true if successful, false if it failed
  */
  public boolean OpenInputDlog(String prompt, String data) {
    GameEvent evh = UserEvHandler;
    GameWgt whom = theInputDlg, nuly = null;
    int nx = 0, whar = 0, why = 0;
    boolean rez = false;
    while (true) {
      why++; // why = 1
      if (DlogIsOpen>0) break; // already have something open
      why++; // why = 2
      if (whom==null) break; // no such dlg created
      nx = whom.GetRefNum();
      why++; // why = 3
      if (!SetInputText(whom,prompt,data,true)) break; // can't set field(s)
      why++; // why = 4
      nuly = DlgInputData;
      if (nuly==null) break; // didn't set it??
      why++; // why = 5
      if (Floatem>0) FloatDlogs();
      whom.HideSho(true);
      if (evh != null) {
        evh.BlurEv(null); // turns ants off
        evh.FocusEv(nuly,3,nuly.GetSize()&0xFFFF);
        if (evh.GetRefWgt(-1) != nuly) break;} //~if // didn't take focus
      DlogIsOpen = nx;
      if (evh != null) evh.OpCloEvDlog(nx);
      rez = true;
      why = 0;
      break;} //~while
    if (why>4) if (whom != null) {
      whom.HideSho(false);
      if (evh != null) evh.OpCloEvDlog(0);
      DlogIsOpen = 0;} //~if
    System.out.println("(OpInpDlog) #" + nx + " '" + prompt + "' = " + why);
    return rez;} //~OpenInputDlog

  /**
  * Open the Alert (macro) widget with a specified message.
  *
  * @param  msg    the message text
  */
  public void GameAlert(String msg) {
    GameEvent evh = UserEvHandler;
    GameWgt whom = theAlertDlg, nuly = null;
    int nx = 0, whar = 0, why = 0;
    boolean rez = false;
    while (true) {
      why++; // why = 1
      if (DlogIsOpen>0) break; // already have something open
      why++; // why = 2
      if (whom==null) break; // no such dlg created
      nx = whom.GetRefNum();
      why++; // why = 3
      if (!SetInputText(whom,msg,"",false)) break; // can't set msg field
      why++; // why = 4
      if (Floatem>0) FloatDlogs();
      whom.HideSho(true);
      DlogIsOpen = nx;
      if (evh != null) evh.OpCloEvDlog(nx);
      rez = true;
      why = 0;
      break;} //~while
    System.out.println("(GmAlert) #" + nx + " '" + msg + "' = " + why);} //~GameAlert

  /**
  * Gets the current or most recent input text (see OpenInputDlog).
  *
  * @return         the most recent text value from the Input Dialog
  */
  public String DlgEvData() { // gets current or most recent input text
    GameWgt whom = DlgInputData;
    if (whom==null) return "";
    return whom.GetTextLn();} //~DlgEvData

  /**
  * Determines if this ClickEvt was directed to an open Alert or Input Dialog,
  * and returns 1 if the OK button was clicked, 2 if the Cancel button was
  * clicked, otherwise 0 if the dialog is still open and used this event,
  * or -1 if it's not an Alert or Input Dialog.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset into this widget where it
  *                 was clicked
  * @param  horz  The horizontal offset into this widget
  *
  * @return       1 if OK, 2 if Cancel, -1 if not Alert or Input, otherwise 0
  */
  public int IsDlgEvent(GameWgt whom, int vert, int horz) {
    int rez = 0, zx = 0, nx = 0, info = 0, why = 0;
    boolean seen = false, impy = false, logy = Logging;
    String msg = ")";
    GameEvent evh = UserEvHandler;
    GameWgt nuly = null;
    while (true) {
      if (DlogIsOpen<0) DlogIsOpen = 0;
      why++; // why = 1
      if (whom==null) break; // huh? MouseDisp screwed up
      why++; // why = 2
      nx = whom.GetRefNum();
      info = whom.GetInfo();
      if (logy) msg = whom.GetIDtxt();
      if (vert<0) { // called from GameEvent base when closing..
        if (whom==theInputDlg) nuly = whom;
        else if (whom==theAlertDlg) nuly = whom;
        zx = horz&0x7FFF;
        seen = true;} //~if
      else zx = info&0xFFFF;
      if (logy) {
        if (StrLength(msg)>0) msg = ") `" + msg + SeeHex("` x",zx,"");
        else if ((info&1023) == info) msg = ") i=" + info;
          else msg = SeeHex(") x",info,"");
        if (seen) msg = msg + " (GE)";} //~if
      if (DlogIsOpen==0) break; // gotta have something open
      why++; // why = 3
      if (nuly==null) nuly = GetIxWgt(whom.GetParent());
      if (nuly==null) break; // no parent: can't be mine
      why++; // why = 4
      if (evh == null) break;
      why++; // why = 5
      if ((info&0x08000000) !=0) impy = true; // input ('I'), not alert ('A')
      if (((info>>24)&0x77) != 0x41) break; // (not 'A' or 'I')
      why++; // why = 6: now known to be either Alert or Input..
      if (zx==0x4F4B) rez++; // "OK" (rez = 1)
      else if (zx==0x436E) rez = 2; // "Cn"
      else if (whom==DlgInputData) evh.FocusEv(whom,vert,horz);
      else if (whom==theInputDlg) evh.FocusEv(whom,vert,horz);
      else if (whom==theAlertDlg) evh.FocusEv(whom,vert,horz);
        else break; // no such dlg
      why = 0;
      if (rez==0) break; // still open & active
      if (impy) {
        if (logy) msg = msg + " I/" + InputWhy;
        if (theInputDlg != null) theInputDlg.HideSho(false);} //~if
      else if (theAlertDlg != null) theAlertDlg.HideSho(false);
      if (evh != null) {
        if (seen) zx = 0; // GotEsc/Ent already saw this
          else zx = -DlogIsOpen; // so GotEsc/Ent knows
        evh.OpCloEvDlog(zx);} //~if
      DlogIsOpen = 0;
      break;} //~while
    if (why>0) rez = -1;
    if (logy) System.out.println("(IsDlgEvt/#" + nx + msg + " = " + rez
        + "/" + why );
    return rez;} //~IsDlgEvent // 1: OK, 2: Cancel, 0: still open, -1: not mine

  private int[] FontBits(int info) {
    if (info==0) return MonoFnt;
    if (info==1) return TextFnt;
    if (info==2) return ItalFnt;
    if (info==3) return TinyFnt;
    if (info==4) return BoldFnt;
    if (info==5) return HeadFnt;
    if (info==6) return NanoFnt;
    if (info==7) return ScorFnt;
    if (info==0x4D) return MonoFnt;
    if (info==0x54) return TextFnt;
    if (info==0x49) return ItalFnt;
    if (info==0x59) return TinyFnt;
    if (info==0x42) return BoldFnt;
    if (info==0x48) return HeadFnt;
    if (info==0x4E) return NanoFnt;
    if (info==0x53) return ScorFnt;
    return null;} //~FontBits

  /**
  * A text list of widgets is retrieved from the user's GameList, and
  * a complete set of game widgets is constructed from it.
  */
  private void ImportGamList() { // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    char xCh; // theAlertDlg = null, theInputDlg = null, mAntsWgt = null;
    int lino, kx, zx, nx = 0, mID = 0, lxx = 0, tye = 0, top = 0, left = 0,
        tall = 0, wide = 0, prnt = 0, color = 0, info = 0, fin = 0, why = 0;
    GameWgt whom;
    int[] data = null;
    String aWord, aLine = "", xID = "", theMacs = "", theList = "";
    GameEvent evh = UserEvHandler;
    if (evh != null) theList = evh.GameList(theWindow);
    lino = NthOffset(0,"\nWgtMacros..\n",theList);
    if (lino>0) { // got new macros..
      theMacs = RestOf(lino+8,theList);
      lino = NthOffset(0,"\n\n",theMacs);
      if (lino>0) theMacs = Substring(0,lino,theMacs);
      while (true) {
        lino = NthOffset(0,"\n",theMacs);
        if (lino<0) break;
        if ((((int)(CharAt(0,NthItemOf(' ',1,theMacs)))-48)&255)>9)
          theMacs = RestOf(lino+1,theMacs);
        if (SafeParseInt(theMacs)>222) break;} //~while
      kx = Countem(",",theMacs);
      if (SafeParseInt(theMacs)>222) if (kx>0) while (true) {
        data = new int[kx+2];
        if (data==null) break;
        for (lino=kx+1; lino >= 0; lino += -1) {
          aWord = NthItemOf(',',lino+1,theMacs);
          if (NthOffset(0,"\n",aWord) >= 0) {
            zx = Countem("\n",aWord);
            aWord = NthItemOf('\n',zx+1,aWord);} //~if
          aWord = NthItemOf(' ',1,aWord);
          if (StrLength(aLine)>80)
            aLine = Substring(0,72," " + aWord + aLine) + "..";
          zx = SafeParseInt(aWord);
          if (lino<0) break;
          if (lino < Wgt_Macros.length)
              if (Wgt_Macros[lino] != zx) {
            System.out.println("**!!** " + lino + " " + zx + " = "
                + Wgt_Macros[lino] + " ??..\n  "+aLine);
            nx++;} //~if
          if (data==null) break;
          if (lino < data.length)
             data[lino] = zx;} //~for
        if (nx>0) {
          System.out.println("**!!** " + nx + " changes");} //~if
        if (Logging||(nx>0))
          System.out.println("[IGL+Mx] " + kx + " '" + Substring(0,42,theMacs)
              );
        WgtMacros = data;
        break;} //~while
      data = null;
      theMacs = "";} //~if // (got new macros)
    if (why !=0) {}
    else if (StrLength(theList)==0) why--; // why = -1
    else for (fin=1; fin <= 999; fin++) {
      data = null;
      aLine = NthItemOf('\n',fin,theList);
      if (StrLength(aLine)==0) break; // normal exit
      why++; // why = 1
      if (Logging) System.out.println("[IGL] " + fin + ": '" + aLine + "'");
      if (SafeParseInt(aLine) != fin) break;
      if (fin==2) {
        theImpoList = theList;
        WantsRollover = false;
        WantsDrag = false;
        MonoFontNo = 0;
        TextFontNo = 0;
        ItalFontNo = 0;
        TinyFontNo = 0;
        BoldFontNo = 0;
        HeadFontNo = 0;
        NanoFontNo = 0;
        ScorFontNo = 0;} //~if
      why++; // why = 2
      mID = 0;
      if (NthOffset(0,"`",aLine)>0) { // got a macro ID..
        mID = EncodeID(aLine);
        xID = myTmpStr;} //~if
      for (nx=2; nx <= 10; nx++) {
        aWord = NthItemOf(' ',nx,aLine);
        if (StrLength(aWord)==0) break; // why = 2
        why++; // why = 3
        zx = (int)(CharAt(0,aWord));
        if (zx != (int)('-')) if (((zx-48)&255)>9) break; // why = 3
          // +0 t=1  0/0 256/512 c=0xFFFF99 i=1 p=0   d=0 'Game Win'
        tye = top;
        top = left;
        left = tall;
        tall = wide;
        wide = color;
        color = info;
        info = prnt;
        prnt = lxx;
        lxx = SafeParseInt(aWord);
        if (Logging) if (fin<8)
          System.out.println("... " + fin + "/" + nx + " = " + lxx);
        if (nx==10) why++; // why = 4
          else why--;} //~for
      zx = tye&255;
      if (info==0) if (mID>0) if ((zx==5)||(zx==0)) // if (tye>255)
         info = mID; // NOT: mID = 0; (might be inner wgts)
      if (Logging) if (fin<8) System.out.println(" => " + fin + ": " + tye
          + " " + top + " " + left + " " + tall + " " + wide + " " + prnt
          + " " + color + " " + info + " +" + lxx + " = " + why);
      if (why<4) break;
      if (lxx>0) {
        data = new int[lxx];
        if (data==null) break; // why = 4
        why++; // why = 5
        for (nx=0; nx <= lxx-1; nx++) {
          aWord = NthItemOf(' ',nx+11,aLine);
          if (StrLength(aWord)==0) break; // why = 5
          why++; // why = 6
          zx = (int)(CharAt(0,aWord));
          if (zx != (int)('-')) if (((zx-48)&255)>9) break; // why = 6
          zx = SafeParseInt(aWord);
          if (Logging) if (fin<8)
            System.out.println(" ++ " + fin + " " + nx + "/" + lxx + " = " + zx);
          if (data != null) if (nx < data.length)
            if (nx >= 0) data[nx] = zx;
          if (nx+1 == lxx) why++; // why = 7
            else why--;}} //~if //~for (nx)
      aWord = "";
      if (NthOffset(0,"'",aLine)>0) {
        aWord = NthItemOf('\'',2,aLine);
        if (StrLength(aWord)>0) if (lxx==0) if ((tye&255)==4) {
          lxx = StrLength(aWord)+4;
          data = new int[lxx>>2];
          if (data==null) break; // why = 7
          for (nx=0; nx <= lxx; nx++) {
            zx = 0;
            if (nx<StrLength(aWord)) zx = (int)(CharAt(nx,aWord))&127;
            kx = nx>>2;
            if (data==null) break;
            if (kx<0) break;
            if (kx >= data.length) break;
            data[kx] = (data[kx]>>8)+(zx<<24);
            if (Logging) if (fin<8) System.out.println(" ++ " + fin + " " + nx
                + "/" + lxx + " = " + zx + " => [" + kx + SeeHex("] x",data[kx],
                IffyStr(nx>0,""," '" + aWord + "'")));} //~for
          aWord = "";}} //~if
      why++; // why = 8
      zx = tye&255;
      if (zx>222) { // queue up this macro for later processing..
        theMacs = theMacs + aLine + "\n";
        tye = (tye&-256)|5;} //~if
      else if (zx==3) if (data==null)  // if (info&15 == info)
        data = FontBits(info);
        //  NewGameWgt(int tye, int top, int left, int tall, int wide,
        //    int prnt, int color, int info, int[] data, GameWgt whom)
      if (fin==1) if (NextRefNo==0) {
        whom = NewGameWgt(1,top,left,tall,wide,0,color,1,null,null);
        if (whom==null) break; // why = 8
        GameRoot = whom;} //~if // (unused)
      why++; // why = 9
      whom = NewGameWgt(tye,top,left,tall,wide,prnt,color,info,data,null);
        // ("..nuGmWgt #" + rnum + "=" + tye + " " + top + "/" + left + " " + tall
        //    + "/" + wide + " " + info + "/" + color + " " + data + " " + prnt
      // => ..nuGmWgt +0=1 0/0 256/512 1/16777113 false 0
      if (whom==null) break; // why = 9
      if (fin>1) {
        if ((tye&4096) !=0) WantsRollover = true;
        if ((tye&2048) !=0) WantsDrag = true;} //~if
      nx = whom.GetRefNum();
      if (zx==3) {
        if (info==0) MonoFontNo = nx;
        else if (info==1) TextFontNo = nx;
        else if (info==2) ItalFontNo = nx;
        else if (info==3) TinyFontNo = nx;
        else if (info==4) BoldFontNo = nx;
        else if (info==5) HeadFontNo = nx;
        else if (info==6) NanoFontNo = nx;
        else if (info==7) ScorFontNo = nx;} //~if
      if (Logging) whom.LogMe(".. ");
      why++; // why = 10
      if (nx != fin) break;
      why = 0;} //~for // (fin)
    if (why==0) for (lino=1; lino <= 999; lino++) {
      data = null;
      aLine = NthItemOf('\n',lino,theMacs);
      if (StrLength(aLine)==0) break; // normal exit
      why = 11;
      if (Logging) System.out.println("[IGLmax] " + lino + ": '" + aLine + "'");
      nx = SafeParseInt(aLine);
      whom = GetIxWgt(nx);
      if (whom==null) break;
      why++; // why = 12
      if (whom.GetRefNum() != nx) break;
      why++; // why = 13
      if ((whom.GetType()&255) != 5) break;
      why++; // why = 14
      why = BuildThisMacro(whom,aLine,why);
      if (why>0) break;
      if (why==0) continue;
      theMacs = theMacs + myTmpStr;
      myTmpStr = "";} //~for // (lino)
    if (why==0) {
      GameWgtsList = FindListWgt("[List of item lists]");
      SpriteList = FindListWgt("[Sprite list]");
      theAlertDlg = FindListWgt("[Std Alert Dialog]");
      theInputDlg = FindListWgt("[Std Input Dialog]");
      if (theInputDlg != null) InputDlgNo = theInputDlg.GetRefNum();
      mAntsWgt = FindListWgt("[Deployable Marching Ants]");
      if (mAntsWgt != null) Antsify(false,mAntsWgt);
      if (theInputDlg != null) {
        Widen1stWgt(theInputDlg,-64);
        theInputDlg.HideSho(false);
        Floatem = 2;} //~if
      if (theAlertDlg != null) {
        theAlertDlg.HideSho(false);
        Floatem++;}} //~if
    if (Logging||true) System.out.println("[IGL] = " + why + " f=" + Floatem
        + " #=" + fin + " m=" + lino + IffyStr(WantsDrag," wD "," -d ")
        + IffyStr(WantsRollover,"wR\n","-r\n") + theMacs);} //~ImportGamList

  /**
  * All known widgets are logged to the system console. Short prefixes are
  * repeated every line, longer prefixes are replaced with ".. " after the
  * first line.
  *
  * @param  why   a prefix string inserted in front of first/each log line
  */
  public static void LogAlWgts(String why) { // caller ensures if (Logging)
    int lino = StrLength(why);
    boolean didit = false;
    GameWgt whom = AllWgts[0];
    if (whom != null) {
      whom.LogMe(why);
      if (lino>4) why = ".. ";
      for (lino=1; lino <= MaxWgts; lino++) {
        whom = GetIxWgt(lino);
        if (whom==null) {
          System.out.println("<AllWgts[" + lino + "]=null>");
          if (didit) break;
          didit = true;
          continue;} //~if
        whom.LogMe(why);}} //~if //~for (lino)
    else System.out.println("<AllWgts[0]=null>");} //~LogAlWgts

  /**
  * Determines if the ASCII code of a character is a valid letter or digit.
  * Only the original ASCII character set is supported. '_' is a letter.
  *
  * @param  whom  the ASCII code of a character
  *
  * @return       true if it's a letter or digit
  */
  public boolean IzLetter(int whom) {
    if (whom<48) return false;
    if (whom<58) return true;
    if (whom<65) return false;
    if (whom<90) return true;
    if (whom<97) if (whom !=95) return false;
    if (whom>122) return false;
    return true;} //~IzLetter

  // special-case class vars, GameList; maybe reduce Head font by usage
  // remove 'myGame.', 'JavaGame.', 'super.', '!= null'; cvt '=null' to =0
  // JSwgtMeths cvt 'xxx.zzz(' => 'zzz(xxx,'; OrideGame cvt 'zzz(' => 'o_zzz('

  // 0:Gtype, 1:Gcolor, 2:Gtop,Gleft, 3:Gtall,Gwide,
  // 4:Ginfo, 5:Gselim, 6:Parent, 7:Gdata;

  /**
  * Constructs an array representation of the widget list for JavaScript.
  *
  * @param  theText  the widget list as returned by GameList
  *
  * @return          the text of an array of numbers for the JS GameEngine
  */
  private String MakeJSwgts(String theText) {
    int lxx = Countem("\n",theText)+myPL, base = 9990000, why = 16, sofar = 8;
    boolean doit, logy = (Dev_Log||Logging)&&(lxx<22);
    int lino, nx, zx, top, left, tall, wide, info, whom, here, thar = 0,
        tye = 0, later = 0, more = 0;
    int[] pix = null;
    String ary = "", soon = "", prio = "", aLine = "",
        aWord, aStr, colo, prnt, fin = NthItemOf('\n',lxx,theText),
        data = NthItemOf('\n',1,theText) + "\n      = \"";
    GameWgt nuly = null;
    while (true) {
      info = GetGameValue(11); // = InputDlgNo
      nuly = GetIxWgt(info); // = theInputDlg
      info = GetGameValue(0); // = MonoFontNo
      nuly = FindSubWgt(nuly,(info<<16)+0x0604,0); // if (nuly != null)
      for (lino=1; lino <= lxx-1; lino++) {
        doit = false;
        aLine = NthItemOf('\n',lino+1,theText);
        if (StrLength(aLine)==0) break; // why = 16: poor form (unexp'd blnk ln)
        if (NthOffset(0,"\"\"",aLine)>0) { // normal exit..
          why = 0;
          break;} //~if
        aLine = NthItemOf('\"',2,aLine);
        zx = SafeParseInt(aLine);
        if (zx != lino) {
          why++; // why = 17
          if (!logy) break;} //~if // why = 17: invalid line +
        // here = (lino-1)*8;
        tye = SafeParseInt(NthItemOf(' ',2,aLine));
        whom = tye&255;
        if (whom>7) {
          why = 18;
          if (!logy) break;} //~if // why = 18: no macros or private wgt types
        top = SafeParseInt(NthItemOf(' ',3,aLine));
        left = SafeParseInt(NthItemOf(' ',4,aLine));
        tall = SafeParseInt(NthItemOf(' ',5,aLine));
        wide = SafeParseInt(NthItemOf(' ',6,aLine));
        colo = NthItemOf(' ',7,aLine);
        aWord = NthItemOf(' ',8,aLine);
        prnt = NthItemOf(' ',9,aLine);
        nx = SafeParseInt(NthItemOf(' ',10,aLine));
        if (StrLength(prnt)==0) why = 20; // why = 20: line is too short
        if (nx<0) why = 21; // why = 21: invalid negative data length
        tall = (tall<<16)+wide;
        top = (top<<16)+left;
        if (Dev_Log||Logging) if (logy) System.out.println("  ..Mk.. "
            + lino + ": " + tye + LogVH(" ",top,LogVH(" ",tall," c="))
            + colo + " i=" + aWord + " '" + prnt + "' " + nx
            + " '" + Substring(0,25,aLine) + "..'");
        if (why>16) break; // why = 17/18/20/21
        if (lino==1) { // special-case 1st line..
          data = data + "1," + colo + HexIfMore(",0,",tall,",1,0,0,0,");
          colo = "-1";} //~if
        if (whom==4) {
          aStr = NthItemOf('\'',2,aLine);
          here = NthOffset(0,"'" + aStr + "'",aLine);
          if (here>0) aLine = Substring(0,here,aLine);
          else for (thar=1; thar <= 16; thar++) { // this shouldn't happen, so OK if slo..
            nx = StrLength(aLine);
            aLine = ReplacAll(" ","  ",aLine);
            if ((int)(CharAt(0,aLine))<33) aLine = RestOf(1,aLine);
            zx = StrLength(aLine);
            if (zx<nx) {nx = zx;
              continue;} //~if
            zx = Countem(" ",aLine)+myPL;
            if (zx==10) break;
              // if (StrLength(aLine)>nx) continue; break;} //~if
            if (zx>10) {
              if (zx>11) nx = nx-(zx-10)*2;
                else nx--;
              aLine = Substring(0,nx,aLine);
              continue;} //~if
            if (nx>0) if ((int)(CharAt(nx-1,aLine))>32)
              aLine = aLine + " ";
            aLine = aLine + "0 ";} //~for
          nx = StrLength(aStr);
          for (zx=0; zx <= nx; zx += 4) {
            info = 0;
            if (zx<nx) // if (nuly != null)
              info = PuTxGetData(nuly,Substring(zx,4,aStr));
            aLine = aLine + HexIfMore(" ",info,"");} //~for
          aLine = aLine + " 0 '" + aStr + "'";
          if (nx>2) {
            soon = soon + aLine + "\n";
            later++;
            nx = (nx>>2)+1;
            nx = -(nx*256+later);} //~if
          else nx = 1;} //~if // (whom=4)
        else if (whom==3) {
          soon = soon + aLine + "\n";
          later++;
          doit = true;
          nx = -later;} //~if
        else if (whom==2) if (SafeParseInt(aWord)+3 ==0) {
          pix = KnownCircleData(wide,false);
          nx = SafeArySz(pix);
          zx = NthOffset(0,aWord,aLine);
          aLine = Substring(0,zx,aLine) + aWord + " " + prnt + " " + nx + " ";
          for (zx=0; zx <= nx; zx++)
            aLine = aLine + HexIfMore(" ",SafeAryElt(zx,pix,0),"");
          soon = soon + aLine + "\n";
          later++;
          nx = -(nx*256+later);} //~if // (whom=2)
        if (nx>0) aStr = "$$" + CvInt2Str((nx<<16)+more);
        else if (doit) aStr = "$" + CvInt2Str(nx);
        else if (nx<0) aStr = "$$" + CvInt2Str(nx);
          else aStr = "0";
        data = data + "\"\n      + \"" + tye + "," + colo + HexIfMore(",",top,
            HexIfMore(",",tall,",")) + aWord + ",0,"
            + prnt + "," + aStr + IffyStr((lino&3)==0,", // #" + lino,",");
        sofar = sofar+8;
        for (zx=1; zx <= nx; zx++) {
          if ((more&15)==0) ary = ary + "\"\n      + \"";
          ary = ary + NthItemOf(' ',zx+10,aLine) + ",";
          more++;}} //~for // why = 16: poor form (no end marker by line +255)
      if (why>0) break;
      if (Dev_Log||Logging) if (logy) System.out.println("  (MkJSwgt) " + sofar
          + " +" + more + " -" + later + ": ---->\n" + soon + "<----\n"
          + theText + " ---->\n" + data + "<----\n");
      later = 0;
      whom = 0;
      while (true) {
        pix = null;
        zx = 2;
        here = NthOffset(0,",$$",data);
        if (here<0) {
          here = NthOffset(0,",$",data); // do the fonts last
          if (here<0) break; // normal exit
          zx--;} //~if
        here++;
        aWord = NthItemOf(',',1,Substring(here+zx,10,data));
        zx = StrLength(aWord)+zx;
        nx = SafeParseInt(aWord);
        if (nx<0) {
          thar = -nx;
          lino = thar&255;
          thar = thar>>8;
          aLine = NthItemOf('\n',lino,soon);
          tye = SafeParseInt(NthItemOf(' ',2,aLine))&255;
          if (tye==3) {
            info = SafeParseInt(NthItemOf(' ',8,aLine));
            pix = FontBits(info);
            thar = SafeArySz(pix);} //~if
          aStr = HexIfMore("",(thar<<16)+sofar+more,"");} //~if
        else aStr = HexIfMore("",nx+sofar,"");
        data = Substring(0,here,data) + aStr + RestOf(here+zx,data);
        if (nx >= 0) continue;
        if (tye==3) {
          if ((Qlog&128) !=0) continue;
          if (pix==null) {
            why = 22;
            break;}} //~if // why = 22: no such font
        doit = true;
        prio = "";
        for (nx=0; nx <= thar; nx++) {
          info = 0;
          if (tye !=3)
            info = SafeParseInt(NthItemOf(' ',nx+11,aLine));
          else if (pix != null) if (nx >= 0)
            if (nx < pix.length)
              info = pix[nx];
          aStr = HexIfMore("",info,",");
          zx = StrLength(aStr);
          if ((zx+later>80)||doit) {
            ary = ary + "\""
                + prio + "\n      + \"";
            later = 11;
            if (whom>60) doit = true;
            if (doit) {
              prio = " // @ " + (sofar+more) + "..";
              later = StrLength(prio)+later;
              whom = 0;} //~if
            else prio = "";
            doit = false;} //~if
          ary = ary + aStr;
          later = later+zx;
          more++;}} //~while //~for (nx)
      data = data + ary + "0\";\n";
      why = 0;
      break;} //~while
    if (why>0) data = CvInt2Str(why) + data; // sho err msg
    return data + fin;} //~MakeJSwgts

  /**
  * Converts the Java code of your game into JavaScript and stores
  * the result in class variable JSinfo, and also writes to fiName.js
  * if successful. If there's an error, the error description is stored
  * in class variable JSinfo (length less than 99).
  *
  * @param  theText  the Java code
  * @param  fiName   the game name, should = the class name
  *
  * @return          true if successful
  */
  public boolean toJS(String theText, String fiName) {
    // char xCh;
    int here, thar, whar, sz, zx, nx, kx = Countem("\n",theText)+myPL,
        lxx = kx, ofx = StrLength(theText), nCh = 0, lino = 0, why = 0;
    boolean dont = false, logy = false, logz = false;
    String xStr, aStr, aLine, aWord = "", ary = "";
    if (lxx==0) if (StrLength(fiName)>0) {
      theText = ReadWholeTextFile(fiName + ".java");
      lxx = Countem("\n",theText)+myPL;} //~if
    here = NthOffset(0,"public " + fiName + "()",theText);
    whar = 0;
    if (here>0) if (here+99>ofx) {
      theText = Substring(0,here,theText) + "\nGameStart();\n";
      whar = ofx-here;
      ofx = here+16;} //~if
    if (Dev_Log) {
      if (ofx<999) xStr = theText;
      else if (SafeCompare(fiName,"Pong")==0) xStr = theText;
        else xStr = Substring(0,999,theText);
      System.out.println("** 2JS  '" + fiName + "' +" + ofx + " -" + whar  // ** 2JS
          + " @ " + here + "....\n" + xStr + "....");} //~if
    while (true) {
      if (Dev_Log||Logging) if (lxx<555) logy = true;
      why++; // why = 1
      here = NthOffset(0,"public class ",theText);
      thar = NthOffset(0," extends GameEvent {",theText);
      if (here<0) break; // why = 1: can't find class name
      lino = Countem("\n",Substring(0,here+1,theText))+myPL;
      if (thar<here) break;
      xStr = NthItemOf(' ',1,Substring(here+12,thar-12-here,theText));
      if (StrLength(xStr)==0) break; // why = 1: can't find class name
      if (SafeCompare(xStr,fiName) !=0) break; // ..or != fiName
      theText = Replace("\r",here,thar+20-here,theText);
      fiName = fiName + ".js";
      here = NthOffset(0,"void main()",theText);
      thar = 0;
      if (here>0) for (here=here-1; here >= 0; here += -1) { // delete the main..
        zx = (int)(CharAt(here,theText));
        if (zx>31) continue;
        if (zx <= 0) break;
        if (thar>0) { // ..by including it in its own Jdoc block-comment..
          aLine = Substring(here,thar-here,theText); // (soon to be del'd)
          zx = NthOffset(0,"*/",aLine);
          if (Dev_Log) System.out.println("  --main " + here + "/" + thar
              + ": " + zx + " '" + aLine + "'");
          if (zx<0) break;
          theText = Replace("\n\r",here,thar-here,theText);
          break;} //~if
        else thar = here;} //~for
      why++; // why = 2
      while (true) { // convert 'Lscore = myGame.FindListWgt("{Lscore}");' to ....
        here = NthOffset(0,"myGame.FindListWgt(\"",theText);        // ..const..
        if (here<0) break; // normal exit
        lino = Countem("\n",Substring(0,here+1,theText))+myPL;
        aStr = NthItemOf(';',1,Substring(here,99,theText));
        sz = StrLength(aStr);
        nx = NthOffset(0,"\"",aStr);
        thar = 0;
        if (nx>0) {
          aWord = NthItemOf('\"',2,aStr);
          thar = NthOffset(0,aWord,theText);} //~if
        if (Dev_Log) System.out.println("  wgt=const " + here + " +" + sz
            + " " + nx + "/" + thar + " '" + aStr + "'");
        if (nx<0) break; // why = 2: invalid search string
        if (thar>99) {
          aLine = Substring(thar-99,99,theText);
          nx = Countem("\n",aLine)+myPL;
          aLine = NthItemOf('\n',nx,aLine)
              + NthItemOf('\n',1,Substring(thar,99,theText));
          aLine = NthItemOf('\"',2,aLine);
          if (NthOffset(0,aWord,aLine)>0) thar = SafeParseInt(aLine);
            else thar = 0;} //~if
        else thar = 0; // didn't find it (presumed OK)
        theText = Replace(CvInt2Str(thar*8),here,sz,theText);} //~while
      here = NthOffset(0,"<$..>",theText);
      thar = NthOffset(0,"<..$>",theText);
      if (Dev_Log||Logging) System.out.println("** 2JS  '" + fiName        // ** 2JS
          + "' +" + lxx + "/" + kx + " (" + ofx + ") " + here + "/" + thar);
      why++; // why = 3
      if (here<0) break; // why = 3: Can't find widget list
      if (thar<here) break;
      aStr = MakeJSwgts(Substring(here,thar-here,theText)); // make wgt array...
      theText = Replace(aStr,here,thar+10-here,theText);
      lxx = Countem("\n",aStr)+lxx+myPL; // probly got bigger
      nx = SafeParseInt(aStr);
      if (nx>0) if (nx<32) {
        why = nx;
        break;} //~if // why = 16+..
      why++; // why = 4
      ofx = StrLength(theText);
      if (NthOffset(0,"\n\n\n\n",theText) >= 0) while (true) {
        theText = ReplacAll("\n\n","\n\n\n",theText);
        sz = StrLength(theText);
        if (sz<ofx) {
          lxx = lxx-ofx+sz; // reduce the line count
          ofx = sz;} //~if
        else break;} //~while
      kx = 0;
      zx = 0;
      for (lino=1; lino <= 9999; lino++) { // remove comments (1*brk, 5*cont) ...............2JS
        aLine = NthItemOf('\n',lino,theText);
        sz = StrLength(aLine);
        if (sz==0) if (lino>33) {
          if (kx+8<lino) break; // why = 4: too many blank lines in a row..
          continue;} //~if  // ..already got rid of them, this must be an error
        kx = lino;
        ofx = 0;
        here = 0;
        if (sz>0) {
          if (!dont) {
            thar = NthOffset(0,"//",aLine);
            if (thar<0) here = NthOffset(0,"/*",aLine);} //~if
          else thar = NthOffset(0,"*/",aLine);    // prior comments already gone..
          if (here >= 0) ofx = NthOffset(0,aLine,theText);} //~if
        if (Dev_Log||Logging) if (logy||logz)
          System.out.println("  .. " + lino + IffyStr(dont,": /* ",IffyStr(thar<0,
                IffyStr(sz>30,": _ ",": (" + lxx + ") "),": // "))
              + here + "/" + thar + " @ " + ofx + " +" + sz + " '"
              + Substring(0,50,aLine) + IffyStr(sz>50,"..' ","' ") + zx);
        if (sz==0) continue;
        if (here<0) continue; // only if no comments on this line
        zx = 0;
        aStr = "\r";
        if (dont) { // thar = NthOffset(0,"*/",aLine);
          if (thar >= 0) {
            aLine = RestOf(thar+2,aLine);
            if (StrLength(NthItemOf(' ',1,aLine))>0) aStr = aLine;
            dont = false;} //~if
          // theText = PutLineOf(aStr,lino,theText);
          theText = Replace(aStr,ofx,sz,theText);
          zx = -2;
          continue;} //~if
        aWord = NthItemOf(' ',1,aLine);
        if (StrLength(aWord)==0) continue;
        here = NthOffset(0,"\"",aLine);
        // thar = NthOffset(0,"//",aLine);
        if (here >= 0) if (here<thar) thar = -4;
        if ((int)(CharAt(0,aWord)) == 47) { // '/'
          nCh = (int)(CharAt(1,aWord));
          zx++; // zx = 1
          if (nCh == 42) { // '*'
            dont = true;
            zx++;} //~if // zx = 2
          else if (nCh == 47) { // comment, is it unused o'ride meth?..
            thar = NthOffset(0,"(",aLine);
            if (thar>16)
                if (SafeCompare(NthItemOf(' ',2,aLine),"public")==0) {
              aStr = Substring(0,thar+1,aLine);
              xStr = "_";
              zx = Countem(" ",aStr)+1; // might be doubled-up spaces..
              while (zx>0) {
                xStr = NthItemOf(' ',zx,aStr);
                if (StrLength(xStr)>0) break;
                zx--;
                if (zx==0) xStr = "_";} //~while // so the search will fail
              aStr = "_" + xStr;
              zx = NthOffset(0,aStr,OrideGameMeths); // zx = >100/-1
              if (zx>0) aStr = "function o" // yup
                  + NthItemOf('\n',1,Substring(zx,99,OrideGameMeths));
              else aStr = "\r";}} //~if
          else if (thar>0) { // thar = NthOffset(0,"//",aLine);
            aStr = Substring(0,thar,aLine);
            zx = StrLength(aStr)+3;} //~if
          else continue; // zx = 1
          theText = Replace(aStr,ofx,sz,theText);} //~if
        else if (thar>0) {
          aStr = Substring(0,thar,aLine);
          theText = Replace(aStr,ofx,sz,theText);
          if (false) System.out.println("  .. " + lino + ": //" + thar
              + " - '" + RestOf(thar,aLine) + "'");}} //~for // (remove comments)
      // if (kx+8<lxx) break; // why = 4: too many blank lines in a row
      lxx = kx+8;
      why++; // why = 5
      dont = false;
      for (lino=1; lino <= lxx; lino++) { // no errors this segment (replace var/fun types)..2JS
        aLine = NthItemOf('\n',lino,theText);
        nx = 0;
        aWord = NthItemOf(' ',1,aLine);
        zx = StrLength(aWord);
        if (zx==0) continue;
        if (dont) {
          if (NthOffset(0,";",aLine)>0) dont = false;
          else if (NthOffset(0,"{",aLine)>0) dont = false;
          if (Logging) if (logz) System.out.println("    , " + lino + "/" + nx
               + ": '" + aWord + "' " + dont);
          continue;} //~if
        if (NthOffset(0,";",aLine)<0) if (NthOffset(0,"{",aLine)<0) dont = true;
        sz = StrLength(aLine);
        ofx = NthOffset(0,aLine,theText); // safe, all type names already gone
        while (true) {
          if (Logging) if (logz) System.out.println("    . " + lino + "/" + nx
               + ": '" + aWord + "' " + zx  + " " + dont);
          if (zx==0) break;
          aStr = " " + aWord + " ";
          if (NthOffset(0,aStr,JavaTypes)<0) break; // normal exit
          here = NthOffset(0,aWord,aLine);
          aLine = Replace("",here,zx+1,aLine);
          aWord = NthItemOf(' ',1,aLine);
          zx = StrLength(aWord);
          nx++;} //~while
        if (nx==0) continue;
        if (zx==0) continue;
        here = NthOffset(0,"(",aLine);
        thar = NthOffset(0,"=",aLine);
        aStr = "var ";
        if (here>0) if ((thar<0)||(thar>here)) {
          aStr = "function ";
          thar = -2;} //~if
        here = NthOffset(0,aWord,aLine);
        aLine = Replace(aStr,here,0,aLine);
        if (Dev_Log||Logging) if (logy||logz)
          System.out.println("  :: " + lino + " => '" + Substring(0,55,aLine)
              + IffyStr(nx>55,"..' ","' ") + dont);
        theText = Replace(aLine,ofx,sz,theText);
        if (here>0) if (thar<0) for (thar=0; thar <= 9; thar++) { // de-type the params..
          nx = 0;
          if (thar>0) {
            aLine = NthItemOf('\n',lino+thar,theText);
            ofx = NthOffset(0,aLine,theText); // safe if needed
            sz = StrLength(aLine);} //~if
          else here = NthOffset(0,"(",aLine);
          if (here<0) break; // can't
          whar = NthOffset(0,")",aLine);
          kx = Countem(" ",JavaTypes); // slightly high in Jv
          if (whar==0) kx = 0;
          else if (whar>0)
            if (StrLength(NthItemOf(' ',1,Substring(0,whar,aLine)))==0) kx = 0;
          if (kx>0) for (kx=kx; kx >= 1; kx += -1) { // all known types..
            aWord = NthItemOf(' ',kx,JavaTypes) + " ";
            zx = StrLength(aWord);
            if (zx<2) continue;
            while (true) {
              here = NthOffset(0,aWord,aLine);
              if (here<0) break;
              if (whar>0) if (here>whar) break;
              if (here>0) if (IzLetter((int)(CharAt(here-1,aLine)))) {
                aLine = Replace("$",here+1,0,aLine);
                continue;} //~if
              aLine = Replace("",here,zx,aLine);
              nx++;} //~while
            aLine = ReplacAll("","$",aLine);} //~for // (kx)
          if (nx>0) {
            if (Dev_Log||Logging) if (logy||logz) System.out.println("  ..." + lino
                + " => '" + Substring(0,55,aLine) + IffyStr(nx>55,"..'","'"));
            theText = Replace(aLine,ofx,sz,theText);} //~if
          if (whar >= 0) break;}} //~for // (thar: de-type the params) (lino)
      why++; // why = 6
      dont = false;
      lino = Countem(" ",CantDoMeths);
      for (lino=lino; lino >= 1; lino += -1) { // unsupported method calls....................2JS
        aWord = NthItemOf(' ',lino,CantDoMeths);
        if (StrLength(aWord)<2) continue;
        if (NthOffset(0,aWord,theText)<0) continue;
        dont = true;
        break;} //~for
      if (dont) {
        lino = 0;
        break;} //~if // why = 6: Can't do this method
      why++; // why = 7
      for (lino=1; lino <= 33; lino++) { // no errors this segment (GameEvent overrides)........
        aWord = NthItemOf(' ',lino,OrideGameMeths);
        lxx = StrLength(aWord);
        if (lxx==1) break; // normal exit
        dont = true; // "doit" = false if it's not really mine
        here = NthOffset(0,aWord,theText);
        if (here<0) continue;
        aStr = Replace("$",1,0,aWord);
        nCh = (int)(CharAt(here+lxx,theText));
        theText = Replace(aStr,here,lxx,theText);
        if (here==0) dont = false;
        else if (IzLetter(nCh)) dont = false;
        else if (IzLetter((int)(CharAt(here-1,theText)))) dont = false;
        if (dont) for (thar=here-1; thar >= 0; thar += -1) {
          nCh = (int)(CharAt(thar,theText));
          if (nCh <= 32) continue;
          if (nCh == 46) break; // '.'
          theText = Replace("o_",here,0,theText);
          break;} //~for // OK either way
        if (Dev_Log||Logging) if (logy||logz)
          System.out.println("  ~~ '" + aWord + "' " + dont + " @ " + here);
        theText = ReplacAll(aWord,aStr,theText);} //~for
      why++; // why = 8
      lino = Countem(" ",JSwgtMeths);
      for (lino=lino; lino >= 1; lino += -1) { // GameWgt method call fixups.......
        xStr = NthItemOf(' ',lino,JSwgtMeths);
        if (StrLength(xStr)<2) continue; // Jv macro counts spaces, not words
        nx = lino;
        aWord = "." + xStr + "(";
        while (true) { // every place it's called
          here = NthOffset(0,aWord,theText);
          if (here<0) break;
          lxx = StrLength(aWord);
          for (nx=0; nx <= 9; nx++) {
            nCh = (int)(CharAt(here+lxx+nx,theText));
            if (nCh <= 32) continue;
            break;} //~for
          if (SafeCompare(xStr,"toString")==0) { // no obj to o'ride, rename it..
            theText = Replace("_",here+3,0,theText);
            lxx++;} //~if
          nx = 0;
          dont = (nCh != 41); // ')'
          for (thar=here-1; thar >= 0; thar += -1) {
            nCh = (int)(CharAt(thar,theText));
            if (IzLetter(nCh)) continue;
            nx = here-1-thar;
            if (nx <= 0) break;
            aStr = NthItemOf(' ',1,Substring(thar+1,nx,theText));
            if (StrLength(aStr)==0) {
              nx = 0;
              break;} //~if
            if (dont) aStr = aStr + ",";
            if (Dev_Log||Logging) if (logy||logz) System.out.println("  -> '"
                + aWord + aStr + "' +" + nx + " @ " + here);
            theText = Replace("",thar+1,nx+1,Replace(aStr,here+lxx,0,theText));
            break;}} //~while //~for (thar) (every place it's called)
        if (nx <= 0) if (Dev_Log||Logging) System.out.println("  ? '" + aWord
            + "' +" + nx + " @ " + here);} //~for // (lino: GameWgt method call..)
      if (nx <= 0) {
        if (here>0) lino = Countem("\n",Substring(0,here+1,theText))+myPL;
          else lino = 0;
        if (!logz) break;} //~if // why = 8: invalid GameWgt object in meth call
      why++; // why = 9
      while (true) { // we know how to do these kinds of things: (null fixes)...
        dont = false;                 // 'if (xxx != null)' => 'if (xxx)'
        aWord = "null";               // '(xxx == null)' => '(xxx == 0)'
        here = NthOffset(0,aWord,theText); // 'xxx = null' => 'xxx = 0'
        lino = 0;                     // ',null,' => ',0,' // '(null)' => '(0)'
        if (here<0) {
          why = 0; // normal exit, no errors
          break;} //~if
        lino = Countem("\n",Substring(0,here+1,theText))+myPL;
        if (IzLetter((int)(CharAt(here+4,theText)))) dont = true;
        else if (here>0) if (IzLetter((int)(CharAt(here-1,theText))))
          dont = true;
        if (dont) {
          theText = Replace("$",here+1,0,theText);
          continue;} //~if
        lino = Countem("\n",Substring(0,here+1,theText))+myPL;
        aStr = Substring(here+4,99,theText);
        nCh = (int)(CharAt(0,NthItemOf(' ',1,aStr)));
        if (nCh == 41) dont = true; // ')'
        else if (nCh != 44) // ','
          if (nCh != 59) break; // ';' why = 9: unrecognized op after 'null'
        if (here<8) break;
        lxx = here-99;
        if (lxx <= 0) lxx = 1;
        for (thar=here-1; thar >= lxx; thar += -1) {
          nCh = (int)(CharAt(thar,theText));
          if (nCh <= 32) if (thar>lxx) continue;
          if (nCh == 44) break; // ','
          if (nCh == 40) break; // '('
          if (nCh != 61) { // '='
            nCh = 46; // '.' can't do this one
            break;} //~if
          if (dont) nCh = (int)(CharAt(thar-1,theText));
          if (nCh == 33) { // '!' dont=true
            aLine = Substring(lxx,thar-1-lxx,theText);
            while (true) {
              lxx = StrLength(aLine);
              nx = NthOffset(0,"if",aLine);
              if (nx <= 0) break;
              kx = NthOffset(1,"if",aLine);
              if (kx>0) {
                aLine = RestOf(kx-1,aLine);
                continue;} //~if
              zx = NthOffset(0,"(",aLine);
              if (zx<0) break;
              kx = NthOffset(1,"(",aLine);
              if (kx <= 0) break;
              aLine = RestOf(kx,aLine);} //~while
            if (nx>0) if (zx>nx) if (zx<lxx) {
              aLine = RestOf(zx+1,aLine);
              if (StrLength(NthItemOf(' ',1,aLine))>0)
                  if (StrLength(NthItemOf(' ',2,aLine))==0) {
                theText = Replace("",thar-1,here+5-thar,theText);
                break;}}} //~if
          if (nCh == 33) nCh = 36; // '$' failed the 'if' test, gotta put 0
          if (nCh == 46) nCh = 60; // '`' shouldn't hap
          break;} //~for
        if (nCh != 33) if (nCh != 46) // '.' is err, already deleted '!'
          theText = Replace("0",here,4,theText);
        if (nCh != 46) continue; // '.'
        why++; // why = 10
        break;} //~while // why = 10: unrecognized op before 'null'
      if (why>0) {
        if (logz) System.out.println("** 2JS error = " + why);
          else break;} //~if
      theText = ReplacAll("console.log","System.out.println",
          ReplacAll("","myGame.",ReplacAll("","JavaGame.",
          ReplacAll("","super.",theText))));
      theText = ReplacAll(" ","\r",ReplacAll("","\n\r",
          ReplacAll("","\r\n",ReplacAll("null","n$ull",theText))));
      lxx = StrLength(theText);
      while (true) {
        theText = ReplacAll("\n\n","\n\n\n",ReplacAll("\n"," \n",theText));
        nx = StrLength(theText);
        if (nx<lxx) lxx = nx;
          else break;} //~while
      theText = "<script type='text/javascript'> // " + fiName + " -- "
          + TimeStamp(false) + "\n" + theText + "\n</script>\n";
      why = 0;
      break;} //~while
    // lxx = StrLength(theText);
    if (why==0) if (lxx<111) why = 15; // Generated text too short
    JSinfo = "Error #" + why;
    if (why>0) switch (why) {
      case 1: // why = 1:
        JSinfo = "Can't find class name";
        break;
      case 3: // why = 3:
        JSinfo = "Can't find widget list";
        break;
      case 6: // why = 6:
        JSinfo = "Can't do this method";
        break;
      case 8: // why = 8:
        JSinfo = "Invalid GameWgt object in method call";
        break;
      case 9: // why = 9:
        JSinfo = "Unrecognized op after 'null'";
        break;
      case 10: // why = 10:
        JSinfo = "Unrecognized op before 'null'";
        break;
      case 15: // why = 15:
        JSinfo = "Generated text too short";
        break;
      case 16: // why = 16:
        JSinfo = "Bad widget list form";
        break;
      case 17: // why = 17:
        JSinfo = "Invalid line number in widget list";
        break;
      case 18: // why = 18:
        JSinfo = "Macros and private widget types not supported";
        break;
      case 20: // why = 20:
        JSinfo = "Widget spec line is too short";
        break;
      case 21: // why = 21:
        JSinfo = "Invalid negative data length";
        break;
      case 22: // why = 22:
        JSinfo = "No such font";
        break;} //~switch
    if (why>0) aLine = "\n  -- " + JSinfo;
      else aLine = "";
    if (Dev_Log||Logging||(why>0)) System.out.println("** 2JS +"
        + StrLength(theText) + IffyStr(why==1,"/" + ofx,"")
        + " @ " + lino + " '" + aWord + "' = " + why + aLine);
    if (why>0) return false;
    JSinfo = theText;
    if ((why==0)||logz) WriteWholeTextFile(fiName,theText);
    return true;} //~toJS

  /**
  * Converts the plain-text user name or JS code into a form suitable for
  * upload over the internet. All non-letters are converted to the form
  * %xx where xx is the hexadecimal code for that character.
  *
  * @param  theText  the raw text
  *
  * @return          the same text as encoded
  */
  private String NetEncode(String theText) { // convert/remove unwanted chars
    char aCh;        // ..suitable for a JS cookie or posting to server.
    int info, nx, lxx = StrLength(theText)-1;
    String nuly = "", aStr = "";
    for (nx=0; nx <= lxx; nx++) { // (omit in BeenHere..)
      aCh = CharAt(nx,theText);
      info = (int)(aCh);
      if (info>126) continue;
      if (info==13) info = 10;
      else if (info<32) if (info !=10) continue;
      if (info>96) if (info<123) info = 0; // keep lower case
      if (info>64) if (info<91) info = 0; // keep caps
      if (info>47) if (info<58) info = 0; // keep digits
      if (info==0) aStr = "" + aCh; // auto-converts to String
        else aStr = SeeHex("%",info,"");
      nuly = nuly + aStr;} //~for
    return nuly;} //~NetEncode

  /**
  * Uploads a particular user's (JS) game code to a game server.
  *
  * @param  whom  the user name and password as name_password
  * @param  data  the JS program as translated by toJS()
  *
  * @return       an error code, or =0 if none
  */
  private int ServerUpload(String whom, String data) { // rtns error +
    int here, thar, why = 0;
    data = "?cmd=18&name=" + NetEncode(whom) + "&data=" + NetEncode(data);
    HttpURLConnection httpCon = null;
    PrintWriter px;  BufferedReader br;  try {
      why++; // why = 1
      URL whar = new URL("http://ittybittycomputers.com/Java/Hack.php?cmd=18");
      why++; // why = 2
      httpCon = (HttpURLConnection) whar.openConnection();
      why++; // why = 3
      httpCon.setDoOutput(true);
      why++; // why = 4
      httpCon.setRequestMethod("POST");
      why++; // why = 5
      httpCon.setRequestProperty("User-Agent","Java client");
      why++; // why = 6
      httpCon.setRequestProperty("Content-Type",
          "application/x-www-form-urlencoded; charset=utf-8");
      why++; // why = 7
      px = new PrintWriter(httpCon.getOutputStream());
      why++; // why = 8
      px.write(data);
      why++; // why = 9
      px.flush();
      why++; // why = 10
      px.close();
      why = httpCon.getResponseCode();
      if (why==200) { // this is doing it the hard way..
        why = 11;
        br = new BufferedReader(new InputStreamReader(httpCon.getInputStream()));
        why++; // why = 12
        data = "";
        for (here=1; here <= 33; here++) {
          whom = br.readLine();
          if (here>11) if (StrLength(whom)==0) break;
          data = data + whom + "\n";} //~for
        why++; // why = 13
        br.close();
        why++; // why = 14
        here = NthOffset(0,"<span>",data);
        thar = NthOffset(0,"</span>",data);
        whom = data;
        if (here >= 0) if (thar>here) {
          why++; // why = 15
          whom = Substring(here+6,thar-6-here,data);
          if ((StrLength(whom)&-2) ==2) why = 0;
            else whom = data;}} //~if
      else whom = "HTTP error code = " + why;}
    catch (Exception ex) {whom = "{oops} = " + why;}
    try {if (httpCon != null) httpCon.disconnect();}
    catch (Exception ex) {}
    if (Dev_Log||Logging)
      System.out.println("** ServerUplo = " + why + " '" + whom + "'");
    if (why>0) {
      if (why<20) JSinfo = "Uploader error #" + why;
        else JSinfo = Substring(0,96,whom);} //~if
    return why;} //~ServerUpload

  /**
  * (Temporary) uploader test.
  */
  void JvUpLdTest() {
    String xStr = "<!doctype html public '-//w3c//dtd html 4.0 transitional//en'>"
        + "\n<html><head><title>Upload Test</title></head><body>\n"
        + "<center><h1>Hello World*</h1></center></body></html>";
    int why = ServerUpload("Test_public",xStr);
    System.out.println("** JavaUploadTest = " + why);} //~JvUpLdTest

/**
 * Subclass to handle events when creating a new game. // . . . . . . (GameMk)
 */
  private class GameMakerEv extends GameEvent { // 2020 July 25

    GameWgt GdChkBx, TypChkBx, ClikChkBx, DragChkBx, BumpChkBx, RollChkBx,
        ColoBx, RxWgt, CxWgt, CxAntz, GameBoard, WgtNamEntry, GameNameWgt,
        ToolHiLi, UsWgHiLi, RexTool, CircTool, BallTool, TexTool, SpriTool,
        OKbtn, BildBtn, JSbtn, CantJS, ToolList, ToolName, ToolPanel,
        FontRad, TopField, LeftField, TallField, WideField, ListofWgts,
        AIdlgChkBx, AntsChkBx, UwgtNamList, ActivWgt, VeloEntry, NullWgt,
        UndoBtn, TextInWgt, SkyTool;
    boolean ToolActive, InChanged, FreshFaster;
    int UwgtNamID, UwgtNamPos, ChosenBtn, NextWgtName, GamePosn, GameSz;
    String GamesName, ValidUserName, UserNames, AddUserWgts, WgtMap,
        WgtList, MkTmpStr;

  /// [more pseudo-global vars near UserEvHandler]

  final static String
      TakenGameNames = " JavaGame GameEvent GameWgt ";

  final static String BaseMakerWgts = ""
    + "1 5 0 0 512 640 0xCCCCFF 0 0 4 34 3 2 21 "
      + "'[List of item lists] in win' \"Build Game\" -- 2020 July 25\n"
    + "2 5 0 0 512 640 -1 0 1 12 11 12 0 0 0 0 0 0 0 0 0 0 '[Sprite list]'\n"
    + "3 5 0 512 512 128 0x999999 0 1 25 5 13 14 15 16 17 18 19 20 "
      + "22 23 24 25 31 32 79 80 81 82 83 84 85 90 105 0 '[Tool panel]'\n"
    + "4 5 0 0 512 512 -1 0 1 12 0 0 0 0 0 0 0 0 0 0 0 0 '[User wgts list]'\n"
    + "5 5 240 56 256 68 0xFFFFFF 0 3 12 26 27 28 29 30 36 0 0 0 0 0 0 '"
      + "[User wgt name list]'\n"
    + "6 3 0 0 12 12 0x0 0 0 0 'Mono font'\n"
    + "7 3 0 0 12 12 0x0 1 0 0 'Text font'\n"
    + "8 3 0 0 12 12 0x0 4 0 0 'Bold font'\n"
    + "9 3 0 0 12 12 0x0 5 0 0 'Head font'\n"
    + "10 3 0 0 12 12 0x0 3 0 0 'tinY font'\n"
    + "11 241 120 40 64 276 0xFFFFFF 0 2 0 'Hello world!' [Std Alert Dialog] "
      + "`AlDg`\n"
    + "12 240 250 40 64 276 0xFFFFFF 0 2 0 'Prompt:' '(..)' "
      + "[Std Input Dialog] `IpDg`\n"
    + "13 1277 4 80 16 40 -1 0 3 0 'Grid' [Grid checkbox] `CkGr`\n"
    + "14 254 48 8 16 112 -1 0 3 0 'Name=' 'Untitled' Game/wgt "
      + "[name entry panel] `Name`\n"
    + "15 254 72 8 16 32 -1 0 3 0 'V=' '256' [Top entry field] `TpFx`\n"
    + "16 254 72 64 16 32 -1 0 3 0 'H=' '512' [Left entry field] `LpFx`\n"
    + "17 254 96 8 16 32 -1 0 3 0 'T=' '256' [Tall entry field] `TaFx`\n"
    + "18 254 96 64 16 32 -1 0 3 0 'W=' '512' [Wide entry field] `WiFx`\n"
    + "19 1275 26 8 16 16 0xCCCC00 0 3 0 '[Color Swatch]' `Colo`\n"
    + "20 4 29 32 12 64 0x0 8 3 0 'GameRect' [Current wgt type name]\n"
    + "21 250 0 0 32 32 -1 0 0 0 '[Deployable Marching Ants]' `Antz`\n"
    + "22 9980 24 80 20 40 0xCCCCCC 3 3 0 ['noJS' button] `noJS`\n"
    + "23 9980 24 80 20 40 0xCCCCCC 3 3 0 ['->JS' button] `MkJS`\n"
    + "24 9980 24 80 20 40 0xCCCCCC 3 3 0 ['Build' button] `Bild`\n"
    + "25 9980 24 92 20 20 0xCCCCCC 1 3 0 ['OK' (accept) button] `OKOK`\n"
    + "26 0 1 1 18 68 0xFF9900 0 5 0 '[Name list hilight]'\n"
    + "27 0 0 0 1 68 0x0 0 5 0 'name list outline (top)'\n"
    + "28 0 0 0 256 1 0x0 0 5 0 'name list outline (left)'\n"
    + "29 0 255 0 1 68 0x0 0 5 0 'name list outline (botm)'\n"
    + "30 0 0 67 268 1 0x0 0 5 0 'name list outline (right)'\n"
    + "31 4 6 16 20 80 0xFF 9 3 0 'BYOG' blue title over tool panel\n"
    + "32 5 240 0 256 48 -1 0 3 8 33 39 44 70 59 50 100 0 '[Tool Palette]' "
      + "`Tpal`\n"
    + "33 0 0 0 30 32 0xFF9900 0 32 0 '[Tool icon hilight]'\n"
    + "34 5 0 0 512 512 -1 0 1 2 35 0 'Game workspace' `GmWS`\n"
    + "35 3845 44 44 250 444 0x66FF66 0 34 8 36 37 38 0 0 0 0 0 '"
      + "[Game board asset list]' `GBDW` *\n"
    + "36 1028 4 4 12 60 0x0 7 5 0 'GameBd' [Game board actual name]\n"
    + "37 4 0 0 12 64 0x0 0 35 0 'GameRect' its wgt type name (TyNm)\n"
    + "38 0 0 5 1 1 -1 0 35 0 'Game board flags' `Flgz`\n"
    + "39 5 0 0 24 24 -1 0 32 2 40 0 'Rect tool icon base' `RcBp`\n"
    + "40 3589 6 4 20 24 0xFF 0 39 4 41 42 43 0 '[Draggable Rect tool] "
      + "`RcDT` => 'RcDW' *\n"
    + "41 1028 4 4 12 60 0x0 7 5 0 'Rect.0' (placeholder for actual name)\n"
    + "42 4 0 0 12 64 0x0 0 40 0 'Rect' its wgt type name (TyNm)\n"
    + "43 0 0 0 1 1 -1 0 40 0 'Rect tool flags' `Flgz`\n"
    + "44 5 32 0 24 32 -1 0 32 2 45 0 'Text tool icon base' `TxBp`\n"
    + "45 3589 6 0 24 32 -1 0 44 5 46 47 48 49 0 '[Draggable Text tool]' "
      + "`TxDT` => 'TxDW' *\n"
    + "46 4 0 0 12 28 0x0 8 45 0 'Text' 'Text tool image' (TxTx)\n"
    + "47 1028 4 4 12 60 0x0 7 5 0 'Text.0' (placeholder for actual name)\n"
    + "48 4 0 0 12 64 0x0 0 45 0 'Text' its wgt type name (TyNm)\n"
    + "49 0 0 4 1 1 -1 0 45 0 'Text tool flags' `Flgz`\n"
    + "50 5 128 0 24 24 -1 0 32 2 51 0 'Circle tool icon base' `CrBp`\n"
    + "51 3589 8 8 24 24 -1 0 50 6 52 53 56 57 58 0 '[Draggable Circle tool] "
      + "`CrDT` => 'CrDW' *\n"
    + "52 2 0 0 16 16 0x6600CC -3 51 0 'Circle tool icon' (CrIc)\n"
    + "53 5 0 0 24 24 -1 0 51 2 54 55 '(hidable) [Circle tool icon ants]' "
      + "`AntI`\n"
    + "54 2 0 0 16 16 0x0 -3 53 0 'Circle tool icon hollow outline' (CrHo)\n"
    + "55 2 0 0 16 16 0xFFFFFF -3 53 0 'Circle tool icon ants' (CrAn)\n"
    + "56 1028 4 4 12 60 0x0 7 5 0 'Circ.0' (placeholder for actual name)\n"
    + "57 4 0 0 12 64 0x0 0 51 0 'Circle' its wgt type name (TyNm)\n"
    + "58 0 0 2 1 1 -1 0 51 0 'Circle tool flags' `Flgz`\n"
    + "59 5 96 0 24 24 -1 0 32 2 60 0 'Sprite tool icon base' `SpBp`\n"
    + "60 3333 4 4 24 24 -1 0 59 6 61 67 68 69 0 0 '[Draggable Sprite tool] "
      + "`SpDT` => 'SpDW' *\n"
    + "61 6 0 0 24 24 -1 0 60 4 62 63 64 0 'Sprite tool token sprite' `SpSp`\n"
    + "62 0 8 0 10 20 0xAA00 0 61 0 'Sprite tool green rect' (Spgr)\n"
    + "63 0 0 8 20 8 0xFF5500 0 61 0 'Sprite tool orange rect' (Spor)\n"
    + "64 5 0 0 24 24 -1 0 61 2 65 66 '(hidable) [Sprite tool icon ants]' "
      + "`AntI`\n"
    + "65 2 0 0 24 24 0x0 -9 64 0 'Sprite tool icon hollow outline' (SpHo)\n"
    + "66 2 0 0 24 24 0xFFFFFF -9 64 0 'Sprite tool icon ants' (SpAn)\n"
    + "67 1028 4 4 12 60 0x0 7 5 0 'Sprit.0' (placeholder for actual name)\n"
    + "68 4 0 0 12 64 0x0 0 60 0 'Sprite' its wgt type name (TyNm)\n"
    + "69 0 0 0 1 1 -1 0 60 0 'Sprite tool flags' `Flgz`\n"
    + "70 5 64 0 24 24 -1 0 32 2 71 0 'Ball sprite tool icon base' `BsBp`\n"
    + "71 3589 8 8 24 24 -1 0 70 6 72 73 76 77 78 0 '[Draggable Ball tool] "
      + "`BsDT` => 'BsDW' *\n"
    + "72 2 0 0 16 16 0xFF0000 -3 71 0 'Ball tool icon' (BsIc)\n"
    + "73 5 0 0 24 24 -1 0 71 2 74 75 '(hidable) [Ball tool icon ants]' "
      + "`AntI`\n"
    + "74 2 0 0 16 16 0x0 -3 73 0 'Ball tool icon hollow outline' (BsHo)\n"
    + "75 2 0 0 16 16 0xFFFFFF -3 73 0 'Ball tool icon ants' (BsAn)\n"
    + "76 1028 4 4 12 60 0x0 7 5 0 'Ball.0' (placeholder for actual name)\n"
    + "77 4 0 0 12 64 0x0 0 71 0 'Ball' its wgt type name (TyNm)\n"
    + "78 0 0 0 1 1 -1 0 71 0 'Ball tool flags' `Flgz`\n"
    + "79 1277 120 4 16 40 -1 0 3 0 'Key' [Typable checkbox] `CkTy`\n"
    + "80 1277 120 42 16 40 -1 0 3 0 'Clik' [Clickable checkbox] `CkCk`\n"
    + "81 1277 120 80 16 40 -1 0 3 0 'Drag' [Dragable checkbox] `CkDg`\n"
    + "82 1277 144 4 16 40 -1 0 3 0 'Bump' [Collidable checkbox] `CkBm`\n"
    + "83 1277 144 56 16 40 -1 0 3 0 'Rollover' [Rollable checkbox] `CkRo`\n"
    + "84 1277 168 4 16 88 -1 0 3 0 'Alert+InputDlog' [Alrt+InDlg checkbox] "
      + "`CkAI`\n"
    + "85 5 168 0 60 127 -1 0 3 7 86 87 88 89 95 96 97 '"
      + "[Font selector radio group]' `RdB!`\n"
    + "86 1273 24 4 16 60 -1 0 85 0 'Large' [\"Head\" font radio btn] `Rd_H`\n"
    + "87 1273 24 64 16 60 -1 0 85 0 'MonoSp' [\"Mono\" font radio btn] "
      + "`Rd_M`\n"
    + "88 1273 4 4 16 60 -1 0 85 0 'Times' [\"Text\" font radio btn] `Rd_T`\n"
    + "89 1273 4 64 16 60 -1 0 85 0 'Bold' [\"Bold\" font radio btn] `Rd_B`\n"
    + "90 5 168 0 24 127 -1 0 3 4 91 92 0 0 '[Velocity entry group]' `Velo`\n"
    + "91 254 0 8 16 60 -1 0 90 0 'D=' '00.00' [Vertical velocity entry field] "
      + "`VvFx`\n"
    + "92 254 0 68 16 60 -1 0 90 0 'R=' '00.00' "
      + "[Horizontal velocity entry field] `HvFx`\n"
    + "93 3 0 0 12 12 0x0 6 0 0 'Nano font'\n"
    + "94 3 0 0 12 12 0x0 2 0 0 'Ital font'\n"
    + "95 1273 44 4 16 44 -1 0 85 0 'Ital' [\"Ital\" font radio btn] `Rd_I`\n"
    + "96 1273 44 42 16 44 -1 0 85 0 'tinY' [\"tinY\" font radio btn] `Rd_Y`\n"
    + "97 1273 44 84 16 44 -1 0 85 0 'Nano' [\"Nano\" font radio btn] `Rd_N`\n"
    + "98 5    0  0 512 640 -1      0 1 0  '[null widget]'\n"
    + "99 3     0  0 12 12 0x0      7 0 0 'Score font'\n"
    + "100 5 96 0 24 24 -1 0 32 2 101 0 'SkyRec sprite tool icon base' `RSBp`\n"
    + "101 3589 4 4 24 24 0x99CCFF 0 100 4 102 103 104 0 '[Rect Sprite tool]"
      + " `RSDT` => 'RSDW'\n"
    + "102 1028  4  4 12 60 0x0 7 5 0 'SkyRx.0' (placeholder for actual name)\n"
    + "103 4     0  0 12 64 0x0     0 101 0 'SkyRec' its wgt type name (TyNm)\n"
    + "104 0     0  0  1  1  -1     0 101 0 'SkyRec tool flags' `Flgz`\n"
    + "105 4 500 88 9 36 0x6600 10 3 0   '20 / 7 / 24' (RevDate)\n";
 /**
  * Return the class name for this object.
  *
  * @return     The reference String for this object
  */
  public String toString() { // . . . . . . . . . . . . . . . . . (GameMk)
    return "(GameMakerEv)" + IffyStr(StrLength(GamesName)==0,"",
        " GamesName = '" + GamesName + "' " + RevDate);} //~toString

  /**
  * IzTool distinguishes between tools in the tool palette and other things
  * the user might click on.
  *
  * @param  whom    the widget clicked on
  *
  * @return         true if it's a tool
  */
  public boolean IzTool(GameWgt whom) {
    if (whom != RexTool) if (whom != CircTool) if (whom != TexTool)
      if (whom != BallTool) if (whom != SpriTool) if (whom != SkyTool)
        return false;
    return true;} //~IzTool

// These methods maintain match the selected widget to the tool panel..

  /**
  * IzUsrWgt examines a group widget clicked on, and returns true if it's
  * a widget already on the game board.
  *
  * @param  whom    the widget clicked on
  *
  * @return         true if it's on the game board
  */
  public boolean IzUsrWgt(GameWgt whom) {
    int rx = 0;
    if (ListofWgts==null) return false;
    if (whom==null) return false;
    rx = whom.GetRefNum();
    if (rx==0) return false;
    if (ListofWgts.FindElemt(rx)<0) return false;
    return true;} //~IzUsrWgt

  /**
  * ChooseBtn exposes one of the four response buttons and hides the others.
  *
  * @param  whom    the chosen button 1..4
  */
  public void ChooseBtn(int whom) {
    boolean doit;
    int why = 0;
    GameWgt nuly = null;
    ChosenBtn = 0;
    if (CantJS != null) {
      doit = (whom==5);
      if (doit) nuly = CantJS;
      CantJS.HideSho(doit);} //~if
    else why = 16;
    if (JSbtn != null) { // JSbtn=4
      doit = (whom==4);
      if (doit) nuly = JSbtn;
      JSbtn.HideSho(doit);} //~if
    else why = why+8;
    if (BildBtn != null) { // BildBtn=3
      doit = (whom==3);
      if (doit) nuly = BildBtn;
      BildBtn.HideSho(doit);} //~if
    else why = why+4;
    if (UndoBtn != null) { // not installed
      doit = (whom==2);
      if (doit) nuly = UndoBtn;
      UndoBtn.HideSho(doit);} //~if
    // else why = why+2;
    if (OKbtn != null) {
      doit = (whom==1);
      if (doit) nuly = OKbtn;
      OKbtn.HideSho(doit);} //~if
    else why++;
    if (why==0) {
      if (whom >= 0) if (whom<5) ChosenBtn = whom;
      if (whom==0) DefaultKeyFocus(GameBoard);
      else if (nuly != null) DefaultKeyFocus(nuly);} //~if
    if (Logging) System.out.println("ChoseBtn " + whom
        + " DF=#" + SafeWgtNum(DefaultFocus) + " = " + why);} //~ChooseBtn

  /**
  * NewData is informed that the user has made some change to the game which
  * should be saved, and displays the OK button to do it.
  *
  * @param  doit    true to force the condition
  */
  public void NewData(boolean doit) {
    if (InChanged) if (!doit) return;
    unSavedWgts = true;
    ChooseBtn(1); // OK
    WgtList = "";
    JSinfo = ""; // set in toJS, used in MakeUserJS
    InChanged = true;} //~NewData

  /**
  * NeedsVelo returns true if the tool or game board widget is a sprite
  * (and therefore needs to have the velocity entry fields shown).
  *
  * @param  whom    the widget clicked on
  *
  * @return         true if it's a sprite
  */
  public boolean NeedsVelo(GameWgt whom) { // sprite rtns T (sho VeloEntry)
    int nx = 0, info = 0, why = 0;
    if (whom==null) return false;
    nx = whom.GetRefNum();
    info = whom.GetInfo()&0x7FFFFFFC; // 'B*DT'/'B*DW' or 'SpDT'/'SpDW'..
    if ((info&0x7F00FFFF) != 0x42004454) if (info != 0x53704454)
      if (info != 0x52534454) return false; // 'RSDT'/'RSDW'
    return true;} //~NeedsVelo

  /**
  * GameMaker sizes and positions are limited to three decimal digits. Fit3
  * tries to format whatever number you give it into three characters.
  *
  * @param  whom    the number to be formatted
  *
  * @return         the 3-digit text as formatted
  */
  private String Fit3(int whom) {
    String msg = "   " + whom;
    int lxx = StrLength(msg);
    return RestOf(lxx-3,msg);} //~Fit3

  /**
  * Whenever a tool or user widget is displayed or dragged, ShoCoords
  * is called to put its position and size up on the tool panel
  *
  * @param  whom    the widget being shown
  * @param  all     true forces display even if no obvious change
  */
  public void ShoCoords(GameWgt whom, boolean all) { // all=F: if diff't
    boolean doit;           // whom: tool or user wgt (or DW being drug)
    int zx, nx = 0, posn = 0, info = 0, why = 0;
    GameWgt nuly = whom;
    String msg, logy = "";
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      nx = whom.GetRefNum();
      info = whom.GetInfo()&0x7FFFFFFC;
      why++; // why = 2
      if ((info&0x7E00FFFC) == 0x42004454) // 'C*DT'/DW' or 'B*DT'/DW'
        nuly = GetIxWgt(whom.GetIxData(0,0)); // (not sprite/text)
      if (nuly==null) break;
      why++; // why = 3
      posn = whom.GetPosn(); // use coords on draggable group wgt
      info = nuly.GetSize(); // ball/circle size from icon
      doit = all;
      zx = posn>>16;
      msg = Fit3(zx);
      logy = "V=" + msg;
      if (TopField != null) {
        if (!doit) if (ParseInFld(TopField) != zx) doit = true;
        if (doit) if (!SetInputText(TopField,"V=",msg,false)) break;} //~if
      why++; // why = 4
      doit = all;
      zx = SignExtend(posn);
      msg = Fit3(zx);
      logy = logy + " H=" + msg;
      if (LeftField != null) {
        if (!doit) if (ParseInFld(LeftField) != zx) doit = true;
        if (doit) if (!SetInputText(LeftField,"H=",msg,false)) break;} //~if
      why++; // why = 5
      doit = all;
      zx = info>>16;
      msg = Fit3(zx);
      logy = logy + " T=" + msg;
      if (TallField != null) {
        if (!doit) if (ParseInFld(TallField) != zx) doit = true;
        if (doit) if (!SetInputText(TallField,"T=",msg,false)) break;} //~if
      why++; // why = 6
      doit = all;
      zx = SignExtend(info);
      msg = Fit3(zx);
      logy = logy + " W=" + msg;
      if (WideField != null) {
        if (!doit) if (ParseInFld(WideField) != zx) doit = true;
        if (doit) if (!SetInputText(WideField,"W=",msg,false)) break;} //~if
      why = 0;
      break;} //~while
    if (Logging || (why>0)) System.out.println("(ShoCords #" + nx + ") "
        + logy + " = " + why);} //~ShoCoords

  /**
  * Whenever you click on the OK button, CapturCoords is called to parse
  * and copy the position and size values to the selected widget group,
  * hopefully not a tool (caller should check).
  *
  * @param  whom    the selected widget
  */
  public void CapturCoords(GameWgt whom) { // set wgt to match panel..
    int nx = 0, rx = 0, posn = 0, info = 0, kin = 0, why = 0;
    String msg;
    GameWgt nuly = whom;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      nx = whom.GetRefNum();           // if (whom.GetType() == 3589)
      kin = whom.GetInfo();
      why++; // why = 2
      if (TopField != null) posn = ParseInFld(TopField)<<16;
      if (LeftField != null) posn = ParseInFld(LeftField)&0xFFFF|posn;
      if (TallField != null) info = ParseInFld(TallField)<<16;
      if (WideField != null) info = ParseInFld(WideField)&0xFFFF|info;
      whom.SetPosn(posn>>16,SignExtend(posn));
      if (whom==GameBoard) {
        GamePosn = posn;
        GameSz = info;} //~if
      if ((kin&0x7E00FFFC)==0x42004454) { // 'C*DT'/`C*DW` or 'B*DT'/`B*DW`
        if (FocusWgt==WideField) rx = info&0xFFFF;
          else rx = info>>16;
        CircleSize(whom,rx);} //~if
      else whom.SetSize(info>>16,info&0xFFFF);
      why++; // why = 3
      if (kin==0x54784457) { // 'TxDW'
        nuly = GetIxWgt(whom.GetIxData(0,0));
        if (nuly==null) break;
        nuly.SetSize(info>>16,info&0xFFFF);} //~if
      why = 0;
      break;} //~while
    if (Logging || (why>0))
      System.out.println("(CapCords #" + nx + SeeHex(") i=x",kin,LogVH(" p=",posn,
          LogVH(" s=",info," = "))) + why);} //~CapturCoords

  /**
  * Whenever a tool or user widget is selected, ShoInfoFrom is called
  * to display it attributes.
  *
  * @param  whom    the widget clicked on
  */
  public void ShoInfoFrom(GameWgt whom) { // whom: tool or user wgt (or DW)
    int logy = SeenLog&0x20000, colo, zx, nx = 0, info = 0, tye = 0, why = 0;
    boolean user = false, abit = false;
    String name = "";
    GameWgt more = null, nuly = FontRad;
    while (true) {
      why++; // why = +1
      if (whom==null) break;
      nx = whom.GetRefNum();
      info = whom.GetInfo()&0x7FFFFFFC;
      if (nuly != null) {
        if (info == 0x54784454) { // 'TxDT'/DW'..
          why = why|16;
          zx = ChoseFont(whom,-1); // gets current wgt font..
          if (zx>0) {
            zx = nuly.RadioBtn(zx&255); // sets the radio btns,
            nuly.HideSho(true);} //~if     // ..then show them
          else nuly.HideSho(false);} //~if
        else nuly.HideSho(false);} //~if
      nuly = FindSubWgt(whom,0x466C677A,0); // 'Flgz'
      if (false) if (Logging) if (nuly==null)
        nuly = FindSubWgt(GameWgtsList,0x466C677A,8000); // (logs same fail)
      why++; // why = +2
      if (nuly != null) if (VeloEntry != null) if ((info == 0x53704454)
          || (info == 0x52534454) || ((info&0x7F00FFFC) == 0x42004454)) {
        why = why|32;               // 'SpDT'/DW' or 'B*DT'/DW' or 'RSDT'/DW'
        info = nuly.GetPosn();
        name = FormFixt("",info>>16);
        more = GetIxWgt(VeloEntry.GetIxData(0,0));
        if (more != null) if (!SetInputText(more,"D=",name,true)) break;
        name = FormFixt("",SignExtend(info));
        more = GetIxWgt(VeloEntry.GetIxData(1,0));
        if (more != null) if (!SetInputText(more,"R=",name,true)) break;
        VeloEntry.HideSho(true);} //~if
      if (nuly != null) tye = nuly.GetType();
      if (TypChkBx != null) {
        abit = (TypChkBx.CheckIt(false) == ((tye&512)==0));
        abit = TypChkBx.CheckIt(abit);
        TypChkBx.HideSho(true);} //~if
      if (BumpChkBx != null) {
        abit = (BumpChkBx.CheckIt(false) == ((tye&256)==0));
        abit = BumpChkBx.CheckIt(abit);
        BumpChkBx.HideSho(true);} //~if
      if (ClikChkBx != null) {
        abit = (ClikChkBx.CheckIt(false) == ((tye&1024)==0));
        abit = ClikChkBx.CheckIt(abit);
        ClikChkBx.HideSho(true);} //~if
      if (DragChkBx != null) {
        abit = (DragChkBx.CheckIt(false) == ((tye&2048)==0));
        abit = DragChkBx.CheckIt(abit);
        DragChkBx.HideSho(true);} //~if
      if (RollChkBx != null) {
        abit = (RollChkBx.CheckIt(false) == ((tye&4096)==0));
        abit = RollChkBx.CheckIt(abit);
        RollChkBx.HideSho(true);} //~if
      if (!IzTool(whom)) user = true; // (unused)
      why++; // why = +3
      colo = whom.GetColor(); // tool or asset may have the color,
      if (colo<0) { // but if not, it will be on rect/icon of 1st sub-wgt..
        nuly = GetIxWgt(whom.GetIxData(0,0));
        if (nuly != null) colo = nuly.GetColor();} //~if
      SetSwatchColor(ColoBx,colo);
      name = " ";
      if (whom != NullWgt) { // look for wgt name field..
        nuly = FindSubWgt(whom,(TextFontNo<<16)+0x0404,0);
        if (Logging) if (nuly==null) // log same search..
          nuly = FindSubWgt(whom,(TextFontNo<<16)+0x0404,9000);
        if (nuly != null) name = nuly.GetTextLn();
        if (StrLength(name)==0) name = "untitled";} //~if
      if (WgtNamEntry != null) {            // look for entry field..
        nuly = FindSubWgt(WgtNamEntry,(MonoFontNo<<16)+0x0604,0);
        if (Logging) if (nuly==null)
          nuly = FindSubWgt(WgtNamEntry,(MonoFontNo<<16)+0x0604,10000);
        if (nuly != null) {
          zx = nuly.PutTextLn(name);
          why = why+4;}} //~if                     // can't use find!?
      // nuly = FindSubWgt(whom,0x54794E6D,11000); // look for tool name 'TyNm'
      if (whom != NullWgt) { // look for tool name..
        nuly = FindSubWgt(whom,4,0); // look for tool name, ty=4, fnt=0
        if (Logging) if (nuly==null) // log same search if failed..
          nuly = FindSubWgt(whom,4,11000);
        if (nuly != null) name = nuly.GetTextLn();} //~if
      if (ToolName != null) {
        if (StrLength(name)>0) {
          zx = ToolName.PutTextLn(name);
          why = why+8;} //~if
        ShoCoords(whom,true);} //~if
      // else show blanks? // **
      why = -why;
      break;} //~while
    if (Logging||(logy==0)) {
      System.out.println("(ShoInFro #" + nx + ") " + user + " = " + why);
      SeenLog = SeenLog|0x20000;}} //~ShoInfoFrom

  /**
  * Whenever you click on the OK button, CapturInfo2Wgt is called to copy
  * the displayed attributes to the selected widget group,
  *
  * @param  whom    the widget clicked on
  * @param  doit    true to update position and size (not a tool)
  * @param  dont    true to stop after updating the font
  */
  public void CapturInfo2Wgt(GameWgt whom, boolean doit, boolean dont) {
    int zx, nx = 0, tye = 0, info = 0, seen = 0, why = 0; // set wgt = panel..
    boolean didit = false;   // doit=T=!ToolActive: update posn/size also
    String name = "";                        // dont=T: only do font
    GameWgt more, nuly = FontRad;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      nx = whom.GetRefNum();
      seen = whom.GetInfo();
      info = seen&0x7FFFFFFC;
      why++; // why = 2
      if (nuly != null) if (info == 0x54784454) { // 'TxDT'/DW'
        zx = ChoseFont(nuly,-1); // gets current radio font..
        if (zx>0) {
          zx = ChoseFont(whom,zx&255); // sets the wgt font
          if (zx>0) why = -why;}} //~if // why = -2 (OK)
      if (dont) break;
      why = 3;   // 'SpDT'/DW' or 'B*DT'/DW' or 'RSDT'/DW', now <0..
      if ((info == 0x53704454) || ((info&0x7F00FFFC) == 0x42004454)
          || (info == 0x52534454)) info = info-0x80000000;
      nuly = FindSubWgt(whom,0x466C677A,0); // 'Flgz'
      if (false) if (Logging) if (nuly==null)
        nuly = FindSubWgt(GameWgtsList,0x466C677A,8000); // (logs same fail)
      if (nuly != null) if (info<0) if (VeloEntry != null) {
        info = 0;
        zx = 0;
        more = GetIxWgt(VeloEntry.GetIxData(0,0));
        if (more != null) info = ParsFixPt(GetInputTxt(more));
        more = GetIxWgt(VeloEntry.GetIxData(1,0));
        if (more != null) zx = ParsFixPt(GetInputTxt(more));
        nuly.SetPosn(info,zx);
        why = why|16;} //~if
      if (nuly != null) {
        tye = nuly.GetType();
        zx = tye&255;
        if (TypChkBx != null) if (TypChkBx.CheckIt(false)) zx = zx|512;
        if (BumpChkBx != null) if (BumpChkBx.CheckIt(false)) zx = zx|256;
        if (ClikChkBx != null) if (ClikChkBx.CheckIt(false)) zx = zx|1024;
        if (DragChkBx != null) if (DragChkBx.CheckIt(false)) zx = zx|2048;
        if (RollChkBx != null) if (RollChkBx.CheckIt(false)) zx = zx|4096;
        // if (zx != tye) NewData();} //~if
        nuly.SetFlags(zx);
        if (Logging) nuly.LogMe(" (set flgs #" + nx
            + SeeHex(") x",zx,SeeHex("/",tye,"")));} //~if
      if (doit) CapturCoords(whom);
      if ((seen&0xFFFC) == 0x4454) { // 'DT'/'DW'..
        info = SwatchGetColor(ColoBx);
        if (info==0) info++; // (this only matters for icon-based wgts)
        else if (info==0xFFFFFF) info--;
        zx = seen>>24;
        if (zx != 0x52) if (zx != 0x47) { // 'R'/'G' // if (whom.GetColor()<0)
          nuly = GetIxWgt(whom.GetIxData(0,0));
          if (nuly != null) if ((nuly.GetType()&249)==0) { // t=0: rect,
            nuly.SetColor(info);        // t=2: icon, t=4: txt, t=6: sprite
            didit = true;}} //~if
        if (!didit) whom.SetColor(info);
        why = why|32;} //~if
      why++; // why = 4
      if (doit) if (WgtNamEntry != null) { // look for entry field..
        why++; // why = 5
        name = GetInputTxt(WgtNamEntry);
        // nuly = FindSubWgt(WgtNamEntry,(MonoFontNo<<16)+0x0604,0);
        // if (nuly != null) {
        //   why++; // why = 6
        //   name = nuly.GetTextLn();
        if (StrLength(name)==0) name = "untitled"; // ** need make-up fun
        nuly = FindSubWgt(whom,(TextFontNo<<16)+0x0404,0); // (name field)
        if (nuly != null) zx = nuly.PutTextLn(name);
          else why++; // why = 6
        if (whom==GameBoard) {
          GamesName = name;} //~if
        else why = why+2;} //~if // why = 7/8
      why = -why; // why = -4/-7/-8 (OK)
      break;} //~while
    if (Logging) System.out.println("(CapWgtIn #" + nx
        + SeeHex(") x",seen," = ") + why);} //~CapturInfo2Wgt

// These methods generate user Java code..

  /**
  * InvalidJava tests if a name is valid Java for a variable name
  * and returns true if not. Reserved words are disallowed.
  *
  * @param  aWord   the proposed variable name
  *
  * @return         true if it's not valid
  */
  public boolean InvalidJava(String aWord) {
    int nx, xCh, lxx = StrLength(aWord)-1;
    final String JvRWds = " abstract assert boolean break byte"
        + " case catch char class const continue default double do"
        + " else enum extends false final finally float for while"
        + " goto implements if import instanceof int interface long"
        + " native new null package private protected public return"
        + " short static strictfp super switch synchronized"
        + " this throw throws transient true try void volatile ";
    if (lxx >= 0) if (NthOffset(0," "+aWord+" ",JvRWds)<0) for (nx=0; nx <= lxx+1; nx++) {
      if (nx>lxx) return false;
      xCh = (int)(CharAt(nx,aWord));
      if (xCh>122) break; // 'z'
      if (xCh>96) continue; // 'a'-1
      if (xCh==95) if (nx>0) if (nx<lxx) continue; // '_'
      if (xCh>90) break; // 'Z'
      if (xCh>64) continue; // 'A'-1
      if (nx==0) break; // no digit in 1st place
      if (xCh>57) break; // '9'
      if (xCh<48) break;} //~for // '0'
    return true;} //~InvalidJava

  /**
  * NameIsTaken compares a proposed variable name against those already
  * used and returns true if it is taken.
  *
  * @param  aWord   the proposed variable name
  *
  * @return         true if it's taken
  */
  public boolean NameIsTaken(String aWord) {
    String data = TakenGameNames + UserNames;
    // if (gm2) data = data + " " + GamesName + " ";
    if (NthOffset(0," " + aWord + " ",data)<0) return false;
    return true;} //~NameIsTaken

  /**
  * When creating an exported widget list, it is important to ignore the
  * "meta-widgets" which carry extra information needed for selection and
  * dragging, but are not part of the user data. IzzitMeta returns true
  * for those widgets.
  *
  * @param  whom    the widget being considered for export
  * @param  prnt    the widget's presumed parent
  *
  * @return         true if it's a meta-widget
  */
  private boolean IzzitMeta(GameWgt whom, int prnt) {
    int tye, info, rx = 0, logy = 0;
    if (whom==null) return false;
    tye = whom.GetType()&255;
    info = whom.GetInfo();
    if (info != 0x466C677A) if (info != 0x416E7449) // 'Flgz'/'AntI'/'Antz'
      if (info != 0x416E747A) if (whom.GetParent() == prnt)
        if ((tye !=4)||(info>0)) return false;
    return true;} //~IzzitMeta

  /**
  * When preparing to export a widget list, it is necessary to pre-scan all
  * candidates to find all fonts that might be in use, and to build a list
  * of exportable widgets so they can be renumbered in the export file.
  * Prescan calls itself recursively from the game list root. The listing
  * is made in instance variable WgtMap.
  *
  * @param  whom  the root group widget to export
  * @param  bitz  one bit for each known used font
  * @param  omit  <0 (draggable group head) to search this level but not list
  *
  * @return       updated bitz
  */
  private int Prescan(GameWgt whom, int bitz, int omit) { // omit<0 not map,
    int ix, zx, nx = 0, lxx = 0, tye = 0, info = 0,       // .. >0 not list
        posn = 0, size = 0, why = 0;
    boolean seen = false, gone = false;
    String aWord;
    GameWgt nuly = null;
    if (whom != null) {
      why++; // why = 1
      nx = whom.GetRefNum();
      tye = whom.GetType()&255;
      info = whom.GetInfo();
      posn = whom.GetPosn();
      size = whom.GetPosn();
      if (GamePosn>0) if (GameSz>0) if ((posn<0)||(posn+size>GameSz)) {
        zx = whom.AddPrntsPos(posn);
        if (((zx+size-GamePosn)&0x80008000) !=0) gone = true;
        else if (((GamePosn+GameSz-zx)&0x80008000) !=0) gone = true;} //~if
      if (!gone) if (tye==5) if ((info&0xFFFF)==0x4457) { // = '__DW'
        if (info==0x54784457) omit--; // 'TxDW'
        else if ((info&0x7F00FFFF) == 0x42004457) omit = 8; // 'B*DW'
        aWord = "\n" + nx + ",";
        if (NthOffset(0,aWord,WgtMap)<0) WgtMap = aWord + WgtMap;} //~if
      lxx = whom.GetIxData(-1,-1);
      if (Logging) {
        System.out.println("  (Pr+NeedFnt #" + nx + ") t=" + tye + " +" + lxx
            + IffyStr(gone," -gone-",""));
        seen = true;} //~if
      if (gone) {}
      else if (tye==4) { // embedded text..
        why++; // why = 2
        if (omit >= 0) {
          aWord = "\n" + nx + ",";
          if (NthOffset(0,aWord,WgtMap)<0) WgtMap = aWord + WgtMap;
          if (omit>0) WgtMap = WgtMap + nx + " ";} //~if
        zx = ChoseFont(whom,-2);
        if (zx >= 0) if (zx <= 8) bitz = (1<<zx)|bitz;
        why = -why;} //~if // why = -2
      else if ((tye&253)==0) { // embedded rect or icon..
        why = 3;
        if (omit >= 0) {
          aWord = "\n" + nx + ",";
          if (NthOffset(0,aWord,WgtMap)<0) WgtMap = aWord + WgtMap;
          if (omit>0) WgtMap = WgtMap + nx + " ";} //~if
        why = -why;} //~if // why = -3
      else if (((tye-5)&255)<3) {
        for (ix=lxx-1; ix >= 0; ix += -1) {
          nuly = GetIxWgt(whom.GetIxData(ix,0));
          if (nuly==null) continue;
          if (IzzitMeta(nuly,nx)) continue;
          bitz = Prescan(nuly,bitz,omit);} //~for
        why = -why;}} //~if // why = -1 // else why=0
    if (Logging) if (seen)
      System.out.println("  ..(Pr+NF #" + nx + ") = " + bitz + "/" + why);
    return bitz;} //~Prescan

  /**
  * MakeWgtLst builds an exportable widget list in instance variable WgtList
  * and writes it to file "GameMaker.txt".
  *
  * @return         true if it failed
  */
  public boolean MakeWgtLst() { // rtns true if failed
    boolean circ;
    int kx, nx, rx, zx, here, thar, whar, nsubs = 0, tye = 0, info = 0,
        posn = 0, size = 0, color = 0, prnt = 0, bitz = 0, lxx = 0, why = 0;
    GameWgt nuly = null, more = null;
    String namex = "", inits = "", aWord = "", aLine, aStr, theList;
    GameWgt whom = GameBoard;
    while (true) {
      why++; // why = 1
      MkTmpStr = "";
      if (whom==null) break; // whom = GameBoard = [Game board asset list]
      why++; // why = 2
      lxx = whom.GetIxData(-1,-1); // => array size
      if (lxx<2) break;
      why++; // why = 3
      more = FindSubWgt(whom,0x466C677A,0); // 'Flgz'
      if (more==null) break;
      WgtMap = "\n" + whom.GetRefNum() + ",2\n\n0 ";
      why++; // why = 4
      // bitz = more.GetPosn(); // unneed, use chkbx directly
      tye = more.GetType()&0xFF00|5;
      size = whom.GetSize();
      color = whom.GetColor();
      if (color==0) color++;
      else if (color<0) color = 0xCCCCFF;
      else if (color==0xFFFFFF) color--;
      theList = " 0 0 " + (size>>16) + " " + (size&0xFFFF);
      aLine = " " + tye + theList;
      theList = "1" + aLine + HexIfMore(" ",color," 0 0 4  2 0 0 0")
          + "  '[List of item lists] in win {" + GamesName
          + "}'\n2 5" + theList + " -1 0 1 " + lxx + " ";
      nsubs = lxx;
      info = 0;
      if (AIdlgChkBx != null) if (AIdlgChkBx.CheckIt(false)) info = 17;
      info = Prescan(whom,info,0)<<1; // adds ref+s to WgtMap
      aLine = " '[Sprite list]' #" + NextWgtName;
      thar = 2;
      for (here=0; here <= 7; here++) { // add used fonts to wgt list..
        info = info>>1;
        if (info==0) break;
        if ((info&1)==0) continue;
        thar++;
        switch (here) {
        case 0:
          aWord = "\n" + MonoFontNo + "," + thar + ",M";
          break;
        case 1:
          aWord = "\n" + TextFontNo + "," + thar + ",T";
          break;
        case 2:
          aWord = "\n" + ItalFontNo + "," + thar + ",I";
          break;
        case 3:
          aWord = "\n" + TinyFontNo + "," + thar + ",Y";
          break;
        case 4:
          aWord = "\n" + BoldFontNo + "," + thar + ",B";
          break;
        case 5:
          aWord = "\n" + HeadFontNo + "," + thar + ",H";
          break;
        case 6:
          aWord = "\n" + NanoFontNo + "," + thar + ",N";
          break;
        case 7:
          aWord = "\n" + ScorFontNo + "," + thar + ",S";
          break;} //~switch
        WgtMap = aWord + WgtMap;
        aLine = aLine + "\n" + thar + " 3 0 0 12 12 0x0 " + here
            + " 0 0 '[Font " + NthItemOf(',',3,aWord) + "]'";} //~for
      info = (size&0xFFFF)>>2;
      if (info>32) info = 32;
      if (AIdlgChkBx != null) if (AIdlgChkBx.CheckIt(false)) {
        nuly = theAlertDlg; // = [Std Alert Dialog]
        if (nsubs>0) if (nuly != null) {
          info = nuly.GetRefNum();
          thar++;
          WgtMap = "\n" + info + "," + thar + WgtMap;
          aLine = aLine + "\n" + thar + " 241 " + (size>>18) + " " + info + " 64"
              + " 240 0xFFFFFF 0 2 0  'Hello world!' [Std Alert Dialog] `AlDg`";
          theList = theList + "  " + info; // (replace later w/new)
          nsubs--;} //~if
        nuly = theInputDlg; // = [Std Input Dialog]
        if (nsubs>0) if (nuly != null) {
          info = nuly.GetRefNum();
          thar++;
          WgtMap = "\n" + info + "," + thar + WgtMap;
          aLine = aLine + "\n" + thar + " 2288 " + (size>>17) + " " + info + " 64"
              + " 240 0xFFFFFF 0 2 0  'Prompt:' '(..)' [Std Input Dialog] `IpDg`";
          theList = theList + "  " + info;
          nsubs--;}} //~if
      why++; // why = 5
      if (Logging) System.out.println(".. .." + WgtMap + "\n.. ..");
      kx = 0;
      for (here=2; here <= 999; here++) { // add actual wgts to list..
        aWord = NthItemOf('\n',here,WgtMap);
        rx = SafeParseInt(aWord);
        if (rx>0) nuly = GetIxWgt(rx);
          else nuly = null;
        circ = false;
        lxx = 0;
        tye = -1;
        if (nuly != null) {
          tye = nuly.GetType();
          info = nuly.GetInfo();
          if ((info&0x7E00FFFF) == 0x42004457) circ = true;} //~if
        if (Logging) System.out.println("  ++wgt " + here + ": '" + aWord + "' = "
            + rx + IffyStr(nuly==null,"",SeeHex(" x",info,""))
            + IffyStr(kx<0," " + kx,""));
        if (StrLength(aWord)==0) break; // normal exit
        if (StrLength(NthItemOf(',',2,aWord))>0) continue; // already got this
        // nuly = GetIxWgt(rx);
        if (nuly==null) continue; // (probly can't fail)
        // info = nuly.GetInfo();
        if (info == 0x47424457) continue; // 'GBDW' (game board already in)
        aWord = "\n" + rx + ",";
        whar = NthOffset(0,aWord,WgtMap);
        if (whar<0) continue; // (probly can't fail)
        whar = StrLength(aWord)+whar; // -> where new wgt+ goes
        aWord = "";
        aStr = "";
        // tye = nuly.GetType();
        kx = tye&255;
        posn = nuly.GetPosn();
        size = nuly.GetSize();
        color = nuly.GetColor();
        prnt = nuly.GetParent();
        more = FindSubWgt(nuly,(TextFontNo<<16)+0x0404,0); // sb wgt name
        if (more != null) {
          aWord = more.GetTextLn();
          zx = StrLength(aWord);
          if (!InvalidJava(aWord)) if (!NameIsTaken(aWord)) {
            if (circ) inits = inits + "\n    if (" + aWord
                + " != null) myGame.CircleSize(" + aWord + ",0);";
            if (StrLength(namex)==0) namex = aWord;
              else namex = namex + " " + aWord;
            UserNames =  UserNames + aWord + " ";
            aWord = " {" + aWord + "}";
            zx = 0;} //~if
          if (zx>0) aWord = " '" + aWord + "'";} //~if
        more = FindSubWgt(nuly,0x466C677A,0); // 'Flgz'
        if (more != null) tye = more.GetType()&0xFF00|kx;
        if (info == 0x52634457) { // 'RcDW' (dragable rectangle)
          if (kx==5) tye = tye-5;
          info = 0;
          more = null;} //~if
        else if (circ) { // 'C*DW'/'B*DW' (dragable ball/circle)..
          if ((info&0x01000000)==0) { // (dragable ball)
            if (kx==5) tye++;
            if (more != null) info = more.GetPosn(); // = velocity
            more = GetIxWgt(nuly.GetIxData(0,0)); // should be the circle icon
            if (more==null) { // (shouldn't fail)
              kx = -kx;
              continue;} //~if
            // size = more.GetSize();
            more = nuly; // so we get the wgt list
            aWord = " (B)" + aWord;
            lxx++; // we want only the 1st sub-wgt, which is the icon
            kx = 6;} //~if
          else { // 'C*DW' (dragable circle) [if (info .==. 0x43004457)]
            if (kx==5) tye = tye-3;
            more = GetIxWgt(nuly.GetIxData(0,0));
            if (more==null) { // (shouldn't fail)
              kx = -kx;
              continue;} //~if
            info = -3;
            size = more.GetSize();
            color = more.GetColor();
            more = null;}} //~if
        else if (info == 0x54784457) { // 'TxDW' (dragable text)     (MakeWgtLs)
          if (kx==5) tye--;
          more = GetIxWgt(nuly.GetIxData(0,0));
          if (more==null) { // (shouldn't fail)
            kx = -kx;
            continue;} //~if
          info = more.GetInfo();
          size = more.GetSize();
          color = more.GetColor();
          kx = more.GetType()&255;
          if (kx==4) {
            aStr = more.GetTextLn();
            aWord = " '" + aStr + "' " + aWord;
            more = null;}} //~if
        else if (info == 0x53704457) { // 'SpDW' (dragable sprite)
          if (kx==5) tye++;
          if (more != null) info = more.GetPosn(); // = velocity
          more = GetIxWgt(nuly.GetIxData(0,0));
          if (more==null) { // (shouldn't fail)
            kx = -kx;
            continue;} //~if
          aWord = " (S)" + aWord;
          // info = more.GetPosn();
          size = more.GetSize();
          color = more.GetColor();
          kx = more.GetType()&255;
          // *** gotta set lxx= + wgts inside this sprite, then..
          more = null; // more = nuly; // so we get the wgt list
          if (kx != 6) {
            kx = -kx;
            continue;}} //~if
        else if (info == 0x52534457) { // 'RSDW' (dragable rect sprite)
          if (kx==5) tye++;
          if (more != null) info = more.GetPosn(); // = velocity
          aWord = " (F)" + aWord;
          size = nuly.GetSize();
          color = nuly.GetColor();
          more = null; // no wgt list
          lxx = 0;
          kx = 6;} //~if
        else if (kx==0) more = null; // rect inside sprite
        else if (kx==2) { // (circle) icon inside sprite..
          info = -3;
          more = nuly;} //~if
        else if (kx==4) { // text inside sprite..
          more = nuly;} //~if
        else { // (shouldn't fail)
          kx = -kx;
          continue;} //~else
        thar++;
        WgtMap = Substring(0,whar,WgtMap) + thar + " " + RestOf(whar,WgtMap);
        if (lxx==0) if (more != null)
          lxx = more.GetIxData(-1,-1); // = array size
        if (lxx<0) lxx = 0;
        aLine = aLine + "\n" + thar + " " + tye + " " + (posn>>16) + " "
            + SignExtend(posn) + " " + (size>>16) + " " + (size&0xFFFF)
            + HexIfMore(" ",color," ");
        if (kx==4) aLine = aLine + HexIfMore(" ",info,"");
        else if (kx==6) aLine = HexIfMore(aLine,info,"");
        else if (more==null) aLine = HexIfMore(aLine,info,"");
        else if (more.IsValidIDtxt()) {
          aWord = " `" + more.GetIDtxt() + "`" + aWord;
          aLine = aLine + "0";} //~if
        else aLine = HexIfMore(aLine,info,"");
        aLine = aLine + "  " + prnt + " " + lxx;
        if (NthOffset(0," "+rx+" ",WgtMap)<0) {
          theList = theList + "  " + rx; // rx = old ref+
          nsubs--;} //~if // otherwise omit from list, don't count it
        if (lxx>0) if (kx>4) if (kx<8)
          for (zx=0; zx <= lxx-1; zx++) {
            bitz = more.GetIxData(zx,0);
            aLine = aLine + IffyStr(bitz>0,"  " + bitz," 0");} //~for
        aLine = aLine + aWord;
        if (nsubs <= 0) break;} //~for // (add actual wgts to list)
      for (nx=nsubs; nx >= 1; nx += -1) aLine = " 0" + aLine;
      theList = theList + aLine + "\n";
      if (Logging) System.out.println(".. .. +" + nsubs + "\n" + theList + "\n.. ..\n"
          + WgtMap + "\n.. ..\n");
      bitz = 0;
      for (here=thar; here >= 2; here += -1) { // fix wgt+s..                         (MakeWgtLs)
        posn = 0;
        bitz = 0;
        aLine = NthItemOf('\n',here,theList);
        nx = StrLength(aLine);
        if (nx==0) break; // shouldn't happen
        tye = SafeParseInt(NthItemOf(' ',2,aLine))&255;
        lxx = SafeParseInt(NthItemOf(' ',10,aLine));
        if (Logging) aStr = " ..(fixWg #" + here + ") t=" + tye + " +" + lxx;
        // if (tye>3) if (tye<8) if ((tye==4)||(lxx>0))
        for (thar=8; thar <= lxx+10; thar++) {
          color = 0;
          bitz = 0;
          if (thar !=9) {
            if (tye<4) continue;
            if (tye>4) if (thar<11) continue;
            if (tye==4) if (thar>8) continue;} //~if
          aWord = NthItemOf(' ',thar,aLine);
          if (SafeParseInt(aWord)<3) continue; // 1,2 both OK
          whar = NthOffset(0,"\n"+aWord+",",WgtMap);
          if (Logging) aStr = aStr + " " + thar + "=" + aWord + "@" + whar;
          bitz = thar;
          if (whar<0) break;
          size = StrLength(aWord)+2;
          color = SafeParseInt(Substring(whar+size,5,WgtMap));
          whar = 0;
          if (color>0) whar = NthOffset(0,"  "+aWord,aLine);
          if (Logging) aStr = aStr + "=" + color + "/" + whar;
          if (whar <= 0) break;
          bitz = 0;
          whar++;
          size--;
          aLine = Substring(0,whar,aLine) + color + RestOf(whar+size,aLine);
          posn--;} //~for
        if (posn<0) {
          posn = NthOffset(0,"\n"+Substring(0,8,aLine),theList)+1;
          if (posn>0) theList = Substring(0,posn,theList) + aLine
              + RestOf(posn+nx,theList);} //~if
        if (Logging) System.out.println(aStr + "\n    @ " + posn + " => " + aLine);
        if (bitz==0) continue;
        posn = here;
        break;} //~for // (fix wgt+s)
      why++; // why = 6
      if (Logging) System.out.println("(MkWgLst)...\n" + theList + "\n");
      if (bitz>0) if (posn>0) {
        System.out.println("!! oops line " + posn + " wd " + bitz);
        break;} //~if
      WriteWholeTextFile("GameMaker.txt",theList); // => AddUserWgts
      MkTmpStr = namex + inits;
      WgtList = theList;
      unSavedWgts = false;
      why = 0;
      break;} //~while
    if (Dev_Log||Logging) System.out.println("(MakeWgtLs) = " + why); // why =
    return why>0;} //~MakeWgtLst

  /**
  * MakeUserJava calls MakeWgtLst to build an exportable widget list,
  * then copies the relevant information to the .java file with the
  * game name. If successful, it also puts up the button to create a JS
  * version of the game.
  */
  private void MakeUserJava() {
    int here = 0, thar = 0, whar = 0, posn = 0, lxx = 0, why = 0;
    boolean gotit = false;
    String data = "", namex = "", inits = "", aLine, aWord,
        aStr = GamesName, fiName = aStr + ".java", theList = UsrGameJv;
    GameWgt whom = GameBoard;
    while (true) {
      why++; // why = 1
      gotit = false;
      here = NthOffset(0,"\n\n/**",theList); // theList = UsrGameJv;
      thar = NthOffset(0,"extends GameEvent",theList);
      if (here>8) if (thar>here) theList = RestOf(here+2,theList);
      UserNames = "";
      if (InvalidJava(aStr)) { // aStr = GamesName
        GameAlert("'" + aStr + "' is not a valid Java class name");
        break;} //~if
      if (NameIsTaken(aStr)) {
        GameAlert("The name '" + aStr + "' is already taken");
        break;} //~if
      UserNames =  " " + aStr + " ";
      // fiName = aStr + ".java";
      data = ReadWholeTextFile(fiName);
      why++; // why = 2
      if (StrLength(data)==0) {
        data = ReplacAll(GamesName,"UserGame",theList);
        why = 4;} //~if
      RecentJava = data;
      while (true) {
        if (NthOffset(0,"public class "+aStr+" extends GameEvent",data)<0)
          break;
        here = NthOffset(0,"($..)\n",data);
        lxx++; // =1
        if (here<0) break;
        thar = NthOffset(0,"\n  /// (..$)",data);
        lxx++; // =2
        if (thar<here) break;
        here = thar;
        thar = NthOffset(0,"<$..>\n",data);
        lxx++; // =3
        if (thar<here) break;
        here = thar;
        thar = NthOffset(0,"\n    /// <..$>",data);
        lxx++; // =4
        if (thar<here) break;
        here = thar;
        thar = NthOffset(0,"[$..]\n",data);
        lxx++; // =5
        if (thar<here) break;
        here = thar;
        thar = NthOffset(0,"\n    /// [..$]",data);
        lxx++; // =6
        if (thar<here) break;
        why++; // why = 3/5
        break;} //~while
      if (Logging) System.out.println("(MkUsJav) = " + why + "/" + lxx   // why =
          + " " + here + "/" + thar);
      if (why==2) {
        GameAlert("File " + fiName
            + " cannot be upgraded, delete/rename it & retry");
        break;} //~if
      else if (why==4) {
        GameAlert("Oops, GameMaker is corrupted, re-install it");
        break;} //~if
      else if (why<5) why = 5;
        // else gotit = true;
      if (MakeWgtLst()) break; // T: failed (logged)
      why++; // why = 6
      namex = MkTmpStr;
      MkTmpStr = "";
      here = NthOffset(0,"\n",namex);
      if (here>0) {
        inits = RestOf(here+1,namex) + "\n";
        namex = Substring(0,here,namex);} //~if
      theList = WgtList;
      aLine = "      = \""
          + ReplacAll("\\n\"\n      + \"","\n",theList) + "\";\n";
      here = NthOffset(0,"<$..>\n",data);
      if (here<0) break; // (already tested)
      why++; // why = 7
      here = here+5;
      thar = NthOffset(0,"\n    /// <..$>",data)-here;
      if (thar<0) break; // (already tested?)
      here++;
      data = Substring(0,here,data) + aLine + RestOf(here+thar,data);
      why++; // why = 8
      if (StrLength(namex)>0) {
        aLine = "    GameWgt " + ReplacAll(" = null, "," ",namex) + " = null;\n";
        thar = 0;
        posn = 0;
        lxx = StrLength(aLine);
        for (whar=0; whar <= 999; whar++) { // split into lines..
          here = NthOffset(whar,", ",aLine);
          if (here<0) {
            if (posn==0) break; // no commas on last line
            here = StrLength(aLine);
            if (here<thar+80) break;} //~if // last line is short (normal exit)
          posn++;
          if (here<thar+80) continue;
          if (posn>1) if (whar>0) {
            here = NthOffset(whar-1,", ",aLine);
            posn = -posn;} //~if
          here++;
          aLine = Substring(0,here,aLine) + "\n     " + RestOf(here,aLine);
          if (posn<0) posn = 1;
            else posn = 0;
          thar = here+5;} //~for // (split into lines)
        here = NthOffset(0,"($..)\n",data);
        if (here<0) break; // (already tested)
        why++; // why = 9
        here = here+5;
        thar = NthOffset(0,"\n  /// (..$)",data)-here;
        if (thar<0) break;
        here++;
        data = Substring(0,here,data) + aLine + RestOf(here+thar,data);
        why++; // why = 10
        aLine = "";
        for (whar=1; whar <= 999; whar++) { // make assignments..
          aWord = NthItemOf(' ',whar,namex);
          if (StrLength(aWord)==0) break; // (normal exit)
          aLine = aLine + "    " + aWord + " = myGame.FindListWgt(\"{"
              + aWord + "}\");\n";} //~for // (make assignments)
        here = NthOffset(0,"[$..]\n",data);
        if (here<0) break; // (already tested)
        why++; // why = 11
        here = here+5;
        thar = NthOffset(0,"\n    /// [..$]",data)-here;
        if (thar<0) break;
        here++;
        data = Substring(0,here,data) + "    if (myGame==null) return;\n" + aLine
            + inits + RestOf(here+thar,data);
        if (Do_JS) {
          if (SafeCompare(GamesName,"Pong")==0) while (true) {
            if (NthOffset(0,"8  4 5 7 8 9 0 0 0 '[Sprite list]'",data)<0) break;
            whar = NthOffset(0,"7 774 1",data);
            if (whar<0) break;
            if ((SafeParseInt(Substring(whar+6,5,data))&1) !=0) break;
            whar = NthOffset(0,"5 0 0 '[Font H]",data);
            if (whar<0) break;
            data = Replace("7 0 0 '[Font S]",whar,15,data); // use a leaner font
            if (Dev_Log) System.out.println("  (using Score font)");
            break;} //~while
          if (StrLength(GamesName)>0)
            gotit = toJS(data,GamesName);}} //~if
      why = 12;
      WriteWholeTextFile(fiName,data);
      if (gotit) {
        RecentJava = data; // now unused
        ChooseBtn(4);} //~if
      else if (Do_JS) ChooseBtn(5);
        else ChooseBtn(0);
      why = 0;
      break;} //~while
    System.out.println("(MkUsrJav) = " + why
        + IffyStr(why<6,""," " + here + "/" + thar));} //~MakeUserJava

  /**
  * Picks out a display version of the user name
  *
  * @param  aLine  the username_password
  *
  * @return        the prefix alphabetic name only
  */
  private String NameOnly(String aLine) { // pick out user name.. (+544)
    char aCh;
    int nx, info, lxx = StrLength(aLine)-1;
    nx = NthOffset(0,"=",aLine);
    if (nx >= 0) if (nx<lxx) {
      aLine = RestOf(nx+1,aLine);
      lxx = lxx-1-nx;} //~if
    for (nx=0; nx <= lxx; nx++) {
      aCh = CharAt(nx,aLine);
      info = (int)(aCh);
      if (info>96) if (info<123) continue;
      if (info>64) if (info<91) continue;
      aLine = Substring(0,nx,aLine);
      break;} //~for
    return aLine;} //~NameOnly

  /**
  * Driver code to convert the user's (Java) game code into JavaScript
  * and upload it to the game server. This method is called twice, the
  * first time with no user name (it asks you), then again with your reply.
  * The user name should be the same name you created for Chomp. If is
  * not, then your game will not upload.
  *
  * @param  whom    the user name, or none to ask for it
  */
  private void MakeUserJS(String whom) {
    String data = RecentJava, title = GamesName, whoz = NameOnly(whom),
        link = "";
    int more = 0, lxx = StrLength(data);
    whom = NthItemOf(' ',1,whom);
    if (Dev_Log||Logging)
      System.out.println("(MkUsrJS) '" + whom + "' / " + title + " + " + lxx);
    if (lxx==0) return;
    if (StrLength(title)==0) return;
    if (StrLength(whom)==0) {
      whom = ValidUserName;
      if (StrLength(whom)==0) {
        if (OpenInputDlog("Username_password:","")) InputWhy = 7;
        return;} //~if
      whoz = NameOnly(whom);} //~if
    more = StrLength(whoz);
    lxx = NthOffset(0,"_",whom);
    if (lxx<0) lxx = StrLength(whom);
    if (lxx>0) {
      link = Substring(0,lxx,whom);
      if (more<3) if (more<lxx) more = 0;
      if (more>0) title = whoz + "'s " + title;
      lxx = StrLength(JSinfo);
      if (lxx==0) {
        if (toJS(data,GamesName)) {}
        lxx = StrLength(JSinfo);} //~if
      data = JSinfo;} //~if
    else data = "Invalid empty user name";
    if (lxx<99) {
      GameAlert(data);
      return;} //~if
    if (StrLength(data)<99) return;
    JSinfo = "";
    data = "<!doctype html public '-//w3c//dtd html 4.0 transitional//en'>"
        + "\n<html><head><title>" + title + "</title></head><body>\n"
        + data + "</body></html>";
    lxx = ServerUpload(whom,data);
    if (lxx>0) {
      JSinfo = "Upload error #" + lxx;
      ChooseBtn(5);
      return;} //~if
    ValidUserName = whom;
    link = "http://ittybittycomputers.com/Java/GE.php?name=" + link;
    GameAlert("OK, URL = '" + link + "'");} //~MakeUserJS

  /**
  * Where ChooseBtn selects one (or none) of four buttons for the user
  * to click on, DoMyButn does whatever that button is specified to do
  * when the user clicks on it, and returns true if it did it.
  *
  * @param  whom    the widget clicked on
  *
  * @return         true if it did it, false if not one of these 4 buttons
  */
  public boolean DoMyButn(GameWgt whom) {
    boolean doit;
    int rx = 0, zx = 0, why = 0;
    GameWgt more = ActivWgt; // might get cleared by Blur
    GameWgt nuly = UsWgHiLi;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      rx = whom.GetRefNum();
      why++; // why = 2
      if (whom.GetPosn()>0x30000000) break; // not visi
      why++; // why = 3
      if (false) if (nuly != null) if (nuly.GetPosn()>0x30000000)
        break; // User Wgt not sel'd
      why++; // why = 4
      if (whom==OKbtn) {
        if (more != null) {
          if (ToolActive) {
            CapturInfo2Wgt(more,false,false);
            // ** drop an instance on the game board
            NewData(true);} //~if
          else {
            CapturInfo2Wgt(more,true,false);
            if (!FreshFaster) doit = MakeWgtLst(); // T: failed (ignored)
            ChooseBtn(3);
            InChanged = false;}} //~if
        why = ~why;
        if (Logging) System.out.println("(OKbtn) A=#" + SafeWgtNum(more)
            + IffyStr(FreshFaster," FF"," nf"));} //~if
      else if (whom==BildBtn) {
        why = -why;
        if (Logging) System.out.println("(BildBtn) A=#" + SafeWgtNum(more));
        MakeUserJava();} //~if
      else if (whom==JSbtn) {
        why = -why;
        if (Logging) System.out.println("(JSbtn)");
        ChooseBtn(0);
        // GameAlert("Coming to a theater near you: JavaScript");
        MakeUserJS("");} //~if
      else if (whom==CantJS) {
        why = -why;
        if (Logging) System.out.println("(CantJS)");
        ChooseBtn(0);
        zx = StrLength(JSinfo);
        if (zx>0) if (zx<99) GameAlert(JSinfo);} //~if
      else if (whom==UndoBtn) {
        why = -why;
        if (Logging) System.out.println("(UndoBtn)");
        } //~if
      else break;
      BlurEv(null);
      if (why>0) why = 0;
      break;} //~while // if (why >= 0) // only exit comes thru here..
    if (Logging)
      System.out.println("(DoMyBtn #" + rx + ") A=" + SafeWgtNum(ActivWgt)
          + " = " + why );
    return (why <= 0);} //~DoMyButn

  /**
  * When the user clicks on a tool or game widget, SelectMe is called
  * to put that widget's information up on the tool panel. It returns
  * false if it is not a tool or game widget.
  *
  * @param  user    >0 if it's a user game widget, =0 if a tool
  * @param  whom    the widget clicked on
  *
  * @return         true if it's now selected, false if failed
  */
  public boolean SelectMe(int user, GameWgt whom) { // rtns true if did
    boolean prio = FreshFaster; // user=0 if tool, >0 game wgt, <0 in transit
    int info, nx = 0, rx = 0, zx = 0, posn = 0, thar = 0, why = 0;
    boolean velo = false, doit = false; // T: hide VeloEntry/AIdlgChkBx
    GameWgt nwgt = null;
    GameWgt nuly = ToolHiLi;
    GameWgt whar = UsWgHiLi;
    while (true) {
      if (ActivWgt != null) if (whom != ActivWgt) if (ChosenBtn==1)
          if (whom != null) if (OKbtn != null) {
        if (IzUsrWgt(whom)) zx++;
        else if (IzTool(whom)) zx--;
        FreshFaster = true; // auto-update: don't write file
        if (zx !=0) if (DoMyButn(OKbtn)) if (!prio) ChooseBtn(0);
        FreshFaster = prio;} //~if
      if (nuly != null) if ((whom==null)||(user !=0)) { // hide tool hili..
        why++; // why = 1
        nuly.HideSho(false); // if (nuly.GetPosn()<0x30000000) // (unneed)
        if (ActivWgt == GameBoard) doit = true;
        else if (NeedsVelo(ActivWgt)) velo = true;
        if (whom==ActivWgt) ActivWgt = null;} //~if // only cleared here (2x)..
      if (whar != null) if ((whom==null)||(user <= 0)) { // hide wgt hili..
        why = why+2;
        if (whar.IsVisible()) { // got a wgt sel'd..
          if (OKbtn != null) if (OKbtn.IsVisible()) { // OK btn shown..
            FreshFaster = true; // auto-update: don't write file
            if (DoMyButn(OKbtn)) if (why==2) why++; // why = 3
            FreshFaster = prio;} //~if
          whar.HideSho(false);} //~if
        if (ActivWgt == GameBoard) doit = true;
        else if (NeedsVelo(ActivWgt)) velo = true;
        if (whom==ActivWgt) ActivWgt = null;} //~if // ..and here (2nd of 2x)
      if (whom==null) break; // why = 1/2 (1st brk)
      if (whom==ActivWgt) break;
      why = why+4; // why = 4..7
      nx = whom.GetRefNum(); // = ID of dragable asset list
      info = whom.GetInfo()&0xFFFF;
      // if ((info&-4) != 0x4454) break; // 'DT' or 'DW'
      zx = nx;
      if (nx <= 0) break; // probly can't
      zx = info;
      if ((whom.GetType()&255) !=5) break;
      if (user !=0) {
        why = why+4; // why = 8..11
        if (info != 0x4457) break; // not 'DW'
        if (whar != null) if (UwgtNamList != null) { // look for name field..
          nwgt = FindSubWgt(whom,(TextFontNo<<16)+0x0404,0);
          if (Logging) if (nwgt==null)
            nwgt = FindSubWgt(whom,(TextFontNo<<16)+0x0404,9000);
          why = why+4; // why = 12..15
          if (nwgt==null) break;
          rx = nwgt.GetRefNum(); // = ID of actual name
          // rx = nwgt.GetParent();
          why = why+4; // why = 16..19
          zx = rx;
          if (rx <= 0) break; // probly can't
          why = why+4; // why = 20..23
          if (user>0) {
            if (UwgtNamList.FindElemt(rx)<0) break; // not an installed user wgt
            why = why+4; // why = 24..27
            posn = nwgt.GetPosn()-0x40004;
            whar.SetPosn(posn>>16,posn&0xFFFF);} //~if // shows hili there
          else if (SpriteList==null) break; // no sprite list for transients
            else if (SpriteList.FindElemt(nx)<0) break; // not a transient
          if (ActivWgt == GameBoard) doit = true;
          else if (NeedsVelo(ActivWgt)) velo = true;
          if (whom == GameBoard) if (AIdlgChkBx != null)
            AIdlgChkBx.HideSho(true);
          ActivWgt = whom;
          ToolActive = false;
          ShoInfoFrom(whom);
          why = -why;}} //~if
      else if (nuly != null) {
        why = why+32; // why = 36..39
        if (info != 0x4454) break; // not 'DT'
        rx = whom.GetParent();
        zx = rx;
        if (rx <= 0) break; // probly can't
        why = why+4; // why = 40..43
        nwgt = GetIxWgt(rx); // this is the locator token in list
        if (nwgt==null) break;
        posn = nwgt.GetPosn();
        nuly.SetPosn(posn>>16,posn&0xFFFF); // auto'ly shows it there
        if (NeedsVelo(whom)) velo = true;
        ActivWgt = whom;
        ToolActive = true;
        ShoInfoFrom(whom);
        why = -why;} //~if
      break;} //~while
    if (ActivWgt==null) { // something went wrong..
      if (nuly != null) nuly.HideSho(false);
      if (whar != null) whar.HideSho(false);
      ToolActive = false;
      ShoInfoFrom(NullWgt);
      why = why+200;
      doit = true;} //~if
    if (velo) if (VeloEntry != null) VeloEntry.HideSho(NeedsVelo(ActivWgt));
    if (doit) if (ActivWgt != GameBoard)
      if (AIdlgChkBx != null) AIdlgChkBx.HideSho(false);
    if (Logging) System.out.println("(SelecMe #" + nx + ") " + user + " " + zx
        + " " + doit + " A=" + SafeWgtNum(ActivWgt) + IffyStr(prio," FF"," nf")
        + " = " + why );
    FreshFaster = prio;
    return why<0;} //~SelectMe

  /**
  * When a tool is dragged to the game board, NewListName is called to
  * add its name to the list of user game widgets and selected.
  *
  * @param  whom    the widget to be added to the list
  *
  * @return         true if it succeeded
  */
  public boolean NewListName(GameWgt whom) { // whom must be asset list
    int ix, nx = 0, rx = 0, zx = 0,          // (but not easy to vfy)
        thar = 0, top = 0, why = 0;       // rtns true if there & sel'd
    String name = "";
    int[] data = null;
    GameWgt nuly = null, nwgt = null;
    GameWgt whar = ListofWgts;
    while (true) {
      why++; // why = 1
      if (whar==null) break;
      why++; // why = 2
      if (whom==null) break;
      nx = whom.GetRefNum();
      why++; // why = 3
      if (whar.FindElemt(nx)<0) { // else already there
        why++; // why = 4               // look for name field..
        nwgt = FindSubWgt(whom,(TextFontNo<<16)+0x0404,0);
        if (Logging) if (nwgt==null)
          nwgt = FindSubWgt(whom,(TextFontNo<<16)+0x0404,9000);
        if (false) if (nwgt==null) { // ad-hoc W/A..
          for (rx=0; rx <= 5; rx++) {
            nwgt = GetIxWgt(whom.GetIxData(rx,0));
            if (nwgt==null) break;
            zx = nwgt.GetType();
            ix = nwgt.GetInfo();
            if (zx==1028) if (ix == TextFontNo) break;
            name = name + " " + rx + ":" + zx + "/" + ix;} //~for
          if (nwgt != null) why = why+10;
          else if (whom==GameBoard) {
            nwgt = FindListWgt("[Game board actual name]");
            if (Logging) System.out.println("  (NLN).." + name);}} //~if
        why++; // why = 5
        if (nwgt==null) break;
        name = "";
        rx = nwgt.GetRefNum(); // this is the actual name wgt
        why++; // why = 6
        name = nwgt.GetTextLn();
        if (StrLength(name)==0) break;
        why++; // why = 7
        top = whar.AddElemt(nx)*WgtNameHi; // index (now posn), or -1 if failed
        if (top<0) break;
        whar = UwgtNamList;
        why++; // why = 8
        if (whar==null) break;
        zx = nwgt.GetPosn();
        nwgt.SetPosn((zx>>16)+top,SignExtend(zx));
        zx = whar.GetRefNum();
        if (false) {
          why = 13;
          data = new int[3];
          if (data==null) break;
          data[0] = rx;
          data[1] = nx;
          why++; // why = 14
          nuly = NewGameWgt(5,top,0,16,56,UwgtNamID,-1,0,data,null);
          if (nuly==null) break;
          thar = nuly.GetRefNum();
          why++; // why = 15
          if (whom.AddElemt(thar)<0) break; // add link to asset list
          why = 8;} //~if
        why++; // why = 9
        if (whar.AddElemt(rx)<0) break; // finally add its name to name list
        nwgt.SetParent(zx); // owned by (so it displays in) the list
        why++;} //~if // why = 10
      else why = 11; // (in case there, but sele failed)
      if (SelectMe(nx,whom)) { // (T: succeeded) also displays it
        if (!FreshFaster) Antsify(nx>0,whom);
        why = -why;} //~if
      break;} //~while
    if (Logging) System.out.println("(NewLstNm #" + nx + ") => #" + thar
        + " '" + name + "' " + " " + rx
        + IffyStr(FreshFaster," FF = "," nf = ") + why);
    return why <= 0;} //~NewListName

 /**
  * If this widget is a group child, move it up or down in its group's data.
  * Called from KeyFld, no early return.
  *
  * @param  whom  The widget to be moved
  * @param  info  true if up
  */
  public void MoveWgtUpDn(GameWgt whom, boolean uppy) {
    int rx = 0, px = 0, inx = 0, why = 0;
    GameWgt prnt;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      rx = whom.GetRefNum();
      px = whom.GetParent();
      why++; // why = 2
      if (px <= 0) break;
      prnt = GetIxWgt(px);
      why++; // why = 3
      if (prnt==null) break;
      why++; // why = 4
      if (((prnt.GetType()-5)&255)>2) break;
      why++; // why = 5
      inx = prnt.FindElemt(rx);
      if (inx<0) break;
      why++; // why = 6
      if (!uppy) { // move down = exchange w/lower in data..
        if (inx==0) {
          why++; // why = 7
          break;} //~if
        inx--;} //~if // why = 6
      else if (prnt.GetIxData(inx+1,0)==0) break; // why = 6
      prnt.XchElemt(inx);
      why = 0;
      break;} //~while
    if (Logging) System.out.println("(MoveWg #" + rx + ") @ " + inx
        + " in #" + px + " " + uppy + " = " + why);} //~MoveWgtUpDn

  // These are enabled by Gtype&512:

 /**
  * KeyFld is called for any widget that has focus when the user
  * types any other key. This override looks for keystrokes that could
  * change data, and marks the game as changed if so. It depends on super
  * to enter those keystrokes into whatever text field has focus.
  *
  * @param  whom  The widget that has focus
  * @param  info  The key that was typed
  */
  public void KeyFld(GameWgt whom, char info) { // frm KeyData (GameMk)
    if (Logging) System.out.println("  (GmMk.KeyFld) #" + SafeWgtNum(whom)
        + " " + (UserMode) );
    if (DlogIsOpen>0) {
      if (whom != DlgInputData) if (DlgInputData != null)
          if (DlgInputData.IsVisible()) if (whom != DlgInputData) {
        whom = DlgInputData;
        if (FocusWgt != whom) if (DefaultFocus != whom)
          FocusEv(whom,-8,-4);}} //~if
    else if ((whom.GetType()&255) !=4) {
      if ((info == '+')||(info == '-')) {
        if (whom != null) MoveWgtUpDn(whom,info == '+');
        return;} //~if
      } //~if
    super.KeyFld(whom,info);
    if ((info >= ' ')||(info == '\b')) NewData(false);} //~KeyFld

 /**
  * GotArrowKey is called for any widget that has focus when the user
  * types one of the arrow keys. This override diverts left/right arrows
  * to the FocusWgt (if it is text) or the input dialog, if it is open.
  *
  * @param  whom  The widget that has focus
  * @param  info  The arrow key that was pressed, one of the constant
  *               values JavaGame.Arro_UP, JavaGame.Arro_DOWN,
  *               JavaGame.Arro_LEFT, or JavaGame.Arro_RIGHT.
  */
  public void GotArrowKey(GameWgt whom, char info) { //        // (GameMk)
    int tye = 0, refn = 0, whar = 0, why = 0;
    GameWgt prio = null;
    if (whom==null) return;
    tye = whom.GetType();
    if ((info == Arro_RIGHT)||(info == Arro_LEFT)) {
      if (DlogIsOpen>0) why--;
      else if ((FocusWgt==whom)||(DefaultFocus==whom))
        if ((tye&767)==516) why--;
      if (why<0) {
        KeyFld(whom,info);
        return;}} //~if
    if ((tye&2048)==0) return;
    whar = DrugWgtOfx; // (still valid after mouseUp)
    if (whar<0) return;
    if (info == Arro_DOWN) whar = whar+0x10000;
    else if (info == Arro_UP) whar = whar-0x10000;
    else if (info == Arro_RIGHT) whar++;
    else if (info == Arro_LEFT) whar--;
      else return;
    refn = whom.GetRefNum();
    prio = WgtBeingDrug;
    WgtBeingDrug = whom;
    Dragging(whom,whar>>16,SignExtend(whar));
    if (Logging)
      System.out.println("** DragKey #" + refn + " ** = " + why);
    WgtBeingDrug = prio;} //~GotArrowKey

 /**
  * BlurEv is called for a widget when it loses focus. This override lets
  * super do whatever needs doing, except that it looks for active marching
  * ants and turns them off.
  *
  * @param  whom  The widget losing focus, or null if unknown by caller
  */
  public void BlurEv(GameWgt whom) { // from MsDispch             (GameMk)
    if (Logging) System.out.println("  (GmMk.BlrEv) " + (mAntsHost != null));
    if (mAntsHost != null) Antsify(false,mAntsHost);
    super.BlurEv(whom);} //~BlurEv

 /**
  * FocusEv is called for any widget that is "Focusable" when the user
  * clicks on it. This override lets super do whatever needs doing,
  * except that it also calls SelectMe and then Antsify activate the
  * marching if appropriate.
  *
  * @param  whom  The widget to be given focus
  * @param  vert  The vertical offset into this widget where
  *                 it was clicked
  * @param  horz  The horizontal offset into this widget
  */
  public void FocusEv(GameWgt whom, int vert, int horz) { //   // (GameMk)
    int nx = 0, yx = 0, zx = 0, offx = 0, posn = 0, info = 0, why = 0;
    GameWgt nuly = null;
    if (DlogIsOpen==0) {
      if (whom != null) {
        nx = whom.GetRefNum();
        info = whom.GetInfo();
        if (UwgtNamPos>0x100000) if (ListofWgts != null) {
          why = 2;
          posn = whom.AddPrntsPos(whom.GetPosn());
          offx = posn-UwgtNamPos;
          if ((offx&0x80008000)==0) { // in the game list,
            zx = (offx>>16)/WgtNameHi; // ..redirect it to asset list..
            nuly = GetIxWgt(ListofWgts.GetIxData(zx,0));
            if (nuly != null) info = nuly.GetInfo();
              else why = 4;
            if ((info&0x4040FFFF) !=0x40404457) { // 'DW'
              why++; // why = 3
              nuly = null;} //~if
            whom = nuly;} //~if
          else why--;}} //~if // why = 1: (not in the game list)
      yx = info&0xFFFF;
      if ((info&0x7FFF0000)==0x52640000) yx = -82; // = 'Rd__'
      else if (yx==0x4F4B) yx = -79; // '__OK'
      else if (yx==0x436E) yx = -67; // '__Cn'
      // else if (whom==OKbtn) yx = 0;
      else if (whom==GdChkBx) yx = -71;
      else if (whom==TypChkBx) yx = -84;
      else if (whom==ClikChkBx) yx = -75;
      else if (whom==DragChkBx) yx = -68;
      else if (whom==BumpChkBx) yx = -66;
      else if (whom==RollChkBx) yx = -76;
      else if (whom==AIdlgChkBx) yx = -65;
      else if (whom==DlgInputData) yx = -73;
      else if (whom==BildBtn) yx = -98;
      else if (whom==JSbtn) yx = -74; // 'J'
      else if (whom==CantJS) yx = -78; // 'N'
      else if (whom==UndoBtn) yx = -85;
        else yx = nx;} //~if
    else if (whom != DlgInputData) if (DlgInputData != null)
      if (DlgInputData.IsVisible()) whom = DlgInputData;
    if (Logging||(why>2)) System.out.println("  (GmMk.FocEv #" + IffyStr(nuly==null,
        "","" + SafeWgtNum(nuly) + "/") + nx + ") " + IffyStr(why>0,
          LogVH("(",posn,LogVH(IffyStr(offx<0,") ",") +"),offx," => ")) + zx,"")
        + " f=" + SafeWgtNum(FocusWgt) + " A=" + SafeWgtNum(ActivWgt)
        + IffyStr(yx>0," T"," " + yx) + SeeHex(" i=x",info,
          IffyStr(FreshFaster," FF = "," nf = ")) + why);
    if (whom==null) {
      BlurEv(null);
      return;} //~if
    super.FocusEv(whom,vert,horz);
    if (whom == DlgInputData) return;
    if (IzTool(whom)) nx = 0;
    else if (FocusWgt != whom) nx = 0;
    else if (!SelectMe(nx,whom)) nx = yx; // F: failed, yx = nx/-reason
    else if (yx<0) nx = yx; // successfully selected but should not do ants
    if (!FreshFaster) Antsify(nx>0,whom);} //~FocusEv

 /**
  * GotEnterKey is called when the user types the EnterKey,
  * if FocusWgt is not null, or else if DefaultFocus is not null,
  * and that widget accepts keystrokes. This override lets super handle
  * an open alert or input dialog, and if true it acts accordingly.
  * Otherwise it directs the action to whatever OK button is showing,
  * or if none, then if the color swatch is selected, it opens the input
  * dialog to accept a new color, or if a text tool or game widget is
  * selected, it opens the input dialog to accept new text.
  *
  * @param  whom  The widget that has focus
  *
  * @return       true if taken (by 2-click or alert or input dlog)
  */
  public boolean GotEnterKey(GameWgt whom) { // from KeyData         (GameMk)
    int kx = EvsDlogOp, nx = 0, info = 0, why = 0;
    boolean gotit = false;
    String msg = "", data = "";
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom != null) {
        nx = whom.GetRefNum();
        info = whom.GetInfo();} //~if
      if (kx !=0) { // = DlogIsOpen, except <0 when dlog about to close
        nuly = GetIxWgt(Math.abs(kx));
        data = DlgEvData();} //~if
      if (Logging) msg = "(GotEntKy #" + nx + " df=#" + SafeWgtNum(DefaultFocus)
          + ") A=" + SafeWgtNum(ActivWgt) + " o=#" + SafeWgtNum(nuly)
          + " cb=" + ChosenBtn + IffyStr(kx==0," ",
            " " + kx + "/" + InputWhy + " '" + data + "' ");
      if (super.GotEnterKey(whom)) {
        if (theInputDlg != null) if (theInputDlg==nuly) {
          whom = nuly;
          gotit = true;} //~if
        if (!gotit) if (theAlertDlg != null) if (theAlertDlg==nuly) {
          whom = nuly;
          gotit = true;} //~if
        if (gotit) // defalt foc was stolen by dlog, restore it to OK btn..
          if (ChosenBtn>0) ChooseBtn(ChosenBtn); // ** better: save/restore
        gotit = false;
        if (whom==null) break;
        why = 10+InputWhy;
        if (whom==theInputDlg) {
          // data = DlgEvData();
          msg = msg + " = '" + data + "'";
          switch (InputWhy) {
          case 1: // got new color..
            if (StrLength(data)>0) {
              SetSwatchColor(ColoBx,SafeParseInt(data));
              NewData(false);} //~if
            else gotit = GotEscKey(whom); // none, treat it as Cancel
            break;
          case 2: // got new text for txt wgt..
            if (TextInWgt==null) break;
            info = TextInWgt.PutTextLn(data);
            TextInWgt = null;
            break;
          case 7: // got new Username_password..
            if (StrLength(data)>0) MakeUserJS(data);
            break;
          case 0: // (none)
            break;} //~switch
          InputWhy = 0;} //~if
        else if (whom != theAlertDlg) break;
        // if (whom != null) whom.HideSho(false); // super did it
        // DlogIsOpen = 0;
        why = why+10;} //~if
      else if (whom==ColoBx) {
        why++; // why = 2
        if (whom==null) break;
        why++; // why = 3
        BlurEv(whom);
        info = SwatchGetColor(whom);
        if (info >= 0)
          data = SeeHex("0x",info,"");
        else data = "-1";
        if (OpenInputDlog("New color: (0xRRGGBB)",data)) InputWhy = 1;} //~if
      else if ((info&0x7FFFFFFC) == 0x54784454) { // 'TxDT'/'TxDW'
        why = 4;
        if (whom != null) nuly = GetIxWgt(whom.GetIxData(0,0));
        if (nuly==null) break;
        why++; // why = 5
        TextInWgt = nuly;
        data = nuly.GetTextLn();
        if (OpenInputDlog("New text:",data)) InputWhy = 2;} //~if
      else if (ChosenBtn==1) {
        why = 31;
        if (OKbtn != null) gotit = OKbtn.IsVisible();
        if (Logging)
          msg = msg + " (OKbtn=1) " + (OKbtn != null) + "/" + gotit;
        if (gotit) {
          if (Logging) msg = msg + " => DBtn";
          FreshFaster = false;
          if (DoMyButn(OKbtn)) break;} //~if
        else if (OKbtn != null) OKbtn.LogMe(" ?? ");} //~if
      else if (ChosenBtn==3) {
        why = 33;
        if (BildBtn != null) if (BildBtn.IsVisible())
          if (DoMyButn(BildBtn)) break;} //~if
      else if (ChosenBtn==4) {
        why = 34;
        if (JSbtn != null) if (JSbtn.IsVisible())
          if (DoMyButn(JSbtn)) break;} //~if
      why = -why;
      break;} //~while                   // "(GotEntKy #"
    if (Logging) System.out.println(msg + " DF=#" + SafeWgtNum(DefaultFocus)
        + " = " + why );
    return false;} //~GotEnterKey

  /**
  * CliKnownStuff tries to find what widget got clicked among the known
  * widgets it manages, such as a tool or a game widget or a radio
  * button (checkboxes and buttons are handled elsewhere).
  *
  * @param  whom    the widget clicked on
  *
  * @return         true if it succeeded
  */
  public boolean CliKnownStuff(GameWgt whom, int vert, int horz) {
    int nx = 0, px = 0, zx = 0, info = 0, why = 0;
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break;
      why++; // why = 2
      nx = whom.GetRefNum();
      FreshFaster = false;
      if (IzTool(whom)) {
        if (SelectMe(0,whom)) why = -why;
        Antsify(false,null);
        break;} //~if
      why++; // why = 3
      if ((whom.GetType()&255)==4) {
        if (whom==FocusWgt) why = -why;
        break;} //~if
      why++; // why = 4
      px = whom.GetParent();
      if (px==0) break;
      why++; // why = 5
      nuly = GetIxWgt(px);
      if (nuly==null) break;
      why++; // why = 6
      info = whom.GetInfo();
      if ((info&-256) == 0x52645F00) { // 'Rd_*' = radio btn..
        if (nuly.RadioBtn(info&255)>0) { // >0: selected it..
          if (ActivWgt != null) { // update font of active text..
            zx = ActivWgt.GetRefNum(); // (for log)
            if ((ActivWgt.GetInfo()&-4)==0x54784454) // 'TxDT'/'TxDW'
              CapturInfo2Wgt(ActivWgt,false,true);} //~if
          why = -why;} //~if // why = -6
        break;} //~if
      why++; // why = 7
      zx = nuly.GetParent();
      if (zx==UwgtNamID) if (SelectMe(nx,whom)) {
        Antsify(nx>0,whom);
        why = 0;} //~if
      break;} //~while
    if (Logging || (why>0)) System.out.println("(CliKnoStf #" + nx + ") p=" + px
        + IffyStr(why+6==0," aw=","/") + zx + " = " + why);
    return why <= 0;} //~CliKnownStuff

  // This event is enabled by Gtype&1024, frm MouseDisp:        // (GameMk)

 /**
  * ClickEvt is called for a widget that is "Clickable" when the user
  * clicks on it (mouse down followed by mouse up on the same widget).
  * This override lets super detect and process double-click and dialog
  * events, and if not that, it tries the various things it knows about,
  * whether one of the known checkboxes, or DoMyButn or CliKnownStuff.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset into this widget where
  *                 it was clicked
  * @param  horz  The horizontal offset into this widget
  *
  * @return       true if taken (by 2-click or alert or input dlog)
  */
  public boolean ClickEvt(GameWgt whom, int vert, int horz) { //  (GameMk)
    int prio = SafeWgtNum(ActivWgt), info = 0, why = 0;
    String msg = "";   // MouseDisp already called FocusEv -> SetFocus
    boolean gotit = false;
    while (true) {                       // (vert,horz are wgt-relative)
      why++; // why = 1
      if (whom==null) break;
      why++; // why = 2
      if (super.ClickEvt(whom,vert,horz)) break; // (2-click or dlog)
      msg = whom.GetIDtxt();
      if (StrLength(msg)==0) msg = " " + whom.GetInfo();
        else msg = " '" + msg + "'";
      msg = IffyStr(CluckTime==0," ctm=0 "," ctm>0 ")
          + vert + "/" + horz + msg;
      // if (CluckTime==0) break; // super converted 2-click to Enter
      why++; // why = 3
      if (DoMyButn(whom)) why = 6; // T: did it
      else if (CliKnownStuff(whom,vert,horz)) why = 7; // T: did it
      else if (whom==GdChkBx) {
        why++; // why = 4
        gotit = whom.CheckIt(true);} //~if
      else if (whom==AIdlgChkBx) {
        why = 20;
        gotit = whom.CheckIt(true);} //~if
      else if (whom==TypChkBx) {
        why = 15;
        gotit = whom.CheckIt(true);} //~if
      else if (whom==ClikChkBx) {
        why = 16;
        gotit = whom.CheckIt(true);} //~if
      else if (whom==DragChkBx) {
        why = 17;
        gotit = whom.CheckIt(true);} //~if
      else if (whom==BumpChkBx) {
        why = 18;
        gotit = whom.CheckIt(true);} //~if
      else if (whom==RollChkBx) {
        why = 19;
        gotit = whom.CheckIt(true);} //~if
      else info = whom.GetInfo(); // (why = 3) // PushButn(whom,4);
      if (why>14) NewData(false);
      break;} //~while
    System.out.println("(GameMk ClickEv) #" + SafeWgtNum(whom) + msg + " DF#"
        + SafeWgtNum(DefaultFocus) + " A=" + SafeWgtNum(ActivWgt)
        + "/" + prio + " " + info + " " + gotit + " = " + why);
    return true;} //~ClickEvt

  // These are enabled by Gtype&2048, frm MouseDisp:

  // public void MouseDn(GameWgt whom, int vert, int horz) {}
  // public void MouseUp(GameWgt whom, int vert, int horz) {}

 /**
  * CancelDrag is called to cancel (terminate without effect) a drag.
  * This override lets super do whatever needs doing, after deleting any
  * transient (dragged tool) from the user widget list(s).
  *
  * @param  whom  The widget being dragged
  */
  public void CancelDrag(GameWgt whom) {
    int ix, nx = 0, px = 0, zx = 0, why = 0;
    GameWgt nuly = null;
    if (whom != null) if (WgtBeingDrug==whom) {
      nx = whom.GetRefNum();     // look for name field..
      nuly = FindSubWgt(whom,(TextFontNo<<16)+0x0404,0);
      while (true) {
        why++; // why = 1
        if (nuly==null) break;
        zx = nuly.GetRefNum();
        px = nuly.GetParent();
        why++; // why = 2
        if (px==0) break;
        why = ~why; // why = -3: OK if me
        if (px==nx) break;
        whom.HideSho(false);
        why = -why; // why = 3
        if (SpriteList != null) { // if transient, remove it from sprite list..
          ix = SpriteList.FindElemt(nx);
          if (ix >= 0) SpriteList.DeleElemt(ix);} //~if
        nuly = GetIxWgt(px);
        why = -why; // why = 3
        if (nuly==null) break;
        why++; // why = 4
        ix = nuly.FindElemt(zx);
        if (ix<0) break;
        nuly.DeleElemt(ix);
        why = ~why; // why = -5: success
        break;} //~while // why = 0: if no actual name (OK)
      if (Logging) System.out.println("  (CancMyDrag #" + nx + ") n= #" + px
          + "/" + zx + " = " + why);} //~if
    NewData(false);
    super.CancelDrag(whom);} //~CancelDrag

 /**
  * After a StartDrag has been called for a Dragable widget and the user
  * continues to move the mouse, Dragging is called for the same widget.
  * This override lets super do whatever needs doing, then shows the new
  * coordinates and marks the widget changed.
  *
  * @param  whom  The widget that is being dragged
  * @param  vert  The vertical offset into this widget before it moves
  * @param  horz  The horizontal offset into this widget
  */
  public void Dragging(GameWgt whom, int vert, int horz) { //     (GameMk)
    super.Dragging(whom,vert,horz);
    NewData(false);      // update displayed coords, if valid..
    if (whom != null) if (whom==ActivWgt) ShoCoords(whom,false);} //~Dragging

 /**
  * After a MouseDn is called for a widget that is "Dragable" and the
  * user moves the mouse at least eight pixels with the button still
  * pressed, then it becomes a drag and StartDrag is called. This
  * override looks to see if it's a tool being dragged, and if so,
  * clones it into a game widget; then lets super do whatever needs
  * doing with the new widget.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset into this widget where it
  *                 was clicked
  * @param  horz  The horizontal offset into this widget
  *
  * @return       The widget being dragged
  */
  public GameWgt StartDrag(GameWgt whom, int vert, int horz) { // (GameMk)
    int info = 0, prio = 0, nx = 0, why = 0;
    String name = "";
    GameWgt nuly = FontRad;
    // we are allowed to substitute a token for whom (null cancels)..
    while (true) {
      why++; // why = 1
      if (DlogIsOpen>0) whom = null;
      if (whom==null) break;
      FreshFaster = false;
      if (Logging) whom.LogMe(" << ");
      why++; // why = 2
      NewData(false);
      nx = whom.GetRefNum();          // clone it only if tool..
      if ((whom.GetInfo()&0xFFFF) != 0x4454) { // 'DT' = Drag Tool
        why = -why; // why = -2
        break;} //~if
      prio = nx;
      nx = 0;
      why++; // why = 3
      whom = CloneWgt(whom,0); // (a) sets new parent to Sprite list,
      if (whom==null) break;   // (b) globalizes wgt coords to match..
      why++; // why = 4
      nx = whom.GetRefNum();        // look for wgt name field..
      nuly = FindSubWgt(whom,(TextFontNo<<16)+0x0404,0);
      if (nuly != null) {
        why++; // why = 5
        name = nuly.GetTextLn();
        info = NthOffset(0,".0",name);
        if (info>2) if (StrLength(name) == info+2) {
          why++; // why = 6
          NextWgtName++;
          name = Substring(0,info+1,name) + NextWgtName;
          info = nuly.PutTextLn(name);}} //~if
      if (SelectMe(-nx,whom)) why = -why;
      Antsify(false,null);
      break;} //~while
    if (Logging || (why>0)) {
      System.out.println("(StartDrag #" + nx + "/" + prio + ") '" + name
          + "' " + info + "/" + NextWgtName + " = " + why);
      if (whom != null) whom.LogMe(" >> ");} //~if
    if (whom==null) return null;
    return super.StartDrag(whom,vert,horz);} //~StartDrag

 /**
  * After a StartDrag has been called for a Dragable widget and the user
  * releases the mouse button, Dropped is called for the same widget.
  * This override looks to see if it's a tool being dragged, and if so,
  * and if the drop is on the game board, adds it to both the game board
  * and user widget list.
  *
  * @param  whom  The widget that was dragged
  */
  public void Dropped(GameWgt whom) { // from MouseUp          (GameMk)
    // (coords already updated by Dragging before call)
    int ix, rx, kx = 0, zx = 0, nx = 0, here = 0, thar = 0, whar = 0,
        posn = 0, more = 0, why = 0;
    GameWgt nuly = null;
    while (true) {
      why++; // why = 1
      if (whom==null) break; // if not in sprite list, everything is OK..
      nx = whom.GetRefNum();
      super.Dropped(whom); // only logs
      why++; // why = 2
      zx = (whom.GetSize()>>1)&0x7FFF7FFF; // = half-size, for bt/rt test
      why++; // why = 3
      if (SpriteList==null) break; // no sprite list for transients
      kx = SpriteList.FindElemt(nx);
      why = ~why; // why = -4
      if (kx<0) break; // not a transient tool clone (OK)
      why = -why; // why = 4
      if (GameBoard==null) break; // no game board to transfer it to
      why++; // why = 5
      here = GameBoard.AddPrntsPos(GameBoard.GetPosn()); // absolute game locn
      posn = whom.AddPrntsPos(whom.GetPosn()); // absolute drop locn of wgt
      // thar = AddPair(posn,(vert<<16)+(horz&0xFFFF));
      whar = AddPair(posn,PairNeg(here)); // game-rel drop locn of wgt
      thar = AddPair(whar,zx); // = game-rel middle of droppee
      more = GameBoard.GetSize();
      if ((thar&0x80008000) !=0) break; // why = 5: mid off game bd up or left
      why++; // why = 6
      zx = AddPair(more,PairNeg(thar)); // >0: mid is inside
      if ((zx&0x80008000) !=0) break; // off the game board down or right
      why++; // why = 7
      SpriteList.DeleElemt(kx); // kx = SpriteList.FindEl
      kx = GameBoard.AddElemt(nx); // = [Game board asset list]
      if (kx<0) break; // failed to add it
      why++; // why = 8
      if (!NewListName(whom)) break; // T: name for wgt is now in wgt name pane
      whom.SetParent(GameBoard.GetRefNum()); // OK, committed
      whom.SetPosn(whar>>16,whar&0xFFFF);
      ShoInfoFrom(whom); // gotta show also info added by import
      NewData(true);
      why = 0;
      break;} //~while
    if (why>1) CancelDrag(whom);
    if (Logging || (why>0)) System.out.println("(Dropd #" + nx
        + LogVH(") @ ",whar,LogVH(" ",here,LogVH(" < ",thar,LogVH(" (",zx,
          LogVH(" ",posn,") = "))))) + why);} //~Dropped // why =

// These methods read in the saved file "GameMaker.txt" and convert its
//   widget lines into game widgets already on the board..

  /**
  * NewDragWgt creates a draggable game widget corresponding to a line from
  * the saved file "GameMaker.txt" and updates its parameters from the
  * attributes on the input line. Only from LoadUserWgt.
  *
  * @param  aLine   one line of import data to convert to game widget
  * @param  lino    its line line number
  */
  private void NewDragWgt(String aLine, int lino) { // create draggable wgt
    int scrn = GameWinSize()>>16,      // ..from AddUserWgts line
        zx, kx, rx = 0, posn = 0, vert = 0, horz = 0, tall = 0, wide = 0,
        tye = 0, info = 0, colo = 0, edge = 0, whar = 0, why = 0;
    GameWgt whom = null, nuly = null, more = null;
    int[] data = null;
    String name = "", aWord = "", aStr = "";
    while (true) {
      why++; // why = 1
      if (GameBoard==null) break; // (caller checked)
      posn = GameBoard.GetPosn();
      if (ListofWgts != null) {
        edge = ListofWgts.GetSize()&0xFFFF;
        why = 9;} //~if // why = why+8
      else edge = GameBoard.GetSize()&0xFFFF;
      why = -why; // why = -1
      tye = SafeParseInt(NthItemOf(' ',2,aLine)); // type+flags
      whar = tye&255;
      if (whar==3) break; // font (all fonts already in)
      why--; // why = -2
      if ((whar&254)==240) break; // Alert/Input Dlog (ditto)
      why = -why; // why = 2
      if (NthOffset(0,"{",aLine)>0)
        name = NthItemOf('{',2,NthItemOf('}',1,aLine));
      else if (NthOffset(1,"'",aLine)>0) {
        zx = Countem("'",aLine);
        if (zx>1) name = NthItemOf('\'',zx&-2,aLine);} //~if
      if (StrLength(name)==0) break;
      why++; // why = 3
      if (whar==0) whom = RexTool; // rect
      else if (whar==2) whom = CircTool; // icon (circle)
      else if (whar==4) whom = TexTool; // text
      else if (whar != 6) break; // why = 3: unrecognized type
      else if (NthOffset(0,"(B)",aLine)>0) { // ball..
        whom = BallTool;
        whar--;} //~if // = 5
      else if (NthOffset(0,"(F)",aLine)>0) { // Skying rect (sprite)..
        whom = SkyTool;
        whar++;} //~if // = 7
      else if (NthOffset(0,"(S)",aLine)>0) whom = SpriTool; // sprite
        else break; // why = 3: unrecognized type
      why++; // why = 4
      if (whom==null) break; // not defined tool
      why++; // why = 5
      whom = CloneWgt(whom,0); // (a) sets new parent to Sprite list,
      if (whom==null) break;   // (b) globalizes wgt coords to match..
      why++; // why = 6
      rx = whom.GetRefNum();
      if (!SelectMe(-rx,whom)) break;
      why++; // why = 7  // vert/horz are game-board-rel..
      vert = SafeParseInt(NthItemOf(' ',3,aLine)); // +(posn>>16);
      horz = SafeParseInt(NthItemOf(' ',4,aLine)); // +SignExtend(posn);
      tall = SafeParseInt(NthItemOf(' ',5,aLine));
      wide = SafeParseInt(NthItemOf(' ',6,aLine));
      colo = SafeParseInt(NthItemOf(' ',7,aLine));
      info = SafeParseInt(NthItemOf(' ',8,aLine));
      if (vert+3>scrn) break; // off-screen bottom
      if (vert+tall<3) break; // off-screen top
      if (horz+wide<3) break; // off-screen left
      if (horz+3>edge) break; // off-screen right
      more = FindSubWgt(whom,0x466C677A,0); // 'Flgz'
      if (more != null) {
        if (whar==7) more.SetPosn(info>>16,SignExtend(info));
        more.SetFlags(tye);} //~if
      else why = why|16;
      why++; // why = +8
      if (whar>4) if (whar<7) { // whar=5/6/7=ball/sprite/SkyRx..
        if (more != null) more.SetPosn(info>>16,SignExtend(info));
        zx = SafeParseInt(NthItemOf(' ',11,aLine));
        if (zx>0) aWord = NthItemOf('\n',zx,AddUserWgts);
        zx = StrLength(aWord);
        if (zx>0) {
          colo = SafeParseInt(NthItemOf(' ',7,aWord));
          aLine = aLine + IffyStr(StrLength(aLine)+zx>72,"'\n + '","'/'")
              + aWord;}} //~if
      if (whar==5) CircleSize(whom,tall); // whar=5: ball
      whom.SetPosn((posn>>16)+vert,SignExtend(posn)+horz); // dropp expects abs
      if (whar !=5) whom.SetSize(tall,wide);
      nuly = GetIxWgt(whom.GetIxData(0,0));
      if (whar==4) if (info>0) if (nuly != null)
          if ((nuly.GetType()&255)==4) { // get/set font before drop..
        aWord = NthItemOf('\n',info,AddUserWgts);
        if (SafeParseInt(NthItemOf(' ',2,aWord))==3) {
          zx = NthOffset(0,"[Font ",aWord);
          if (zx>0) zx = (int)(CharAt(zx+6,aWord)); // prefer name letter
          if (zx>84) zx = 0;
          if (zx<66) zx = SafeParseInt(NthItemOf(' ',8,aWord)); // MOS font+
          info = ChoseFont(nuly,zx);} //~if // sets the font wgt
        aStr = NthItemOf('\'',2,aLine);
        info = (nuly.PutTextLn(aStr)<<16)+(info&0xFFFF);} //~if
      // ShoCoords(whom,false); // Dropp does it..
      Dropped(whom);
      SetSwatchColor(ColoBx,colo);
      why++; // why = 9
      // if (whar==0) whom.SetInfo(info); else // rect
      if (whar>0) if (whar !=2) if (whar !=4) if (whar<5)
        break; // why = 9: (can't, already checked)
      NewData(true);
      if (WgtNamEntry != null) if (SetInputText(WgtNamEntry,"",name,true))
        if (DoMyButn(OKbtn)) ChooseBtn(1);
      why++; // why = 10
      zx = StrLength(theImpoList)-1;
      if (zx<9) break;
      for (kx=zx; kx <= zx-8; kx++) if ((int)(CharAt(kx,theImpoList))>31) {
        if (kx==zx) break;
        theImpoList = Substring(0,kx+1,theImpoList);
        break;} //~if
      zx = NthOffset(0," ",aLine);
      theImpoList = theImpoList
          + ("\n" + rx + RestOf(zx,NthItemOf('\n',1,aLine)));
      why = 0;
      break;} //~while
    if (Logging || (why>0)) {
      System.out.println("(NuDrgWg #" + rx + " => " + whar + IffyStr(tye>whar,
          SeeHex("/",tye," @ ")," @ ") + vert + "/" + horz + " +" + tall
          + "/" + wide + SeeHex(" c=",colo,SeeHex(" i=",info,
          IffyStr(StrLength(name)>0," '" + name + "' "," ")))
          + (whom != null) + IffyStr(FreshFaster," FF"," nf")
          + LogVH(" s=",scrn," e=") + edge
          + "\n  '" + aLine + "' = " + why);
      if (why==0) if (whom != null) whom.LogMe("(++) ");}} //~NewDragWgt

  /**
  * LoadUserWgts takes the import file "GameMaker.txt" (currently in
  * instance variable AddUserWgts) and feeds it one line at a time to
  * NewDragWgt to populate the game board as it was when the file was saved.
  */
  public void LoadUserWgts() {
    boolean didit = NthOffset(0,"[Std Alert Dialog]",AddUserWgts)>0;
    GameWgt whom = FindSubWgt(GameBoard,0x466C677A,0); // = 'Flgz'
    String aLine = NthItemOf('\n',1,AddUserWgts);
    int lino, colo = 0, tops = 0, lift = 0, area = 0,
        info = SafeParseInt(NthItemOf(' ',2,aLine)),
        tall = SafeParseInt(NthItemOf(' ',5,aLine)),
        wide = SafeParseInt(NthItemOf(' ',6,aLine)), size = tall*0x10000+wide,
        whar = NthOffset(0,"{",aLine);
    if (GameBoard==null) return;
    if (OKbtn==null) return;
    if (info<5) return;
    BlurEv(null); // make sure ants are turned off
    if (whom != null) if ((info>255)||(GameBoard.GetSize() != size)) {
      if (tall>4) if (wide>4) {
        if (ListofWgts != null) {
          area = ListofWgts.GetSize();
          if (((area-(GameBoard.GetPosn()+size))&0x80008000) !=0) {
            tops = (area>>16)-tall;
            if (tops<0) tops = 4;
            lift = (area&0xFFFF)-wide;
            if (lift<0) lift = 4;
            GameBoard.SetPosn(tops,lift);}} //~if
        GameBoard.SetSize(tall,wide);
        whom.SetSize(tall,wide);} //~if
      whom.SetFlags(info);
      ShoInfoFrom(GameBoard);} //~if
    if (didit) if (AIdlgChkBx != null) didit = AIdlgChkBx.CheckIt(true);
    colo = SafeParseInt(NthItemOf(' ',7,aLine));
    if (colo>0) SetSwatchColor(ColoBx,colo);
    if (WgtNamEntry==null) return; // (last early exit)
    FreshFaster = true;
    if (whar>0) aLine = NthItemOf('{',2,NthItemOf('}',1,aLine));
      else aLine = "";
    if (StrLength(aLine)>0) if (!InvalidJava(aLine)) {
      GamesName = aLine;
      // GamePosn = GameBoard.Posn();
      GameSz = tall*0x10000+wide;
      if (SetInputText(WgtNamEntry,"",aLine,true))
        if (DoMyButn(OKbtn)) ChooseBtn(1);} //~if
    for (lino=2; lino <= 999; lino++) {
      aLine = NthItemOf('\n',lino,AddUserWgts);
      if (StrLength(aLine)==0) break; // normal exit
      if (SafeParseInt(aLine)==lino) NewDragWgt(aLine,lino);} //~for
    FreshFaster = false;
    BlurEv(null); // make sure ants are turned off
    lino = NthOffset(0," '[Sprite list]' #",AddUserWgts);
    if (lino>0) lino = SafeParseInt(Substring(lino+18,5,AddUserWgts));
    if (lino>0) NextWgtName = lino;
    lino = SafeWgtNum(GameBoard);
    if (lino>0) didit = SelectMe(lino,GameBoard);
      // if (false) if (didit) Antsify(true,GameBoard);} //~if
    if (Logging) System.out.println("(LoadUserWgts) " + info
        + SeeHex(" c=",colo," ...") + theImpoList);
    AddUserWgts = "";} //~LoadUserWgts

 /**
  * GameEnding is called once when the game window is about to close, so
  * you can do whatever needs to be done at the end. This override tries
  * to save the current game state in the output file "GameMaker.txt"
  * [but it doesn't seem to be working].
  */
  public void GameEnding() { // called when window is closing (GameMk)
    if (unSavedWgts) unSavedWgts = MakeWgtLst();} //~GameEnding

 /**
  * After the initial widgets have been created, StartUp is called to do
  * any other initialization the client game may have.
  *
  * This override searches the input widgets for the named widgets needed
  * to operate GameMaker, does some initilization, then calls NewListName
  * to put the game board into the visible widget list.
  */
  public void StartUp() { // called once to set it up             (GameMk)
    GameWgt whom = FindListWgt("[Circle tool icon ants]");
    int zx, lino, whar, why = 0;
    boolean didit = false;
    String aStr, aLine;
    if (whom != null) didit = AntsIcon(false,whom);
    whom = FindListWgt("[Sprite tool icon ants]");
    if (whom != null) didit = AntsIcon(false,whom);
    whom = FindListWgt("[Ball tool icon ants]");
    if (whom != null) didit = AntsIcon(false,whom);
    GdChkBx = FindListWgt("[Grid checkbox]");
    TypChkBx = FindListWgt("[Typable checkbox]");
    ClikChkBx = FindListWgt("[Clickable checkbox]");
    DragChkBx = FindListWgt("[Dragable checkbox]");
    BumpChkBx = FindListWgt("[Collidable checkbox]");
    RollChkBx = FindListWgt("[Rollable checkbox]");
    AntsChkBx = FindListWgt("[Ants checkbox]");
    if (AntsChkBx != null) AntsChkBx.HideSho(false);
    AIdlgChkBx = FindListWgt("[Alrt+InDlg checkbox]");
    ColoBx = FindListWgt("[Color Swatch]");
    NullWgt = FindListWgt("[null widget]");
    GameBoard = FindListWgt("[Game board asset list]");
    ToolHiLi = FindListWgt("[Tool icon hilight]");
    UsWgHiLi = FindListWgt("[Name list hilight]");
    RexTool = FindListWgt("[Draggable Rect tool]");
    SpriTool = FindListWgt("[Draggable Sprite tool]");
    CircTool = FindListWgt("[Draggable Circle tool]");
    BallTool = FindListWgt("[Draggable Ball tool]");
    TexTool = FindListWgt("[Draggable Text tool]");
    SkyTool = FindListWgt("[Rect Sprite tool]");
    ListofWgts = FindListWgt("[User wgts list]");
    ToolList = FindListWgt("[Tool Palette]");
    ToolPanel = FindListWgt("[Tool panel]");
    TopField = FindListWgt("[Top entry field]");
    LeftField = FindListWgt("[Left entry field]");
    TallField = FindListWgt("[Tall entry field]");
    WideField = FindListWgt("[Wide entry field]");
    VeloEntry = FindListWgt("[Velocity entry group]");
    if (VeloEntry != null) VeloEntry.HideSho(false);
    whom = FindListWgt("[User wgt name list]");
    if (whom != null) {
      UwgtNamID = whom.GetRefNum();
      UwgtNamPos = whom.AddPrntsPos(whom.GetPosn());
      UwgtNamList = whom;} //~if
    FontRad = FindListWgt("[Font selector radio group]");
    GameNameWgt = FindListWgt("[Game board actual name]");
    if (GameNameWgt != null) {
      aStr = GameNameWgt.GetTextLn();
      GamesName = aStr;} //~if
    if (GameBoard != null) {
      GamePosn = GameBoard.GetPosn();
      GameSz = GameBoard.GetSize();} //~if
    // CxAntz = null;
    // if (true) { // this W/A OK..
    OKbtn = FindListWgt("['OK' (accept) button]");
    // UndoBtn = FindListWgt("['Undo' button]"); // not there now
    BildBtn = FindListWgt("['Build' button]");
    JSbtn = FindListWgt("['->JS' button]");
    CantJS = FindListWgt("['noJS' button]");
    WgtNamEntry = FindListWgt("[name entry panel]");
    ToolName = FindListWgt("[Current wgt type name]"); // } //~if
    if (BallTool != null) CircleSize(BallTool,0);
    if (CircTool != null) {
      CircleSize(CircTool,0);
      CircTool.HideSho(false);} //~if
    if (SpriTool != null) {
      didit = MkSpriteAnts(SpriTool,false);
      SpriTool.HideSho(false);} //~if
    if (FontRad != null) {
      why = ChoseFont(FontRad,(int)('T'));
      FontRad.HideSho(false);} //~if
    if (TexTool != null) why = ChoseFont(TexTool,(int)('T'));
    if (WgtNamEntry != null) { // widen name panel..
      // didit = SetInputText(WgtNamEntry,"","            ",false);
      whom = FindSubWgt(WgtNamEntry,0x4E616D66,0); // = 'Namf'
      if (whom != null) whom.SetSize(0,72);} //~if
    NewData(true); // ChooseBtn(1);
    didit = NewListName(GameBoard); // also selects & displays it
    if (GameBoard != null) LoadUserWgts();
    FloatDlogs();
    if (Logging||true) System.out.println("(GameMaker StartUp) " + didit
        + IffyStr(UwgtNamList==null," -uwnl",
          " +UwNmL=" + UwgtNamID + LogVH(" @ ",UwgtNamPos,""))
        + IffyStr(GdChkBx==null," -ckbx"," +CkBx")
        + IffyStr(ColoBx==null," -cosw"," +ColoSw")
        // + IffyStr(ColoDlog==null," -codlg"," +CoDlg")
        // + IffyStr(AskColo==null," -asco"," +AsCo")
        + IffyStr(RxWgt==null," -rxwg"," +RecWg")
        + IffyStr(CxWgt==null," -cxwg"," +CirWg")
        + IffyStr(OKbtn==null," -nOKbn"," +OKbtn")
        + "\n  fnts = " + MonoFontNo + "/" + TextFontNo + "/" + TinyFontNo
        + "/" + BoldFontNo + "/" + HeadFontNo);} //~StartUp

 /**
  * GameList returns a text descriptor of all the specified widgets in
  * a game. This override returns the constant String BaseMakerWgts.
  *
  * @param  whom  A reference to the host JavaGame, which is sent to super
  *
  * @return       BaseMakerWgts
  */
  public String GameList(JavaGame whom) { //                   // (GameMk)
    int whar, tye, nx, zx = 0,
        lxx = 0, lino = 0, seen = 0, info = 0, why = 0;
    String data = "", theList = "", aWord, aLine = super.GameList(whom);
    theList = BaseMakerWgts;
    data = super.GameList(whom);
    AddUserWgts = ReadWholeTextFile("GameMaker.txt");
    if (NthOffset(0,"[List of item lists]",AddUserWgts)<0) AddUserWgts = "";
    return theList;} //~GameList

 /**
  * Constructor..
  */
  public GameMakerEv() { // . . . . . . . . . . . . . . . . . . . (GameMk)
    NextWgtName = 0;
    ChosenBtn = 0;
    UwgtNamID = 0;
    UwgtNamPos = 0;
    GamePosn = 0;
    GameSz = 0;
    InChanged = false;
    ToolActive = false;
    FreshFaster = false;
    NullWgt = null;
    FontRad = null;
    GdChkBx = null;
    TypChkBx = null;
    ClikChkBx = null;
    DragChkBx = null;
    BumpChkBx = null;
    RollChkBx = null;
    AIdlgChkBx = null;
    AntsChkBx = null;
    ColoBx = null;
    RxWgt = null;
    CxWgt = null;
    CxAntz = null;
    OKbtn = null;
    JSbtn = null;
    CantJS = null;
    BildBtn = null;
    UndoBtn = null;
    ToolHiLi = null;
    UsWgHiLi = null;
    RexTool = null;
    SkyTool = null;
    TexTool = null;
    BallTool = null;
    CircTool = null;
    SpriTool = null;
    GameBoard = null;
    ListofWgts = null;
    UwgtNamList = null;
    WgtNamEntry = null;
    ToolPanel = null;
    ToolName = null;
    ActivWgt = null;
    ToolList = null;
    TopField = null;
    LeftField = null;
    TallField = null;
    WideField = null;
    TextInWgt = null;
    VeloEntry = null;
    GameNameWgt = null;
    ValidUserName = "";
    AddUserWgts = "";
    GamesName = "";
    UserNames = "";
    MkTmpStr = "";
    WgtList = "";
    WgtMap = "";
    System.out.println("(new GamMkEv)");}} //~GameMakerEv

  // Required Java class (Java version)..
  private class PaintAction implements ActionListener { @Override
    public String toString() {return "(JavaGame PaintAction)";}  @Override
    public void actionPerformed(ActionEvent evt) {
      if (theWindow != null)
        theWindow.repaint();}} //~PaintAction

// Private variables used by low-level game engine code..

  private static GameMakerEv myMaker = null;

  private boolean BusyPaint = false;
  private PaintAction JavaOften = null; // for periodic events..
  // private BufferedImage theImag = null; // use theBuff for both
  private BufferedImage theBuff = null;

// [Shared] methods used only hereafter..

  /**
  * Starts the game timer.
  *
  * @param  ms    =100, nominally the number of milliseconds
  */
  private static void JstartTimer(int ms) { // =100ms
    if (ms>0) {
      if (TickTock != null) TickTock.start(); // posts "Timr"
        else System.out.println("** Oops, no Timer **");}} //~JstartTimer

//  [Shared] JavaGame methods..

 /**
  * Start the animation timer.
  *
  * @param why    The initial timer count, >0 when animating
  */
  public void TimerStart(int why) { // Start timer (all vers)
    int tm =   FrameTime;
    if (why==0) return;
    if (TimerTix>0) return;
    if (TimerTix==0) {
      TimerTix = why;
      JstartTimer(tm);
      if (Logging||true)
        System.out.println("[start timer = " + tm + "ms @ " + why + "]");} //~if
    else if (TimerTix<0) if (why>0) {
      if (Logging||true) System.out.println("[restart timer @ " + why + "]");
      TimerTix = why;}} //~TimerStart

 /**
  * Test for GameWgt collision.
  * Only sprites can move, so they only can collide, but only if collidable
  * and only with objects marked collidable (+256)
  *
  * @param whom    A sprite or list that wants collision testing
  */
  void Collider(GameWgt whom) { // Collision tester, from Every100m.
    int rx, size, far, more, here, thar, whar = 0, refn = 0, info = 0;
    GameWgt twgt;
    GameEvent evh = UserEvHandler;
    if (evh == null) return;
    if (whom==null) return;
    refn = whom.GetRefNum();
    here = whom.AddPrntsPos(whom.GetPosn()); // abs posn of sprite being tested
    size = whom.GetSize();
    far = PairNeg(AddPair(here,size)); // (neg)bottom-right of same wgt
    info = whom.GetType();
    if ((info&0x81FF)==0x8106) { // ran off the edge, say so
      whom.SetFlags(info&~0x8000);
      if (evh.Collided(whom,null)) return;} //~if
    if ((info&255)==5) for (whar=0; whar <= MaxWgts; whar++) {
      twgt = GetIxWgt(whom.GetIxData(whar,0)); // try all sprites..
      if (twgt==null) return; // normal exit: end of list
      if (twgt != whom) if (twgt != DontBump) // ignore all but Bump+Sprite wgt..
        if ((twgt.GetType()&511)==262) Collider(twgt);} //~for
    for (whar=NextRefNo-1; whar >= 1; whar += -1) { // Bump+Sprite wgt, try all Bumpable wgts
      if (whar==refn) continue; // can't collide with self
      twgt = AllWgts[whar];
      if (twgt==null) continue;
      thar = twgt.AddPrntsPos(twgt.GetPosn());
      if (twgt==whom) continue;
      if (twgt==DontBump) continue;
      if ((twgt.GetType()&256)==0) continue;
      if (twgt != GameWgtsList) {   // is whom below or to right of twgt..
        if ((AddPair(thar,far)&0x80008000) != 0x80008000) continue;
        else if ((AddPair(here,PairNeg(AddPair(thar,twgt.GetSize())))
            & 0x80008000)!= 0x80008000) continue;} //~if // above or left of twgt
      else if ((AddPair(twgt.GetSize(),far)&0x80008000) !=0)
        twgt = null; // whom is off botm or right edge
      else if ((here&0x80008000) !=0) twgt = null; // off top or left
        else continue;
      if (evh != null) if (evh.Collided(whom,twgt)) break;} //~for
    // DoDeferredOps();
    if (Logging) System.out.println("   -Collr #" + refn + "");} //~Collider

  /**
  * MouseDispatch accepts mouse activity from whatever source, then chooses
  * one or more event handlers to direct it to.
  *
  * @param  msg    a number representing what happened to the mouse:
  *                  =0: rollover/idle, 1: mouse down, 2: drag, 3: mouse up,
  *                  +4: right-click, +8* gives reason/source for log
  */
  private void MouseDispatch(int msg, int vert, int horz) {
    // msg: =0: rollover/idle, 1: mouse down, 2: drag, 3: mouse up, +4: right
    //      // +8* gives reason/source
    // vert,horz are window-relative (or 0,0 if unknown)
    //
    //   boolean ValidMouse = false, WantsRollover = false, WantsDrag = false,
    //       MouseIsDown = false;
    //   int MouseVert = 0, MouseHorz = 0;
    // Gtype: 0= static rect, 1= pix image, 2= 1-color icon, 3= font,
    //   4= text line, 5= wgt list, 6= sprite, 7= scrolled (text); +flags:
    //   +8: user-defined subclass, +256: collision, +512: takes keyboard evts,
    //  +1K: click evts, +2K: drag evts, +4K: rollover evts, +8K: no focus(scro)
    //
    // These enabled by Gtype&512:
    //   void KeyData(GameWgt whom, char info)
    //   void FocusEv(GameWgt whom)
    //   void BlurEv(GameWgt whom)
    // This is enabled by Gtype&1024: (vert,horz are wgt-relative)
    //   boolean ClickEvt(GameWgt whom, int vert, int horz)
    // These are enabled by Gtype&2048:
    //   void MouseDn(GameWgt whom, int vert, int horz)
    //   void Dragging(GameWgt whom, int vert, int horz)
    //   void MouseUp(GameWgt whom, int vert, int horz)
    //   void CancelDrag(GameWgt whom)
    // These are enabled by Gtype&4096:
    //   void RollIn(GameWgt whom, int vert, int horz)
    //   void RollOver(GameWgt whom, int vert, int horz)
    //   void RollOut(GameWgt whom, int vert, int horz)
    int prio = msg, locn, here, thar, whar = 0, preDr = 0,
        mloc = 0, refn = 0, whif = 0, bitz = 0, info = 0, why = 0;
    boolean clikt = false, okfoc = false, doit = false, txt = false;
    GameWgt whom, ClkWg = null, RoWgt = null, WgBeDrg = null, FocWg = null;
    GameEvent evh = UserEvHandler;
    if (!ValidMouse) why = 16;
    if (MouseIsDown) why = why+8;
    if (msg>12) msg = msg-16;
    while (true) {
      why++; // why = +1
      if (evh == null) break; // can't do anything if no handler
      why++; // why = +2
      if (msg+2 ==0) { // frm Every100, somebody wants rollover evts..........
        if (!ValidMouse) break;
        msg = 0;
        if (MouseIsDown) {
          if (!WantsDrag) break;
          msg = 2;} //~if
        else if (!WantsRollover) break;
        horz = MouseInfo.getPointerInfo().getLocation().x
            - getLocationOnScreen().x - SignExtend(EdgeOffs);
        vert = MouseInfo.getPointerInfo().getLocation().y-(EdgeOffs>>16)
            - getLocationOnScreen().y;} //~if
      else if (msg<0) if (ValidMouse) if (evh != null) { // (Java) mouse exit
        WgBeDrg = evh.GetRefWgt(2);
        if (WgBeDrg != null) {
          preDr = WgBeDrg.GetRefNum();
          bitz++; // bitz = +1
          evh.CancelDrag(WgBeDrg); // also sets WBD=null
          WgBeDrg = null;
          doit = true;} //~if
        if (false) if (MouseIsDown) { // maybe better to require esc keypress
          if (WantsDrag) MouseDispatch(8+3,vert,horz);    // ..or else ignore
          MouseIsDown = false;} //~if
        ValidMouse = false;} //~if
      why++; // why = +3
      if (msg<0) break;
      msg = msg&3; // higher bits to log where it came from (now in prio)
      MouseVert = vert; // these are window-relative coordinates
      MouseHorz = horz; // (used for 2-click & drag substitution)
      mloc = (vert<<16)+(horz&0xFFFF);
      // here = (vert<<16)+(horz&0xFFFF);
      ValidMouse = true;
      if (MouseIsDown) if (msg>1) if (evh != null) { // mouse still down
        ClkWg = evh.GetRefWgt(1); // (only set below)
        if (msg==3) MouseIsDown = false;
        if (ClkWg != null) {
          // whom = ClkWg;
          info = ClkWg.GetType();
          if ((info&255)==4) txt = true;
          bitz = bitz|info&0xF00;
          okfoc = ((info&0x2000)==0);
          if ((info&2048) !=0) {
            horz = LocalizePos(ClkWg,vert,horz);
            vert = LocalVert;
            bitz = bitz+2;
            if (MouseIsDown) { // ClickLocn = mloc;
              bitz = bitz+2; // bitz = +4
              if (WgBeDrg==null) if (!KillDrag) {
                WgBeDrg = evh.GetRefWgt(2);
                if (WgBeDrg==null) { // start drag if moved far enough..
                  locn = evh.GetClickLoc(); // (set by FocEv)
                  // thar = AddPair((vert<<16)+horz,PairNeg(locn));
                  thar = ((vert<<16)+horz)-locn; // (faster, less accurate)
                  if (!txt) { // calc iAbs of each half..
                    info = thar&0x80008000; // (slight error in LSB)
                    info = info+info-(info>>15); // =FFFF if <0
                    thar = (thar^info)&0xFFF8FFF8;} //~if // !=0 if moved >8px
                  if (locn >= 0) if (thar !=0) { // (only place drag starts)
                    WgBeDrg = evh.StartDrag(ClkWg,locn>>16,locn&0xFFFF);
                    if (WgBeDrg==null) {
                      evh.BlurEv(null);
                      KillDrag = true;}}}} //~if
              if (WgBeDrg != null) {
                if (preDr==0) preDr = WgBeDrg.GetRefNum();
                horz = LocalizePos(WgBeDrg,MouseVert,MouseHorz);
                vert = LocalVert;
                evh.Dragging(WgBeDrg,vert,horz);}} //~if
            else if (WgBeDrg != null) {
              if (preDr==0) preDr = WgBeDrg.GetRefNum();
              horz = LocalizePos(WgBeDrg,MouseVert,MouseHorz);
              vert = LocalVert;
              evh.MouseUp(WgBeDrg,vert,horz);} //~if
            else {
              if ((info&1024) !=0) {
                whif = (whif<<4)+7;
                FocWg = evh.GetRefWgt(-1);
                if (okfoc) if (FocWg != ClkWg) {
                  whif = (whif<<4)+8;
                  evh.FocusEv(ClkWg,vert,horz); // also calls SetFocus
                  FocWg = evh.GetRefWgt(-1);} //~if // FocEv may redirect it!
                bitz = bitz|32;
                clikt = evh.ClickEvt(ClkWg,vert,horz);} //~if // 1st of 2 ClkEv
              evh.MouseUp(ClkWg,vert,horz);} //~else // bitz = +2
            clikt = true;
            info = 0;} //~if
          else if (MouseIsDown) info = 0;
            else info = info&1024;
          if (info !=0) {
            thar = ClkWg.AddPrntsPos(ClkWg.GetPosn());
            if ((AddPair(thar,PairNeg(mloc))&0x80008000) != 0x80008000)
              info = 0; // not still inside it..
            else if ((AddPair(mloc,PairNeg(AddPair(thar,ClkWg.GetSize())))
                &0x80008000) != 0x80008000) info = 0;
            if (info !=0) {
              whif = (whif<<4)+2;
              FocWg = evh.GetRefWgt(-1);
              horz = LocalizePos(ClkWg,vert,horz);
              vert = LocalVert;
              if (okfoc) if (FocWg != ClkWg) {
                whif = (whif<<4)+1;
                evh.FocusEv(ClkWg,vert,horz); // also calls SetFocus
                FocWg = evh.GetRefWgt(-1);} //~if // FocEv may redirect it!
              bitz = bitz|32;
              clikt = evh.ClickEvt(ClkWg,vert,horz); // 2nd of 2 ClickEv calls
              clikt = true;}}} //~if
        if (!MouseIsDown) {
          WgBeDrg = null; // local only, for the log
          evh.SetRefWgt(5,null);
          ClkWg = null;}} //~if // (mouse still down)
      why++; // why = +4
      if (clikt) break;
      why++; // why = +5
      if (msg==3) {
        MouseIsDown = false; // could be missed, if no handler
        break;} //~if
      why++; // why = +6
      if (evh != null) RoWgt = evh.GetRefWgt(0);
      if (RoWgt != null) {
        info = msg;
        if ((RoWgt.GetType()&4096)==0) info++;
        else if (evh != null) {
          if (info==0) { // otherwise >0 if msdn
            thar = RoWgt.AddPrntsPos(RoWgt.GetPosn());
            if ((AddPair(thar,PairNeg(mloc))&0x80008000) != 0x80008000)
              info--; // not still inside it..
            else if ((AddPair(mloc,PairNeg(AddPair(thar,RoWgt.GetSize())))
                &0x80008000) != 0x80008000) info--;} //~if
          horz = LocalizePos(RoWgt,vert,horz);
          vert = LocalVert;
          bitz = bitz+8;
          if (info !=0) { // not still inside & rollover..
            bitz = bitz+8; // bitz = +16
            evh.RollOut(RoWgt,vert,horz);} //~if
          else evh.RollOver(RoWgt,vert,horz); // bitz = +8
          doit = true;} //~if
        else info++;
        if (info==0) break; // did the rollover notification, so done
        if (evh != null) evh.SetRefWgt(4,null);
        RoWgt = null;} //~if
      why++; // why = +7
      MouseIsDown = msg>0;      // search back-to-front to get deepest 1st..
      for (whar=NextRefNo; whar >= 1; whar += -1) {
        whom = GetIxWgt(whar);
        if (whom==null) continue;
        info = whom.GetType(); // bitz = bitz+((info&0xF00)<<4);
        txt = ((info&255)==4);
        okfoc = ((info&0x2000)==0);
        if (info<1024) continue; // no mouse evts wanted here
        if (msg==0) if ((info&4096)==0) continue; // rollover unwanted
        thar = whom.AddPrntsPos(whom.GetPosn());
        if ((AddPair(thar,PairNeg(mloc))&0x80008000) != 0x80008000) continue;
        if ((AddPair(mloc,PairNeg(AddPair(thar,whom.GetSize())))
            &0x80008000) != 0x80008000) continue;
        thar = LocalizePos(whom,vert,horz); // horx
        here = LocalVert;
        if (msg==0) {
          evh.SetRefWgt(8,whom);
          RoWgt = whom;
          bitz = bitz|64;
          evh.RollOver(whom,here,thar);
          doit = true;
          break;} //~if
        ClkWg = whom;
        evh.SetRefWgt(9,whom); // = ClickWgt
        if (FocWg==null) {
          whif = (whif<<4)+4;
          FocWg = evh.GetRefWgt(-1);} //~if
        if ((info&2048) !=0) {
          bitz = bitz|128;
          KillDrag = false;
          if (okfoc) if ((FocWg != whom)||txt) { // scro drags but not foc..
            whif = (whif<<4)+3;
            evh.FocusEv(whom,here,thar);
            FocWg = evh.GetRefWgt(-1);} //~if // FocEv may redirect it!
          evh.MouseDn(ClkWg,here,thar); // this is the only msDn call
          clikt = true;} //~if
        else if ((info&1024) !=0) {
          if (okfoc) if ((FocWg != whom)||txt) {
            whif = (whif<<4)+5;
            evh.FocusEv(whom,here,thar);
            FocWg = evh.GetRefWgt(-1);}} //~if // FocEv may redirect it!
        break;} //~for // (whar: search back-to-front)
      break;} //~while
    if (evh != null) {
      if (FocWg==null) {
        whif = (whif<<4)+6;
        FocWg = evh.GetRefWgt(-1);} //~if
      if (ClkWg==null) ClkWg = evh.GetRefWgt(1);} //~if
    whom = FocWg;
    if (false) if (msg>0) if (!clikt) if (evh != null) if ((info&512) !=0)
        if (whom != null) if ((whom.GetType()&0x2000)==0) { // =okfoc
      whif = (whif<<4)+9;
      FocWg = null;
      evh.BlurEv(whom);
      // whom.SetFocus(false,0,0); // BlurEv did it
      clikt = true;} //~if
    if (TimerTix==0) if (doit||clikt||((whif&0x11111111) !=0)) // FocEv called
      TimerStart(8);
    if (Logging) {     // bitz = |info&0xF00 // if (!ValidMouse) why = +16;
      myTmpStr = SeeHex(" x",bitz," = "); // if (MouseIsDown) why = +8;
      if (ClkWg != null)
        myTmpStr = " Cw#" + ClkWg.GetRefNum() + myTmpStr;
      if (FocWg != null)
        myTmpStr = " Fw#" + FocWg.GetRefNum() + myTmpStr;
      if (RoWgt != null)
        myTmpStr = " Rw#" + RoWgt.GetRefNum() + myTmpStr;
      if (WgBeDrg != null)
        myTmpStr = " Dw#" + WgBeDrg.GetRefNum() + myTmpStr;
      if (preDr !=0) myTmpStr = " Pw#" + preDr + myTmpStr;
      System.out.println("(MsDispch) =" + msg + "/" + prio + LogVH(" ",mloc," => ")
          + vert + "/" + horz + " " + clikt + IffyStr(ValidMouse," Vm"," -v")
          + IffyStr(MouseIsDown," mD"," mU") + IffyStr(WantsDrag," wD"," -dr")
          + IffyStr(WantsRollover," wR"," -ro") + IffyStr(okfoc," okF "," -of ")
          + info + myTmpStr + why + SeeHex(" f=",whif,"")); // whif = // why =
      myTmpStr = "";} //~if
    // DoDeferredOps();
    } //~MouseDispatch

 /**
  * Timer-driven animation engine. Timer is started by TimerStart.
  */
  private void Every100ms() { // Timer-driven anim'n engine (all vers'ns)
    int logy = SeenLog&0x10000, info, wide = WinWi;
    GameWgt whom;
    GameEvent evh = UserEvHandler;
    if (TimerTix<0) TimerTix--;
    else if (TimerTix>0) TimerTix++;
    if (evh != null) evh.DoOften();
    if (ValidMouse) {
      if (MouseIsDown) {if (WantsDrag) MouseDispatch(-2,0,0);} //~if
      else if (WantsRollover) MouseDispatch(-2,0,0);} //~if
    // else if (MouseIsDown) // (cancel drag)
    whom = AllWgts[0];
    if (Logging||(logy==0)) if ((MaxLogWgts>0)||(TimerTix<99)||(logy==0)) {
      if (whom != null) {
        if ((TimerTix>55)||(TimerTix+22<0)) SeenLog = SeenLog|0x10000;
          else wide--;
        myTmpStr = whom.toString();} //~if
      else myTmpStr = "000";
      System.out.println("(100ms) " + (evh != null) + " " + myTmpStr
          + IffyStr(TimerTix<0," "," +") + TimerTix);
      myTmpStr = "";} //~if
    MaxLogWgts--;
    if (whom==null) return;
    whom.DrawMe(0,0,WinHi,WinWi); // = AllWgts[0] = wPixM
    whom = GetIxWgt(whom.GetInfo()); // => +1 = [List of item lists]
    if (whom==null) return;
    if (logy==0) whom.LogMe("  [LoL] ");
    whom = GetIxWgt(whom.GetIxData(0,0)); // => +2 = [Sprite list]
    if (whom==null) return;
    if (logy==0) whom.LogMe("  [Spr] ");
    if ((whom.GetType()&255)==5) Collider(whom);
    // DoDeferredOps();
    } //~Every100ms

  /**
  * KeyDispatch accepts keystroke activity from whatever source,
  * then directs it to the KeyData event handler.
  * xxx. +4 private public
  *
  * @param  xCh    the key code
  * @param  mods   modifiers (unused, always =0)
  */
  private void KeyDispatch(char xCh, int mods) {
    GameEvent evh = UserEvHandler;
    GameWgt FocWg = null;
    int why = 0;
    if (evh != null) FocWg = evh.GetRefWgt(-2);
    if (Logging) {
      myTmpStr = " = ";
      if (FocWg != null)
        myTmpStr = " Fw#" + FocWg.GetRefNum() + " +"
            + (FocWg.GetType()&512) + myTmpStr;
      System.out.println("(KyDispch) " + (int)xCh
          + " " + mods + myTmpStr + why);
      myTmpStr = "";} //~if
    if (evh != null) if (FocWg != null) {
      if ((FocWg.GetType()&512) !=0) evh.KeyData(FocWg,xCh);
      else if (xCh < ' ') evh.KeyData(FocWg,xCh);} //~if
    // DoDeferredOps();
    } //~KeyDispatch

 /**
  * Install user event handler.
  *
  * @param whom    A subclass of GameEvent that handles the desired events
  */
  public void InstallMyEvts(GameEvent whom) {
    UserEvHandler = whom;
    if (whom != null) System.out.println("** Installed "
        + whom.toString() + " **");} //~InstallMyEvts

  // Java wants these if 'implements WindowListener'..
  @Override public void windowOpened(WindowEvent evt) {}
  @Override public void windowClosed(WindowEvent evt) {}
  @Override public void windowIconified(WindowEvent evt) {}
  @Override public void windowDeiconified(WindowEvent evt) {}
  @Override public void windowActivated(WindowEvent evt) {}
  @Override public void windowDeactivated(WindowEvent evt) {}

 /**
  * Accepts window-about-to-close notification.
  */
  @Override public void windowClosing(WindowEvent evt) {
    GameEvent evh = UserEvHandler;
    if (evh != null) evh.GameEnding();} //~windowClosing

  // Java may also want this if 'implements KeyListener'..
  @Override public void keyReleased(KeyEvent evt) {} // (+2688)

 /**
  * Accepts keystrokes to input text.
  */
  @Override public void keyTyped(KeyEvent evt) { // (in Game)
    char aCh = evt.getKeyChar();
    // int xCh = (int)aCh;
    if (Logging) if ((Qlog&4) !=0) System.out.println("(Game) key Typed"
       + " " + ((int)aCh));
    KeyDispatch(aCh,0);} //~keyTyped

 /**
  * Accepts keystrokes to control action, because Java does not
  * do these as keyTyped.
  */
  @Override public void keyPressed(KeyEvent evt) { // (Game)
    int whom = evt.getKeyCode();
    char aCh;
    if (whom== KeyEvent. VK_ENTER) aCh = '\n';
    else if (whom== KeyEvent. VK_TAB)
        aCh = '\t';
    else if (whom== KeyEvent. VK_BACK_SPACE)
        aCh = '\b';
    else if (whom== KeyEvent. VK_RIGHT) aCh = Arro_RIGHT;
    else if (whom== KeyEvent. VK_KP_LEFT) aCh = Arro_LEFT;
    else if (whom== KeyEvent. VK_LEFT) aCh = Arro_LEFT;
    else if (whom== KeyEvent. VK_KP_DOWN) aCh = Arro_DOWN;
    else if (whom== KeyEvent. VK_DOWN) aCh = Arro_DOWN;
    else if (whom== KeyEvent. VK_KP_UP) aCh = Arro_UP;
    else if (whom== KeyEvent. VK_UP) aCh = Arro_UP;
      else return;
    if (aCh > '\n') KeyDispatch(aCh,0);} //~keyPressed

  /**
  * JavaMouse accepts mouse activity from whatever source, translates the
  * coordinates to window-content-relative, then directs it to MouseDispatch.
  *
  * @param  whom    activity kind: 0=enter/roll, 1=press/click, 3=up, -1=exit
  * @param  evt     some kind of Java magic containing the click location
  */
  private void JavaMouse(int whom, MouseEvent evt) {
    Insets edges = getInsets();
    int logz = SeenLog&0x4000, kx = 0, Vx = 0, Hx = 0, Vz = 0, Hz = 0;
    boolean didit = false, rite = false ;
    if (whom<8) // mouseClick is redundant
        if (evt != null) if (edges != null) { // we only implement/o'ride this'n
      if (evt.getButton()==MouseEvent.BUTTON3) whom = whom|4;
      Hz = edges.left;
      Vz = edges.top;
      kx = (Vz<<16)+(Hz&0xFFFF);
      if (kx !=0) if (EdgeOffs==0) EdgeOffs = kx;
      Hx = evt.getX()-Hz; // now window(content)-relative
      Vx = evt.getY()-Vz;} //~if
    if (Logging||(logz==0)) {
      System.out.println("(Game) Got mouse @ " + Vx + "/" + Hx
          + " + " + Vz + "/" + Hz + " => " + whom);
      if (whom==3) SeenLog = SeenLog|0x4000;} //~if
    if (whom<8) MouseDispatch(whom+16,Vx,Hx);} //~JavaMouse

 /**
  * Accepts clicks on screen image to control operation
  */
  @Override public void mouseClicked(MouseEvent evt) { // (Game)
    JavaMouse(9,evt);} //~mouseClicked // possibly redundant

  // Java wants these if 'implements MouseListen' ..
  @Override public void mouseExited(MouseEvent evt) {
    JavaMouse(-1,evt);} //~mouseExited
  @Override public void mouseEntered(MouseEvent evt) {
    JavaMouse(0,evt);} //~mouseEntered
  @Override public void mousePressed(MouseEvent evt) {
    JavaMouse(1,evt);} //~mousePressed
  @Override public void mouseReleased(MouseEvent evt) {
    JavaMouse(3,evt);} //~mouseReleased

 /**
  * Converts a RGB pixel array to BufferedImage used in painting.
  * Adapted from example code found on StackExchange.
  *
  * @param  pixels  The pixel array
  * @param  width   Its width
  * @param  height  Its height
  *
  * @return         The BufferedImage result
  */
  public BufferedImage Int2BufImg(int[] pixels, int width, int height)
      throws IllegalArgumentException {
    int lxx = 0;
    int[] myDat = null; // Raster raz = null;  DataBufferInt DBI = nell;
    BufferedImage bufIm = null;
    if (pixels != null) lxx = pixels.length;

    if (lxx==0) return null;
    if (width==WinWi) if (height == WinHi) bufIm = theBuff;
    if (bufIm == null) // (should be never)
      bufIm = new BufferedImage(width,height, BufferedImage. TYPE_INT_RGB);
    if (bufIm == null) return null;
    myDat = ((DataBufferInt) bufIm. getRaster().getDataBuffer()). getData();
    System. arraycopy(pixels,0,myDat,0,lxx);
    return bufIm;} //~Int2BufImg

 /**
  * All the heavy lifting happens here. Called on timer activation.
  */
  @Override public void paint(Graphics graf) { // (in Game)
    Insets edges;
    if (BusyPaint) return; // Java won't accept local static method vars
    PendingDraw = false;
    BusyPaint = true;
    edges = getInsets();
    // super.paint(graf);
    if (thePixels != null) try {
      theBuff = Int2BufImg(thePixels,WinWi,WinHi);
      if (theBuff != null) if (graf != null) if (edges != null)
        graf.drawImage(theBuff,edges.left,edges.top,null);
      Every100ms();} //~if // calls: AllWgts[0].DrawMe();
    catch (Exception ex) {theBuff = null;}
    BusyPaint = false;} //~paint

  /**
  * StartGame is called from either the user game's main() or else
  * from JavaGame's main(), in either case with the desired window
  * title, size, and base color, plus an instance of the GameEvent
  * subclass that will be implementing the game (or GameMaker) logic.
  * StartGame creates the specified window, builds all the widgets,
  * calls the user StartUp method, then begins processing timer (and
  * mouse and keyboard) events.
  *
  * @param  tall    the window height
  * @param  wide    the window width
  * @param  color   the window background color
  * @param  title   the window title
  * @param  evh     the GameEvent subclass instance
  */
  public void StartGame(int tall, int wide, int color, String title,
      GameEvent evh) {
    int nx = tall*wide;
    String sayso = "= " + tall + "x" + wide;
    GameWgt whom;
    int[] myPix = null;
    boolean soon = true;
    AllWgts = new GameWgt[MaxWgts];
    if (Logging||true) System.out.println("(StartGame) " + sayso);
    InstallMyEvts(evh);
    if (nx>0) if (color>0) if (title != "") soon = false;
    if (soon) { // no win info, do import early..
      ImportGamList(); // gets "GameImports" from wherever
      whom = AllWgts[0];
      if (whom != null) {
        nx = whom.GetSize();
        color = whom.GetColor();
        tall = nx>>16;
        wide = nx&0xFFFF;
        nx = 0;
        if (tall>0) if (wide>0) {
          WinHi = tall;
          WinWi = wide;
          nx = tall*wide;
          sayso = " = " + tall + "x" + wide + " !";}} //~if
      else System.out.println("ImportGamList failed");} //~if
    else  if (nx>0) {
      WinHi = tall;
      WinWi = wide;} //~if
    sayso = sayso + " px=" + nx;
    // theStudent = null;
    if (nx>0) myPix = NewArray(nx); // nx = WinHi*WinWi
    if (myPix == null) {
      System.out.println("new pix or Canvas failed" + sayso);
      System.exit(9);
      return;} //~if // 1st of 2 early retn
    // PixOfx = LeftOfx;
    thePixels = myPix;
    GameWgt.  SetMyScreen(myPix,WinHi,WinWi);
     if (color <= 0) color = 0xFFFFFF; // prefill white if unspec'd
    while (nx>0) {
      nx--;
      if (nx<0) break;
      if (nx < myPix.length)
        myPix[nx] = color;} //~while
    // reduce memory-manager burden by pre-allocating this..
    theBuff = new BufferedImage(WinWi,WinHi, BufferedImage. TYPE_INT_RGB);
    // ** if (StartMe(FrameTime,null)) {} // copy in validation?
    if (StrLength(title)==0) title = "JavaGame";
    setSize(WinWi+18,WinHi+40); // make it larger for insets to come
    setTitle(title); // was: this.setTitle etc..   // did these in const'r..
    // setDefaultCloseOperation( JFrame. EXIT_ON_CLOSE);
    // addMouseListener(this); // if 'implements MouseListen..'
    // addKeyListener(this); // if 'implements KeyListen..'
    setVisible(true);
    if (!soon) ImportGamList();
    if (evh != null) evh.StartUp();
    MaxLogWgts = Max_Log;
    whom = AllWgts[0];
    if (whom==null) {
      System.out.println("*** no AllWgs[0] ***");
      return;} //~if // 2nd of 2 early retn
    myPix = whom.SwapData(myPix); // root needs -> Screen pixels
    whom.  DrawMe(0,0,WinHi,WinWi);
    TimerStart(-8);} //~StartGame

 /**
  * NewMaker is an access function to create a new GameMakerEv instance
  * when JavaGame is started from its own main().
  *
  * @return       the new GameMakerEv instance
  */
  public GameMakerEv NewMaker() {
    myMaker = new GameMakerEv();
    return myMaker;} //~NewMaker

 /**
  * This main should be called to run JavaGame in GameMaker mode. To run
  * in user game mode, use the main() code in the generated Java file.
  */
  public static void main() { // (in JavaGame) (Java only)
    JavaGame whom = new JavaGame();
    GameEvent evh = null;
    String data = "", title = "";
    int tall = 0, wide = 0, colo = 0xFFFFFF;
    while (true) {
      if (whom == null) break;
      title = "(Oops!)";
      evh = whom.NewMaker();
      if (evh == null) break;
      // theWindow = whom; // now in const'r
      data = evh.GameList(whom);
      title = NthItemOf('\"',2,data);
      tall = SafeParseInt(NthItemOf(' ',5,data));
      wide = SafeParseInt(NthItemOf(' ',6,data));
      colo = SafeParseInt(NthItemOf(' ',7,data));
      whom.StartGame(tall,wide,colo,title,evh);
      LogAlWgts("[main] ");
      break;} //~while
    System.out.println("{main} " + tall + "x" + wide + " '" + title + "'");}

 /**
  * Return the class name for this object.
  *
  * @return     The reference String for this object
  */
  public String toString() {return "(JavaGame)";} //~toString

 /**
  * This is the constructor, which sets everything up.
  */
  public JavaGame() { // outer class constructor..       // (+2788)
    MaxWinSize = Win_Hi*0x10000
        + Win_Wi;
    theWindow = this;
    JavaOften = new PaintAction();
    TimeBase = GetMilliSecs();
    TickTock = new Timer( FrameTime,JavaOften); // (not start it)
    if (Logging) System.out.println("iGame ");
    setDefaultCloseOperation( JFrame. EXIT_ON_CLOSE);
    addMouseListener(this); // if 'implements MouseListen..'
    addKeyListener(this); // if 'implements KeyListen..'
    if (false) setVisible(true);}} //~JavaGame

// Index to Selected Useful JavaGame Methods (BlueJ has none)..
//   // [Predefined] GameEngine Macros & Data..
//   // [Shared] GameEngine working data (static)..
//   CharAt  NthOffset  AddPair   class GameIco
//   UsrGameJv  AllWgts  GetMilliSecs  SafeParseInt
//   ReadWholeTextFile  ReplacAll  NewGameWgt  ParsFixPt
//   BuildThisMacro  FindSubWgt  Antsify  SetInputText
//   OpenInputDlog  IsDlgEvent  PaintAction  ImportGamList
//   class GameMakerEv  InvalidJava  ChoseFont  SetSwatchColor
//   ShoInfoFrom  MakeWgtLst  MakeUserJava  SelectMe  KeyFld
//   CliKnownStuff  StartDrag  Collided  StartUp
//   // [Shared] methods used only hereafter..
//   Collider  Every100ms  KeyDispatch  JvUpLdTest  toJS
//   windowClosing  JavaMouse  paint  StartGame  JavaGame()

  // (javagame) (BJ)
